/*************************************************************************
*  Program:       Recurr.c                                               *
*  Description:   General Recurrent-Network - Example for VieNet2 V2.03  *
*  Last modified: 25.01.97                                               *
*  Author:        Guenter Linhart                                        *
*  Copyright (C) OEFAI 1994-96 All Rights Reserved.                      *
**************************************************************************/

/* 25.1.97 GL 1) bugfix in NN_RN_Update() */

#include <iostream.h>
#include <malloc.h>
#include <time.h>

#include "neurodef.h"
#include "neurons.h"
#include "nngraph.h"
#ifdef PC
#include "biblio\nn_bp.h"
#else
#include "biblio/nn_bp.h"
#endif

#define SOURCENAME "RECURR"

#define MaxStates 10

long steps =15000;  /* nnumber of learning steps */
long actpat=0;      /* actual pattern number */

pINT  IN,
      FULL_IN,
      HIDDEN,
      OUT,
      BIAS,
      STATE_IN[MaxStates],
      STATE_HIDDEN[MaxStates],
      STATE_OUT[MaxStates];

pINT LayerSize[3];
REAL InitWeights[3];
REAL LEARNRATE;
pINT ActStates=1;

pINT UseFeedback[3];
REAL state_W_self[3][MaxStates];
REAL state_W_back[3][MaxStates];

REAL Error=0.0, MSE=0.0, RMSE=0.0;

/*----- reset all activations to default values -----*/
void reset_all(void)
{ reset_layers();
  first_pattern(IN);
  first_pattern(OUT);
  actpat=0;
  draw_layers();
}

/*----- connect network layers, initializes weights -----*/
void connect_all_new(void)
{ 
  int s;
  NN_BP_Init(FULL_IN,HIDDEN,OUT,BIAS,InitWeights[0],InitWeights[1]);

  for (s=0;s<ActStates;s++)
    {
    /* self recurrent connections */
    if (UseFeedback[0])
      {
      nnconnect(STATE_IN[s], STATE_IN[s], IDENT, 
		    state_W_self[0][s],state_W_self[0][s],1.0);
      nnconnect(IN, STATE_IN[s], IDENT, 
		    state_W_back[0][s],state_W_back[0][s],1.0);
      }
    if (UseFeedback[1])
      {
      nnconnect(STATE_HIDDEN[s], STATE_HIDDEN[s], IDENT, 
		    state_W_self[1][s],state_W_self[1][s],1.0);
      nnconnect(HIDDEN, STATE_HIDDEN[s], IDENT, 
		    state_W_back[1][s],state_W_back[1][s],1.0);
      }
    if (UseFeedback[2])
      {
      nnconnect(STATE_OUT[s], STATE_OUT[s], IDENT, 
		    state_W_self[2][s],state_W_self[2][s],1.0);
      nnconnect(OUT, STATE_OUT[s], IDENT, 
		    state_W_back[2][s],state_W_back[2][s],1.0);
      }
    }
}

/*----- calculates mean absolut error and	mean-squared-error
	for actual target values -----*/
void NNCalcError(pINT l)
{ pINT i;
  float err,mse=0.0;
  Error=0.0;
  for (L_ALL(l,i))
	 {
	 Error+=fabs(err = (X_target(l,i) - X(l,i)));
	 mse+=err*err;
	 }
  Error/=L_Tab[l].size;
  MSE=mse/L_Tab[l].size;
  RMSE=sqrt(MSE);
}


/*----- full recurrent update -----*/
void NN_RN_Update(void)
{ int s;
  pINT i,offs=0;

  /* 1) forward */
  update(IN,IN,EXT,SINGLE);

  for (L_ALL(IN,i))
    X(FULL_IN,i)=X(IN,i);
  offs+=i;

  for (s=0;s<ActStates;s++)
    {
    if (UseFeedback[0])
      {
      for (L_ALL(STATE_IN[s],i))
        X(FULL_IN,i+offs)=X(STATE_IN[s],i);
      offs+=i;
      }
    if (UseFeedback[1])
      {
      for (L_ALL(STATE_HIDDEN[s],i))
        X(FULL_IN,i+offs)=X(STATE_HIDDEN[s],i);
      offs+=i;
      }
    if (UseFeedback[2])
      {
      for (L_ALL(STATE_OUT[s],i))
        X(FULL_IN,i+offs)=X(STATE_OUT[s],i);
      offs+=i;
      }
    }

  NN_BP_Update(FULL_IN,HIDDEN,OUT);

  /* 2) recurrent */
  for (s=0;s<ActStates;s++)
    {    
    if (UseFeedback[0])
       {
       update(STATE_IN[s], STATE_IN[s], COPY_X_W, SINGLE);/* 2a) self-weightings */
       update(IN, STATE_IN[s], COPY_ADD_X_W, SINGLE);     /* 2b) add weighted actual activations */
       }
    if (UseFeedback[1])
       {
       update(STATE_HIDDEN[s], STATE_HIDDEN[s], COPY_X_W, SINGLE);
       update(HIDDEN, STATE_HIDDEN[s], COPY_ADD_X_W, SINGLE);
       }
    if (UseFeedback[2])
       {
       update(STATE_OUT[s], STATE_OUT[s], COPY_X_W, SINGLE);
       update(OUT, STATE_OUT[s], COPY_ADD_X_W, SINGLE);
       }
    }
}

/*----- one update step -----*/
void upd()
{
  update(IN,IN,EXT,SINGLE); /* load external pattern */
  NNSetTarget(OUT,NULL);   /* calc target values from actual pattern */
  NN_RN_Update();
}

/*----- one update step with display -----*/
void upd1(void)
{ upd();
  draw_layers();
  draw_ext(IN);
  draw_ext(OUT);
  NNCalcError(OUT);
  pprintf("Actual RMSE: %.3f\n",RMSE);
}

/*----- learn n steps, break with key "ESC" -----*/
void learn_n(void)
{time_t start,stop;
 long i;
 float last_rmse,rmse=0.0;
 pprintf("Learning (break = 'ESC')\n");
 time(&start);
 actpat=0;
 reset_layers();
 for (i=0;i<steps;i++)
     {
     nth_pat(OUT,nth_pat(IN,actpat));
     upd();
     NN_BP_Learn(FULL_IN,HIDDEN,OUT,LEARNRATE);
     if (nn_break()) break;
     actpat=(actpat<patanz(IN)-1)?++actpat:0;
     }
 time(&stop);
 if (start<stop)
    pprintf("Learning time for %li steps: %li s (%.2f kCUPS)\n",i,(stop-start),
                      ((float)(i*weights_used/1000.0))/(float)(stop-start));
}

/*----- select next pattern and update all -----*/
void test(void)
{ nth_pat(IN,actpat);
  nth_pat(OUT,actpat);
  upd1();
  actpat=(actpat<patanz(IN)-1)?++actpat:0;
}

/*----- calc error values for all patterns -----*/
void calc_error()
{ float rmse=0.0;
  pINT i;
  for (i=0;i<patanz(IN);i++)
    {
    nth_pat(OUT,nth_pat(IN,i));
    upd();
    NNCalcError(OUT);
    rmse+=RMSE;
  }
  pprintf("Mean RMSE: %.4f\n",rmse/i);
}

void set_pattern(void)
{ nth_pat(IN,actpat); nth_pat(OUT,actpat); upd1(); }

/*==== init new net with actual parameters =====*/
void setup_net()
{ char title[20];
  int i,size;

  IN     = init_layer(LayerSize[0],1,GRID,130,300,"In");
  HIDDEN = init_layer(LayerSize[1],1,DELTA_LAYER, 155,160,"Hidden");
  OUT	 = init_layer(LayerSize[2],1,DELTA_LAYER,180,110,"Out");

  init_ext(IN, -80,0,"extern");
  init_ext(OUT,-80,0,"target");

  size=LayerSize[0];
  if (UseFeedback[0]) size+=LayerSize[0]*ActStates;
  if (UseFeedback[1]) size+=LayerSize[1]*ActStates;
  if (UseFeedback[2]) size+=LayerSize[2]*ActStates;

  FULL_IN = init_layer( size,1,GRID,130,220,"Full-In");
  BIAS  =  init_layer( 1,1,BIAS_LAYER,80,190,"Bias");

  for (i=0;i<ActStates;i++)
      {
      sprintf(title,"State_IN_%u",i);
      if (UseFeedback[0])
      	 STATE_IN[i] = init_layer( LayerSize[0],1,GRID,130,350+i*40,title);
      sprintf(title,"State_HIDDEN_%u",i);
      if (UseFeedback[1])
         STATE_HIDDEN[i] = init_layer( LayerSize[1],1,GRID,205,350+i*40,title);
      sprintf(title,"State_OUT_%u",i);
      if (UseFeedback[2])
         STATE_OUT[i] = init_layer( LayerSize[2],1,GRID,400,350+i*40,title);
      }
  
  connect_all_new();

}

/*===== setup net and I/O =====*/
void nn_init(int argc, char** argv)
{ int s,i;

  /* set default values */
  LayerSize[0]=2; 
  LayerSize[1]=10;
  LayerSize[2]=3;
  InitWeights[0]=-0.5;
  InitWeights[1]=0.5;
  InitWeights[2]=1.0;
  LEARNRATE = 0.05;

  UseFeedback[0]=0; /* dont use feedback from input layer */
  UseFeedback[1]=1; /* use feedback from hidden layer */
  UseFeedback[2]=1; /* use feedback from output layer */
  for (s=0;s<MaxStates;s++)
    for (i=0;i<3;i++)
	    {
      state_W_self[i][s]=0.1;
	    state_W_back[i][s]=0.4;
	    }
	
  setup_net();

  load_patterns("input.pat",IN);
  load_patterns("output.pat",OUT);

  make_button("Reset",reset_all,'R');
  make_button("Update",upd1,'u');
  make_button("Learn",learn_n,'l');
  make_button("Test",test,'t');
  make_button("Calc Error",calc_error,'E');

  make_slider("Learnsteps:", NULL,' ',&steps,1,100000L);
  make_slider_e("Rate:", NULL,' ',"%.2f",&LEARNRATE, 0.0,1.0,0.01);
  make_slider_e("Gain:", NULL,' ',"%.2f",&GAIN, 0.0,10.0,0.1);

  make_slider("Pattern:",set_pattern,' ',&actpat,0L,(long)L_Tab[IN].patanz-1);

  init_xgraph(argc,argv);

  pprintf("This is a demo of VieNet2 %s\n\n",VERSION);
  pprintf("Recurrent-Example\n");
  pprintf("=================\n\n");

  wait_for_action();
}

#if !defined(WIN)
void main(int argc, char** argv)
{ nn_setup(argc,argv);
  nn_init(argc,argv);}
#endif


