/*************************************************************************
*  Program:       BACKPRP1.C                                             *
*  Description:   Simple Backprop-example for VieNet2                    *
*  Last modified: 21.02.94                                               *
*  Author:        LG                                                     *
**************************************************************************/

#include "neurodef.h"
#include "neurons.h"
#include "nngraph.h"
#include "biblio\nn_bp.h"
#include <time.h>

#define SOURCENAME "BACKPRP1"

pINT 	IN,
			HIDDEN,
			OUT;

/* Learning Steps */
long steps=200;

/* Learning rate */
float rate  = 0.9;

#ifndef WIN
/* standard-functions, already defined in windows-menus */
void savew(void)
{ pprintf("saving \"%s\" ...\n",WeightFile);
	if (save_weights(WeightFile)) pprintf("done\n");}

void loadw(void)
{ pprintf("loading \"%s\" ...\n",WeightFile);
	if (load_weights(WeightFile)) pprintf("done\n");}
#endif

/* reset all activations to default values */
void reset_all(void)
{ reset_layers();
  draw_layers();
}

/* functions for choosing input-patterns */
void nextpat(void)
{ next_pattern(IN); next_pattern(OUT); draw_ext(IN); draw_ext(OUT); }
void prevpat(void)
{ prev_pattern(IN); prev_pattern(OUT); draw_ext(IN); draw_ext(OUT); }
void randpat(void)
{ nth_pat(OUT,rand_pat(IN)); draw_ext(IN); draw_ext(OUT); }

/* connect network layers, initializes weights */
void connect_all_new(void)
{
  nnconnect(    IN, HIDDEN, FULL,-0.6,0.6,1.0);
  nnconnect(HIDDEN, OUT,    FULL,-0.6,0.6,1.0);
}

/* full update through network */
void upd(void) 
{
	update(IN,IN,EXT,SINGLE);
  update(IN,HIDDEN,STANDARD,SINGLE);
	copy_layer(HIDDEN);
  update(HIDDEN,OUT,STANDARD,SINGLE);
  copy_layer(OUT);
}

/* update and show */
void upd1(void)
{	upd(); draw_layers(); }

/* learn n steps, break with key "ESC" */
void learn_n(void)
{time_t start,stop;
 long i;
 pprintf("Learning (break = 'ESC')\n");
 time(&start);
 for (i=0;i<steps;i++)
     { nth_pat(OUT,rand_pat(IN));
       upd();
       NN_BP_Learn(IN,HIDDEN,OUT,rate);
			 if (nn_break()) break;
		 }
 time(&stop);
 pprintf("done\n");
 if (start<stop) pprintf("Learning time for %li steps: %li s (%.2f Cps)\n",i,(stop-start),((float)(i*weights_used))/(float)(stop-start));
}

/* select next pattern and update all */
void test(void)
{
 static pINT i=0;
 nth_pat(IN,i);
 nth_pat(OUT,i);
 upd1();
 i=(i<patanz(IN)-1)?++i:0;
}

/* main init function */
void nn_init(int argc, char** argv)
{
	DX  =  12; /* set Unit gridsize */
	UWID = 10; /* set Unit width */

	IN	   = init_layer( 3,3,GRID,100,220,"In");
	HIDDEN = init_layer(10,1,DELTA_LAYER, 40,160,"Hidden");
	OUT	   = init_layer( 3,3,DELTA_LAYER,100, 80,"Out");

	init_ext(IN, -80,0,"extern");
	init_ext(OUT,-80,0,"target");

	connect_all_new();

#ifndef WIN
	make_button("QUIT",shutdown,'Q');
	make2button("loadw",loadw,'l',"savew",savew,'s');
	make_button("NEW WEIGHTS",connect_all_new,'W');
	make_button("RESET",reset_all,'R');
#endif
	make_button("Update",upd1,'u');
	make_button("Learn",learn_n,'l');
	make_button("Test",test,'t');
	make2button("- I/O",prevpat,'-',"+",nextpat,'+');
	make_button("rand I/O",randpat,'r');

	make_slider("Learnsteps:", NULL,' ',&steps,1,1000);
	make_slider_e("Learnrate:",NULL,' ',"%.2f",&rate, 0.0,1.0,0.02);

	make_pattern(IN,"111000000");
	make_pattern(IN,"110100000");
	make_pattern(IN,"010110000");
	make_pattern(IN,"000110001");
	make_pattern(IN,"000010101");
	make_pattern(IN,"000010110");
	make_pattern(IN,"001000110");
	make_pattern(IN,"001001010");
	make_pattern(IN,"001001001");
	make_pattern(IN,"100001001");
	make_pattern(IN,"100100001");
	make_pattern(IN,"100110000");
	make_pattern(IN,"000110010");
	make_pattern(IN,"010010010");
	make_pattern(IN,"011010000");
	first_pattern(IN);
	make_pattern(OUT,"110100000");
	make_pattern(OUT,"010110000");
	make_pattern(OUT,"000110001");
	make_pattern(OUT,"000010101");
	make_pattern(OUT,"000010110");
	make_pattern(OUT,"001000110");
	make_pattern(OUT,"001001010");
	make_pattern(OUT,"001001001");
	make_pattern(OUT,"100001001");
	make_pattern(OUT,"100100001");
	make_pattern(OUT,"100110000");
	make_pattern(OUT,"000110010");
	make_pattern(OUT,"010010010");
	make_pattern(OUT,"011010000");
	make_pattern(OUT,"111000000");
	first_pattern(OUT);

#if defined(sun4) || defined(WIN)
	init_xgraph(argc,argv);
#endif

	pprintf("This is a demo of VieNet2 %s\n\n",VERSION);
	pprintf("Backprop-Example 1\n");
	pprintf("==================\n\n");

	wait_for_action();
}

#ifndef WIN
void main(int argc, char** argv)
{ nn_setup(argc,argv);
	nn_init(argc,argv);}
#endif

