/*************************************************************************
*  Program:       NNGRAPH.C                                              *
*  Description:   Graphicfunctions for VieNet2 on Apollo                 *
*  Version:       1.2                                                    *
*  Last modified: 20.9.93                                                *
*  Author:        LG                                                     *
*  Copyright (C) Austrian Research Institute for Artificial Intelligence *
**************************************************************************/

#include "neurodef.h"

#include "/usr/apollo/include/base.h"
#include "/usr/apollo/include/gpr.h"  
#include "/usr/apollo/include/pad.h"
#include "/usr/apollo/include/kbd.h"

#define MODULE
#include "neurons.h"
#include "nngraph.h"

#define BACK(col) ((col & 0x00f0) >> 4)   /* Aufteilen des Farbwertes in */
#define FRONT(col) (col & 0x000f)            /* Hintergrund und Vordergrund */
                                             /* (col=front+16*back)     */

void show_shadow(pINT,pINT,pINT,pINT,pINT);
 
int bitmap_x_size;
short int               pathname_length = 4, font_id;
name_$pname_t           pathname_of_font = {"f5x9"};
gpr_$event_t            event_type;
gpr_$keyset_t           mouse_buttons, key_set;
unsigned char           event_data;
gpr_$position_t         position;
/* gpr_$mask_32_t          hi_plane; */

status_$t            status;
gpr_$bitmap_desc_t   display_bitmap;
gpr_$offset_t        display_bitmap_size;
gpr_$rgb_plane_t     hi_plane;
gpr_$disp_char_t     display_characteristics;
   

void check(char *messagex)
{
   if (status.all)
   {   /* error_$print (status); !!! */
       printf("Error occurred while %s.\n", messagex);
   }
}
     
void g_acquire(void)
{ gpr_$acquire_display(&status); }

void g_release(void)
{ gpr_$release_display(&status); }

void bit_blt(pINT fromx,
             pINT fromy,
             pINT sizex,
             pINT sizey,
             pINT to_x,
             pINT to_y
            )
{ gpr_$window_t source_window;
  gpr_$position_t dest_origin;
  source_window.window_base.x_coord = fromx;
  source_window.window_base.y_coord = fromy;
  source_window.window_size.x_size = sizex;
  source_window.window_size.y_size = sizey;
  dest_origin.x_coord = to_x;
  dest_origin.y_coord = to_y;
  gpr_$pixel_blt(display_bitmap,source_window,dest_origin,&status);
}

void bit_blt_inverted(pINT fromx,
                      pINT fromy,
                      pINT sizex,
                      pINT sizey,
                      pINT tox,
                      pINT toy
                     )
{ gpr_$raster_op_t raster_op;
  raster_op=12; /* INVERT */
  gpr_$set_raster_op(hi_plane,raster_op,&status);
  bit_blt(fromx,fromy,sizex,sizey,tox,toy);
  raster_op=3; /* default */
  gpr_$set_raster_op(hi_plane,raster_op,&status);
}

void rectangle(pINT x, pINT y, pINT x1, pINT y1)
{ gpr_$draw_box((unsigned short)x,(unsigned short)y,
                (unsigned short)x1,(unsigned short)y1,&status);
}

void box(pINT x, pINT y)
{ gpr_$draw_box((unsigned short)x,(unsigned short)y,
                (unsigned short)(x+UWID),(unsigned short)(y+UWID),&status);
}

/* loescht Rechteck mit col=0/1 (weiss/schwarz) */
void clear_box(pINT col, pINT x1, pINT y1, pINT x2, pINT y2)
{  gpr_$window_t rect;
   gpr_$set_fill_value((long)col,&status);
   rect.window_base.x_coord = x1;
   rect.window_base.y_coord = y1;
   rect.window_size.x_size = x2-x1+1;
   rect.window_size.y_size = y2-y1+1;
   gpr_$rectangle(rect,&status);
   gpr_$set_fill_value((long)1,&status);
}

void draw_unit(float val, pINT x, pINT y, pINT ux, pINT uy)
/* x,y are taken to be the center of that unit's rectangle  */
/* val is truncated into 0..1 */
{ int b2=ux>>1, h2=uy>>1, at, negative=0;
  if (val>1.0)   val = 1.0;  else
  if (val<-1.0)  val = -1.0; else
  if (val<0.0)   { val *= -1.0; negative = 1; }
  at = DX*(int)(val * 15.0);
  bit_blt(at,0,ux+1,uy+1,x-b2,y-h2);
  rectangle(x-b2,y-h2,x-b2+ux,y-h2+uy);
   if(negative)
      { h2=(int)(uy/6); /* Halbe Strichdicke abhaengig von Unitgroesse */
        clear_box(0,x-b2,y-h2,x+b2,y+h2);
        rectangle(x-b2,y-h2,x+b2,y+h2); 
      }
}


 
void set_linestyle(pINT, pINT style, pINT width)
{ /* Color ignored on APOLLO */
   if (style) gpr_$set_linestyle(gpr_$dotted,2,&status);
         else gpr_$set_linestyle(gpr_$solid,1,&status);
   gpr_$set_draw_width((short int)width,&status);
}

void set_draw_op(pINT val)
{ gpr_$set_raster_op(hi_plane,(gpr_$raster_op_t)val,&status); }


void draw_line(pINT x, pINT y, pINT x1, pINT y1)
{ gpr_$move((unsigned short) x,(unsigned short) y,&status);
  gpr_$line((unsigned short) x1,(unsigned short) y1,&status);
}

void draw_circle(pINT x, pINT y, pINT rd)
{  gpr_$position_t center;
   center.x_coord = x;
   center.y_coord = y;
   gpr_$circle(center,(short int)rd,&status);
}

void draw_bar(pINT col, pINT x, pINT y, pINT x1, pINT y1)
{ register short i,j;
   for (i=x;i<=x1;i++)
      for (j=y;j<=y1;j++)
         { if (col>dither[i%4][j%4]) gpr_$set_draw_value((long)1,&status);
                                          else gpr_$set_draw_value(0,&status);
           gpr_$move(i,j,&status); /* Set/Reset Pixel */
           gpr_$line(i,j,&status);
         }
   gpr_$set_draw_value((long)1,&status);
}

void draw_rahmen(pINT l1, pINT ext)
{  layer L;
   char titel[30];
   int x,y;
   L=L_Tab[l1];
   x=L.x; y=L.y;
   strcpy(titel,L.titel);
   if (ext)
      {strcat(titel,"-Extern");
      x+=L.extsep_x;
      y+=L.extsep_y;}
   gprintf(x-L.ux/2,y-L.uy/2-14,COL_TEXT,titel);
}


void enable_events(void)
{ char c;
   event_type = gpr_$locator_update;
   gpr_$enable_input(event_type,key_set,&status);

  event_type = gpr_$keystroke;
   lib_$init_set(key_set,SETSIZE);
   for (c='a';c<='z';c++)
      lib_$add_to_set(key_set,SETSIZE,(short)c);
   for (c='A';c<='Z';c++)
      lib_$add_to_set(key_set,SETSIZE,(short)c);
   for (c='0';c<='9';c++)
      lib_$add_to_set(key_set,SETSIZE,(short)c);
   gpr_$enable_input(event_type,key_set,&status);

  event_type = gpr_$buttons;
  lib_$init_set(mouse_buttons,SETSIZE);
  lib_$add_to_set(mouse_buttons,SETSIZE,KBD_$M1D);
  lib_$add_to_set(mouse_buttons,SETSIZE,KBD_$M2D);
  lib_$add_to_set(mouse_buttons,SETSIZE,KBD_$M3D);
  gpr_$enable_input(event_type, mouse_buttons, &status);
}

char nnwait(void)
/* this version does not return until a mouse click occurs */
{ do { gpr_$event_wait(&event_type,&event_data,&position,&status);
     } while(event_type!=gpr_$buttons && event_type!=gpr_$keystroke);
  switch(event_type) 
   {
   case gpr_$buttons:
      mouse_x = position.x_coord;
      mouse_y = position.y_coord;
      switch (event_data) 
        {
        case KBD_$M1D: mouse_button = 1; break;
        case KBD_$M2D: mouse_button = 2; break;
        case KBD_$M3D: mouse_button = 3; break;
        default: mouse_button = 0;
        }
      return('@'); /* means a mouse event */
   case gpr_$keystroke:  return(event_data);
   default: return(' '); /* means no event */
   }
}

void init_colors(void)
{ int i,f;
   g_acquire();
   for (i=0;i<16;i++)
      {f=(i<15)?i:16;
      draw_bar(f,i*DX,0,i*DX+UWID,UWID);
      box(i*DX,0);}
   g_release();
}

/* Init graphcis, DX must be set to the greatest value in network
   (cause of unit-prototyps with grayvalues) */

void init(gpr_$display_mode_t mode)
{
static short int         unit = 1;
static short int         disp_len = sizeof(gpr_$disp_char_t);
       short int         disp_len_returned;
   
   gpr_$inq_disp_characteristics(mode, unit, disp_len,
                                 &display_characteristics,
                                 &disp_len_returned, &status);
   check("in init after inquiring");
  
   display_bitmap_size.x_size = display_characteristics.x_window_size;
   display_bitmap_size.y_size = display_characteristics.y_window_size;
   hi_plane                   = display_characteristics.n_planes - 1;
   
   gpr_$init(mode, unit, display_bitmap_size,
             hi_plane, &display_bitmap, &status);
   check("in init after initializing");
}



void init_graph(void)
{  init(gpr_$direct);
   gpr_$set_auto_refresh(true,&status);
   if( (display_bitmap_size.x_size < MIN_WINDOW_X_SIZE)
      ||(display_bitmap_size.y_size < MIN_WINDOW_Y_SIZE))
      { close_graph();
        printf("Look, son, your window is too small.\n");
        printf("It should be %d by %d at least, but it is only %d by %d.\n",
                MIN_WINDOW_X_SIZE,MIN_WINDOW_Y_SIZE,
                display_bitmap_size.x_size,display_bitmap_size.y_size);
        exit(1);
      }
   else
      {
      UWID=DX-2;
      bitmap_x_size = display_bitmap_size.x_size;
      button_x_pos = bitmap_x_size - BUTTON_WIDTH - 10;
      gpr_$load_font_file(pathname_of_font, pathname_length, &font_id, &status);
      check("loading a load_font_file");
      gpr_$set_text_font(font_id, &status);
      check("setting the text font");
      enable_events();
      check("setting the refresh-function");
      init_colors();
      gpr_$set_cursor_active(true,&status);
      } 
        if (GP_BORDER) gp_border();
}

void close_graph(void)
{   g_release();
                gpr_$terminate(false,&status);
}

void cursor_on(void)
{ gpr_$set_cursor_active(true,&status); }

void cursor_off(void)
{ gpr_$set_cursor_active(false,&status); }

/*===================================================================
   displays text in graphicsmode, format-arguments like printf()
   col<0 .. textbackground unchanged
 ====================================================================*/
int gprintf(pINT x, pINT y, int col, char *fmt, ...)
{
  va_list args;
  gpr_$string_t str;       /* Buffer to build string into   */
  /* gpr_$offset_t size; */
  int cnt;              /* Result of SPRINTF for return */
  va_start(args,fmt);
  cnt = vsprintf(str,fmt,args);  /* prints string to buffer */
  if (cnt)
     { g_acquire();
       if (col<0) 
          gpr_$set_text_background_value((long)(-1),&status); /* transparent */
       gpr_$move((unsigned short)x,(unsigned short)(y+9),&status); /* Fonthoehe=9 */
       gpr_$text(str,(unsigned short)strlen(str),&status);
       gpr_$set_text_background_value((long)(-2),&status); /* default */
       g_release();
      }
  va_end(args);
  return(cnt);          /* Return the conversion count  */
}

/*=========================================================================
    displays button for state stat=EIN/AUS/NEU
 =========================================================================*/
void show_button(pINT id, pINT stat)
{ pINT x,y;
   int i;
   char *text=buttons[id].text;
   char hotkey=buttons[id].key;
   pINT width=buttons[id].width;
   pINT height=buttons[id].height;
   gpr_$offset_t size,size1;
   x=buttons[id].x;
   y=buttons[id].y;
   g_acquire();
   cursor_off();
   switch (stat) {
      case NEU:
         rectangle(x,y,x+width,y+height);
         gprintf(x+3,y+6,-1,text);
         i = strchr(text,hotkey)-text;
         if (i>=0)
            {/* Hotkey im Buttontext unterstreichen */
            gpr_$inq_text_extent(text,(unsigned short)i,&size,&status);
            gpr_$inq_text_extent(text,(unsigned short)(i+1),&size1,&status);
            draw_line(x+3+size.x_size,y+17,x+3+size.x_size+5,y+17);
            }
         break;
      case AUS:
      bit_blt(x+3,y+3,width+4,height+4,x,y);
         break;
      case EIN:
         bit_blt(x-3,y-3,width+4,height+4,x,y);
         break;
   } /* end switch */
   if (stat==NEU) stat=AUS;
   show_shadow (x,y,width,height,stat);
   buttons[id].status=stat;
   cursor_on();
   g_release();
}
/*=========================================================================
    displays buttonshadow for state stat=EIN/AUS
 =========================================================================*/
void show_shadow(pINT x, pINT y, pINT width, pINT height, pINT stat)
{  pINT i;
   /* Top Shadow */
  if (stat==EIN)
      for(i=1; i<4; i++)
         {draw_line(x,y+i-1,x+i+width,y+i-1);
         draw_line(x+i-1,y+3,x+i-1,y+i+height);}
   if (stat==AUS)
      /* Bottom Shadow */
      for(i=1; i<4; i++)
         {draw_line(x+i,y+i+height,x+3+width,y+i+height);
         draw_line(x+i+width,y+i,x+i+width,y+height);
         }
}

pINT load_layout(char *fname) {}
BOOL nn_break(void) {return 0;};