/*************************************************************************
*  Program:       BACKPRP2.C                                             *
*  Description:   Backprop with BIAS - Example for VieNet2               *
*  Last modified: 15.02.94                                               *
*  Author:        LG                                                     *
**************************************************************************/

#include "neurodef.h"
#include "neurons.h"
#include "nngraph.h"
#ifdef PC
#include "biblio\nn_bp.h"
#else
#include "biblio/nn_bp.h"
#endif
#include <time.h>

#define SOURCENAME "BACKPROP2"

pINT  IN,
      BIAS1,
      HIDDEN,
			BIAS2,
			OUT;
      
/* Learning Steps */
long steps=1000;

/* Learning rate */
float rate  = 0.5;

#ifndef WIN
/* standard-functions, already defined in windows-menus */
void savew(void)
{ pprintf("saving \"%s\" ...\n",WeightFile);
  if (save_weights(WeightFile)) pprintf("done\n");
}
void loadw(void)
{ pprintf("loading \"%s\" ...\n",WeightFile);
	if (load_weights(WeightFile)) pprintf("done\n");
}
#endif

/* reset all activations to default values */
void reset_all(void)
{ reset_layers();
  draw_layers();
}

/* functions for choosing input-patterns */
void nextpat(void)
{ next_pattern(IN); next_pattern(OUT); draw_ext(IN); draw_ext(OUT); }
void prevpat(void)
{ prev_pattern(IN); prev_pattern(OUT); draw_ext(IN); draw_ext(OUT); }
void randpat(void)
{ nth_pat(OUT,rand_pat(IN)); draw_ext(IN); draw_ext(OUT); }

/* connect network layers, initializes weights */
void connect_all_new(void)
{ 
  nnconnect(    IN, HIDDEN, FULL,-0.6,0.6,1.0);
  nnconnect( BIAS1, HIDDEN, FULL,-0.6,0.6,1.0);
  nnconnect(HIDDEN, OUT,    FULL,-0.6,0.6,1.0);
  nnconnect( BIAS2, OUT,    FULL,-0.6,0.6,1.0);
}

/* full update through network */
void upd(void) 
{
	update(IN,IN,EXT,SINGLE);
  update(IN,HIDDEN,STANDARD,ALL);
	copy_layer(HIDDEN);
  update(HIDDEN,OUT,STANDARD,ALL);
  copy_layer(OUT);
}

/* update and show */
void upd1(void)
{	upd(); draw_layers(); }

/* select next pattern and update all */
void test(void)
{static pINT i=0;
 nth_pat(IN,i);
 nth_pat(OUT,i);
 upd1();
 i=(i<patanz(IN)-1)?++i:0;
}

/* calculates mean-square-error for all patterns */
void calc_error()
{ pINT i;
	float err,mse=0.0;
  for(i=0;i<patanz(OUT);i++)
		{nth_pat(IN,i);
		 nth_pat(OUT,i);
		 upd();
     err = L_Tab[OUT].p->pat[0] - X(OUT,0);
     mse+=err*err;
		}
	mse/=(REAL)patanz(OUT);
	pprintf("MSE: %.4f\n",mse);
}

/* learn n steps, break with key "ESC" */
void learn_n(void)
{time_t start,stop;
 long i;
 pprintf("Learning (break = 'ESC')\n");
 time(&start);
 for (i=0;i<steps;i++)
     { nth_pat(OUT,rand_pat(IN));
       upd();
       NN_BP_Learn(IN,HIDDEN,OUT,rate);
			 if (nn_break()) break;
		 }
 time(&stop);
 pprintf("done\n");
 if (start<stop) pprintf("Learning time for %li steps: %li s (%.2f Cps)\n",i,(stop-start),((float)(i*weights_used))/(float)(stop-start));
 calc_error();
}

void nn_init(int argc, char** argv)
{ GAIN = 6.0; /* parameter for activation-function */

	DX = 12;
	UWID = 10;

	IN	   = init_layer( 9,1,GRID,130,220,"In");
  BIAS1  = init_layer( 1,1,BIAS_LAYER,80,190,"Bias1");
	HIDDEN = init_layer(10,1,DELTA_LAYER, 120,160,"Hidden");
  BIAS2  = init_layer( 1,1,BIAS_LAYER,80,120,"Bias2");
	OUT	   = init_layer( 1,1,DELTA_LAYER,180,110,"Out");

	init_ext(IN, 0,40,"extern");
	init_ext(OUT,0,-55,"target");

#ifdef WIN
	load_layout(LayoutFile);
#endif

	connect_all_new();

#ifndef WIN
	make_button("QUIT",shutdown,'Q');
	make2button("loadw",loadw,'l',"savew",savew,'s');
	make_button("NEW WEIGHTS",connect_all_new,'W');
	make_button("RESET",reset_all,'R');
#endif
	make_button("Update",upd1,'U');
	make_button("Learn",learn_n,'L');
	make_button("Test",test,'T');
	make_button("Calc Error",calc_error,'E');
	make2button("- I/O",prevpat,'-',"+",nextpat,'+');
	make_button("rand I/O",randpat,'r');

	make_slider("Steps:",NULL,' ',&steps,100,100000L);
	make_slider_e("Rate:", NULL,' ',"%.2f",&rate, 0.0,1.0,0.01);
	make_slider_e("Gain:", NULL,' ',"%.2f",&GAIN, 0.0,10.0,0.1);

  load_data("input1.pat",IN);
	first_pattern(IN);
  load_data("output1.pat",OUT);
	first_pattern(OUT);
	
#if defined(sun4) || defined(WIN)
	init_xgraph(argc,argv);
#endif

	pprintf("This is a demo of VieNet2 %s\n\n",VERSION);
	pprintf("Backprop-Example 2\n");
	pprintf("==================\n\n");

	wait_for_action();
}

#ifndef WIN
void main(int argc, char** argv)
{ nn_setup(argc,argv);
	nn_init(argc,argv);}
#endif

