/*************************************************************************
*  Program:       DDEBSP.C                                               *
*  Description:   Exampleapplication for VieNet2 DDE Functions           *
*  Last modified: 21.02.94                                               *
*  Author:        LG                                                     *
**************************************************************************/

/*
	Dieses Beispiel ist im wesentlichen identisch mit HEBB.C,
  zeigt aber die notwendigen Erweiterungen fuer die Verwendung
  eigener Funktionen als DDE-Funktionen.
  Die notwendigen Erweiterungen sind:
  	upd_DDE(..)
    NN_DDE_RegisterFunction(..)
*/

/* Reihenfolge der Headerfiles muss eingehalten werden! */
#include "neurodef.h"
#include "neurons.h"
#include "nngraph.h"

#define SOURCENAME "DDEBSP" /* Aktueller Sourcename fuer Fehlermeldungen */

pINT A,B;        /* Layervariablen global definieren */

long steps = 50; /* Long-Variable fuer Slider-Einstellung der Lernschrittanzahl */

/* naechstes Pattern aus Patternliste auswaehlen */
void nextpat_A()
{ next_pattern(A); draw_ext(A); }

/* voriges Pattern aus Patternliste auswaehlen */
void prevpat_A()
{ prev_pattern(A); draw_ext(A); }

/* zufaelliges Pattern aus Patternliste auswaehlen */
void randpat_A()
{ rand_pat(A); draw_ext(A); }

/* Alle gewuenschten Verbindungen aufbauen und initialisieren,
   diese Funktion MUSS immer definiert werden! */
void connect_all_new()
{ nnconnect(A,B,FULL,-1.0,1.0,0.7); }

/* Standard-Update ausfuehren */
void upd()
{ update(A,A,EXT,SINGLE);   /* Kopiert externes Pattern nach X() im Layer A */
  update(A,B,STANDARD,ALL); /* Berechnet XNEW() im Layer B */
  copy_layer(B);            /* Kopiert alle XNEW(B,i) nach X(B,i) */
  border_x(B);	            /* X(B,i) begrenzen auf [XMIN,XMAX] */
  draw_layers();            /* alle Layer neu anzeigen */
}

/* Update-Funktion fuer DDE */
void upd_DDE()
{ update(A,B,STANDARD,ALL);
  copy_layer(B);;
	draw_layers();
}

/* einfaches Hebb-Lernen */
void learn()
{ pINT i,j;
	pprintf("Learning\n"); /* Ausgabe eines Textes im Messagewindow, nicht printf(..)! */
	for (L_ALL(A,i))
		for (L_ALL(B,j))
      if ((X(A,i) > 0.5) && (X(B,j) > 0.5))
        W(A,B,i,j) += 0.02;
      else
        W(A,B,i,j) -= 0.02;
	pprintf("done\n");
}


/* NNInit() initialisiert alle Layer, Verbindungen, Buttons und sonstiges */
void nn_init(int argc, char** argv)
{
	A = init_layer(10,1,GRID,100,200,"Layer_A"); /* Layer erzeugen - noch keine Anzeige */
	B = init_layer(20,3,GRID,40,80,"Layer_B");

#ifdef WIN
	/* Layout laden (optional), ueberschreibt die Daten aus init_layer() und init_ext() */
	load_layout(LayoutFile);
#endif

	connect_all_new(); /* Alle Verbindungen aufbauen und gewichten */

	/* Alle gewuenschten Buttons definieren: */
	make_button("Update",upd,'U');
	make_button("Learn",learn,'L');
	make2button("- IN",prevpat_A,'-',"+ IN",nextpat_A,'+');
	make_button("randIN",randpat_A,'r');

	/* mehrere Bilder laut Schedule-File laden  */
	load_schedule("ddebsp.scd",A); /* new.scd muss existieren! */
	first_pattern(A);

#if defined(WIN) || defined(sun4)
	init_xgraph(argc,argv);  /* X/Windows-Graphik einschalten */
#endif

#ifdef WIN
  /* DDE */
	NN_DDE_RegisterFunction(upd_DDE,"UPDATE");
#endif

	pprintf("This is %s  --  hello out there\n",ProgName);
	pprintf("VieNet2 is up to version %s!\n\n",VERSION);

 	wait_for_action();     /* Graphik-Eingabeschleife */
}

/*  main mit Argumenten */
#ifndef WIN
void main(int argc, char** argv)
{	nn_setup(argc,argv);}
	nn_init(argc,argv);}
#endif



