/*************************************************************************
*  Program:       Diagram.c                                              *
*  Description:   Example for VieNet2 diagram functions                  *
*  Last modified: 25.02.94                                               *
*  Author:        LG                                                     *
*  Remarks:       for MS Windows only!                                   *
*                                                                        *
*                 This 3-Layer-Backprop Network loads 100 training- and  *
*                 100 test-datasets sequentially into the external       *
*                 pattern chain.                                         *
*                 You may choose how many data is used for training by   *
*                 varying the last slider. The remaining part of data    *
*                 is used for testing.                                   *
*                                                                        *
*                 The diagramwindow shows curves for training and        *
*                 testset after learning is completed.                   *
**************************************************************************/

#include <math.h>
#include <stdlib.h>

#include "neurodef.h"
#include "neurons.h" 
#include "nngraph.h"
#include "nndiagr.h"        /* diagram module */
#ifdef PC
#include "biblio\nn_bp.h"
#else
#include "biblio/nn_bp.h"
#endif

#define SOURCENAME "DIAGRAM"

// Layergroessen
#define IN_SIZE 60
#define IN_LINES 4
#define HIDDEN_SIZE 5
#define OUT_SIZE 2

#define DATA_TRN_FILE "diagram.trn"
#define DATA_TST_FILE "diagram.tst"

int TRN_SIZE=100;
int TST_SIZE=100;

#define TEST_INT  200
#define MAX_LEARN 30000L
#define ERR_SIZE  MAX_LEARN/TEST_INT

pINT 	IN, HIDDEN, OUT, BIAS1, BIAS2;

REAL eta = 0.1;          /* Lernrate */
long steps = 10000;      /* Lernschritte */
long patnr = 0;          /* aktuelle Patternnummer */

REAL trainerr[ERR_SIZE], testerr[ERR_SIZE];  /* Arrays fuer Lernkurven */

int Diagram1,Curve1,Curve2;   /* Diagramm- und Lernkurvenhandles */

/* baut Verbindungen auf, bzw. setzt Gewichte neu */
void connect_all_new(void)
{	nnconnect(IN,HIDDEN,FULL,-0.5,0.5,1.0);
	nnconnect(BIAS1,HIDDEN,FULL,-0.5, 0.5, 1.0);
	nnconnect(HIDDEN,OUT,FULL,-0.5,0.5,1.0);
	nnconnect(BIAS2,OUT,FULL,-0.5, 0.5, 1.0);
}

void upd(void)
{ update(IN,IN,EXT,SINGLE);
	update(IN,HIDDEN,STANDARD,ALL); 
  copy_layer(HIDDEN);
  update(HIDDEN,OUT,STANDARD,ALL);
	copy_layer(OUT);                 
}

void upd_draw(void)
{ upd(); draw_layers();}


/* waehlt zufaelliges Muster aus Trainigsset */
pINT rand_train_pat(pINT l)
{ return nth_pat(l,(pINT)(random_nr()*(REAL)TRN_SIZE)); }

/* testet, ob die gewuenschte Klasse die Klasse 1 ist */
BOOL target1(pINT out)
{ if (L_Tab[out].p->pat[0]==1.0) return 1; else return 0; }

/* testet, ob das Netzwerk den Input Klasse 1 zuordnet */
BOOL klasse1(pINT out)
{ if (X(out,0)>X(out,1)) return 1; else return 0; }

/* berechnet den quadratischen Fehler des Outputlayers */
float allerr(pINT out)
{ pINT i;
	float err=0.0;
	for (L_ALL(out,i))
		err+=sqr(L_Tab[out].p->pat[i] - X(out,i));
	return err;
}

/* Teste alle Muster im Trainings- bzw. Testset auf richtigen Output*/
/* Returniere den durchschnittlichen quadratischen Fehler */
float test_pats (BOOL set, pINT in, pINT out, pINT printflag)
{int i,r1,r2,f12,f21,size=(set==0)?TRN_SIZE:TST_SIZE;
 float err;

 r1=r2=f12=f21=0;
 err=0.0;

 if (set==0)
  {first_pattern(in); first_pattern(out);}
 else 
  {/* first pattern of Testset */
  nth_pat(in,TRN_SIZE);nth_pat(out,TRN_SIZE);}

 for (i=0;i<size;i++)
     {
		 upd();
     if (target1(out))
       if (klasse1(out)) r1++;
			 else f12++;
     else
       if (klasse1(out)) f21++;
     	 else r2++;
     err+=allerr(out);
		 next_pattern(in);
     next_pattern(out);
		 }
	if (printflag)
     {
     if (set==0)
          pprintf("\nTeste alle %i Trainingsmuster:\n",size);
     else
          pprintf("\nTeste alle %i Testmuster:\n",size);
     pprintf("Klasse 1 richtig: %i, falsch: %i\nKlasse 2 richtig: %i, falsch: %i\nDurchschnittlicher quadratischer Fehler: %f \n",r1,f12,r2,f21,err/(float)size);
     }
  return err/(float)size;
}


/* n Lernzyklen durch das Netzwerk
   zeichnet Lernkurven */
void learn_n(void)
{time_t start,stop;
 long i,nr_weights;
 pINT j=0;

 pprintf("\nLearning (break = 'ESC') ...\n");
 time(&start);             	/* zur Rechenzeitmessung */

 for (i=0;i<steps;i++)      /* Schleife fuer Lernzyklen */
     {
		 if (nn_break()) break; /* erlaubt Unterbrechung */
     nth_pat(OUT, rand_train_pat(IN)); /* waehle zufaelliges Musterpaar */

     NNSetTarget(OUT,NULL); /* setze Target mit externem Pattern */

     upd();                 /* Update durchs Netz */

     NN_BP_Learn(IN,HIDDEN,OUT,eta);

		 if (!(i%TEST_INT) && i)
        { /* Punkte fuer Fehlerkurven berechnen */
				trainerr[j]= test_pats(0,IN,OUT,0);
				testerr[j] = test_pats(1,IN,OUT,0);
				j++;
        }
     if (!(i%1000) && i)
       pprintf("%lu Steps learned ... \n",i);
		 }

 time(&stop);
 pprintf("done\n");

 /* Anzahl der Gewichte im Netzwerk, Bias-Verbindungen nicht mitgerechnet
  (globale Variable weights_used enthaelt Anzahl ALLER Gewichte im Netz) */
 nr_weights=L_Tab[HIDDEN].size*(L_Tab[IN].size+L_Tab[OUT].size);

 if (start<stop)
    pprintf("\nLearning time for %li steps: %li s (%.0f Cps)\n",i,(stop-start),
			((float)(i*nr_weights))/(float)(stop-start));

 /* Kurven erst jetzt zeichnen - ist schneller als punktweise Ausgabe */
 dia_set_curve(Curve1,ERR_SIZE,trainerr,NULL);
 dia_set_curve(Curve2,ERR_SIZE,testerr,NULL);
}

/* Teste alle Muster des Trainigsset, mit Ausgabe des Resultats */
void test_all_train() { test_pats(0,IN,OUT,1); }

/* Teste alle Muster des Testsets, mit Ausgabe des Resultats */
void test_all_test()  { test_pats(1,IN,OUT,1); }

/* ladet alle Trainingsdaten vom vordefinierten File */
void train_load()     { load_data2(DATA_TRN_FILE,IN,OUT); }

/* ladet alle Testdaten vom vordefinierten File hinter die Trainingsdaten */
void test_load()      { load_data2(DATA_TST_FILE,IN,OUT); }

/* Funktion fuer Slider, Aenderung von TRN_SIZE bewirkt eine
   Verschiebung der zum Training und Testen verwendeten Daten */
void change_train()
{ TST_SIZE=200-TRN_SIZE;
  pprintf("Trainingssetgre %i, Testsetgre %i\n", TRN_SIZE, TST_SIZE);
}

void set_pattern(void)
{ nth_pat(IN,patnr); nth_pat(OUT,patnr);
  update(IN,IN,EXT,SINGLE);
  upd_draw();
}

/* Init network */
void nn_init(int argc, char** argv)
{
	IN     = init_layer(IN_SIZE/IN_LINES,IN_LINES,GRID,100,200,"Muster");
	HIDDEN = init_layer(HIDDEN_SIZE,1,DELTA_LAYER,138,160,"Hidden");   
	OUT    = init_layer(OUT_SIZE,   1,DELTA_LAYER,158,120,"Klasse");   
	BIAS1  = init_layer(1,1,BIAS_LAYER,40, 190, "Bias1");
	BIAS2  = init_layer(1,1,BIAS_LAYER,40, 150, "Bias2");

	init_ext(IN, 0,100," ");   /* rel. Koordinaten fuer Externlayer (Input) */
	init_ext(OUT,0,-60," ");   /* rel. Koordinaten fuer Externlayer (Output) */

	connect_all_new();         /* Verbindungen aufbauen */

  train_load();
  test_load();
  
	make_button("Ende",shutdown,'Q');
	make_button("Update",upd_draw,'U');
	make_button("Lernen",learn_n,'L');
	make2button("Check Trn",test_all_train,'r',"Check Tst",test_all_test,'s');

	make_slider("Pattern:",set_pattern,'P',&patnr,0,L_Tab[IN].patanz-1);

	make_slider_e("Lernen:",NULL,'L',"%li",&steps,500L,MAX_LEARN,500L);
	make_slider_e("Rate:",NULL,'R',"%.2f",&eta,0.0,0.9,0.05);
	make_slider_e("Training Set:",change_train,'T',"%i",&TRN_SIZE,0,TRN_SIZE+TST_SIZE-1,1);

  /* zuerst X/Windows-Graphik einschalten */
  init_xgraph(argc,argv);
  
  /* dann Diagram und Lernkurven definieren! */
  Diagram1 = init_diagram (0.0, (float)(MAX_LEARN), 1000.0, 0.0, "Lernschritte",
				 0.0, 1.0, 0.1, 0.0, "Quadratischer Fehler",
				 X_AUTOSCALE | Y_AUTOSCALE,
				 "Lernkurven", "Trainings- und Testset", "Diagramm");
  
	Curve1 = init_curve ("Training Set", 0, LINE_SOLID, (float)TEST_INT, Diagram1);
	Curve2 = init_curve ("Test Set", LINE_RED, LINE_SOLID, (float)TEST_INT, Diagram1);

  /* Windows-Positionen und Sichtbarkeit definieren */
	NNSetWindowPos(INFO_WINDOW,NN_HIDEWINDOW);
	NNSetWindowPos(MSG_WINDOW,NN_SETWINDOWPOS,0,400,400,200);
	NNSetWindowPos(MAIN_WINDOW,NN_SETWINDOWPOS,250,0,540,480);
	NNSetWindowPos(USER_WINDOW,NN_SETWINDOWPOS,0,0,400,400);
}


#ifndef WIN
// fuer Sun-Version
void main(int argc, char** argv)
{ nn_setup(argc,argv);
	nn_init(argc,argv);}
#endif

