/*************************************************************************
*  Program:       NNDIAGR.CPP                                            *
*  Description:   Diagram-Windows for VieNet2                            *
*  Version:       2.0                                                    *
*  Last modified: 08.03.94                                               *
*  Author:        Wolfgang Reh, revised by Guenter Linhart               *
*  Copyright (C) Austrian Research Institute for Artificial Intelligence *
**************************************************************************/
#define MODULE
#include "neurodef.h"
#include "neurons.h" 
#include "nngraph.h"
#include "nndiagr.h"
#define SOURCENAME "NNDIAGR"

class TMarker
 {
  public:
   float    Position;
   int      Orientation;
   int      LineType;
   COLORREF Color;
   BOOL     DisplayValue;

   TMarker (float MyPosition, int MyOrientation, int MyLineType, COLORREF MyColor,
            BOOL MyValue)
    {
     Position     = MyPosition;
     Orientation  = MyOrientation;
     LineType     = MyLineType;
     Color        = MyColor;
     DisplayValue = MyValue;
    };
 };

// Turn off warning 'Function hides virtual function'
#pragma warn -hid

class TMarkerList : public TNSCollection
 {
  public:
    TMarkerList ();
    virtual int insert (TMarker *Marker);
    virtual void free (TMarker *Marker);
    virtual TMarker *at (int Index);
    virtual void *keyOf (TMarker *Item);
    virtual int indexOf (TMarker *Item);
    virtual TMarker *firstThat (Boolean (*TestFunc) (void *, void*), void *Data);
 };

#pragma warn +hid

class TGraphWindow : public TWindow
 {
  friend TCurve;

  private:
    int           MyHandle;
    int           DiagramType;
    TAxisRange    AxisRange;
    TDiagramTitle DiagramTitle;
    TBitmapDim    BitmapDimension;
    TBitmapDim    FixedDimension;
    TScale        Scale;
    TCurve        *Curves;
    TMarkerList   Markers;
    TFloatPoint   ZoomFactor;
    HBITMAP       MemorySheet;
    HFONT         TitleFont;
    HFONT         LegendFontNorm;
    HFONT         LegendFontTilt;
    BOOL          SelectActive;
    BOOL          ZoomActive;
    BOOL          FitToWindow;
    RECT          SelectArea;

    virtual void   CreateMemorySheet (TFloatPoint ZoomIn, BOOL ForceNew);
    virtual void   DrawAxis (BOOL Redraw);
    virtual void   DrawCurves ();
    virtual void   CreateMyFonts (char * FontName);
    virtual void   MarkCurve (int MenuID);
    virtual HANDLE RetrieveBitmap ();

  public:
    TGraphWindow (LPSTR MyTitle, int GraphHandle, TAxisRange AxisValues, int Type,
                  TDiagramTitle Title, POINT UpperLeft, POINT WindowSize, char *FontName);
    virtual ~TGraphWindow ();
    virtual void SetupWindow ();
    virtual LPSTR GetClassName ();
    virtual void GetWindowClass (WNDCLASS& TheWndClass);
    virtual void Paint (HDC PaintDC, PAINTSTRUCT&);
    virtual int  NewCurve (char *Name, COLORREF Color, int LineStyle, float Increment);
    virtual int  AddCoordinate (float XValue, float YValue, BOOL XValid, int CurveHandle,
                                BOOL Redraw);
    virtual int  ClearCurve (int CurveHandle);
    virtual int  SetCurveStart (int CurveHandle, float NewXStart);
    virtual int  SetMarker (float MarkerPosition, int MarkerOrientation, int MarkerLineType,
                            COLORREF MarkerColor, BOOL DisplayValue);
    virtual void CloseGraph (TMessage& TheMessage) = [CM_FIRST + CM_CLOSEGRAPH];
    virtual void CopyToClipboard () = [CM_FIRST + CM_COPYGRAPH];
    virtual void Zoom () = [CM_FIRST + CM_ZOOMGRAPH];
    virtual void ZoomX () = [CM_FIRST + CM_ZOOMGRAPHX];
    virtual void ZoomY () = [CM_FIRST + CM_ZOOMGRAPHY];
    virtual void DisplayOptions () = [CM_FIRST + CM_GRAPH_DISPLAY];
    virtual void Load () = [CM_FIRST + CM_LOADGRAPH];
    virtual void Save () = [CM_FIRST + CM_SAVEGRAPH];
    virtual void CalcMean () = [CM_FIRST + CM_GRAPHMEAN];
    virtual void CalcSDev () = [CM_FIRST + CM_GRAPHSDEV];
    virtual void CalcMSE () = [CM_FIRST + CM_GRAPHMSE];
    virtual void SelectAll () = [CM_FIRST + CM_SELECTALL];
    virtual void UnselectAll () = [CM_FIRST + CM_UNSELECTALL];
    virtual void Select0 () = [CM_FIRST + CM_CURVE_0];
    virtual void Select1 () = [CM_FIRST + CM_CURVE_1];
    virtual void Select2 () = [CM_FIRST + CM_CURVE_2];
    virtual void Select3 () = [CM_FIRST + CM_CURVE_3];
    virtual void Select4 () = [CM_FIRST + CM_CURVE_4];
    virtual void Select5 () = [CM_FIRST + CM_CURVE_5];
    virtual void Select6 () = [CM_FIRST + CM_CURVE_6];
    virtual void Select7 () = [CM_FIRST + CM_CURVE_7];
    virtual void Select8 () = [CM_FIRST + CM_CURVE_8];
    virtual void Select9 () = [CM_FIRST + CM_CURVE_9];
    virtual void WMSize (TMessage& TheMessage) = [WM_FIRST + WM_SIZE];
    virtual void WMLButtonDown (TMessage& TheMessage) = [WM_FIRST + WM_LBUTTONDOWN];
    virtual void ResizeSelect (TMessage& TheMessage) = [WM_FIRST + WM_MOUSEMOVE];
    virtual void EndSelect (TMessage& TheMessage) = [WM_FIRST + WM_LBUTTONUP];
    virtual void CancelSelect () = [WM_FIRST + WM_RBUTTONUP];
 };

typedef TGraphWindow *PTGraphWindow;

class TWindowRef
 {
  public:
    int           Handle;
    PTGraphWindow Window;

    TWindowRef ()
     {
      Handle = 0;
      Window = NULL;
     };
 };

class TPoint
 {
  public:
   float x;
   float y;

   TPoint (float NewX, float NewY)
    {
     x = NewX;
     y = NewY;
    };
 };

 // Turn off warning 'Function hides virtual function'
#pragma warn -hid

class TWindowList : public TNSSortedCollection
 {
  private:
    virtual int compare (void *First, void *Second);

  public:
    TWindowList ();
    virtual int insert (TWindowRef *WindowRef);
    virtual void free (TWindowRef *WindowRef);
    virtual TWindowRef *at (int Index);
    virtual void *keyOf (TWindowRef *Item);
    virtual int indexOf (TWindowRef *Item);
    virtual TWindowRef *firstThat (Boolean (*TestFunc) (void *, void*),
				   void *Data);
 };

class TPointList : public TNSSortedCollection
 {
  private:
    virtual int compare (void *First, void *Second);

  public:
    TPointList ();
    virtual int insert (TPoint *Point);
    virtual void free (TPoint *Point);
    virtual TPoint *at (int Index);
    virtual void *keyOf (TPoint *Item);
    virtual int indexOf (TPoint *Item);
    virtual TPoint *firstThat (Boolean (*TestFunc) (void *, void*),
			       void *Data);
 };

#pragma warn +hid

class TCurve
 {
  private:
   int           LineType;
   float         DefaultXStart;
   float         XIncrement;
   COLORREF      Color;
   TPointList    Points;
   TCurveBounds  Bounds;
   PTGraphWindow ChartWindow;

  public:
   int    Handle;
   char   Name [MAX_CURV_NAME + 1];
   BOOL   Mark;
   TCurve *Next;
   TCurve *Previous;

   TCurve (int MyHandle, char *MyName, COLORREF MyColor, int MyLineType,
           float MyIncrement, TCurve *Last, PTGraphWindow Chart);
   ~TCurve ();
   virtual void   AddPoint (float Y);
   virtual void   AddCoordinate (float X, float Y);
   virtual void   ClearPoints ();
   virtual void   SetXStart (float NewXStart);
   virtual void   Draw (HDC ChartDC);
   virtual void   Legend (HDC ChartDC, POINT& Offset, HFONT Font);
   virtual void   WritePoints (ofstream *OutputFile);
   virtual HANDLE RetrieveYValues ();
   virtual float  LowerX ();
   virtual float  UpperX ();
   virtual float  LowerY ();
   virtual float  UpperY ();
 };

class TSaveDlg : public TDialog
 {
  private:
   char *ChosenDirectory;
   int  *ChosenType;
   BOOL *SelectedCurves;

  public:
   TSaveDlg (char *SelectedDirectory, int *Type, BOOL *Selected, PTWindowsObject MyParent);
   virtual void SetupWindow ();
   virtual void ChangeDirectory (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_DIRLIST];
   virtual void TypeChange (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_TYPELIST];
   virtual void SelectedOnly (TMessage& TheMessage) = [ID_FIRST + ID_SELECTEDONLY];
   virtual void Ok (TMessage& TheMessage) = [ID_FIRST + IDOK];
 };

class TStyleComboBox : public TComboBox
 {
  public:
   TStyleComboBox (int ResourceID, PTWindowsObject MyParent);
   virtual void ODADrawEntire (DRAWITEMSTRUCT _FAR &DrawInfo);
   virtual void ODAFocus (DRAWITEMSTRUCT _FAR &DrawInfo);
   virtual void ODASelect (DRAWITEMSTRUCT _FAR &DrawInfo);
 };

class TLoadDlg : public TDialog
 {
  private:
   char           *NewName;
   char           *FileName1;
   char           *FileName2;
   float          *XStart;
   float          *XStep;
   int            *Type;
   int            *LineStyle;
   COLORREF       *NewColor;
   TStyleComboBox *ColorBox;
   TStyleComboBox *LineBox;
   TStyleComboBox *DotBox;

   virtual void RadioButtonChange ();

  public:
   TLoadDlg (char *Name, COLORREF *Color, int *Style, char *File1,
	     char *File2, float *XBegin, float *XIncrement, int *FileType,
	     PTWindowsObject MyParent);
   ~TLoadDlg ();
   virtual void SetupWindow ();
   virtual void Browse1 () = [ID_FIRST + ID_GRAPH_BROWSE1];
   virtual void Browse2 () = [ID_FIRST + ID_GRAPH_BROWSE2];
   virtual void CombinedFile (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_COMBINED];
   virtual void File1XValues (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_XVAL1];
   virtual void File1YValues (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_YVAL1];
   virtual void File2XValues (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_XVAL2];
   virtual void File2YValues (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_YVAL2];
   virtual void NoFile2 (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_NOXVAL];
   virtual void ColorChanged (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_CURVECOLOR];
   virtual void TypeChanged (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_CURVESTYLE];
   virtual void DotChanged (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_DOTSTYLE];
   virtual void Ok (TMessage& TheMessage) = [ID_FIRST + IDOK];
 };

class TDisplayDlg : public TDialog
 {
  private:
   TBitmapDim *NewDimension;
   TAxisRange *NewRange;
   int        *ScaleType;
   BOOL       *BitmapSizeable;
   BOOL       XLogScaling;
   BOOL       YLogScaling;

  public:
   TDisplayDlg (BOOL *Fit, TBitmapDim *Dimension, int *Type, BOOL XScaleToLog,
                BOOL YScaleToLog, TAxisRange *Range, PTWindowsObject MyParent);
   virtual void SetupWindow ();
   virtual void FitInWindow (TMessage& TheMessage) = [ID_FIRST + ID_FIT_TO_WINDOW];
   virtual void AutoX (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_XAUTO];
   virtual void AutoY (TMessage& TheMessage) = [ID_FIRST + ID_GRAPH_YAUTO];
   virtual void Ok (TMessage& TheMessage) = [ID_FIRST + IDOK];
 };

class TResultBox : public TDialog
 {
  private:
   char *ResultTitle;
   char *ResultText;
   char *BoxCaption;

  public:
   TResultBox (char *Title, char *Result, char *Caption,
	       PTWindowsObject MyParent);
   virtual void SetupWindow ();
 };

//*************************************************************************
//
//  Global variables
//
//*************************************************************************

extern TApplication *PNeuroApp;

TWindowList GraphWindowList;

//*************************************************************************
//
//  Function to capture errors in 'log10'-calls
//
//*************************************************************************

int matherr (struct exception *Exception)
{
 switch (Exception->type)
  {
   case DOMAIN:
   case SING:
     if (stricmp (Exception->name, "log10") == 0)
      {
       Exception->retval = -MAXFLOAT;
       return (1);
      }
     else
       return (0);

   default:
     return (0);
  }
}

//*************************************************************************
//
//  Utilityfunctions for 'firstThat'- & 'forEach'-calls
//
//*************************************************************************

Boolean FindGraphWindow (void *Item, void *GraphHandle)
{
 TWindowRef *WindowRef = (TWindowRef *) Item;
 int        *SearchFor = (int *) GraphHandle;

 return ((Boolean) (WindowRef->Handle == *SearchFor));
}

Boolean FindCoordinate (void *Item, void *XValue)
{
 TPoint *Point     = (TPoint *) Item;
 float  *SearchFor = (float *) XValue;

 return ((Boolean) (Point->x == *SearchFor));
}

Boolean FindMarker (void *Item, void *Position)
{
 TMarker       *Marker    = (TMarker *) Item;
 TSearchMarker *SearchFor = (TSearchMarker *) Position;

 return ((Boolean) ((Marker->Position    == SearchFor->Position) &&
                    (Marker->Orientation == SearchFor->Orientation)));
}

void DrawConnection (void *Item, void *Data)
{
 TPoint    *Point    = (TPoint *) Item;
 TDrawData *DrawData = (TDrawData *) Data;

 double XValue = ((DrawData->ScaleType & X_LOGARITHMIC) != 0) ? log10 (Point->x) : Point->x;
 double YValue = ((DrawData->ScaleType & Y_LOGARITHMIC) != 0) ? log10 (Point->y) : Point->y;
 double Infinite = -MAXFLOAT;

 int XEndPoint = (XValue > Infinite) ? (XValue * DrawData->Scaling.PixelsPerXUnit +
		                        DrawData->Scaling.Origin.x) : -MAXINT;
 int YEndPoint = (YValue > Infinite) ? (-YValue * DrawData->Scaling.PixelsPerYUnit +
		                        DrawData->Scaling.Origin.y) : MAXINT;

 if (DrawData->DotDC != NULL)
   BitBlt (DrawData->DrawDC, XEndPoint - 3, YEndPoint - 3, 7, 7, DrawData->DotDC, 0, 0,
	   SRCCOPY);

 if (DrawData->DrawLine)
   LineTo (DrawData->DrawDC, XEndPoint, YEndPoint);
}

void PrintPoint (void *Item, void *Data)
{
 TPoint   *Point      = (TPoint *) Item;
 ofstream *OutputFile = (ofstream *) Data;
 char     OutputLine [80];

 sprintf (OutputLine, "%f\t%f\n", Point->x, Point->y);
 *OutputFile << OutputLine; 
}

void DrawMarker (void *Item, void *Data)
{
 TMarker     *Marker   = (TMarker *) Item;
 TMarkerData *DrawData = (TMarkerData *) Data;
 int         XStartPoint, XEndPoint, YStartPoint, YEndPoint;
 double      PosValue;
 double      Infinite = -MAXFLOAT;
 char        PosString [MAX_VALUE_LEN + 1];
 HFONT       PreviousFont;
 DWORD       TextExtent;

 HPEN MarkerPen   = CreatePen (Marker->LineType, 1, Marker->Color);
 HPEN PreviousPen = SelectObject (DrawData->DrawDC, MarkerPen);

 if (Marker->Orientation == MARKER_HORIZONTAL)
  {
   XStartPoint = DrawData->TicSpace;
   XEndPoint   = DrawData->DrawDimension.Width - DrawData->LegendSpace;
   PosValue    = ((DrawData->ScaleType & Y_LOGARITHMIC) != 0) ? log10 (Marker->Position) :
                                                                Marker->Position;
   YStartPoint = (PosValue > Infinite) ? (-PosValue * DrawData->Scaling.PixelsPerYUnit +
                                          DrawData->Scaling.Origin.y) : MAXINT;
   YEndPoint   = YStartPoint;
  }
 else
  {
   PosValue    = ((DrawData->ScaleType & X_LOGARITHMIC) != 0) ? log10 (Marker->Position) :
                                                                Marker->Position;
   XStartPoint = (PosValue > Infinite) ? (PosValue * DrawData->Scaling.PixelsPerXUnit +
		                          DrawData->Scaling.Origin.x) : -MAXINT;
   XEndPoint   = XStartPoint;
   YStartPoint = DrawData->HeaderSpace;
   YEndPoint   = DrawData->DrawDimension.Height - DrawData->TicSpace;
  }

 MoveTo (DrawData->DrawDC, XStartPoint, YStartPoint);
 LineTo (DrawData->DrawDC, XEndPoint, YEndPoint);

 if (Marker->DisplayValue)
  {
   COLORREF PreviousColor = SetTextColor (DrawData->DrawDC, Marker->Color);
   HBRUSH   PreviousBrush = SelectObject (DrawData->DrawDC, GetStockObject (WHITE_BRUSH));
   if (Marker->Position == (int) Marker->Position)
     sprintf (PosString, "%i", (int) Marker->Position);
   else
     sprintf (PosString, "%.2f", Marker->Position);

   if (Marker->Orientation == MARKER_HORIZONTAL)
    {
     PreviousFont = SelectObject (DrawData->DrawDC, DrawData->HorizontalFont);
     TextExtent = GetTextExtent (DrawData->DrawDC, PosString, strlen (PosString));
     Rectangle (DrawData->DrawDC, XStartPoint + 48,
                                  YStartPoint - HIWORD (TextExtent) / 2 - 2,
	                          XStartPoint + 48 + LOWORD (TextExtent) + 4,
	                          YStartPoint + HIWORD (TextExtent) / 2 + 2);
     TextOut (DrawData->DrawDC, XStartPoint + 50, YStartPoint - HIWORD (TextExtent) / 2,
              PosString, strlen (PosString));
    }
   else
    {
     PreviousFont = SelectObject (DrawData->DrawDC, DrawData->VerticalFont);
     TextExtent = GetTextExtent (DrawData->DrawDC, PosString, strlen (PosString));
     Rectangle (DrawData->DrawDC, XStartPoint - HIWORD (TextExtent) / 2 - 2,
                                  YEndPoint - 48,
	                          XStartPoint + HIWORD (TextExtent) / 2 + 2,
				  YEndPoint - 48 - LOWORD (TextExtent) - 4);
     TextOut (DrawData->DrawDC, XStartPoint - HIWORD (TextExtent) / 2, YEndPoint - 50,
              PosString, strlen (PosString));
    }

   SetTextColor (DrawData->DrawDC, PreviousColor);
   SelectObject (DrawData->DrawDC, PreviousBrush);
   SelectObject (DrawData->DrawDC, PreviousFont);
  }

 SelectObject (DrawData->DrawDC, PreviousPen);
 DeleteObject (MarkerPen);
}

//*************************************************************************
//
//  External access-functions (called from the main program)
//
//*************************************************************************

int init_diagram (float XStart, float XEnd, float XStep, float XOffset, char *XName,
		  float YStart, float YEnd, float YStep, float YOffset, char *YName,
		  int Type, char *Header1, char *Header2, char *Title,
		  int UpperLeftX, int UpperLeftY, int WindowWidth, int WindowHeight,
		  char *Font)
{
 TWindowRef    *NewWindow;
 TAxisRange    AxisValues;
 TDiagramTitle Header;
 char          WindowTitle [MAX_WINDOW_TITLE + 8];
 POINT         WindowPosition, WindowSize;

 if (!NNGraphInit)
    nn_fatalerror(SOURCENAME, "init_xgraph(..) must be called before init_diagr(..) in your program!\n");

 // Check if parameters are within allowed bounds
 if ((XStart == XEnd) || (YStart == YEnd) ||
     (XStep == 0.0) || (YStep == 0.0) ||
     (strlen (XName) > MAX_AXIS_NAME) ||
     (strlen (YName) > MAX_AXIS_NAME) ||
     (strlen (Header1) > MAX_DIAG_NAME) ||
     (strlen (Header2) > MAX_DIAG_NAME) ||
     ((Title != NULL) && (strlen (Title) > MAX_WINDOW_TITLE)))
   return (GRAPHERR_PARAMINVAL);

 strncpy (AxisValues.X.Title, XName, sizeof (AxisValues.X.Title));
 AxisValues.X.Start  = XStart;
 AxisValues.X.End    = XEnd;
 AxisValues.X.Step   = XStep;
 AxisValues.X.Offset = XOffset;
 strncpy (AxisValues.Y.Title, YName, sizeof (AxisValues.Y.Title));
 AxisValues.Y.Start  = YStart;
 AxisValues.Y.End    = YEnd;
 AxisValues.Y.Step   = YStep;
 AxisValues.Y.Offset = YOffset;

 strncpy (Header.Title1, Header1, sizeof (Header.Title1));
 strncpy (Header.Title2, Header2, sizeof (Header.Title2));

 WindowPosition.x = (UpperLeftX   >= 0) ? UpperLeftX   : -1;
 WindowPosition.y = (UpperLeftY   >= 0) ? UpperLeftY   : -1;
 WindowSize.x     = (WindowWidth  >  0) ? WindowWidth  : -1;
 WindowSize.y     = (WindowHeight >  0) ? WindowHeight : -1;

 if ((NewWindow = new TWindowRef ()) == NULL)
  {
   TErrorMessage BoxMessage, BoxTitle;
   LoadString (PNeuroApp->hInstance, STR_BOX_TITLE, BoxTitle,
               sizeof (BoxTitle));
   LoadString (PNeuroApp->hInstance, ERR_NO_WINDOW, BoxMessage,
	       sizeof (BoxMessage));
   MessageBeep (-1);
   BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONSTOP | MB_OK);
   return (GRAPHERR_NOMEM);
  }
 else
  {
   if (GraphWindowList.getCount () == 0)
     NewWindow->Handle = 1;
   else
     NewWindow->Handle =
       (GraphWindowList.at (GraphWindowList.getCount () - 1))->Handle + 1;
   if (NewWindow->Handle > MAX_GRAPHS)
    {
     // Maximum number of graphics-windows exceeded
     TErrorMessage BoxMessage, BoxTitle;
     LoadString (PNeuroApp->hInstance, STR_BOX_TITLE, BoxTitle,
                 sizeof (BoxTitle));
     LoadString (PNeuroApp->hInstance, ERR_WINDOW_COUNT, BoxMessage,
                 sizeof (BoxMessage)); 
     MessageBeep (-1);
     BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                     MB_ICONSTOP | MB_OK);
     delete NewWindow;
     return (GRAPHERR_MAXGRAPH);
    }
   if (Title == NULL)
     sprintf (WindowTitle, "VieNet2-Diagramm %i", NewWindow->Handle);
   else
     sprintf (WindowTitle, "VieNet2: %s", Title);
   if ((NewWindow->Window = new TGraphWindow (WindowTitle, NewWindow->Handle,
					      AxisValues, Type, Header, WindowPosition,
					      WindowSize, Font)) == NULL)
    {
     TErrorMessage BoxMessage, BoxTitle;
     LoadString (PNeuroApp->hInstance, STR_BOX_TITLE, BoxTitle,
                 sizeof (BoxTitle));
     LoadString (PNeuroApp->hInstance, ERR_NO_WINDOW, BoxMessage,
                 sizeof (BoxMessage)); 
     MessageBeep (-1);
     BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                     MB_ICONSTOP | MB_OK);
     delete NewWindow;
     return (GRAPHERR_NOMEM);
    }
   else
    { 
     GraphWindowList.insert (NewWindow);
     NewWindow->Window->Create ();
     NewWindow->Window->Show (SW_SHOW);
     AppendWindowMenu ((void *) NewWindow->Window);
     return (NewWindow->Handle);
    }
  }
}

void close_diagram (int GraphHandle)
{
 TWindowRef *WindowToDelete;

 if (GraphHandle != CLOSE_ALL)
  {
   WindowToDelete = GraphWindowList.firstThat (FindGraphWindow, &GraphHandle);
   if (WindowToDelete != NULL)
     SendMessage (WindowToDelete->Window->HWindow, WM_CLOSE, 0, 0);
  }
 else
   while (GraphWindowList.getCount () > 0)
     SendMessage (GraphWindowList.at (0)->Window->HWindow, WM_CLOSE, 0, 0);
}

int init_curve (char *Name, COLORREF Color, int LineStyle, float Increment, int GraphHandle)
{
 TWindowRef *GraphWindow;

 // Check if parameters are within allowed bounds
 if ((strlen (Name) > MAX_CURV_NAME) || (Increment == 0))
   return (GRAPHERR_PARAMINVAL);

 GraphWindow = GraphWindowList.firstThat (FindGraphWindow, &GraphHandle);

 if (GraphWindow == NULL)
   // No graphics-window for the given handle exists
   return (GRAPHERR_HINVAL);

 int CurveHandle = GraphWindow->Window->NewCurve (Name, (COLORREF) Color, LineStyle,
                                                  Increment);

 return (CurveHandle);
}

int dia_set_point (int CurveHandle, float Y, float X)
{
 TWindowRef *GraphWindow;
 float      Test   = MINFLOAT;    // Correct comparison!!
 BOOL       XValid = (X != Test);

 int GraphHandle = CurveHandle >> 8;
 GraphWindow = GraphWindowList.firstThat (FindGraphWindow, &GraphHandle);

 if (GraphWindow == NULL)
   // No graphics-window for the given handle exists
   return (GRAPHERR_HINVAL);

 return (GraphWindow->Window->AddCoordinate (X, Y, XValid, CurveHandle, TRUE));
}

int dia_set_curve (int CurveHandle, int PointCount, float Y [], float X [], float XStart)
{
 TWindowRef *GraphWindow;
 int        ReturnValue;
 int        Index;

 int GraphHandle = CurveHandle >> 8;
 GraphWindow = GraphWindowList.firstThat (FindGraphWindow, &GraphHandle);

 if (GraphWindow == NULL)
   // No graphics-window for the given handle exists
   return (GRAPHERR_HINVAL);

 ReturnValue = GraphWindow->Window->ClearCurve (CurveHandle);
 if ((PointCount <= 0) || (ReturnValue != GRAPHERR_NOERR))
   return (ReturnValue);

 if (X == NULL)
  {
   GraphWindow->Window->AddCoordinate (XStart, Y [0], TRUE, CurveHandle, FALSE);
   for (Index = 1; Index < PointCount - 1; Index++)
     GraphWindow->Window->AddCoordinate (XStart, Y [Index], FALSE,
					 CurveHandle, FALSE);
   ReturnValue = GraphWindow->Window->AddCoordinate (XStart, Y [PointCount - 1], FALSE,
                                                     CurveHandle, TRUE);
  }
 else
  {
   for (Index = 0; Index < PointCount - 1; Index++)
     GraphWindow->Window->AddCoordinate (X [Index], Y [Index], TRUE, CurveHandle, FALSE);
   ReturnValue = GraphWindow->Window->AddCoordinate (X [PointCount - 1], Y [PointCount - 1],
						     TRUE, CurveHandle, TRUE);
  }

 return (ReturnValue);
}

int dia_set_xstart (int CurveHandle, float NewXStart)
{
 TWindowRef *GraphWindow;

 int GraphHandle = CurveHandle >> 8;
 GraphWindow = GraphWindowList.firstThat (FindGraphWindow, &GraphHandle);

 if (GraphWindow == NULL)
   // No graphics-window for the given handle exists
   return (GRAPHERR_HINVAL);

 return (GraphWindow->Window->SetCurveStart (CurveHandle, NewXStart));
}

int dia_clear_curve (int CurveHandle)
{
 TWindowRef *GraphWindow;

 int GraphHandle = CurveHandle >> 8;
 GraphWindow = GraphWindowList.firstThat (FindGraphWindow, &GraphHandle);

 if (GraphWindow == NULL)
   // No graphics-window for the given handle exists
   return (GRAPHERR_HINVAL);

 return (GraphWindow->Window->ClearCurve (CurveHandle));
}

int dia_set_marker (int GraphHandle, float MarkerPosition, int MarkerOrientation,
                    int MarkerLineType, COLORREF MarkerColor, BOOL DisplayValue)
{
 TWindowRef *GraphWindow;

 // Check if parameters are within allowed bounds
 if ((MarkerOrientation != MARKER_HORIZONTAL) &&
     (MarkerOrientation != MARKER_VERTICAL))
   return (GRAPHERR_PARAMINVAL);

 GraphWindow = GraphWindowList.firstThat (FindGraphWindow, &GraphHandle);

 if (GraphWindow == NULL)
   // No graphics-window for the given handle exists
   return (GRAPHERR_HINVAL);

 return (GraphWindow->Window->SetMarker (MarkerPosition, MarkerOrientation, MarkerLineType,
                                         MarkerColor, DisplayValue));
}

float calc_mean (float Values [], int ValueCount)
{
 float ValueSum = 0.0;
 for (int Index = 0; Index < ValueCount; Index++)
   ValueSum += Values [Index];
 return (ValueSum / (float) ValueCount);
}

float calc_sdev (float Values [], int ValueCount)
{
 float ValueSum  = 0.0;
 float ValueSum2 = 0.0;
 for (int Index = 0; Index < ValueCount; Index++)
  {
   ValueSum  += Values [Index];
   ValueSum2 += square (Values [Index]);
  }
 return (sqrt ((1.0 / (float) ValueCount) * (ValueSum2 -
	       (square (ValueSum) / (float) ValueCount))));
}

float calc_mse (float Reference [], float Actual [], int ValueCount)
{
 float DiffSum2 = 0.0;
 for (int Index = 0; Index < ValueCount; Index++)
   DiffSum2 += square (Reference [Index] - Actual [Index]);
 return (DiffSum2 / (float) ValueCount);      
}

//*************************************************************************
//
//  Memberfunctions of TWindowList
//
//*************************************************************************

TWindowList::TWindowList () : TNSSortedCollection (10, 5)
{
 duplicates = False;
}

int TWindowList::insert (TWindowRef *WindowRef)
{
 return (TNSSortedCollection::insert (WindowRef));
}

void TWindowList::free (TWindowRef *WindowRef)
{
 TNSSortedCollection::free (WindowRef);
}

TWindowRef *TWindowList::at (int Index)
{
 return ((TWindowRef *) TNSSortedCollection::at (Index));
}

void *TWindowList::keyOf (TWindowRef *Item)
{
 return ((void *) &(Item->Handle));
}

int TWindowList::indexOf (TWindowRef *Item)
{
 return (TNSSortedCollection::indexOf ((void *) Item));
}

int TWindowList::compare (void *First, void *Second)
{
 int FirstArg  = *((int *) keyOf ((TWindowRef *) First));
 int SecondArg = *((int *) keyOf ((TWindowRef *) Second));

 if (FirstArg == SecondArg)
   return (0);
 else
   return ((FirstArg < SecondArg) ? -1 : 1);
}

TWindowRef *TWindowList::firstThat (Boolean (*TestFunc) (void *, void*),
				    void *Data)
{
 return ((TWindowRef *) TNSSortedCollection::firstThat (TestFunc, Data));
}

//*************************************************************************
//
//  Memberfunctions of TPointList
//
//*************************************************************************

TPointList::TPointList () : TNSSortedCollection (100, 50)
{
 duplicates = False;
}

int TPointList::insert (TPoint *Point)
{
 return (TNSSortedCollection::insert (Point));
}

void TPointList::free (TPoint *Point)
{
 TNSSortedCollection::free (Point);
}

TPoint *TPointList::at (int Index)
{
 return ((TPoint *) TNSSortedCollection::at (Index));
}

void *TPointList::keyOf (TPoint *Item)
{
 return ((void *) &(Item->x));
}

int TPointList::indexOf (TPoint *Item)
{
 return (TNSSortedCollection::indexOf ((void *) Item));
}

int TPointList::compare (void *First, void *Second)
{
 float FirstArg  = *((float *) keyOf ((TPoint *) First));
 float SecondArg = *((float *) keyOf ((TPoint *) Second));

 if (FirstArg == SecondArg)
   return (0);
 else
   return ((FirstArg < SecondArg) ? -1 : 1);
}

TPoint *TPointList::firstThat (Boolean (*TestFunc) (void *, void*),
			       void *Data)
{
 return ((TPoint *) TNSSortedCollection::firstThat (TestFunc, Data));
}

//*************************************************************************
//
//  Memberfunctions of TMarkerList
//
//*************************************************************************

TMarkerList::TMarkerList () : TNSCollection (10, 5)
{
}

int TMarkerList::insert (TMarker *Marker)
{
 return (TNSCollection::insert (Marker));
}

void TMarkerList::free (TMarker *Marker)
{
 TNSCollection::free (Marker);
}

TMarker *TMarkerList::at (int Index)
{
 return ((TMarker *) TNSCollection::at (Index));
}

void *TMarkerList::keyOf (TMarker *Item)
{
 return ((void *) &(Item->Position));
}

int TMarkerList::indexOf (TMarker *Item)
{
 return (TNSCollection::indexOf ((void *) Item));
}

TMarker *TMarkerList::firstThat (Boolean (*TestFunc) (void *, void*), void *Data)
{
 return ((TMarker *) TNSCollection::firstThat (TestFunc, Data));
}

//*************************************************************************
//
// Memberfunctions of TCurve
//
//*************************************************************************

TCurve::TCurve (int MyHandle, char *MyName, COLORREF MyColor,
		int MyLineType, float MyIncrement, TCurve *Last,
		PTGraphWindow Chart)
{
 strcpy (Name, MyName);
 Handle        = MyHandle;
 Color         = MyColor;
 LineType      = MyLineType;
 XIncrement    = MyIncrement;
 Next          = NULL;
 Previous      = Last;
 ChartWindow   = Chart;
 Mark          = FALSE;
 DefaultXStart = ChartWindow->AxisRange.X.Start; 
 Bounds.MinX   = ChartWindow->AxisRange.X.Start;
 Bounds.MaxX   = ChartWindow->AxisRange.X.End;
 Bounds.MinY   = ChartWindow->AxisRange.Y.Start;
 Bounds.MaxY   = ChartWindow->AxisRange.Y.End;
}

TCurve::~TCurve ()
{
 // Free memory of point-list
 Points.freeAll ();

 // Ensure correct linkage within list of curves
 if (Previous != NULL)
   Previous->Next = Next;
 if (Next != NULL)
   Next->Previous = Previous;
}

void TCurve::AddPoint (float Y)
{
 float X;

 if (Points.getCount () == 0)
   X = DefaultXStart;
 else
   X = (Points.at (Points.getCount () - 1))->x + XIncrement;
 Points.insert (new TPoint (X, Y));

 if (X < Bounds.MinX)
   Bounds.MinX = X;
 else if (X > Bounds.MaxX)
   Bounds.MaxX = X;

 if (Y < Bounds.MinY)
   Bounds.MinY = Y;
 else if (Y > Bounds.MaxY)
   Bounds.MaxY = Y;
}

void TCurve::AddCoordinate (float X, float Y)
{
 Points.insert (new TPoint (X, Y));

 if (X < Bounds.MinX)
   Bounds.MinX = X;
 else if (X > Bounds.MaxX)
   Bounds.MaxX = X;

 if (Y < Bounds.MinY)
   Bounds.MinY = Y;
 else if (Y > Bounds.MaxY)
   Bounds.MaxY = Y;
}

void TCurve::ClearPoints ()
{
 Points.freeAll ();
 DefaultXStart = ChartWindow->AxisRange.X.Start;
}

void TCurve::SetXStart (float NewXStart)
{
 DefaultXStart = NewXStart;
}

void TCurve::Draw (HDC ChartDC)
{
 TDrawData DrawData;
 HBITMAP   DotBitmap, PreviousBitmap;
 HPEN      CurvePen, PreviousPen;

 if (Points.getCount () == 0)
   return;

 if ((LineType & LINE_NONE) != LINE_NONE)
  {
   CurvePen          = CreatePen (LineType & 0x000F, 1, Color);
   PreviousPen       = SelectObject (ChartDC, CurvePen);
   DrawData.DrawLine = TRUE;
  }
 else
   DrawData.DrawLine = FALSE;

 switch (LineType & 0xFF00)
  {
   case DOT_HOLLOW_SQUARE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotHollowSquare");
     break;
   case DOT_HOLLOW_CIRCLE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotHollowCircle");
     break;
   case DOT_HOLLOW_TRIANGLE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotHollowTriangle");
     break;
   case DOT_FILLED_SQUARE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotFilledSquare");
     break;
   case DOT_FILLED_CIRCLE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotFilledCircle");
     break;
   case DOT_FILLED_TRIANGLE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotFilledTriangle");
     break;
   case DOT_ASTERISK:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotAsterisk");
     break;
   default:
     DotBitmap      = NULL;
     DrawData.DotDC = NULL;
  }

 if (DotBitmap != NULL)
  {
   HDC ScreenDC   = GetDC (NULL);
   DrawData.DotDC = CreateCompatibleDC (ScreenDC);
   PreviousBitmap = SelectObject (DrawData.DotDC, DotBitmap);
   ReleaseDC (NULL, ScreenDC);
   // Colorize the bitmap with the curve-color
   for (int DotX = 0; DotX < 7; DotX++)
     for (int DotY = 0; DotY < 7; DotY++)
       if (GetPixel (DrawData.DotDC, DotX, DotY) == RGB (0, 0, 0))
         SetPixel (DrawData.DotDC, DotX, DotY, Color);
  }

 DrawData.DrawDC = ChartDC;
 memcpy (&(DrawData.Scaling), &(ChartWindow->Scale),
	 sizeof (DrawData.Scaling));
 DrawData.ScaleType = ChartWindow->DiagramType;

 double Infinite = -MAXFLOAT;
 double XValue = ((DrawData.ScaleType & X_LOGARITHMIC) != 0) ? log10 ((Points.at (0))->x) :
                                                               (Points.at (0))->x;
 double YValue = ((DrawData.ScaleType & Y_LOGARITHMIC) != 0) ? log10 ((Points.at (0))->y) :
                                                               (Points.at (0))->y;

 int XStartPoint = (XValue > Infinite) ? (XValue * ChartWindow->Scale.PixelsPerXUnit +
		                          ChartWindow->Scale.Origin.x) : -MAXINT;;
 int YStartPoint = (YValue > Infinite) ? (-YValue * ChartWindow->Scale.PixelsPerYUnit +
		                          ChartWindow->Scale.Origin.y) : MAXINT;
 MoveTo (ChartDC, XStartPoint, YStartPoint);

 Points.forEach (DrawConnection, (void *) &DrawData); 

 if ((LineType & LINE_NONE) != LINE_NONE)
  {
   SelectObject (ChartDC, PreviousPen);
   DeleteObject (CurvePen);
  }

 if (DotBitmap != NULL)
  {
   SelectObject (DrawData.DotDC, PreviousBitmap);
   DeleteObject (DotBitmap);
   DeleteDC     (DrawData.DotDC);
  }
}

void TCurve::Legend (HDC ChartDC, POINT& Offset, HFONT Font)
{
 TEXTMETRIC FontInfo;
 HPEN       CurvePen, PreviousPen;
 HBITMAP    DotBitmap, PreviousBitmap;
 BOOL       DrawLine;

 HFONT PreviousFont = SelectObject (ChartDC, Font);

 if ((LineType & LINE_NONE) != LINE_NONE)
  {
   CurvePen    = CreatePen (LineType & 0x000F, 1, Color);
   PreviousPen = SelectObject (ChartDC, CurvePen);
   DrawLine    = TRUE;
  }
 else
   DrawLine = FALSE;

 switch (LineType & 0xFF00)
  {
   case DOT_HOLLOW_SQUARE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotHollowSquare");
     break;
   case DOT_HOLLOW_CIRCLE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotHollowCircle");
     break;
   case DOT_HOLLOW_TRIANGLE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotHollowTriangle");
     break;
   case DOT_FILLED_SQUARE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotFilledSquare");
     break;
   case DOT_FILLED_CIRCLE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotFilledCircle");
     break;
   case DOT_FILLED_TRIANGLE:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotFilledTriangle");
     break;
   case DOT_ASTERISK:
     DotBitmap = LoadBitmap (ChartWindow->GetApplication ()->hInstance, "DotAsterisk");
     break;
   default:
     DotBitmap = NULL;
  }

 GetTextMetrics (ChartDC, &FontInfo);
 int YLineOffset = Offset.y + (FontInfo.tmHeight / 2);

 if (DotBitmap != NULL)
  {
   HDC ScreenDC   = GetDC (NULL);
   HDC DotDC      = CreateCompatibleDC (ScreenDC);
   PreviousBitmap = SelectObject (DotDC, DotBitmap);
   ReleaseDC (NULL, ScreenDC);
   // Colorize the bitmap with the curve-color
   for (int DotX = 0; DotX < 7; DotX++)
     for (int DotY = 0; DotY < 7; DotY++)
       if (GetPixel (DotDC, DotX, DotY) == RGB (0, 0, 0))
	 SetPixel (DotDC, DotX, DotY, Color);

   // Put the bitmap into the legend
   BitBlt (ChartDC, Offset.x, YLineOffset - 3, 7, 7, DotDC, 0, 0, SRCCOPY);
   BitBlt (ChartDC, Offset.x + 28, YLineOffset - 3, 7, 7, DotDC, 0, 0, SRCCOPY);

   SelectObject (DotDC, PreviousBitmap);
   DeleteObject (DotBitmap);
   DeleteDC     (DotDC);
  }

 if (DrawLine)
  {
   MoveTo (ChartDC, Offset.x, YLineOffset);
   LineTo (ChartDC, Offset.x + 35, YLineOffset);
  }
 TextOut (ChartDC, Offset.x + 40, Offset.y, Name, strlen (Name));
 Offset.y += FontInfo.tmHeight + FontInfo.tmExternalLeading; 

 SelectObject (ChartDC, PreviousFont);

 if ((LineType & LINE_NONE) != LINE_NONE)
  {
   SelectObject (ChartDC, PreviousPen);
   DeleteObject (CurvePen);
  }
}

void TCurve::WritePoints (ofstream *OutputFile)
{
 Points.forEach (PrintPoint, (void *) OutputFile);
}

HANDLE TCurve::RetrieveYValues ()
{
 HANDLE       Result;
 TYValueArray *ResultArray;
 int          PointCount = Points.getCount ();

 Result = GlobalAlloc (GHND, (PointCount * sizeof (float)) +
			      sizeof (TYValueArray));
 if (Result == NULL)
   return (NULL);
 ResultArray = (TYValueArray *) GlobalLock (Result);
 ResultArray->ElementCount = PointCount;

 for (int Index = 0; Index < PointCount; Index++)
   ResultArray->Values [Index] = Points.at (Index)->y;

 GlobalUnlock (Result);

 return (Result);
}

float TCurve::LowerX ()
{
 return (Bounds.MinX);
}

float TCurve::UpperX ()
{
 return (Bounds.MaxX);
}

float TCurve::LowerY ()
{
 return (Bounds.MinY);
}

float TCurve::UpperY ()
{
 return (Bounds.MaxY);
}

//*************************************************************************
//
// Memberfunctions of TSaveDlg
//
//*************************************************************************

TSaveDlg::TSaveDlg (char *SelectedDirectory, int *Type, BOOL *Selected,
		    PTWindowsObject MyParent) : TDialog (MyParent, "GraphSave")
{
 ChosenDirectory = SelectedDirectory;
 ChosenType      = Type;
 SelectedCurves  = Selected;

 // Set default values
 *ChosenType = SAVE_FILE_VALUES;
 // Probably we're in the root-directory, so check for a terminating '\'
 if (ChosenDirectory [strlen (ChosenDirectory) - 1] == '\\')
   strcat (ChosenDirectory, "*.*");
 else 
   strcat (ChosenDirectory, "\\*.*");
}

void TSaveDlg::SetupWindow ()
{
 char Description [255];
 char ShowDirectory [MAXPATH];

 TDialog::SetupWindow ();

 LoadString (GetApplication ()->hInstance, STR_CURVE_VALUES, Description,
	     sizeof (Description));
 SendDlgItemMsg (ID_GRAPH_TYPEDESCR, WM_SETTEXT, 0, (DWORD) Description);
 SendDlgItemMsg (ID_GRAPH_TYPELIST, CB_ADDSTRING, 0,
		 (DWORD) "Value-files (*.xy)");
 SendDlgItemMsg (ID_GRAPH_TYPELIST, CB_ADDSTRING, 0,
		 (DWORD) "Bitmap-file (*.bmp)");
 strcpy (ShowDirectory, ChosenDirectory);
 DlgDirList (HWindow, ShowDirectory, ID_GRAPH_DIRLIST, ID_GRAPH_CURDIR,
	     0xC010);
 SendDlgItemMsg (ID_GRAPH_TYPELIST, CB_SETCURSEL, 0, 0);
 SendDlgItemMsg (ID_GRAPH_DIRLIST, LB_SETCURSEL, 0, 0);
 if (*SelectedCurves)
   SendDlgItemMsg (ID_SELECTEDONLY, BM_SETCHECK, *SelectedCurves, 0);
 else
   // No curves selected, this option doesn't make sense
   ShowWindow (GetItemHandle (ID_SELECTEDONLY), SW_HIDE);
}

void TSaveDlg::ChangeDirectory (TMessage& TheMessage)
{
 int  StringIndex;
 char ListBoxEntry [MAXFILE + MAXEXT];
 char CurrentDirectory [MAXPATH];
 char ShowDirectory [MAXPATH];

 if (TheMessage.LP.Hi == LBN_DBLCLK)
  {
   int ListIndex = SendDlgItemMsg (ID_GRAPH_DIRLIST, LB_GETCURSEL, 0, 0);
   if (ListIndex != LB_ERR)
    {
     SendDlgItemMsg (ID_GRAPH_DIRLIST, LB_GETTEXT, ListIndex,
		    (DWORD) ListBoxEntry);
     switch (ListBoxEntry [1])
      {
       case '.':
         // We can start to eliminate the last directory-name before the
	 // terminating \*.* 
	 for (StringIndex = strlen (ChosenDirectory) - 5;
	      ChosenDirectory [StringIndex] != '\\';
	      StringIndex--)
	   ChosenDirectory [StringIndex] = '\0';
	 strcat (ChosenDirectory, "*.*");
         break;

       case '-':
	 if (getcurdir (toupper (ListBoxEntry [2]) - '@', CurrentDirectory) == 0)
          {
	   strcpy (ChosenDirectory, "A:");
	   ChosenDirectory [0] = toupper (ListBoxEntry [2]);
	   if (CurrentDirectory [0] != '\\')
	     strcat (ChosenDirectory, "\\");
	   strcat (ChosenDirectory, CurrentDirectory);
	   // Probably we're in the root-directory, so check for a terminating '\'
	   if (ChosenDirectory [strlen (ChosenDirectory) - 1] == '\\')
	     strcat (ChosenDirectory, "*.*");
	   else 
	     strcat (ChosenDirectory, "\\*.*");
	  }
	 break;

       default:
         StringIndex = strlen (ChosenDirectory) - 3;
	 for (int EntryIndex = 1; EntryIndex < strlen (ListBoxEntry) - 1; EntryIndex ++)
	   ChosenDirectory [StringIndex++] = ListBoxEntry [EntryIndex];
	 ChosenDirectory [StringIndex] = '\0';
	 // Probably we're in the root-directory, so check for a terminating '\'
	 if (ChosenDirectory [strlen (ChosenDirectory) - 1] == '\\')
	   strcat (ChosenDirectory, "*.*");
	 else 
	   strcat (ChosenDirectory, "\\*.*");
      }
     strcpy (ShowDirectory, ChosenDirectory);
     DlgDirList (HWindow, ShowDirectory, ID_GRAPH_DIRLIST, ID_GRAPH_CURDIR,
		 0xC010);
     SendDlgItemMsg (ID_GRAPH_DIRLIST, LB_SETCURSEL, 0, 0);
    }
  }
 else
   DefWndProc (TheMessage);
}

void TSaveDlg::TypeChange (TMessage& TheMessage)
{
 char Description [255];

 if (TheMessage.LP.Hi == CBN_SELCHANGE)
   switch (SendDlgItemMsg (ID_GRAPH_TYPELIST, CB_GETCURSEL, 0, 0))
    {
     case 0:
       *ChosenType = SAVE_FILE_VALUES;
       LoadString (GetApplication ()->hInstance, STR_CURVE_VALUES, Description,
                   sizeof (Description));
       SendDlgItemMsg (ID_GRAPH_TYPEDESCR, WM_SETTEXT, 0, (DWORD) Description);
       EnableWindow (GetItemHandle (ID_SELECTEDONLY), TRUE);
       break;

     case 1:
       *ChosenType = SAVE_FILE_BITMAP;
       LoadString (GetApplication ()->hInstance, STR_GRAPH_BITMAP, Description,
                   sizeof (Description));
       SendDlgItemMsg (ID_GRAPH_TYPEDESCR, WM_SETTEXT, 0, (DWORD) Description);
       EnableWindow (GetItemHandle (ID_SELECTEDONLY), FALSE); 
       break;
    }  
 else
   DefWndProc (TheMessage);
}

void TSaveDlg::SelectedOnly (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   *SelectedCurves = SendDlgItemMsg (ID_SELECTEDONLY, BM_GETCHECK, 0, 0);
 else
   DefWndProc (TheMessage); 
}

void TSaveDlg::Ok (TMessage& TheMessage)
{
 ChosenDirectory [strlen (ChosenDirectory) - 4] = '\0';

 TDialog::Ok (TheMessage);
}

//*************************************************************************
//
// Memberfunctions of TLoadDlg
//
//*************************************************************************

TLoadDlg::TLoadDlg (char *Name, COLORREF *Color, int *Style, char *File1,
		    char *File2, float *XBegin, float *XIncrement,
		    int *FileType, PTWindowsObject MyParent) :
	  TDialog (MyParent, "GraphLoad")
{
 NewName   = Name;
 NewColor  = Color;
 LineStyle = Style;
 FileName1 = File1;
 FileName2 = File2;
 XStart    = XBegin;
 XStep     = XIncrement;
 Type      = FileType;

 // Set default-values
 strcpy (FileName1, "");
 strcpy (FileName2, "");
 *NewColor  = LINE_BLACK;
 *LineStyle = LINE_SOLID | DOT_NONE;
 *Type      = OPEN_FILE_COMBINED;

 // Attach owner-drawn comboboxes to their class
 ColorBox = new TStyleComboBox (ID_GRAPH_CURVECOLOR, this);
 LineBox  = new TStyleComboBox (ID_GRAPH_CURVESTYLE, this);
 DotBox   = new TStyleComboBox (ID_GRAPH_DOTSTYLE, this);
}

TLoadDlg::~TLoadDlg ()
{
 // Remove the owner-drawn combobox-elements
 delete ColorBox;
 delete LineBox;
 delete DotBox;
}

void TLoadDlg::SetupWindow ()
{
 char NumberString [15];

 TDialog::SetupWindow ();

 SendDlgItemMsg (ID_GRAPH_CURVENAME, EM_LIMITTEXT, MAX_CURV_NAME, 0);
 SendDlgItemMsg (ID_GRAPH_CURVENAME, WM_SETTEXT, 0, (DWORD) NewName);

 SendDlgItemMsg (ID_GRAPH_CURVECOLOR, CB_ADDSTRING, 0, 0); // Black
 SendDlgItemMsg (ID_GRAPH_CURVECOLOR, CB_ADDSTRING, 0, 0); // Red
 SendDlgItemMsg (ID_GRAPH_CURVECOLOR, CB_ADDSTRING, 0, 0); // Green
 SendDlgItemMsg (ID_GRAPH_CURVECOLOR, CB_ADDSTRING, 0, 0); // Blue
 SendDlgItemMsg (ID_GRAPH_CURVECOLOR, CB_SETCURSEL, 0, 0);

 SendDlgItemMsg (ID_GRAPH_CURVESTYLE, CB_ADDSTRING, 0, 0); // Solid
 SendDlgItemMsg (ID_GRAPH_CURVESTYLE, CB_ADDSTRING, 0, 0); // Dashes
 SendDlgItemMsg (ID_GRAPH_CURVESTYLE, CB_ADDSTRING, 0, 0); // Dotted
 SendDlgItemMsg (ID_GRAPH_CURVESTYLE, CB_ADDSTRING, 0, 0); // Dash-Dot
 SendDlgItemMsg (ID_GRAPH_CURVESTYLE, CB_ADDSTRING, 0, 0); // Dash-Dot-Dot
 SendDlgItemMsg (ID_GRAPH_CURVESTYLE, CB_ADDSTRING, 0, 0); // None
 SendDlgItemMsg (ID_GRAPH_CURVESTYLE, CB_SETCURSEL, 0, 0);

 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_ADDSTRING, 0, 0); // Hollow Square
 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_ADDSTRING, 0, 0); // Hollow Circle
 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_ADDSTRING, 0, 0); // Hollow Triangle
 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_ADDSTRING, 0, 0); // Filled Square
 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_ADDSTRING, 0, 0); // Filled Circle
 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_ADDSTRING, 0, 0); // Filled Triangle
 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_ADDSTRING, 0, 0); // Asterisk
 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_ADDSTRING, 0, 0); // None
 SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_SETCURSEL, 7, 0);

 SendDlgItemMsg (ID_GRAPH_COMBINED, BM_SETCHECK, TRUE, 0);
 SendDlgItemMsg (ID_GRAPH_XVAL2, BM_SETCHECK, TRUE, 0);
 sprintf (NumberString, "%.2f", *XStart);
 SendDlgItemMsg (ID_GRAPH_XSTART, WM_SETTEXT, 0, (DWORD) NumberString);
 sprintf (NumberString, "%.2f", *XStep);
 SendDlgItemMsg (ID_GRAPH_XINCR, WM_SETTEXT, 0, (DWORD) NumberString);
 EnableWindow (GetItemHandle (ID_GRAPH_XSTART), FALSE);
 EnableWindow (GetItemHandle (ID_GRAPH_XINCR), FALSE);
}

void TLoadDlg::Browse1 ()
{
 OPENFILENAME OpenStruct;
 char         *DefaultExtension [] = {"#", "xy", "x", "y"};
 int          Index;

 if (SendDlgItemMsg (ID_GRAPH_COMBINED, BM_GETCHECK, 0, 0))
   Index = 1;
 else if (SendDlgItemMsg (ID_GRAPH_XVAL1, BM_GETCHECK, 0, 0))
   Index = 2;
 else
   Index = 3;

 memset (&OpenStruct, 0, sizeof (OpenStruct));

 OpenStruct.lStructSize  = sizeof (OpenStruct);
 OpenStruct.hwndOwner    = HWindow;
 OpenStruct.lpstrFilter  = "Combined Files\0*.xy\0X-Value Files\0*.x\0Y-Value Files\0*.y\0All Files\0*.*\0";
 OpenStruct.nFilterIndex = Index;
 OpenStruct.lpstrFile    = FileName1;
 OpenStruct.nMaxFile     = MAXPATH;
 OpenStruct.lpstrDefExt  = DefaultExtension [Index];
 OpenStruct.lpstrTitle   = "Load Curve Data";
 OpenStruct.Flags        = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;

 if (GetOpenFileName (&OpenStruct))
   SendDlgItemMsg (ID_GRAPH_FILE1, WM_SETTEXT, 0,
		   (DWORD) AnsiLower (FileName1));
}

void TLoadDlg::Browse2 ()
{
 OPENFILENAME OpenStruct;
 char         *DefaultExtension [] = {"#", "x", "y"};
 int          Index;

if (SendDlgItemMsg (ID_GRAPH_XVAL2, BM_GETCHECK, 0, 0))
   Index = 1;
 else
   Index = 2;

 memset (&OpenStruct, 0, sizeof (OpenStruct));

 OpenStruct.lStructSize  = sizeof (OpenStruct);
 OpenStruct.hwndOwner    = HWindow;
 OpenStruct.lpstrFilter  = "X-Value Files\0*.x\0Y-Value Files\0*.y\0All Files\0*.*\0";
 OpenStruct.nFilterIndex = Index;
 OpenStruct.lpstrFile    = FileName2;
 OpenStruct.nMaxFile     = MAXPATH;
 OpenStruct.lpstrDefExt  = DefaultExtension [Index];
 OpenStruct.lpstrTitle   = "Load Curve Data";
 OpenStruct.Flags        = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;

 if (GetOpenFileName (&OpenStruct))
   SendDlgItemMsg (ID_GRAPH_FILE2, WM_SETTEXT, 0,
		   (DWORD) AnsiLower (FileName2));
}

void TLoadDlg::CombinedFile (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   RadioButtonChange ();
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::File1XValues (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   RadioButtonChange ();
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::File1YValues (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   RadioButtonChange ();
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::File2XValues (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   RadioButtonChange ();
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::File2YValues (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   RadioButtonChange ();
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::NoFile2 (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   RadioButtonChange ();
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::ColorChanged (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == CBN_SELCHANGE)
   switch (SendDlgItemMsg (ID_GRAPH_CURVECOLOR, CB_GETCURSEL, 0, 0))
    {
     case 0:
       *NewColor = LINE_BLACK;
       break;
     case 1:
       *NewColor = LINE_RED;
       break;
     case 2:
       *NewColor = LINE_GREEN;
       break;
     case 3:
       *NewColor = LINE_BLUE;
       break;
     default:
       ; // Leave old value, no new value selected
    }
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::TypeChanged (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == CBN_SELCHANGE)
  {
   switch (SendDlgItemMsg (ID_GRAPH_CURVESTYLE, CB_GETCURSEL, 0, 0))
    {
     case 0:
       *LineStyle = (*LineStyle & 0xFF00) | LINE_SOLID;
       break;
     case 1:
       *LineStyle = (*LineStyle & 0xFF00) | LINE_DASH;
       break;
     case 2:
       *LineStyle = (*LineStyle & 0xFF00) | LINE_DOT;
       break;
     case 3:
       *LineStyle = (*LineStyle & 0xFF00) | LINE_DASH_DOT;
       break;
     case 4:
       *LineStyle = (*LineStyle & 0xFF00) | LINE_DASH_DOT_DOT;
       break;
     case 5:
       *LineStyle = (*LineStyle & 0xFF00) | LINE_NONE;
       break;
    }
  }
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::DotChanged (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == CBN_SELCHANGE)
  {
   switch (SendDlgItemMsg (ID_GRAPH_DOTSTYLE, CB_GETCURSEL, 0, 0))
    {
     case 0:
       *LineStyle = (*LineStyle & 0x00FF) | DOT_HOLLOW_SQUARE;
       break;
     case 1:
       *LineStyle = (*LineStyle & 0x00FF) | DOT_HOLLOW_CIRCLE;
       break;
     case 2:
       *LineStyle = (*LineStyle & 0x00FF) | DOT_HOLLOW_TRIANGLE;
       break;
     case 3:
       *LineStyle = (*LineStyle & 0x00FF) | DOT_FILLED_SQUARE;
       break;
     case 4:
       *LineStyle = (*LineStyle & 0x00FF) | DOT_FILLED_CIRCLE;
       break;
     case 5:
       *LineStyle = (*LineStyle & 0x00FF) | DOT_FILLED_TRIANGLE;
       break;
     case 6:
       *LineStyle = (*LineStyle & 0x00FF) | DOT_ASTERISK;
       break;
     case 7:
       *LineStyle = (*LineStyle & 0x00FF) | DOT_NONE;
       break;
    }
  }
 else
   DefWndProc (TheMessage);
}

void TLoadDlg::Ok (TMessage& TheMessage)
{
 char          *DecimalPoint;
 char          NumberString [15];
 BOOL          XOk;
 TErrorMessage BoxMessage, BoxTitle;

 SendDlgItemMsg (ID_GRAPH_XSTART, WM_GETTEXT, sizeof (NumberString),
		 (DWORD) NumberString);
 // Make '.' and ',' equal for input
 if ((DecimalPoint = strchr (NumberString, ',')) != NULL)
   *DecimalPoint = '.';
 XOk = (sscanf (NumberString, "%f", XStart) == 1) ? TRUE : FALSE;
 SendDlgItemMsg (ID_GRAPH_XINCR, WM_GETTEXT, sizeof (NumberString),
		 (DWORD) NumberString);
 // Make '.' and ',' equal for input
 if ((DecimalPoint = strchr (NumberString, ',')) != NULL)
   *DecimalPoint = '.';
 XOk = (sscanf (NumberString, "%f", XStep) == 1) ? XOk : FALSE;

 if (SendDlgItemMsg (ID_GRAPH_CURVENAME, WM_GETTEXT, MAX_CURV_NAME,
		     (DWORD) NewName) == 0)
  {
   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
	       sizeof (BoxTitle));
   LoadString (GetApplication ()->hInstance, ERR_NO_CURVENAME, BoxMessage,
	       sizeof (BoxMessage));
   MessageBeep (-1);
   BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
		   MB_ICONINFORMATION | MB_OK);
   return;
  }
 else if ((strlen (FileName1) == 0) ||
	  ((*Type != OPEN_FILE_ONLY_Y) && (*Type != OPEN_FILE_COMBINED) &&
	   (strlen (FileName2) == 0)))
  {
   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
	       sizeof (BoxTitle));
   LoadString (GetApplication ()->hInstance, ERR_NO_FILE, BoxMessage,
	       sizeof (BoxMessage));
   MessageBeep (-1);
   BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
		   MB_ICONINFORMATION | MB_OK);
   return;
  }
 else if ((*Type == OPEN_FILE_ONLY_Y) && !XOk)
  {
   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
	       sizeof (BoxTitle));
   LoadString (GetApplication ()->hInstance, ERR_NO_XVALUES, BoxMessage,
	       sizeof (BoxMessage));
   MessageBeep (-1);
   BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
		   MB_ICONINFORMATION | MB_OK);
   return;
  }

 TDialog::Ok (TheMessage);
}

void TLoadDlg::RadioButtonChange ()
{
 BOOL Combined = (BOOL) SendDlgItemMsg (ID_GRAPH_COMBINED, BM_GETCHECK, 0,
					0);
 BOOL XValue1  = (BOOL) SendDlgItemMsg (ID_GRAPH_XVAL1, BM_GETCHECK, 0, 0);
 BOOL XValue2  = (BOOL) SendDlgItemMsg (ID_GRAPH_XVAL2, BM_GETCHECK, 0, 0);
 BOOL YValue2  = (BOOL) SendDlgItemMsg (ID_GRAPH_YVAL2, BM_GETCHECK, 0, 0);

 if (Combined)
  {
   EnableWindow (GetItemHandle (ID_GRAPH_BROWSE2), FALSE);
   EnableWindow (GetItemHandle (ID_GRAPH_XVAL2), FALSE);
   EnableWindow (GetItemHandle (ID_GRAPH_YVAL2), FALSE);
   EnableWindow (GetItemHandle (ID_GRAPH_NOXVAL), FALSE);
   EnableWindow (GetItemHandle (ID_GRAPH_XSTART), FALSE);
   EnableWindow (GetItemHandle (ID_GRAPH_XINCR), FALSE);
   SendDlgItemMsg (ID_GRAPH_FILE2, WM_SETTEXT, 0, (DWORD) "");
   *Type = OPEN_FILE_COMBINED;
  }
 else
   if (XValue1)
    {
     EnableWindow (GetItemHandle (ID_GRAPH_XVAL2), FALSE);
     EnableWindow (GetItemHandle (ID_GRAPH_NOXVAL), FALSE);
     EnableWindow (GetItemHandle (ID_GRAPH_XSTART), FALSE);
     EnableWindow (GetItemHandle (ID_GRAPH_XINCR), FALSE);
     EnableWindow (GetItemHandle (ID_GRAPH_BROWSE2), TRUE);
     EnableWindow (GetItemHandle (ID_GRAPH_YVAL2), TRUE);
     SendDlgItemMsg (ID_GRAPH_XVAL2, BM_SETCHECK, FALSE, 0);
     SendDlgItemMsg (ID_GRAPH_NOXVAL, BM_SETCHECK, FALSE, 0);
     SendDlgItemMsg (ID_GRAPH_YVAL2, BM_SETCHECK, TRUE, 0);
     SendDlgItemMsg (ID_GRAPH_FILE2, WM_SETTEXT, 0, (DWORD) FileName2);
     *Type = OPEN_FILE_SPLIT_XY;
    }
   else
    {
     EnableWindow (GetItemHandle (ID_GRAPH_YVAL2), FALSE);
     EnableWindow (GetItemHandle (ID_GRAPH_XVAL2), TRUE);
     EnableWindow (GetItemHandle (ID_GRAPH_NOXVAL), TRUE);
     if (YValue2)
      {
       SendDlgItemMsg (ID_GRAPH_YVAL2, BM_SETCHECK, FALSE, 0);
       SendDlgItemMsg (ID_GRAPH_XVAL2, BM_SETCHECK, TRUE, 0);
       YValue2 = FALSE;
       XValue2 = TRUE;
      }
     if (XValue2)
      {
       EnableWindow (GetItemHandle (ID_GRAPH_XSTART), FALSE);
       EnableWindow (GetItemHandle (ID_GRAPH_XINCR), FALSE);
       EnableWindow (GetItemHandle (ID_GRAPH_BROWSE2), TRUE);
       SendDlgItemMsg (ID_GRAPH_FILE2, WM_SETTEXT, 0, (DWORD) FileName2);
       *Type = OPEN_FILE_SPLIT_YX;
      }
     else
      {
       EnableWindow (GetItemHandle (ID_GRAPH_BROWSE2), FALSE);
       EnableWindow (GetItemHandle (ID_GRAPH_XSTART), TRUE);
       EnableWindow (GetItemHandle (ID_GRAPH_XINCR), TRUE);
       SendDlgItemMsg (ID_GRAPH_FILE2, WM_SETTEXT, 0, (DWORD) "");
       *Type = OPEN_FILE_ONLY_Y;
      }
    }
}

//*************************************************************************
//
// Memberfunctions of TDisplayDlg
//
//*************************************************************************

TDisplayDlg::TDisplayDlg (BOOL *Fit, TBitmapDim *Dimension, int *Type, BOOL XScaleToLog,
			  BOOL YScaleToLog, TAxisRange *Range, PTWindowsObject MyParent) :
	     TDialog (MyParent, "GraphDisplayOptions")
{
 BitmapSizeable = Fit;
 NewDimension   = Dimension;
 ScaleType      = Type;
 XLogScaling    = XScaleToLog;
 YLogScaling    = YScaleToLog;
 NewRange       = Range;
}

void TDisplayDlg::SetupWindow ()
{
 char NumberString [15];

 TDialog::SetupWindow ();

 sprintf (NumberString, "%i", NewDimension->Width);
 SendDlgItemMsg (ID_GRAPH_WIDTH, WM_SETTEXT, 0, (DWORD) NumberString);
 sprintf (NumberString, "%i", NewDimension->Height);
 SendDlgItemMsg (ID_GRAPH_HEIGHT, WM_SETTEXT, 0, (DWORD) NumberString);

 sprintf (NumberString, "%.2f", NewRange->X.Start);
 SendDlgItemMsg (ID_GRAPH_XLOW, WM_SETTEXT, 0, (DWORD) NumberString);
 sprintf (NumberString, "%.2f", NewRange->X.End);
 SendDlgItemMsg (ID_GRAPH_XHIGH, WM_SETTEXT, 0, (DWORD) NumberString);
 sprintf (NumberString, "%.2f", NewRange->X.Offset);
 SendDlgItemMsg (ID_GRAPH_XOFFSET, WM_SETTEXT, 0, (DWORD) NumberString);
 sprintf (NumberString, "%.2f", NewRange->Y.Start);
 SendDlgItemMsg (ID_GRAPH_YLOW, WM_SETTEXT, 0, (DWORD) NumberString);
 sprintf (NumberString, "%.2f", NewRange->Y.End);
 SendDlgItemMsg (ID_GRAPH_YHIGH, WM_SETTEXT, 0, (DWORD) NumberString);
 sprintf (NumberString, "%.2f", NewRange->Y.Offset);
 SendDlgItemMsg (ID_GRAPH_YOFFSET, WM_SETTEXT, 0, (DWORD) NumberString);


 if ((*ScaleType & X_LOGARITHMIC) != 0)
   SendDlgItemMsg (ID_GRAPH_LOG_X, BM_SETCHECK, TRUE, 0);
 else
   SendDlgItemMsg (ID_GRAPH_LIN_X, BM_SETCHECK, TRUE, 0); 
 if ((*ScaleType & Y_LOGARITHMIC) != 0)
   SendDlgItemMsg (ID_GRAPH_LOG_Y, BM_SETCHECK, TRUE, 0);
 else
   SendDlgItemMsg (ID_GRAPH_LIN_Y, BM_SETCHECK, TRUE, 0);

 if ((*ScaleType & X_AUTOSCALE) != 0)
  {
   SendDlgItemMsg (ID_GRAPH_XAUTO, BM_SETCHECK, TRUE, 0);
   EnableWindow (GetItemHandle (ID_GRAPH_XLOW), FALSE);
   EnableWindow (GetItemHandle (ID_GRAPH_XHIGH), FALSE);
  }
 if ((*ScaleType & Y_AUTOSCALE) != 0)
  {
   SendDlgItemMsg (ID_GRAPH_YAUTO, BM_SETCHECK, TRUE, 0);
   EnableWindow (GetItemHandle (ID_GRAPH_YLOW), FALSE);
   EnableWindow (GetItemHandle (ID_GRAPH_YHIGH), FALSE);
  }

 if (*BitmapSizeable)
  {
   SendDlgItemMsg (ID_FIT_TO_WINDOW, BM_SETCHECK, TRUE, 0);
   EnableWindow (GetItemHandle (ID_GRAPH_WIDTH), FALSE);
   EnableWindow (GetItemHandle (ID_GRAPH_HEIGHT), FALSE);
  }

 if (!XLogScaling)
   EnableWindow (GetItemHandle (ID_GRAPH_LOG_X), FALSE);
 if (!YLogScaling)
   EnableWindow (GetItemHandle (ID_GRAPH_LOG_Y), FALSE);
}

void TDisplayDlg::FitInWindow (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   if (SendDlgItemMsg (ID_FIT_TO_WINDOW, BM_GETCHECK, 0, 0))
    {
     EnableWindow (GetItemHandle (ID_GRAPH_WIDTH), FALSE);
     EnableWindow (GetItemHandle (ID_GRAPH_HEIGHT), FALSE);
    }
   else
    {
     EnableWindow (GetItemHandle (ID_GRAPH_WIDTH), TRUE);
     EnableWindow (GetItemHandle (ID_GRAPH_HEIGHT), TRUE);
    }
 else
   DefWndProc (TheMessage); 
}

void TDisplayDlg::AutoX (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   if (SendDlgItemMsg (ID_GRAPH_XAUTO, BM_GETCHECK, 0, 0))
    {
     EnableWindow (GetItemHandle (ID_GRAPH_XLOW), FALSE);
     EnableWindow (GetItemHandle (ID_GRAPH_XHIGH), FALSE);
    }
   else
    {
     EnableWindow (GetItemHandle (ID_GRAPH_XLOW), TRUE);
     EnableWindow (GetItemHandle (ID_GRAPH_XHIGH), TRUE);
    }
 else
   DefWndProc (TheMessage); 
}

void TDisplayDlg::AutoY (TMessage& TheMessage)
{
 if (TheMessage.LP.Hi == BN_CLICKED)
   if (SendDlgItemMsg (ID_GRAPH_YAUTO, BM_GETCHECK, 0, 0))
    {
     EnableWindow (GetItemHandle (ID_GRAPH_YLOW), FALSE);
     EnableWindow (GetItemHandle (ID_GRAPH_YHIGH), FALSE);
    }
   else
    {
     EnableWindow (GetItemHandle (ID_GRAPH_YLOW), TRUE);
     EnableWindow (GetItemHandle (ID_GRAPH_YHIGH), TRUE);
    }
 else
   DefWndProc (TheMessage); 
}

void TDisplayDlg::Ok (TMessage& TheMessage)
{
 char NumberString [15];

 if (!(*BitmapSizeable = SendDlgItemMsg (ID_FIT_TO_WINDOW, BM_GETCHECK, 0, 0)))
  {
   SendDlgItemMsg (ID_GRAPH_WIDTH, WM_GETTEXT, sizeof (NumberString), (DWORD) NumberString);
   NewDimension->Width = atoi (NumberString);
   SendDlgItemMsg (ID_GRAPH_HEIGHT, WM_GETTEXT, sizeof (NumberString), (DWORD) NumberString);
   NewDimension->Height = atoi (NumberString);
  }

 SendDlgItemMsg (ID_GRAPH_XOFFSET, WM_GETTEXT, sizeof (NumberString), (DWORD) NumberString);
 sscanf (NumberString, "%f", &(NewRange->X.Offset));
 if (SendDlgItemMsg (ID_GRAPH_XAUTO, BM_GETCHECK, 0, 0))
   *ScaleType |= X_AUTOSCALE;
 else
  {
   *ScaleType &= ~X_AUTOSCALE;
   SendDlgItemMsg (ID_GRAPH_XLOW, WM_GETTEXT, sizeof (NumberString), (DWORD) NumberString);
   sscanf (NumberString, "%f", &(NewRange->X.Start));
   SendDlgItemMsg (ID_GRAPH_XHIGH, WM_GETTEXT, sizeof (NumberString), (DWORD) NumberString);
   sscanf (NumberString, "%f", &(NewRange->X.End));
  }

 SendDlgItemMsg (ID_GRAPH_YOFFSET, WM_GETTEXT, sizeof (NumberString), (DWORD) NumberString);
 sscanf (NumberString, "%f", &(NewRange->Y.Offset));
 if (SendDlgItemMsg (ID_GRAPH_YAUTO, BM_GETCHECK, 0, 0))
   *ScaleType |= Y_AUTOSCALE;
 else
  {
   *ScaleType &= ~Y_AUTOSCALE;
   SendDlgItemMsg (ID_GRAPH_YLOW, WM_GETTEXT, sizeof (NumberString), (DWORD) NumberString);
   sscanf (NumberString, "%f", &(NewRange->Y.Start));
   SendDlgItemMsg (ID_GRAPH_YHIGH, WM_GETTEXT, sizeof (NumberString), (DWORD) NumberString);
   sscanf (NumberString, "%f", &(NewRange->Y.End));
  }

 if (SendDlgItemMsg (ID_GRAPH_LIN_X, BM_GETCHECK, 0, 0))
   *ScaleType &= ~X_LOGARITHMIC;
 else
   *ScaleType |= X_LOGARITHMIC;

 if (SendDlgItemMsg (ID_GRAPH_LIN_Y, BM_GETCHECK, 0, 0))
   *ScaleType &= ~Y_LOGARITHMIC;
 else
   *ScaleType |= Y_LOGARITHMIC;

 TDialog::Ok (TheMessage);
}

//*************************************************************************
//
// Memberfunctions of TResultBox
//
//*************************************************************************

TResultBox::TResultBox (char *Title, char *Result, char *Caption,
			PTWindowsObject MyParent) :
	    TDialog (MyParent, "GraphResultBox")
{
 ResultTitle = Title;
 ResultText  = Result;
 BoxCaption  = Caption;
}

void TResultBox::SetupWindow ()
{
 TDialog::SetupWindow ();

 SetCaption (BoxCaption);
 SendDlgItemMsg (ID_GRAPH_RESULTDESCR, WM_SETTEXT, 0, (DWORD) ResultTitle);
 SendDlgItemMsg (ID_GRAPH_RESULT, WM_SETTEXT, 0, (DWORD) ResultText);
}

//*************************************************************************
//
// Memberfunctions of TStyleComboBox
//
//*************************************************************************

TStyleComboBox::TStyleComboBox (int ResourceID, PTWindowsObject MyParent) :
		TComboBox (MyParent, ResourceID, 0)
{
}

void TStyleComboBox::ODADrawEntire (DRAWITEMSTRUCT _FAR & ItemInfo)
{
 RECT ItemArea;

 ItemArea.top    = ItemInfo.rcItem.top    + 4;
 ItemArea.left   = ItemInfo.rcItem.left   + 4;
 ItemArea.bottom = ItemInfo.rcItem.bottom - 4;
 ItemArea.right  = ItemInfo.rcItem.right  - 4;

 int PreviousBkMode = SetBkMode (ItemInfo.hDC, TRANSPARENT);

 switch (GetId ())
  {
   case ID_GRAPH_CURVECOLOR:
     HBRUSH ItemBrush;
     switch (ItemInfo.itemID)
      {
       case 0:
         ItemBrush = CreateSolidBrush (LINE_BLACK);
         break;
       case 1:
         ItemBrush = CreateSolidBrush (LINE_RED); 
         break;
       case 2:
         ItemBrush = CreateSolidBrush (LINE_GREEN);
         break;
       case 3:
         ItemBrush = CreateSolidBrush (LINE_BLUE);
         break;
      }
     FillRect (ItemInfo.hDC, &ItemArea, ItemBrush);
     DeleteObject (ItemBrush);
     break;

   case ID_GRAPH_CURVESTYLE:
     HPEN LinePen;
     switch (ItemInfo.itemID)
      {
       case 0:
	 LinePen = CreatePen (PS_SOLID, 1, LINE_BLACK);
         break;
       case 1:
	 LinePen = CreatePen (PS_DASH, 1, LINE_BLACK);
         break;
       case 2:
	 LinePen = CreatePen (PS_DOT, 1, LINE_BLACK);
         break;
       case 3:
	 LinePen = CreatePen (PS_DASHDOT, 1, LINE_BLACK);
         break;
       case 4:
	 LinePen = CreatePen (PS_DASHDOTDOT, 1, LINE_BLACK);
         break;
       case 5:
	 LinePen = NULL;
         break;
      }
     if (LinePen != NULL)
      {
       HPEN PreviousPen = SelectObject (ItemInfo.hDC, LinePen);
       int CenterY = ItemArea.top + (ItemArea.bottom - ItemArea.top) / 2;
       MoveTo (ItemInfo.hDC, ItemArea.left, CenterY);
       LineTo (ItemInfo.hDC, ItemArea.right, CenterY);
       SelectObject (ItemInfo.hDC, PreviousPen);
       DeleteObject (LinePen);
      }
     else
      {
       HFONT PreviousFont = SelectObject (ItemInfo.hDC, GetStockObject (SYSTEM_FONT));
       DrawText (ItemInfo.hDC, "None", -1, &(ItemInfo.rcItem), DT_CENTER | DT_SINGLELINE);
       SelectObject (ItemInfo.hDC, PreviousFont);
      }
     break;

   case ID_GRAPH_DOTSTYLE:
     HBITMAP DotBitmap;
     switch (ItemInfo.itemID)
      {
       case 0:
         DotBitmap = LoadBitmap (GetApplication ()->hInstance, "DotHollowSquare");
         break;
       case 1:
         DotBitmap = LoadBitmap (GetApplication ()->hInstance, "DotHollowCircle");
         break;
       case 2:
         DotBitmap = LoadBitmap (GetApplication ()->hInstance, "DotHollowTriangle");
         break;
       case 3:
         DotBitmap = LoadBitmap (GetApplication ()->hInstance, "DotFilledSquare");
         break;
       case 4:
         DotBitmap = LoadBitmap (GetApplication ()->hInstance, "DotFilledCircle");
         break;
       case 5:
         DotBitmap = LoadBitmap (GetApplication ()->hInstance, "DotFilledTriangle");
         break;
       case 6:
         DotBitmap = LoadBitmap (GetApplication ()->hInstance, "DotAsterisk");
         break;
       case 7:
	 DotBitmap = NULL;
         break;
      }
     if (DotBitmap != NULL)
      {
       HDC BitmapDC = CreateCompatibleDC (ItemInfo.hDC);
       SelectObject (BitmapDC, DotBitmap);
       int UpperLeftX = (ItemArea.left + (ItemArea.right - ItemArea.left) / 2) - 3;
       int UpperLeftY = (ItemArea.top + (ItemArea.bottom - ItemArea.top) / 2) - 3;
       BitBlt (ItemInfo.hDC, UpperLeftX, UpperLeftY, 7, 7, BitmapDC, 0, 0, SRCCOPY);
       DeleteDC (BitmapDC);
       DeleteObject (DotBitmap);
      }
     else
      {
       HFONT PreviousFont = SelectObject (ItemInfo.hDC, GetStockObject (SYSTEM_FONT));
       DrawText (ItemInfo.hDC, "None", -1, &(ItemInfo.rcItem), DT_CENTER | DT_SINGLELINE);
       SelectObject (ItemInfo.hDC, PreviousFont);
      }
     break;
  }

 SetBkMode (ItemInfo.hDC, PreviousBkMode);

 if ((ItemInfo.itemState & ODS_SELECTED) == ODS_SELECTED)
   DrawFocusRect (ItemInfo.hDC, &(ItemInfo.rcItem));
}

void TStyleComboBox::ODAFocus (DRAWITEMSTRUCT _FAR & ItemInfo)
{
 HBRUSH SelectBrush = CreateSolidBrush (GetSysColor (COLOR_HIGHLIGHT));
 if ((ItemInfo.itemState & ODS_FOCUS) == ODS_FOCUS)
   FillRect (ItemInfo.hDC, &(ItemInfo.rcItem), SelectBrush);
 else
   FillRect (ItemInfo.hDC, &(ItemInfo.rcItem), GetStockObject (WHITE_BRUSH));
 DeleteObject (SelectBrush);

 ODADrawEntire (ItemInfo);
}

void TStyleComboBox::ODASelect (DRAWITEMSTRUCT _FAR & ItemInfo)
{
 ODAFocus (ItemInfo);
}

//*************************************************************************
//
// Memberfunctions of TGraphWindow
//
//*************************************************************************

TGraphWindow::TGraphWindow (LPSTR MyTitle, int GraphHandle, TAxisRange AxisValues, int Type,
			    TDiagramTitle Title, POINT UpperLeft, POINT WindowSize,
			    char *FontName) : TWindow (NULL, MyTitle)
{
 // Set initial position and size if desired
 if (UpperLeft.x >= 0)
   Attr.X = UpperLeft.x;
 if (UpperLeft.y >= 0)
   Attr.Y = UpperLeft.y;
 if (WindowSize.x > 0)
   Attr.W = WindowSize.x;
 if (WindowSize.y > 0)
   Attr.H = WindowSize.y;

 // Initialize internal class-variables
 MyHandle               = GraphHandle;
 MemorySheet            = NULL;
 DiagramType            = Type;
 Curves                 = NULL;
 BitmapDimension.Width  = 0;
 BitmapDimension.Height = 0;
 Scale.PixelsPerXUnit   = 0.0;
 Scale.PixelsPerYUnit   = 0.0;
 Scale.Origin.x         = 0;
 Scale.Origin.y         = 0;
 SelectActive           = FALSE;
 ZoomActive             = FALSE;
 SelectArea.left        = 0;
 SelectArea.top         = 0;
 SelectArea.right       = 0;
 SelectArea.bottom      = 0;
 ZoomFactor.x           = 1.0;
 ZoomFactor.y           = 1.0;
 FitToWindow            = TRUE;
 memcpy (&AxisRange, &AxisValues, sizeof (AxisRange));
 memcpy (&DiagramTitle, &Title, sizeof (DiagramTitle));

 CreateMyFonts (FontName);
}

TGraphWindow::~TGraphWindow ()
{
 TCurve *CurveToDelete;

 // Destroy bitmap
 if (MemorySheet != NULL)
  {
   DeleteObject (MemorySheet);
   MemorySheet   = NULL;
  }

 // Delete the curves of this diagram
 while (Curves != NULL)
  {
   CurveToDelete = Curves;
   Curves        = Curves->Next;
   delete CurveToDelete;
  }

 // Delete the markers of this diagram
 Markers.freeAll ();

 // Delete the created fonts
 DeleteObject (TitleFont);
 DeleteObject (LegendFontNorm);
 DeleteObject (LegendFontTilt);

 // Remove the window from the list of graphics-windows
 GraphWindowList.free (GraphWindowList.firstThat (FindGraphWindow,
                                                  &MyHandle));
}

LPSTR TGraphWindow::GetClassName ()
{
 return ("NNGraphWindow");
}

void TGraphWindow::GetWindowClass (WNDCLASS& TheWndClass)
{
 TWindow::GetWindowClass (TheWndClass);

 TheWndClass.hIcon        = LoadIcon (GetApplication ()->hInstance, "GraphWindow");
 TheWndClass.hCursor      = NULL;
 TheWndClass.style       |= CS_NOCLOSE;
 TheWndClass.lpszMenuName = "GraphMenu";
}

void TGraphWindow::SetupWindow ()
{
 TWindow::SetupWindow ();

 Attr.Style |= WS_HSCROLL | WS_VSCROLL;
 Scroller    = new TScroller (this, 10, 10, 0, 0);

 CreateMemorySheet (ZoomFactor, TRUE);
}

void TGraphWindow::CloseGraph (TMessage& TheMessage)
{
 TWindow::WMClose (TheMessage);
}

void TGraphWindow::WMSize (TMessage& TheMessage)
{
 TWindow::WMSize (TheMessage);

 if (!IsIconic (HWindow))
   CreateMemorySheet (ZoomFactor, FALSE);
}

void TGraphWindow::WMLButtonDown (TMessage& TheMessage)
{
 if (ZoomActive)
  {
   SelectActive      = TRUE;
   SelectArea.left   = TheMessage.LP.Lo;
   SelectArea.top    = TheMessage.LP.Hi;
   SelectArea.right  = TheMessage.LP.Lo;
   SelectArea.bottom = TheMessage.LP.Hi;
  }

 TWindow::WMLButtonDown (TheMessage);
}

void TGraphWindow::ResizeSelect (TMessage& TheMessage)
{
 if (SelectActive)
  {
   HDC    ScreenDC      = GetDC (HWindow);
   HPEN   SelectPen     = CreatePen (PS_DOT, 1, RGB (0, 0, 0));
   HPEN   PreviousPen   = SelectObject (ScreenDC, SelectPen);
   HBRUSH NullBrush     = GetStockObject (NULL_BRUSH);
   HBRUSH PreviousBrush = SelectObject (ScreenDC, NullBrush);
   HDC    MemorySheetDC = CreateCompatibleDC (ScreenDC);

   SelectObject (MemorySheetDC, MemorySheet);

   BitBlt (ScreenDC, 0, 0, BitmapDimension.Width, BitmapDimension.Height,
	   MemorySheetDC, 0, 0, SRCCOPY);

   SelectArea.right  = TheMessage.LP.Lo;
   SelectArea.bottom = TheMessage.LP.Hi;

   Rectangle (ScreenDC, SelectArea.left, SelectArea.top, SelectArea.right,
	      SelectArea.bottom);

   SelectObject (ScreenDC, PreviousPen);
   SelectObject (ScreenDC, PreviousBrush);
   DeleteObject (SelectPen);
   ReleaseDC (HWindow, ScreenDC);
   DeleteDC (MemorySheetDC);
   InvalidateRect (HWindow, NULL, FALSE);
   UpdateWindow (HWindow);
  }
 else if (ZoomActive)
   SetCursor (LoadCursor (GetApplication ()->hInstance, "ZoomCursor"));
 else
   SetCursor (LoadCursor (NULL, IDC_ARROW));
}

void TGraphWindow::EndSelect (TMessage& TheMessage)
{
 if (!ZoomActive)
   return;

 HMENU Menu = GetMenu (HWindow);

 SelectArea.right  = TheMessage.LP.Lo;
 SelectArea.bottom = TheMessage.LP.Hi;
 SelectActive      = FALSE;

 // Transform the device coordinates into logical coordinates!
 HDC WindowDC = GetDC (HWindow); 
 DPtoLP (WindowDC, (LPPOINT) &SelectArea, 2);
 ReleaseDC (HWindow, WindowDC);
  
 int SelectedWidth  = abs (SelectArea.right  - SelectArea.left);
 int SelectedHeight = abs (SelectArea.bottom - SelectArea.top);
 SelectedWidth  = (SelectedWidth  > 0) ? SelectedWidth  : 1;
 SelectedHeight = (SelectedHeight > 0) ? SelectedHeight : 1;

 // Set new zoom-factor
 if ((GetMenuState (Menu, CM_ZOOMGRAPH,  MF_BYCOMMAND) & MF_CHECKED) != 0)
  {
   ZoomFactor.x = (float) BitmapDimension.Width  / (float) SelectedWidth;
   ZoomFactor.y = (float) BitmapDimension.Height / (float) SelectedHeight;
  }
 else if ((GetMenuState (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND) & MF_CHECKED) != 0)
  {
   ZoomFactor.x = (float) BitmapDimension.Width / (float) SelectedWidth;
   ZoomFactor.y = 1.0;
  }
 else if ((GetMenuState (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND) & MF_CHECKED) != 0)
  {
   ZoomFactor.x = 1.0;
   ZoomFactor.y = (float) BitmapDimension.Height / (float) SelectedHeight;
  }
 else
  {
   ZoomFactor.x = 1.0;
   ZoomFactor.y = 1.0;
  }
   
 SetCursor (LoadCursor (NULL, IDC_ARROW));
 ClipCursor (NULL);
 ZoomActive = FALSE;

 POINT Corner;
 Corner.x = min (SelectArea.left, SelectArea.right);
 Corner.y = min (SelectArea.top, SelectArea.bottom);
 CreateMemorySheet (ZoomFactor, FALSE);
 Scroller->ScrollTo ((long) ((float) Corner.x * ZoomFactor.x / (float) Scroller->XUnit),
		     (long) ((float) Corner.y * ZoomFactor.y / (float) Scroller->YUnit));
}

void TGraphWindow::CancelSelect ()
{
 if (ZoomActive)
  {
   HMENU Menu = GetMenu (HWindow);

   SetCursor (LoadCursor (NULL, IDC_ARROW));
   ClipCursor (NULL);
   CheckMenuItem (Menu, CM_ZOOMGRAPH,  MF_BYCOMMAND | MF_UNCHECKED);
   CheckMenuItem (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND | MF_UNCHECKED);
   CheckMenuItem (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND | MF_UNCHECKED);
   SelectActive = FALSE;
   ZoomActive   = FALSE;
  }
}

void TGraphWindow::CopyToClipboard ()
{
 BITMAPINFOHEADER DIBInfo;

 HDC ScreenDC = GetDC (HWindow);

 HDC MemorySheetDC = CreateCompatibleDC (ScreenDC);
 SelectObject (MemorySheetDC, MemorySheet);

 // Create a new bitmap, as it will be owned by the clipboard. Access to
 // data sent to the clipboard may not be possible if another application
 // puts its own data into the clipboard.
 DIBInfo.biSize          = sizeof (DIBInfo);;
 DIBInfo.biWidth         = BitmapDimension.Width;
 DIBInfo.biHeight        = BitmapDimension.Height;
 DIBInfo.biPlanes        = 1;
 DIBInfo.biBitCount      = 8;       // 256 colors possible
 DIBInfo.biCompression   = BI_RGB;  // No compression
 DIBInfo.biSizeImage     = 0;
 DIBInfo.biXPelsPerMeter = 0;
 DIBInfo.biYPelsPerMeter = 0;
 DIBInfo.biClrUsed       = 0;
 DIBInfo.biClrImportant  = 0;

 HBITMAP ClipboardData = CreateDIBitmap (ScreenDC, &DIBInfo, 0, NULL, NULL, DIB_RGB_COLORS);
 if (ClipboardData == NULL)
  {
   TErrorMessage BoxMessage, BoxTitle;
   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
               sizeof (BoxTitle));
   LoadString (GetApplication ()->hInstance, ERR_NO_COPY, BoxMessage,
	       sizeof (BoxMessage));
   MessageBeep (-1);
   BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONSTOP | MB_OK);
   ReleaseDC (HWindow, ScreenDC);
   DeleteDC (MemorySheetDC);
   return;
  }

 HDC ClipboardDC = CreateCompatibleDC (ScreenDC);
 SelectObject (ClipboardDC, ClipboardData);
 BitBlt (ClipboardDC, 0, 0, BitmapDimension.Width, BitmapDimension.Height,
	 MemorySheetDC, 0, 0, SRCCOPY);
 ReleaseDC (HWindow, ScreenDC);
 DeleteDC (ClipboardDC);
 DeleteDC (MemorySheetDC);

 if (!OpenClipboard (HWindow))
  {
   TErrorMessage BoxMessage, BoxTitle;
   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
	       sizeof (BoxTitle));
   LoadString (GetApplication ()->hInstance, ERR_CLP_IN_USE, BoxMessage,
	       sizeof (BoxMessage));
   MessageBeep (-1);
   BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONSTOP | MB_OK);
   DeleteObject (ClipboardData);
  }
 else
  {
   EmptyClipboard (); 
   SetClipboardData (CF_BITMAP, ClipboardData);
   // The bitmap is freed by the clipboard!
   CloseClipboard ();
  }
}

void TGraphWindow::Zoom ()
{
 union
  {
   RECT ClientArea;
   struct
    {
     POINT UpperLeft;
     POINT LowerRight;
    } ClientCorners;
  } ClipArea;

 HMENU Menu = GetMenu (HWindow);

 if (((GetMenuState (Menu, CM_ZOOMGRAPH,  MF_BYCOMMAND) & MF_CHECKED) == 0) &&
     ((GetMenuState (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND) & MF_CHECKED) == 0) &&
     ((GetMenuState (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND) & MF_CHECKED) == 0)) 
  {
   CheckMenuItem (Menu, CM_ZOOMGRAPH, MF_BYCOMMAND | MF_CHECKED);
   // Change cursor to spy-glass and capture it within this window
   GetClientRect (HWindow, &(ClipArea.ClientArea));
   ClientToScreen (HWindow, &(ClipArea.ClientCorners.UpperLeft));
   ClientToScreen (HWindow, &(ClipArea.ClientCorners.LowerRight));
   ClipCursor (&(ClipArea.ClientArea));
   SetCursor (LoadCursor (GetApplication ()->hInstance, "ZoomCursor"));
   ZoomActive = TRUE;
  }
 else if (GetMenuState (Menu, CM_ZOOMGRAPH, MF_BYCOMMAND) == MF_CHECKED)
  {
   CheckMenuItem (Menu, CM_ZOOMGRAPH, MF_BYCOMMAND | MF_UNCHECKED);
   ZoomFactor.x = 1.0;
   ZoomFactor.y = 1.0;
   CreateMemorySheet (ZoomFactor, FALSE);
  }
 else
   MessageBeep (-1);

}

void TGraphWindow::ZoomX ()
{
 union
  {
   RECT ClientArea;
   struct
    {
     POINT UpperLeft;
     POINT LowerRight;
    } ClientCorners;
  } ClipArea;

 HMENU Menu = GetMenu (HWindow);

 if (((GetMenuState (Menu, CM_ZOOMGRAPH,  MF_BYCOMMAND) & MF_CHECKED) == 0) &&
     ((GetMenuState (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND) & MF_CHECKED) == 0) &&
     ((GetMenuState (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND) & MF_CHECKED) == 0)) 
  {
   CheckMenuItem (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND | MF_CHECKED);
   // Change cursor to spy-glass and capture it within this window
   GetClientRect (HWindow, &(ClipArea.ClientArea));
   ClientToScreen (HWindow, &(ClipArea.ClientCorners.UpperLeft));
   ClientToScreen (HWindow, &(ClipArea.ClientCorners.LowerRight));
   ClipCursor (&(ClipArea.ClientArea));
   SetCursor (LoadCursor (GetApplication ()->hInstance, "ZoomCursor"));
   ZoomActive = TRUE;
  }
 else if (GetMenuState (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND) == MF_CHECKED)
  {
   CheckMenuItem (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND | MF_UNCHECKED);
   ZoomFactor.x = 1.0;
   ZoomFactor.y = 1.0;
   CreateMemorySheet (ZoomFactor, FALSE);
  }
 else
   MessageBeep (-1);
}

void TGraphWindow::ZoomY ()
{
 union
  {
   RECT ClientArea;
   struct
    {
     POINT UpperLeft;
     POINT LowerRight;
    } ClientCorners;
  } ClipArea;

 HMENU Menu = GetMenu (HWindow);

 if (((GetMenuState (Menu, CM_ZOOMGRAPH,  MF_BYCOMMAND) & MF_CHECKED) == 0) &&
     ((GetMenuState (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND) & MF_CHECKED) == 0) &&
     ((GetMenuState (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND) & MF_CHECKED) == 0))
  {
   CheckMenuItem (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND | MF_CHECKED);
   // Change cursor to spy-glass and capture it within this window
   GetClientRect (HWindow, &(ClipArea.ClientArea));
   ClientToScreen (HWindow, &(ClipArea.ClientCorners.UpperLeft));
   ClientToScreen (HWindow, &(ClipArea.ClientCorners.LowerRight));
   ClipCursor (&(ClipArea.ClientArea));
   SetCursor (LoadCursor (GetApplication ()->hInstance, "ZoomCursor"));
   ZoomActive = TRUE;
  }
 else if (GetMenuState (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND) == MF_CHECKED)
  {
   CheckMenuItem (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND | MF_UNCHECKED);
   ZoomFactor.x = 1.0;
   ZoomFactor.y = 1.0;
   CreateMemorySheet (ZoomFactor, FALSE);
  }
 else
   MessageBeep (-1);
}

void TGraphWindow::DisplayOptions ()
{
 TCurve *CurrentCurve;

 BOOL XLogPossible = !((AxisRange.X.Start <= 0) || (AxisRange.X.End <= 0));
 BOOL YLogPossible = !((AxisRange.Y.Start <= 0) || (AxisRange.Y.End <= 0));

 if (GetApplication ()->ExecDialog (new TDisplayDlg (&FitToWindow, &FixedDimension,
						     &DiagramType, XLogPossible,
						     YLogPossible, &AxisRange, this)) == IDOK)
  {
   if (!FitToWindow)
     SetWindowPos (HWindow, NULL, 0, 0,
		   FixedDimension.Width + 2 * GetSystemMetrics (SM_CXFRAME),
		   FixedDimension.Height + 2 * GetSystemMetrics (SM_CYFRAME) +
					       GetSystemMetrics (SM_CYCAPTION) +
					       GetSystemMetrics (SM_CYMENU),
		   SWP_NOMOVE | SWP_NOZORDER);

   if ((DiagramType & X_AUTOSCALE) != 0)
     for (CurrentCurve = Curves; CurrentCurve != NULL; CurrentCurve = CurrentCurve->Next)
      {
       if (CurrentCurve->LowerX () < AxisRange.X.Start)
         AxisRange.X.Start = CurrentCurve->LowerX ();
       if (CurrentCurve->UpperX () > AxisRange.X.End)
         AxisRange.X.End = CurrentCurve->UpperX ();
      }
   if ((DiagramType & Y_AUTOSCALE) != 0)
     for (CurrentCurve = Curves; CurrentCurve != NULL; CurrentCurve = CurrentCurve->Next)
      {
       if (CurrentCurve->LowerY () < AxisRange.Y.Start)
         AxisRange.Y.Start = CurrentCurve->LowerY ();
       if (CurrentCurve->UpperY () > AxisRange.Y.End)
         AxisRange.Y.End = CurrentCurve->UpperY ();
      }

   CreateMemorySheet (ZoomFactor, TRUE);
  }
}

void TGraphWindow::CalcMean ()
{
 TErrorMessage BoxMessage, BoxTitle;
 TYValueArray  *YValues;
 HANDLE        YValueMem;
 char          ResultString [MAX_RESULT_BUFFER];
 char          SingleResult [50];

 memset (ResultString, 0, sizeof (ResultString));

 for (TCurve *CurrentCurve = Curves; CurrentCurve != NULL;
      CurrentCurve = CurrentCurve->Next)
   if (CurrentCurve->Mark)
    {
     YValueMem = CurrentCurve->RetrieveYValues ();
     if (YValueMem == NULL)
      {
       LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		   sizeof (BoxTitle));
       LoadString (GetApplication ()->hInstance, ERR_CALC_MEM, BoxMessage,
		   sizeof (BoxMessage));
       MessageBeep (-1);
       BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                       MB_ICONINFORMATION | MB_OK);
       return;
      }
     YValues = (TYValueArray *) GlobalLock (YValueMem);

     if (YValues->ElementCount == 0)
      {
       LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		   sizeof (BoxTitle));
       LoadString (GetApplication ()->hInstance, ERR_CALC_NO_VAL,
                   BoxMessage, sizeof (BoxMessage));
       MessageBeep (-1);
       BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                       MB_ICONINFORMATION | MB_OK);
      }
     else
      {
       sprintf (SingleResult, "%s:  %.3f,   %i values\n",
		CurrentCurve->Name,
		calc_mean (YValues->Values, YValues->ElementCount),
		YValues->ElementCount);
       strcat (ResultString, SingleResult);
      }

     GlobalUnlock (YValueMem);
     GlobalFree (YValueMem);
    }

 if (strlen (ResultString) > 0)
   GetApplication ()->ExecDialog (new TResultBox ("Mean", ResultString,
				     "Graphics-Window Mean-values", this));
}

void TGraphWindow::CalcSDev ()
{
 TErrorMessage BoxMessage, BoxTitle;
 TYValueArray  *YValues;
 HANDLE        YValueMem;
 char          ResultString [MAX_RESULT_BUFFER];
 char          SingleResult [50];

 memset (ResultString, 0, sizeof (ResultString));

 for (TCurve *CurrentCurve = Curves; CurrentCurve != NULL;
      CurrentCurve = CurrentCurve->Next)
   if (CurrentCurve->Mark)
    {
     YValueMem = CurrentCurve->RetrieveYValues ();
     if (YValueMem == NULL)
      {
       LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		   sizeof (BoxTitle));
       LoadString (GetApplication ()->hInstance, ERR_CALC_MEM, BoxMessage,
		   sizeof (BoxMessage));
       MessageBeep (-1);
       BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                       MB_ICONINFORMATION | MB_OK);
       return;
      }
     YValues = (TYValueArray *) GlobalLock (YValueMem);

     if (YValues->ElementCount == 0)
      {
       LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		   sizeof (BoxTitle));
       LoadString (GetApplication ()->hInstance, ERR_CALC_NO_VAL,
                   BoxMessage, sizeof (BoxMessage));
       MessageBeep (-1);
       BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                       MB_ICONINFORMATION | MB_OK);
      }
     else
      {
       sprintf (SingleResult, "%s:  %.3f,   %i values\n",
		CurrentCurve->Name,
		calc_sdev (YValues->Values, YValues->ElementCount),
		YValues->ElementCount);
       strcat (ResultString, SingleResult);
      }

     GlobalUnlock (YValueMem);
     GlobalFree (YValueMem);
    }

 if (strlen (ResultString) > 0)
   GetApplication ()->ExecDialog (new TResultBox ("Standard Deviation",
     ResultString, "Graphics-Window Standard-deviation-values", this));
}

void TGraphWindow::CalcMSE ()
{
 TErrorMessage BoxMessage, BoxTitle;
 TYValueArray  *Y1Values, *Y2Values;
 TCurve        *Curve1 = NULL;
 HANDLE        Y1ValueMem, Y2ValueMem;
 BOOL          CalcDone = FALSE;
 int           Elements;
 char          ResultString [MAX_RESULT_BUFFER];
 char          SingleResult [50];

 memset (ResultString, 0, sizeof (ResultString));

 for (TCurve *CurrentCurve = Curves; CurrentCurve != NULL;
      CurrentCurve = CurrentCurve->Next)
   if (CurrentCurve->Mark)
     if (Curve1 == NULL)
      {
       CalcDone   = TRUE;
       Curve1     = CurrentCurve;
       Y1ValueMem = CurrentCurve->RetrieveYValues ();
       if (Y1ValueMem == NULL)
        {
         LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		      sizeof (BoxTitle));
         LoadString (GetApplication ()->hInstance, ERR_CALC_MEM, BoxMessage,
		     sizeof (BoxMessage));
         MessageBeep (-1);
         BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                         MB_ICONINFORMATION | MB_OK);
         return;
        }
       Y1Values = (TYValueArray *) GlobalLock (Y1ValueMem);
      }
     else
      {
       Y2ValueMem = CurrentCurve->RetrieveYValues ();
       if (Y2ValueMem == NULL)
        {
         LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		      sizeof (BoxTitle));
         LoadString (GetApplication ()->hInstance, ERR_CALC_MEM, BoxMessage,
		     sizeof (BoxMessage));
         MessageBeep (-1);
         BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                         MB_ICONINFORMATION | MB_OK);
         GlobalUnlock (Y1ValueMem);
         GlobalFree (Y1ValueMem);
	 return;
        }
       Y2Values = (TYValueArray *) GlobalLock (Y2ValueMem);

       if (Y1Values->ElementCount != Y2Values->ElementCount)
	{
	 LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		     sizeof (BoxTitle));
	 LoadString (GetApplication ()->hInstance, ERR_CALC_DIFF_COUNT,
		     BoxMessage, sizeof (BoxMessage));
         MessageBeep (-1);
	 BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                         MB_ICONINFORMATION | MB_OK);
	}

       Elements = min (Y1Values->ElementCount, Y2Values->ElementCount);
       if (Elements == 0)
        {
         LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		     sizeof (BoxTitle));
         LoadString (GetApplication ()->hInstance, ERR_CALC_NO_VAL,
                     BoxMessage, sizeof (BoxMessage));
         MessageBeep (-1);
         BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                         MB_ICONINFORMATION | MB_OK);
        }
       else
        {
	 sprintf (SingleResult, "%s - %s:  %.3f,   %i values\n",
		  Curve1->Name, CurrentCurve->Name,
		  calc_mse (Y1Values->Values, Y2Values->Values, Elements),
		  Elements);
	 strcat (ResultString, SingleResult);
	}

       GlobalUnlock (Y2ValueMem);
       GlobalFree (Y2ValueMem);
      }

 GlobalUnlock (Y1ValueMem);
 GlobalFree (Y1ValueMem);

 if (CalcDone)
   if (strlen (ResultString) > 0)
     GetApplication ()->ExecDialog (new TResultBox ("Mean Square Error",
       ResultString, "Graphics-Window Mean-Square-Error", this));
 else
  {
   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
	       sizeof (BoxTitle));
   LoadString (GetApplication ()->hInstance, ERR_NO_SELECT, BoxMessage,
	       sizeof (BoxMessage));
   MessageBeep (-1);
   BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
                   MB_ICONINFORMATION | MB_OK);
  }
}

void TGraphWindow::SelectAll ()
{
 HMENU GraphMenu  = GetMenu (HWindow);
 int   CurveCount = 0;

 for (TCurve *CurrentCurve = Curves; CurrentCurve != NULL;
      CurrentCurve = CurrentCurve->Next)
  {
   CurrentCurve->Mark = TRUE;
   CheckMenuItem (GraphMenu, CM_CURVE_0 + CurveCount++,
                  MF_BYCOMMAND | MF_CHECKED); 
  }
 EnableMenuItem (GraphMenu, CM_SELECTALL, MF_BYCOMMAND | MF_GRAYED);
 EnableMenuItem (GraphMenu, CM_UNSELECTALL, MF_BYCOMMAND | MF_ENABLED); 
 EnableMenuItem (GraphMenu, CM_GRAPHMEAN, MF_BYCOMMAND | MF_ENABLED);
 EnableMenuItem (GraphMenu, CM_GRAPHSDEV, MF_BYCOMMAND | MF_ENABLED);
 EnableMenuItem (GraphMenu, CM_GRAPHMSE, MF_BYCOMMAND | MF_ENABLED);
}

void TGraphWindow::UnselectAll ()
{
 HMENU GraphMenu  = GetMenu (HWindow);
 int   CurveCount = 0;

 for (TCurve *CurrentCurve = Curves; CurrentCurve != NULL;
      CurrentCurve = CurrentCurve->Next)
  {
   CurrentCurve->Mark = FALSE;
   CheckMenuItem (GraphMenu, CM_CURVE_0 + CurveCount++,
                  MF_BYCOMMAND | MF_UNCHECKED); 
  }
 EnableMenuItem (GraphMenu, CM_SELECTALL, MF_BYCOMMAND | MF_ENABLED);
 EnableMenuItem (GraphMenu, CM_UNSELECTALL, MF_BYCOMMAND | MF_GRAYED); 
 EnableMenuItem (GraphMenu, CM_GRAPHMEAN, MF_BYCOMMAND | MF_GRAYED);
 EnableMenuItem (GraphMenu, CM_GRAPHSDEV, MF_BYCOMMAND | MF_GRAYED);
 EnableMenuItem (GraphMenu, CM_GRAPHMSE, MF_BYCOMMAND | MF_GRAYED);
}

void TGraphWindow::Select0 ()
{
 MarkCurve (CM_CURVE_0);
}

void TGraphWindow::Select1 ()
{
 MarkCurve (CM_CURVE_1);
}

void TGraphWindow::Select2 ()
{
 MarkCurve (CM_CURVE_2);
}

void TGraphWindow::Select3 ()
{
 MarkCurve (CM_CURVE_3);
}

void TGraphWindow::Select4 ()
{
 MarkCurve (CM_CURVE_4);
}

void TGraphWindow::Select5 ()
{
 MarkCurve (CM_CURVE_5);
}

void TGraphWindow::Select6 ()
{
 MarkCurve (CM_CURVE_6);
}

void TGraphWindow::Select7 ()
{
 MarkCurve (CM_CURVE_7);
}

void TGraphWindow::Select8 ()
{
 MarkCurve (CM_CURVE_8);
}

void TGraphWindow::Select9 ()
{
 MarkCurve (CM_CURVE_9);
}

void TGraphWindow::Load ()
{
 char          CurveName [MAX_CURV_NAME + 1];
 char          File1 [MAXPATH], File2 [MAXPATH], TextLine [255];
 float         XStart     = AxisRange.X.Start;
 float         XIncrement = AxisRange.X.Step;
 float         XValue, YValue;
 int           CurveCount = 1;
 int           FileType, CurveStyle, CurveHandle;
 COLORREF      CurveColor;
 BOOL          XValid = TRUE;
 BOOL          ValuesOk;
 ifstream      *InputFile1, *InputFile2;
 TErrorMessage BoxMessage, BoxTitle;

 for (TCurve *CurrentCurve = Curves; CurrentCurve != NULL;
      CurrentCurve = CurrentCurve->Next)
   CurveCount++;
 sprintf (CurveName, "Curve%02i", CurveCount);

 if (GetApplication ()->ExecDialog (new TLoadDlg (CurveName, &CurveColor,
                                                  &CurveStyle, File1, File2,
                                                  &XStart, &XIncrement,
						  &FileType, this))
     == IDOK)
  {
   if ((CurveHandle = NewCurve (CurveName, CurveColor, CurveStyle,
			        XIncrement)) < 0)
    {
     LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		 sizeof (BoxTitle));
     LoadString (GetApplication ()->hInstance, ERR_NO_CURVE, BoxMessage,
		 sizeof (BoxMessage));
     MessageBeep (-1);
     BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
		     MB_ICONSTOP | MB_OK);
     return;
    }
   strcpy (TextLine, "");
   switch (FileType)
    {
     case OPEN_FILE_ONLY_Y:
       XValid = FALSE;

     case OPEN_FILE_COMBINED:
       InputFile1 = new ifstream (File1, ios::in | ios::nocreate);
       while (((strlen (TextLine) == 0) || (TextLine [0] == ';')) &&
              !(InputFile1->eof()))
	 InputFile1->getline (TextLine, 255);
       if (InputFile1->eof ())
	{
	 LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
		     sizeof (BoxTitle));
	 LoadString (GetApplication ()->hInstance, ERR_EMPTY_FILE,
		     BoxMessage, sizeof (BoxMessage));
	 MessageBeep (-1);
	 BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
			 MB_ICONINFORMATION | MB_OK);
	 InputFile1->close ();
	 return;  
        }
       if (FileType == OPEN_FILE_ONLY_Y)
	{
         XValue = XStart;
	 ValuesOk = sscanf (TextLine, "%f", &YValue) == 1;
        }
       else
	 ValuesOk = sscanf (TextLine, "%f %f", &XValue, &YValue) == 2;
       if (!ValuesOk)
	{
	 LoadString (GetApplication ()->hInstance, STR_BOX_TITLE,
	             BoxTitle, sizeof (BoxTitle));
	 LoadString (GetApplication ()->hInstance, ERR_ILL_VALUE,
	             BoxMessage, sizeof (BoxMessage));
         MessageBeep (-1);
         BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
			 MB_ICONSTOP | MB_OK);
	 InputFile1->close ();
	 DrawCurves ();
         return;
	}
       if (AddCoordinate (XValue, YValue, TRUE, CurveHandle, FALSE) < 0)
        {
	 LoadString (GetApplication ()->hInstance, STR_BOX_TITLE,
	             BoxTitle, sizeof (BoxTitle));
	 LoadString (GetApplication ()->hInstance, ERR_NOT_ALL_READ,
	             BoxMessage, sizeof (BoxMessage));
         MessageBeep (-1);
         BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
			 MB_ICONSTOP | MB_OK);
	 InputFile1->close ();
	 DrawCurves ();
         return;
	}
       while (!(InputFile1->eof ()))
	{
         // Force read of a new line
         strcpy (TextLine , "");
	 while (((strlen (TextLine) == 0) || (TextLine [0] == ';')) &&
                !(InputFile1->eof()))
	   InputFile1->getline (TextLine, 255);
	 if (InputFile1->eof ())
           break;
	 if (FileType == OPEN_FILE_ONLY_Y)
	   ValuesOk = sscanf (TextLine, "%f", &YValue) == 1;
	 else
           ValuesOk = sscanf (TextLine, "%f %f", &XValue, &YValue) == 2;
         if (!ValuesOk)
    	  {
	   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE,
	               BoxTitle, sizeof (BoxTitle));
	   LoadString (GetApplication ()->hInstance, ERR_ILL_VALUE,
	               BoxMessage, sizeof (BoxMessage));
           MessageBeep (-1);
           BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
			   MB_ICONSTOP | MB_OK);
           break;
	  }
	 if (AddCoordinate (XValue, YValue, XValid, CurveHandle, FALSE) < 0)
          {
	   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE,
	               BoxTitle, sizeof (BoxTitle));
	   LoadString (GetApplication ()->hInstance, ERR_NOT_ALL_READ,
	               BoxMessage, sizeof (BoxMessage));
           MessageBeep (-1);
           BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
			   MB_ICONSTOP | MB_OK);
           break;
          }
	}
       InputFile1->close ();
       DrawCurves ();
       break;

     case OPEN_FILE_SPLIT_YX:
       // Just swap the filenames, then continue with OPEN_FILE_SPLIT_XY
       char Buffer [MAXPATH];
       strcpy (Buffer, File1);
       strcpy (File1, File2);
       strcpy (File2, Buffer);

     case OPEN_FILE_SPLIT_XY:
       InputFile1 = new ifstream (File1, ios::in | ios::nocreate);
       InputFile2 = new ifstream (File2, ios::in | ios::nocreate);
       while (!(InputFile1->eof ()))
	{
	 // Force read of a new line
         strcpy (TextLine , "");
	 while (((strlen (TextLine) == 0) || (TextLine [0] == ';')) &&
                !(InputFile1->eof()))
	   InputFile1->getline (TextLine, 255);
	 if (InputFile1->eof ())
           break;
	 ValuesOk = sscanf (TextLine, "%f", &XValue) == 1;
	 // Force read of a new line
         strcpy (TextLine , "");
	 while (((strlen (TextLine) == 0) || (TextLine [0] == ';')) &&
                !(InputFile2->eof()))
	   InputFile2->getline (TextLine, 255);
	 if (InputFile2->eof ())
	  {
	   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE,
	               BoxTitle, sizeof (BoxTitle));
	   LoadString (GetApplication ()->hInstance, ERR_DIFF_COUNT,
	               BoxMessage, sizeof (BoxMessage));
           MessageBeep (-1);
           BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
			   MB_ICONSTOP | MB_OK);
           break;
	  }
	 ValuesOk = (sscanf (TextLine, "%f", &YValue) == 1) ? ValuesOk :
	            FALSE;
	 if (!ValuesOk)
	  {
	   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE,
	               BoxTitle, sizeof (BoxTitle));
	   LoadString (GetApplication ()->hInstance, ERR_ILL_VALUE,
	               BoxMessage, sizeof (BoxMessage));
           MessageBeep (-1);
           BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
			   MB_ICONSTOP | MB_OK);
           break;
	  }
	 if (AddCoordinate (XValue, YValue, TRUE, CurveHandle, FALSE) < 0)
	  {
	   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE,
	               BoxTitle, sizeof (BoxTitle));
	   LoadString (GetApplication ()->hInstance, ERR_NOT_ALL_READ,
	               BoxMessage, sizeof (BoxMessage));
           MessageBeep (-1);
           BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle,
			   MB_ICONSTOP | MB_OK);
           break;
          }
	}
       InputFile1->close ();
       InputFile2->close ();
       DrawCurves ();
       break;
    }
  }
}

void TGraphWindow::Save ()
{
 char    OutputDirectory [MAXPATH];
 char    CurrentDirectory [MAXPATH];
 char    FileName [MAXPATH];
 char    CleanName [MAXPATH];
 int     SaveType;
 int     OrgNameIndex;
 int     CleanNameIndex;
 HCURSOR PreviousCursor;
 BOOL    SelectedOnly = FALSE;
 TCurve  *CurrentCurve;

 strcpy (OutputDirectory, "A:\\");
 OutputDirectory [0] = getdisk () + 'A';
 getcurdir (0, CurrentDirectory);
 strcat (OutputDirectory, CurrentDirectory);

 for (CurrentCurve = Curves; CurrentCurve != NULL; CurrentCurve = CurrentCurve->Next)
   SelectedOnly |= CurrentCurve->Mark;

 if (GetApplication ()->ExecDialog (new TSaveDlg (OutputDirectory, &SaveType, &SelectedOnly,
                                                  this)) == IDOK)
  {
   switch (SaveType)
    {
     case SAVE_FILE_VALUES:
       ofstream *OutputFile;

       PreviousCursor = SetCursor (LoadCursor (NULL, IDC_WAIT));
       // Save all available curves
       for (CurrentCurve = Curves; CurrentCurve != NULL; CurrentCurve = CurrentCurve->Next)
	{
	 if (SelectedOnly && !CurrentCurve->Mark)
	   continue;

	 // Replace all blanks with underscores, delete all non-alphanumeric characters
	 memset (CleanName, 0, sizeof (CleanName));
	 CleanNameIndex = 0;  
	 for (OrgNameIndex = 0; OrgNameIndex < strlen (CurrentCurve->Name); OrgNameIndex++)
	   if (isalnum (CurrentCurve->Name [OrgNameIndex]))
	     CleanName [CleanNameIndex++] = CurrentCurve->Name [OrgNameIndex];
	   else if (CurrentCurve->Name [OrgNameIndex] == ' ')
             CleanName [CleanNameIndex++] = '_';

         sprintf (FileName, "%s\\%.8s.xy", OutputDirectory, CleanName);
	 OutputFile = new ofstream (FileName);
         if (OutputFile == NULL)
          {
           TErrorMessage BoxMessage, BoxTitle, FormatString;
	   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
	               sizeof (BoxTitle));
	   LoadString (GetApplication ()->hInstance, ERR_NO_SAVE, FormatString,
	               sizeof (FormatString));
           sprintf (BoxMessage, FormatString, FileName);
           MessageBeep (-1);
           BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONINFORMATION | MB_OK);
           SetCursor (PreviousCursor);
           return;  
	  }

	 char *MonthName [] = {"#", "-Jan-", "-Feb-", "-Mar-", "-Apr-", "-May-", "-Jun-",
				    "-Jul-", "-Aug-", "-Sep-", "-Okt-", "-Nov-", "-Dec-"};
	 char   TimeString [6];
	 struct time Now;
	 struct date Today;
	 gettime (&Now);
   getdate (&Today);
	 sprintf (TimeString, "%02i:%02i", Now.ti_hour, Now.ti_min);

	 *OutputFile << ";\n; Curve-name: " << CurrentCurve->Name;
	 *OutputFile << "\tSaved at: " << (int) Today.da_day << MonthName [Today.da_mon];
	 *OutputFile << Today.da_year << "  " << TimeString << "\n;\n";
         CurrentCurve->WritePoints (OutputFile);
	 OutputFile->close ();
         // Check if everything went ok
	 if (!OutputFile->good ())
	  {
	   TErrorMessage BoxMessage, BoxTitle, FormatString;
	   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle,
	               sizeof (BoxTitle));
	   LoadString (GetApplication ()->hInstance, ERR_NO_DISK_SPACE, FormatString,
	               sizeof (FormatString));
           sprintf (BoxMessage, FormatString, FileName);
           MessageBeep (-1);
           BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONINFORMATION | MB_OK);
	   remove (FileName);
	   // If save of one curve failed, the others won't work too!
           break;
	  }
	}
	SetCursor (PreviousCursor);
       break;

     case SAVE_FILE_BITMAP:
       HANDLE           DIBMemory;
       BITMAPFILEHEADER DIBFileHeader;

       PreviousCursor = SetCursor (LoadCursor (NULL, IDC_WAIT));
       if ((DIBMemory = RetrieveBitmap ()) == NULL)
	{
	 TErrorMessage BoxMessage, BoxTitle;
	 LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle, sizeof (BoxTitle));
	 LoadString (GetApplication ()->hInstance, ERR_NO_MEM_SAVE_BMP, BoxMessage,
	             sizeof (BoxMessage));
         MessageBeep (-1);
         BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONINFORMATION | MB_OK);
	 SetCursor (PreviousCursor);
         return;  
	}

       char huge *DIBInfo = (char huge *) GlobalLock (DIBMemory);

       DIBFileHeader.bfType      = (WORD) (('M' << 8) | 'B');  // "BM"
       DIBFileHeader.bfSize      = sizeof (DIBFileHeader) + GlobalSize (DIBMemory);
       DIBFileHeader.bfReserved1 = 0;
       DIBFileHeader.bfReserved2 = 0;
       DIBFileHeader.bfOffBits   = sizeof (DIBFileHeader) +
                                   ((BITMAPINFOHEADER *) DIBInfo)->biSize +
                                   256 * sizeof (RGBQUAD);

       // Strip "Graph: " form window-title for the filename
       // Replace all blanks with underscores, delete all non-alphanumeric characters
       memset (CleanName, 0, sizeof (CleanName));
       CleanNameIndex = 0;  
       for (OrgNameIndex = 7; OrgNameIndex < strlen (Title); OrgNameIndex++)
	 if (isalnum (Title [OrgNameIndex]))
	   CleanName [CleanNameIndex++] = Title [OrgNameIndex];
	 else if (Title [OrgNameIndex] == ' ')
	   CleanName [CleanNameIndex++] = '_';

       sprintf (FileName, "%s\\%.8s.bmp", OutputDirectory, CleanName);

       OFSTRUCT OpenStruct;

       int BitmapFile = OpenFile (FileName, &OpenStruct, OF_CREATE | OF_WRITE);
       if (BitmapFile == -1)
	{
	 TErrorMessage BoxMessage, BoxTitle, FormatString;
	 LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle, sizeof (BoxTitle));
	 LoadString (GetApplication ()->hInstance, ERR_NO_SAVE, FormatString,
	             sizeof (FormatString));
         sprintf (BoxMessage, FormatString, FileName);
         MessageBeep (-1);
         BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONINFORMATION | MB_OK);
	 GlobalUnlock (DIBMemory);
         GlobalFree   (DIBMemory);
	 SetCursor (PreviousCursor);
         return;  
	}
	 
       unsigned long BytesWritten = _lwrite (BitmapFile, (char *) &DIBFileHeader,
                                    sizeof (DIBFileHeader));
       // Write to file in chunks of no more than 32kB
       for (long SizeToWrite = GlobalSize (DIBMemory); SizeToWrite > 0; SizeToWrite -= 32767)
         if (SizeToWrite < 32767)
	   BytesWritten += _lwrite (BitmapFile, (char *) DIBInfo, SizeToWrite);
	 else
          {
	   BytesWritten += _lwrite (BitmapFile, (char *) DIBInfo, 32767);
	   DIBInfo += 32767;
          }

       _lclose (BitmapFile);
       if (BytesWritten < (sizeof (DIBFileHeader) + GlobalSize (DIBMemory)))
	{
	 TErrorMessage BoxMessage, BoxTitle, FormatString;
	 LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle, sizeof (BoxTitle));
	 LoadString (GetApplication ()->hInstance, ERR_NO_DISK_SPACE, FormatString,
	             sizeof (FormatString));
         sprintf (BoxMessage, FormatString, FileName);
         MessageBeep (-1);
         BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONINFORMATION | MB_OK);
	 remove (FileName);
	} 

       GlobalUnlock (DIBMemory);
       GlobalFree   (DIBMemory);

       SetCursor (PreviousCursor);
       break;
    }
  }
}

void TGraphWindow::Paint (HDC PaintDC, PAINTSTRUCT&)
{
 if (!SelectActive)
  {
   HDC MemorySheetDC = CreateCompatibleDC (PaintDC);
   SelectObject (MemorySheetDC, MemorySheet);

   BitBlt (PaintDC, 0, 0, BitmapDimension.Width, BitmapDimension.Height,
	   MemorySheetDC, 0, 0, SRCCOPY);

   DeleteDC (MemorySheetDC);

   SelectArea.left   = 0;
   SelectArea.top    = 0;
   SelectArea.right  = 0;
   SelectArea.bottom = 0;
  } 
}

int TGraphWindow::NewCurve (char *Name, COLORREF Color, int LineStyle,
                            float Increment)
{
 int    NewHandle;
 char   MenuEntry [MAX_CURV_NAME + 6];
 char   CurveName [MAX_CURV_NAME + 1];
 int    CurveCount = 1;
 TCurve *LastCurve = Curves;

 // Make sure the curve gets a name, even if the user didn't supply one
 if (Name != NULL)
   strcpy (CurveName, Name);
 else
   strcpy (CurveName, "");

 if (Curves == NULL)
  {
   // First curve for this diagram
   if (strlen (CurveName) == 0)
     strcpy (CurveName, "Curve01");
   NewHandle = (MyHandle << 8) | 1;
   Curves = new TCurve (NewHandle, CurveName, Color, LineStyle, Increment, NULL, this);
   if (Curves == NULL)
     // Couldn't create curve; out of memory
     return (GRAPHERR_NOMEM);
  }
 else
  {
   // Append curve to end of list
   while (LastCurve->Next != NULL)
    {
     LastCurve = LastCurve->Next;
     CurveCount++;
    }
   if (CurveCount >= MAX_CURVES)
     // No more then MAX_CURVES curves per diagram
     return (GRAPHERR_MAXCURV);
   CurveCount++;   // We're initializing a new curve!
   if (strlen (CurveName) == 0)
     sprintf (CurveName, "Curve%02i", CurveCount);
   NewHandle = LastCurve->Handle + 1;
   LastCurve->Next = new TCurve (NewHandle, CurveName, Color, LineStyle, Increment, LastCurve,
                                 this);
   if (LastCurve->Next == NULL)
     // Couldn't create curve; out of memory
     return (GRAPHERR_NOMEM);
  }

 // Get the menu-handle for the 'Curve'-menu
 HMENU GraphMenu = GetMenu (HWindow);
 HMENU CurveMenu = GetSubMenu (GraphMenu, 3);
 sprintf (MenuEntry, "&%1i   %s", CurveCount - 1, CurveName);
 AppendMenu (CurveMenu, MF_STRING, CM_CURVE_0 + CurveCount - 1, MenuEntry);
 EnableMenuItem (GraphMenu, CM_SAVEGRAPH, MF_BYCOMMAND | MF_ENABLED);
 EnableMenuItem (GraphMenu, CM_SELECTALL, MF_BYCOMMAND | MF_ENABLED);

 return (NewHandle);
}

int TGraphWindow::AddCoordinate (float XValue, float YValue, BOOL XValid,
                                 int CurveHandle, BOOL Redraw)
{
 TCurve *CurrentCurve = Curves;
 BOOL   CurveFound    = FALSE;
 while ((CurrentCurve != NULL) && !CurveFound)
   if (CurrentCurve->Handle == CurveHandle)
     CurveFound = TRUE;
   else
     CurrentCurve = CurrentCurve->Next;

 if (!CurveFound)
   return (GRAPHERR_HINVAL);

 if (XValid)
   CurrentCurve->AddCoordinate (XValue, YValue);
 else
   CurrentCurve->AddPoint (YValue);

 if ((DiagramType & X_AUTOSCALE) != 0)
  {
   if (CurrentCurve->LowerX () < AxisRange.X.Start)
     AxisRange.X.Start = CurrentCurve->LowerX ();
   if (CurrentCurve->UpperX () > AxisRange.X.End)
     AxisRange.X.End = CurrentCurve->UpperX ();
  }
 if ((DiagramType & Y_AUTOSCALE) != 0)
  {
   if (CurrentCurve->LowerY () < AxisRange.Y.Start)
     AxisRange.Y.Start = CurrentCurve->LowerY ();
   if (CurrentCurve->UpperY () > AxisRange.Y.End)
     AxisRange.Y.End = CurrentCurve->UpperY ();
  }

 if (Redraw)
   DrawCurves ();

 return (GRAPHERR_NOERR);
}

int TGraphWindow::ClearCurve (int CurveHandle)
{
 TCurve *CurrentCurve = Curves;
 BOOL   CurveFound    = FALSE;
 while ((CurrentCurve != NULL) && !CurveFound)
   if (CurrentCurve->Handle == CurveHandle)
     CurveFound = TRUE;
   else
     CurrentCurve = CurrentCurve->Next;

 if (CurveFound)
  {
   CurrentCurve->ClearPoints ();
   DrawCurves ();
   return (GRAPHERR_NOERR);
  }
 else
   return (GRAPHERR_HINVAL);
}

int TGraphWindow::SetCurveStart (int CurveHandle, float NewXStart)
{
 TCurve *CurrentCurve = Curves;
 BOOL   CurveFound    = FALSE;
 while ((CurrentCurve != NULL) && !CurveFound)
   if (CurrentCurve->Handle == CurveHandle)
     CurveFound = TRUE;
   else
     CurrentCurve = CurrentCurve->Next;

 if (CurveFound)
  {
   CurrentCurve->SetXStart (NewXStart);
   return (GRAPHERR_NOERR);
  }
 else
   return (GRAPHERR_HINVAL);
}

void TGraphWindow::MarkCurve (int MenuID)
{
 char  MenuEntry [MAX_CURV_NAME + 6];
 char  *CurveName;
 BOOL  NoneSelected = TRUE;
 BOOL  AllSelected  = TRUE; 
 HMENU GraphMenu    = GetMenu (HWindow);

 GetMenuString (GraphMenu, MenuID, MenuEntry, sizeof (MenuEntry),
                MF_BYCOMMAND);
 CurveName = &(MenuEntry [5]);
 for (TCurve *CurrentCurve = Curves; CurrentCurve != NULL;
      CurrentCurve = CurrentCurve->Next)
   if (strcmp (CurveName, CurrentCurve->Name) == 0)
     if (CurrentCurve->Mark)
      {
       CurrentCurve->Mark = FALSE;
       CheckMenuItem (GraphMenu, MenuID, MF_BYCOMMAND | MF_UNCHECKED);
       AllSelected  = FALSE; 
      }
     else
      {
       CurrentCurve->Mark = TRUE;
       CheckMenuItem (GraphMenu, MenuID, MF_BYCOMMAND | MF_CHECKED);
       NoneSelected = FALSE;

    }
   else
    {
     NoneSelected &= !(CurrentCurve->Mark);
     AllSelected  &= CurrentCurve->Mark;
    }

 if (NoneSelected)
  {
   EnableMenuItem (GraphMenu, CM_SELECTALL, MF_BYCOMMAND | MF_ENABLED);
   EnableMenuItem (GraphMenu, CM_UNSELECTALL, MF_BYCOMMAND | MF_GRAYED);
   EnableMenuItem (GraphMenu, CM_GRAPHMEAN, MF_BYCOMMAND | MF_GRAYED);
   EnableMenuItem (GraphMenu, CM_GRAPHSDEV, MF_BYCOMMAND | MF_GRAYED);
   EnableMenuItem (GraphMenu, CM_GRAPHMSE, MF_BYCOMMAND | MF_GRAYED);
  }
 else if (AllSelected)
  {
   EnableMenuItem (GraphMenu, CM_SELECTALL, MF_BYCOMMAND | MF_GRAYED);
   EnableMenuItem (GraphMenu, CM_UNSELECTALL, MF_BYCOMMAND | MF_ENABLED);
   EnableMenuItem (GraphMenu, CM_GRAPHMEAN, MF_BYCOMMAND | MF_ENABLED);
   EnableMenuItem (GraphMenu, CM_GRAPHSDEV, MF_BYCOMMAND | MF_ENABLED);
   EnableMenuItem (GraphMenu, CM_GRAPHMSE, MF_BYCOMMAND | MF_ENABLED);
  }
 else
  {
   EnableMenuItem (GraphMenu, CM_SELECTALL, MF_BYCOMMAND | MF_ENABLED);
   EnableMenuItem (GraphMenu, CM_UNSELECTALL, MF_BYCOMMAND | MF_ENABLED);
   EnableMenuItem (GraphMenu, CM_GRAPHMEAN, MF_BYCOMMAND | MF_ENABLED);
   EnableMenuItem (GraphMenu, CM_GRAPHSDEV, MF_BYCOMMAND | MF_ENABLED);
   EnableMenuItem (GraphMenu, CM_GRAPHMSE, MF_BYCOMMAND | MF_ENABLED);
  }
}

int TGraphWindow::SetMarker (float MarkerPosition, int MarkerOrientation,
			     int MarkerLineType, COLORREF MarkerColor, BOOL DisplayValue)
{
 TSearchMarker SearchCriterion;

 SearchCriterion.Position    = MarkerPosition;
 SearchCriterion.Orientation = MarkerOrientation; 
 TMarker *NewMarker = Markers.firstThat (FindMarker, &SearchCriterion);

 if (NewMarker == NULL)
  {
   NewMarker = new TMarker (MarkerPosition, MarkerOrientation, MarkerLineType, MarkerColor,
                            DisplayValue);
   if (NewMarker == NULL)
     return (GRAPHERR_NOMEM);
   else
     Markers.insert (NewMarker);
  }
 else
   Markers.free (NewMarker);

 // Display or remove the marker
 if (Curves != NULL)
   DrawCurves ();
 else
   DrawAxis (TRUE);

 return (GRAPHERR_NOERR);
}

void TGraphWindow::CreateMemorySheet (TFloatPoint ZoomIn, BOOL ForceNew)
{
 RECT             ClientArea;
 BITMAPINFOHEADER DIBInfo;
 int              NewWidth;
 int              NewHeight;

 GetClientRect (HWindow, &ClientArea);
 HDC ScreenDC  = GetDC (HWindow);
 int XRange    = 0;
 int YRange    = 0;

 if (FitToWindow)
  {
   NewWidth              = ClientArea.right  * ZoomIn.x;
   NewHeight             = ClientArea.bottom * ZoomIn.y;
   // Remember the new dimension in case the user decides to change to fixed dimensions
   FixedDimension.Width  = ClientArea.right;
   FixedDimension.Height = ClientArea.bottom;
  }
 else
  {
   NewWidth  = FixedDimension.Width  * ZoomIn.x;
   NewHeight = FixedDimension.Height * ZoomIn.y;
  }

 if (NewWidth < MIN_BITMAP_WIDTH)
   NewWidth = MIN_BITMAP_WIDTH;
 if (NewHeight < MIN_BITMAP_HEIGHT)
   NewHeight = MIN_BITMAP_HEIGHT;

 if (NewWidth > ClientArea.right)
   XRange = (NewWidth - ClientArea.right +
	     GetSystemMetrics (SM_CXVSCROLL)) / 10 + 1;
 if (NewHeight > ClientArea.bottom)
   YRange = (NewHeight - ClientArea.bottom +
	     GetSystemMetrics (SM_CYHSCROLL)) / 10 + 1;

 if ((BitmapDimension.Width  == NewWidth)  &&
     (BitmapDimension.Height == NewHeight) &&
     !ForceNew)
  {
   // Bitmap-dimension hasn't changed; ignore this call
   ReleaseDC (HWindow, ScreenDC);
   // If the size is fixed, we have to re-adjust the scroller
   if (!FitToWindow)
     Scroller->SetRange (XRange, YRange);
   return;
  }

 // Destroy bitmap if there is one existing
 if (MemorySheet != NULL)
  {
   DeleteObject (MemorySheet);
   MemorySheet   = NULL;
  }

 DIBInfo.biSize          = sizeof (DIBInfo);;
 DIBInfo.biWidth         = NewWidth;
 DIBInfo.biHeight        = NewHeight;
 DIBInfo.biPlanes        = 1;
 DIBInfo.biBitCount      = 8;       // 256 colors possible
 DIBInfo.biCompression   = BI_RGB;  // No compression
 DIBInfo.biSizeImage     = 0;
 DIBInfo.biXPelsPerMeter = 0;
 DIBInfo.biYPelsPerMeter = 0;
 DIBInfo.biClrUsed       = 0;
 DIBInfo.biClrImportant  = 0;

 if ((MemorySheet = CreateDIBitmap (ScreenDC, &DIBInfo, 0, NULL, NULL, DIB_RGB_COLORS)) ==
     NULL)
  {
   TErrorMessage BoxMessage, BoxTitle;
   LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle, sizeof (BoxTitle));
   LoadString (GetApplication ()->hInstance, ERR_NO_BITMAP, BoxMessage, sizeof (BoxMessage)); 
   MessageBeep (-1);
   BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONSTOP | MB_OK);
   if ((ZoomIn.x != 1.0) || (ZoomIn.y != 1.0))
    {
     HMENU Menu = GetMenu (HWindow);
     CheckMenuItem (Menu, CM_ZOOMGRAPH,  MF_BYCOMMAND | MF_UNCHECKED);
     CheckMenuItem (Menu, CM_ZOOMGRAPHX, MF_BYCOMMAND | MF_UNCHECKED);
     CheckMenuItem (Menu, CM_ZOOMGRAPHY, MF_BYCOMMAND | MF_UNCHECKED);
     ZoomFactor.x = 1.0;
     ZoomFactor.y = 1.0;
    }
   // Recreate the old bitmap
   DIBInfo.biWidth  = BitmapDimension.Width;
   DIBInfo.biHeight = BitmapDimension.Height;
   XRange           = 0;
   YRange           = 0;
   if ((MemorySheet = CreateDIBitmap (ScreenDC, &DIBInfo, 0, NULL, NULL, DIB_RGB_COLORS)) ==
        NULL)
    {
     // Try with the smallest possible bitmap
     DIBInfo.biWidth  = MIN_BITMAP_WIDTH;
     DIBInfo.biHeight = MIN_BITMAP_HEIGHT;
     if ((MemorySheet = CreateDIBitmap (ScreenDC, &DIBInfo, 0, NULL, NULL, DIB_RGB_COLORS)) ==
          NULL)
      {
       // Not even the smallest bitmap can be created; force program exit!
       LoadString (GetApplication ()->hInstance, ERR_MEM_TOO_LOW_EXIT, BoxMessage,
		   sizeof (BoxMessage));
       MessageBeep (-1);
       BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONSTOP | MB_OK);
       ReleaseDC (HWindow, ScreenDC);
       exit (EXIT_FAILURE);
      }
    }
  }
 else
  {
   // Store the new bitmap-dimension
   BitmapDimension.Width  = NewWidth;
   BitmapDimension.Height = NewHeight;
  }

 HDC MemorySheetDC = CreateCompatibleDC (ScreenDC);
 SelectObject (MemorySheetDC, MemorySheet);
 // Clear the new bitmap to all white
 PatBlt (MemorySheetDC, 0, 0, BitmapDimension.Width, BitmapDimension.Height, WHITENESS);
 DeleteDC (MemorySheetDC);

 Scroller->SetRange (XRange, YRange);
 ReleaseDC (HWindow, ScreenDC);

 DrawCurves ();
}

void TGraphWindow::DrawAxis (BOOL Redraw)
{
 RECT       TitleArea;
 TEXTMETRIC FontInfo;
 TAxisRange TransValues;
 char       Title [MAX_DIAG_NAME * 2 + 1];
 char       StartValue [MAX_VALUE_LEN + 1];
 char       EndValue [MAX_VALUE_LEN + 1];
 int        TicPosition;

 HDC ScreenDC      = GetDC (HWindow);
 HDC MemorySheetDC = CreateCompatibleDC (ScreenDC);
 SelectObject (MemorySheetDC, MemorySheet);

 // Select the font and set the color for the title
 HFONT    PreviousFont  = SelectObject (MemorySheetDC, TitleFont);
 COLORREF PreviousColor = SetTextColor (MemorySheetDC, RGB (0, 0, 255));
 GetTextMetrics (MemorySheetDC, &FontInfo);
 sprintf (Title, "%s\n%s", DiagramTitle.Title1, DiagramTitle.Title2);
 int MaxTextWidth = MAX_DIAG_NAME * FontInfo.tmAveCharWidth;
 TitleArea.left   = (BitmapDimension.Width / 2) - (MaxTextWidth / 2);
 TitleArea.top    = 10;
 TitleArea.right  = (BitmapDimension.Width / 2) + (MaxTextWidth / 2);
 TitleArea.bottom = 10 + 2 * FontInfo.tmHeight +
			 2 * FontInfo.tmExternalLeading;
 // Add the title to the diagram
 DrawText (MemorySheetDC, Title, -1, &TitleArea,
	   DT_CENTER | DT_NOPREFIX | DT_WORDBREAK);

 // Select the font and set the color for the axis- and tic-description
 SelectObject (MemorySheetDC, LegendFontNorm);
 SetTextColor (MemorySheetDC, RGB (0, 0, 0));
 GetTextMetrics (MemorySheetDC, &FontInfo);
 int TicSpace    = 3 + 2 * FontInfo.tmHeight + 2 * FontInfo.tmExternalLeading;
 int DescrHeight = FontInfo.tmHeight + FontInfo.tmExternalLeading;
 int LegendSpace = 10 + FontInfo.tmAveCharWidth * MAX_CURV_NAME + 40;


 // Transform the necessary values to the logarithmic value if necessary. Change to linear
 // if axis-range is out of the domain for logarithmic values!
 BOOL XLogarithmic = (DiagramType & X_LOGARITHMIC) != 0;
 BOOL YLogarithmic = (DiagramType & Y_LOGARITHMIC) != 0;

 if ((AxisRange.X.Start <= 0) || (AxisRange.X.End <= 0))
  if (XLogarithmic)
   {
    TErrorMessage BoxMessage, BoxTitle, FormatString;
    DiagramType &= ~X_LOGARITHMIC;
    XLogarithmic = FALSE;
    LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle, sizeof (BoxTitle));
    LoadString (GetApplication ()->hInstance, ERR_SCALE_TO_LINEAR, FormatString,
		sizeof (FormatString));
    sprintf (BoxMessage, FormatString, 'X');
    MessageBeep (-1);
    BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONINFORMATION | MB_OK);
   }
 if ((AxisRange.Y.Start <= 0) || (AxisRange.Y.End <= 0))
  if (YLogarithmic)
   { 
    TErrorMessage BoxMessage, BoxTitle, FormatString;
    DiagramType &= ~Y_LOGARITHMIC;
    YLogarithmic = FALSE;
    LoadString (GetApplication ()->hInstance, STR_BOX_TITLE, BoxTitle, sizeof (BoxTitle));
    LoadString (GetApplication ()->hInstance, ERR_SCALE_TO_LINEAR, FormatString,
		sizeof (FormatString));
    sprintf (BoxMessage, FormatString, 'Y');
    MessageBeep (-1);
    BWCCMessageBox (GetFocus (), BoxMessage, BoxTitle, MB_ICONINFORMATION | MB_OK);
   }
 TransValues.X.Start  = XLogarithmic ? log10 (AxisRange.X.Start)  : AxisRange.X.Start;
 TransValues.X.End    = XLogarithmic ? log10 (AxisRange.X.End)    : AxisRange.X.End;
 TransValues.X.Offset = YLogarithmic ? log10 (AxisRange.X.Offset) : AxisRange.X.Offset;
 TransValues.Y.Start  = YLogarithmic ? log10 (AxisRange.Y.Start)  : AxisRange.Y.Start;
 TransValues.Y.End    = YLogarithmic ? log10 (AxisRange.Y.End)    : AxisRange.Y.End;
 TransValues.Y.Offset = XLogarithmic ? log10 (AxisRange.Y.Offset) : AxisRange.Y.Offset;

 // Calculate the scaling of this diagram. The origin on the device is the upper left corner!
 Scale.PixelsPerXUnit  = (float) (BitmapDimension.Width - LegendSpace - TicSpace) /
                                 (TransValues.X.End - TransValues.X.Start);
 Scale.PixelsPerYUnit  = (float) (BitmapDimension.Height - TitleArea.bottom - TicSpace) /
                                 (TransValues.Y.End - TransValues.Y.Start);
 Scale.Origin.x        = (- TransValues.X.Start * Scale.PixelsPerXUnit) + TicSpace;
 Scale.Origin.y        = (  TransValues.Y.End   * Scale.PixelsPerYUnit) + TitleArea.bottom;

 // Draw the axis
 HPEN ThinBlackPen = CreatePen (PS_SOLID, 1, RGB (0, 0, 0));
 HPEN PreviousPen  = SelectObject (MemorySheetDC, ThinBlackPen);
 int XAxisY = - TransValues.X.Offset * Scale.PixelsPerYUnit + Scale.Origin.y;
 int YAxisX =   TransValues.Y.Offset * Scale.PixelsPerXUnit + Scale.Origin.x;
 MoveTo (MemorySheetDC, TicSpace, XAxisY);
 LineTo (MemorySheetDC, BitmapDimension.Width - LegendSpace, XAxisY);
 MoveTo (MemorySheetDC, YAxisX, TitleArea.bottom);
 LineTo (MemorySheetDC, YAxisX, BitmapDimension.Height - TicSpace);

 // Draw tic-marks for x-axis
 float PreviousTicMark = AxisRange.Y.Offset + AxisRange.X.Step;
 for (float TicMark = AxisRange.Y.Offset; TicMark <= AxisRange.X.End;
      TicMark += AxisRange.X.Step)
  {
   // Check if the new tic-mark-position differs from the previous one.
   // If not, skip drawing of tic-marks!
   if (PreviousTicMark == TicMark)
     break;
   else
     PreviousTicMark = TicMark;
   if (XLogarithmic)
     TicPosition = (int) (log10 (TicMark) * Scale.PixelsPerXUnit) + Scale.Origin.x;
   else
     TicPosition = (int) (TicMark * Scale.PixelsPerXUnit) + Scale.Origin.x;
   MoveTo (MemorySheetDC, TicPosition, XAxisY);
   LineTo (MemorySheetDC, TicPosition, XAxisY + 3); 
  }
 PreviousTicMark = AxisRange.Y.Offset + AxisRange.X.Step;
 for (TicMark = AxisRange.Y.Offset; TicMark >= AxisRange.X.Start;
      TicMark -= AxisRange.X.Step)
  {
   // Check if the new tic-mark-position differs from the previous one.
   // If not, skip drawing of tic-marks!
   if (PreviousTicMark == TicMark)
     break;
   else
     PreviousTicMark = TicMark;
   if (XLogarithmic)
     TicPosition = (int) (log10 (TicMark) * Scale.PixelsPerXUnit) + Scale.Origin.x;
   else
     TicPosition = (int) (TicMark * Scale.PixelsPerXUnit) + Scale.Origin.x;
   MoveTo (MemorySheetDC, TicPosition, XAxisY);
   LineTo (MemorySheetDC, TicPosition, XAxisY + 3); 
  }

 // Draw tic-marks for y-axis
 PreviousTicMark = AxisRange.X.Offset + AxisRange.Y.Step;
 for (TicMark = AxisRange.X.Offset; TicMark <= AxisRange.Y.End;
      TicMark += AxisRange.Y.Step)
  {
   // Check if the new tic-mark-position differs from the previous one.
   // If not, skip drawing of tic-marks!
   if (PreviousTicMark == TicMark)
     break;
   else
     PreviousTicMark = TicMark;
   if (YLogarithmic)
     TicPosition = (int) (-log10 (TicMark) * Scale.PixelsPerYUnit) + Scale.Origin.y;
   else
     TicPosition = (int) (-TicMark * Scale.PixelsPerYUnit) + Scale.Origin.y;
   MoveTo (MemorySheetDC, YAxisX, TicPosition);
   LineTo (MemorySheetDC, YAxisX - 3, TicPosition);
  }
 PreviousTicMark = AxisRange.X.Offset + AxisRange.Y.Step;
 for (TicMark = AxisRange.X.Offset; TicMark >= AxisRange.Y.Start;
      TicMark -= AxisRange.Y.Step)
  {
   // Check if the new tic-mark-position differs from the previous one.
   // If not, skip drawing of tic-marks!
   if (PreviousTicMark == TicMark)
     break;
   else
     PreviousTicMark = TicMark;
   if (YLogarithmic)
     TicPosition = (int) (-log10 (TicMark) * Scale.PixelsPerYUnit) + Scale.Origin.y;
   else
     TicPosition = (int) (-TicMark * Scale.PixelsPerYUnit) + Scale.Origin.y;
   MoveTo (MemorySheetDC, YAxisX, TicPosition);
   LineTo (MemorySheetDC, YAxisX - 3, TicPosition);
  }

 // Add descriptions to x-axis
 char AxisTitle [MAX_AXIS_NAME + 7];
 strcpy (AxisTitle, AxisRange.X.Title);
 if (XLogarithmic)
   strcat (AxisTitle, " (log)");

 DWORD TextExtent = GetTextExtent (MemorySheetDC, AxisTitle, strlen (AxisTitle));
 int TextX = (BitmapDimension.Width - TicSpace - LegendSpace - LOWORD (TextExtent)) / 2 +
             TicSpace;
 TextOut (MemorySheetDC, TextX, BitmapDimension.Height - DescrHeight, AxisTitle,
          strlen (AxisTitle));
 if ((AxisRange.X.Start == (int) AxisRange.X.Start) &&
     (AxisRange.X.End   == (int) AxisRange.X.End))
  {
   sprintf (StartValue, "%i", (int) AxisRange.X.Start);
   sprintf (EndValue, "%i", (int) AxisRange.X.End);
  }
 else
  {
   sprintf (StartValue, "%.2f", AxisRange.X.Start);
   sprintf (EndValue, "%.2f", AxisRange.X.End);
  }
 TextOut (MemorySheetDC, TicSpace, XAxisY + 3, StartValue,
          strlen (StartValue));
 TextExtent = GetTextExtent (MemorySheetDC, EndValue, strlen (EndValue));
 TextX = BitmapDimension.Width - LegendSpace - LOWORD (TextExtent);
 TextOut (MemorySheetDC, TextX, XAxisY + 3, EndValue, strlen (EndValue));

 // Add descriptions to y-axis
 strcpy (AxisTitle, AxisRange.Y.Title);
 if (YLogarithmic)
   strcat (AxisTitle, " (log)");
 SelectObject (MemorySheetDC, LegendFontTilt);

 TextExtent = GetTextExtent (MemorySheetDC, AxisTitle, strlen (AxisTitle));
 int TextY = (BitmapDimension.Height - TicSpace - TitleArea.bottom - LOWORD (TextExtent)) / 2 +
              (TitleArea.bottom + LOWORD (TextExtent));
 TextOut (MemorySheetDC, 0, TextY, AxisTitle, strlen (AxisTitle));
 if ((AxisRange.Y.Start == (int) AxisRange.Y.Start) &&
     (AxisRange.Y.End   == (int) AxisRange.Y.End))
  {
   sprintf (StartValue, "%i", (int) AxisRange.Y.Start);
   sprintf (EndValue, "%i", (int) AxisRange.Y.End);
  }
 else
  {
   sprintf (StartValue, "%.2f", AxisRange.Y.Start);
   sprintf (EndValue, "%.2f", AxisRange.Y.End);
  }
 TextOut (MemorySheetDC, YAxisX - DescrHeight - 3,
	  BitmapDimension.Height - TicSpace, StartValue,
	  strlen (StartValue));
 TextExtent = GetTextExtent (MemorySheetDC, EndValue, strlen (EndValue));
 TextY = TitleArea.bottom + LOWORD (TextExtent);
 TextOut (MemorySheetDC, YAxisX - DescrHeight - 3, TextY, EndValue,
          strlen (EndValue));

 // Select old objects into the DC
 SelectObject (MemorySheetDC, PreviousPen);
 SelectObject (MemorySheetDC, PreviousFont);
 SetTextColor (MemorySheetDC, PreviousColor);

 // Destroy unnecessary objects (free resources!)
 DeleteObject (ThinBlackPen);

 // Draw existing markers
 if (Markers.getCount () != 0)
  {
   TMarkerData MarkerData;

   MarkerData.DrawDC         = MemorySheetDC;
   MarkerData.HorizontalFont = LegendFontNorm;
   MarkerData.VerticalFont   = LegendFontTilt;
   MarkerData.TicSpace       = TicSpace;
   MarkerData.HeaderSpace    = TitleArea.bottom;
   MarkerData.LegendSpace    = LegendSpace;
   MarkerData.ScaleType      = DiagramType;
   memcpy (&(MarkerData.Scaling), &Scale, sizeof (Scale));
   memcpy (&(MarkerData.DrawDimension), &BitmapDimension, sizeof (BitmapDimension));

   Markers.forEach (DrawMarker, (void *) &MarkerData);
  }

 DeleteDC (MemorySheetDC);
 ReleaseDC (HWindow, ScreenDC);

 if (Redraw)
  {
   // Make sure the window is redrawn
   InvalidateRect (HWindow, NULL, FALSE);
   UpdateWindow (HWindow);
  }
}

void TGraphWindow::DrawCurves ()
{
 POINT      LegendOffset;
 TEXTMETRIC FontInfo;

 if (Curves != NULL)
  {
   HDC ScreenDC      = GetDC (HWindow);
   HDC MemorySheetDC = CreateCompatibleDC (ScreenDC);
   SelectObject (MemorySheetDC, MemorySheet);
   // Clear the new bitmap to all white
   PatBlt (MemorySheetDC, 0, 0, BitmapDimension.Width,
           BitmapDimension.Height, WHITENESS);

   HFONT PreviousFont = SelectObject (MemorySheetDC, LegendFontNorm);
   GetTextMetrics (MemorySheetDC, &FontInfo);
   LegendOffset.x = BitmapDimension.Width - 10 -
		    FontInfo.tmAveCharWidth * MAX_CURV_NAME - 40;
   SelectObject (MemorySheetDC, TitleFont);
   GetTextMetrics (MemorySheetDC, &FontInfo);
   LegendOffset.y = 10 + 2 * FontInfo.tmHeight +
			 2 * FontInfo.tmExternalLeading;
   SelectObject (MemorySheetDC, PreviousFont);

   // Remove DC's because 'DrawAxis' creates it's own set!
   DeleteDC (MemorySheetDC);
   ReleaseDC (HWindow, ScreenDC);

   // Redraw the axis but don't redraw the screen. The screen will be redrawn after the curves
   // have been drawn!
   DrawAxis (FALSE);

   // Draw all available curves
   ScreenDC      = GetDC (HWindow);
   MemorySheetDC = CreateCompatibleDC (ScreenDC);
   SelectObject (MemorySheetDC, MemorySheet);

   for (TCurve *CurrentCurve = Curves; CurrentCurve != NULL;
	CurrentCurve = CurrentCurve->Next)
    {
     CurrentCurve->Legend (MemorySheetDC, LegendOffset, LegendFontNorm);
     CurrentCurve->Draw (MemorySheetDC);
    }

   DeleteDC (MemorySheetDC);
   ReleaseDC (HWindow, ScreenDC);

   // Make sure the window is redrawn
   InvalidateRect (HWindow, NULL, FALSE);
   UpdateWindow (HWindow);
  }
}

void TGraphWindow::CreateMyFonts (char *FontName)
{
 LOGFONT    MyFont;
 TEXTMETRIC SystemMetrics;
 HDC        WindowDC     = GetDC (HWindow);
 HFONT      PreviousFont = SelectObject (WindowDC, GetStockObject (SYSTEM_FONT));

 // Retrieve the metrics of the system-font to get correct measures for the own font
 GetTextMetrics (WindowDC, &SystemMetrics);
 SelectObject (WindowDC, PreviousFont);
 ReleaseDC (HWindow, WindowDC);

 MyFont.lfHeight = SystemMetrics.tmHeight * 1.2;
 MyFont.lfWidth = 0;
 MyFont.lfEscapement = 0;
 MyFont.lfOrientation = 0;
 MyFont.lfWeight = FW_HEAVY;
 MyFont.lfItalic = FALSE;
 MyFont.lfUnderline = FALSE;
 MyFont.lfStrikeOut = FALSE;
 MyFont.lfCharSet = ANSI_CHARSET;
 MyFont.lfOutPrecision = OUT_TT_PRECIS;
 MyFont.lfClipPrecision = CLIP_CHARACTER_PRECIS;
 MyFont.lfQuality = PROOF_QUALITY;
 MyFont.lfPitchAndFamily = VARIABLE_PITCH | FF_DONTCARE;
 if (FontName == NULL)
   strcpy (MyFont.lfFaceName, "Arial");
 else
   strcpy (MyFont.lfFaceName, FontName);
 TitleFont = CreateFontIndirect (&MyFont);

 MyFont.lfHeight = SystemMetrics.tmHeight * 0.8;
 MyFont.lfWeight = FW_LIGHT;
 LegendFontNorm = CreateFontIndirect (&MyFont);

 MyFont.lfEscapement = 900;
 LegendFontTilt = CreateFontIndirect (&MyFont);
}

HANDLE TGraphWindow::RetrieveBitmap ()
{
 BITMAPINFOHEADER BitmapInfo;
 BITMAPINFOHEADER *BitmapMemory;
 HANDLE           DIBMemory;
 int              PaletteSize = 256 * sizeof (RGBQUAD);

 BitmapInfo.biSize          = sizeof (BITMAPINFOHEADER);
 BitmapInfo.biWidth         = BitmapDimension.Width;
 BitmapInfo.biHeight        = BitmapDimension.Height;
 BitmapInfo.biPlanes        = 1;
 BitmapInfo.biBitCount      = 8;
 BitmapInfo.biCompression   = BI_RGB;
 BitmapInfo.biSizeImage     = widthbytes ((DWORD) BitmapDimension.Width * 8) *
			      BitmapDimension.Height;
 BitmapInfo.biXPelsPerMeter = 0;
 BitmapInfo.biYPelsPerMeter = 0;
 BitmapInfo.biClrUsed       = 0;
 BitmapInfo.biClrImportant  = 0;

 DIBMemory = GlobalAlloc (GHND, BitmapInfo.biSize + PaletteSize + BitmapInfo.biSizeImage);

 if (!DIBMemory)
   return (NULL);

 BitmapMemory  = (BITMAPINFOHEADER *) GlobalLock (DIBMemory);
 *BitmapMemory = BitmapInfo;
 char *BitField = (char *) BitmapMemory;
 BitField = &(BitField [BitmapMemory->biSize + PaletteSize]);

 HDC ScreenDC = GetDC (NULL);

 // Copy the bits into our buffer
 if (GetDIBits (ScreenDC, MemorySheet, 0, BitmapDimension.Height, BitField,
		(LPBITMAPINFO) BitmapMemory, DIB_RGB_COLORS) == 0)
  {
   GlobalUnlock (DIBMemory);
   GlobalFree   (DIBMemory);
   ReleaseDC (NULL, ScreenDC);
   return (NULL);
  }

 BitmapInfo = *BitmapMemory;
 GlobalUnlock (DIBMemory);

 ReleaseDC (NULL, ScreenDC);
 return (DIBMemory);
}