/*************************************************************************
*  Program:       NNDIAGR.H                                              *
*  Description:   Headers for NNDIAGR.CPP (Diagram-Windows for VieNet2)  *
*  Version:       1.4                                                    *
*  Last modified: 26.6.93                                                *
*  Author:        Wolfgang Reh                                           *
*  Copyright (C) Austrian Research Institute for Artificial Intelligence *
**************************************************************************/

#ifndef NNDIAGR_H
  #define NNDIAGR_H
  #define WIN31

  #include <owl.h>
  #include <mem.h>
  #include <dir.h>
  #include <math.h>
  #include <bwcc.h>
  #include <ctype.h>
  #include <stdio.h>
  #include <string.h>
  #include <values.h>
  #include <fstream.h>
  #include <windows.h>
  #include <commdlg.h>
  #include <tcollect.h>
  #include <scroller.h>
  #include <combobox.h>

  #include "neurores.h"            // Resource-IDs

  #define GRAPHERR_NOERR       0   // No error, success
  #define GRAPHERR_NOMEM      -1   // No memory (new)
  #define GRAPHERR_MAXGRAPH   -2   // Maximum number of graphs exceeded
  #define GRAPHERR_MAXCURV    -3   // Maximum number of curves exceeded
  #define GRAPHERR_HINVAL     -4   // Handle invalid
  #define GRAPHERR_PARAMINVAL -5   // Invalid parameter passed to function

  #define X_AUTOSCALE   0x0001
  #define X_LINEAR      0x0000
  #define X_LOGARITHMIC 0x0002
  #define Y_AUTOSCALE   0x0100
  #define Y_LINEAR      0x0000
  #define Y_LOGARITHMIC 0x0200

  #define LINE_SOLID        0x0000
  #define LINE_DASH         0x0001
  #define LINE_DOT          0x0002
  #define LINE_DASH_DOT     0x0003
  #define LINE_DASH_DOT_DOT 0x0004
  #define LINE_NONE         0x0080

  #define DOT_HOLLOW_SQUARE   0x0100
  #define DOT_HOLLOW_CIRCLE   0x0200
  #define DOT_HOLLOW_TRIANGLE 0x0300
  #define DOT_FILLED_SQUARE   0x0400
  #define DOT_FILLED_CIRCLE   0x0500
  #define DOT_FILLED_TRIANGLE 0x0600
  #define DOT_ASTERISK        0x0700
  #define DOT_NONE            0x0000

  #define LINE_RED   RGB (255,   0,   0)
  #define LINE_GREEN RGB (  0, 255,   0)
  #define LINE_BLUE  RGB (  0,   0, 255)
  #define LINE_BLACK RGB (  0,   0,   0)

  #define MARKER_HORIZONTAL 0x0001
  #define MARKER_VERTICAL   0x0002

  #define CLOSE_ALL -1

  #define SAVE_FILE_VALUES 0x0001
  #define SAVE_FILE_BITMAP 0x0002

  #define OPEN_FILE_COMBINED 0x0001
  #define OPEN_FILE_SPLIT_XY 0x0002
  #define OPEN_FILE_SPLIT_YX 0x0003
  #define OPEN_FILE_ONLY_Y   0x0004

  #define MAX_AXIS_NAME      30
  #define MAX_DIAG_NAME      50
  #define MAX_CURV_NAME      20
  #define MAX_WINDOW_TITLE   30
  #define MAX_VALUE_LEN      10
  #define MAX_GRAPHS         50
  #define MAX_CURVES         10
  #define MAX_RESULT_BUFFER 500
  #define MIN_BITMAP_WIDTH  300
  #define MIN_BITMAP_HEIGHT 200

  #define square(x)        ((x) * (x))
  #define min(x,y)         (((x) < (y)) ? (x) : (y))
  #define widthbytes(bits) (((bits) + 31) / 32 * 4)

  typedef char TErrorMessage [128];

  typedef struct _AxisValues
   {
    char  Title [MAX_AXIS_NAME + 1];
    float Start;
    float End;
    float Step;
    float Offset;
   } TAxisValues;

  typedef struct _TAxisRange
   {
    TAxisValues X;
    TAxisValues Y;
   } TAxisRange;

  typedef struct _TScale
   {
    float PixelsPerXUnit;
    float PixelsPerYUnit;
    POINT Origin;
   } TScale;

  typedef struct _TFloatPoint
   {
    float x;
    float y;
   } TFloatPoint;

  typedef struct _TDiagramTitle
   {
    char Title1 [MAX_DIAG_NAME + 1];
    char Title2 [MAX_DIAG_NAME + 1];
   } TDiagramTitle;

  typedef struct _TBitmapDim
   {
    int Width;
    int Height;
   } TBitmapDim;

  typedef struct _TDrawData
   {
    HDC    DrawDC;
    HDC    DotDC;
    BOOL   DrawLine;
    TScale Scaling;
    int    ScaleType;
   } TDrawData;

  typedef struct _TMarkerData
   {
    HDC        DrawDC;
    HFONT      HorizontalFont;
    HFONT      VerticalFont;
    TScale     Scaling;
    TBitmapDim DrawDimension;
    int        TicSpace;
    int        HeaderSpace;
    int        LegendSpace;
    int        ScaleType;
   } TMarkerData;

  typedef struct _TSearchMarker
   {
    float Position;
    int   Orientation;
   } TSearchMarker;

  typedef struct _TYValueArray
   {
    int   ElementCount;
    float Values [1];
   } TYValueArray;

  typedef struct _TCurveBounds
   {
    float MinX;
    float MaxX;
    float MinY;
    float MaxY;
   } TCurveBounds;

  extern void AppendWindowMenu (void *);

  int init_diagram (float XStart, float XEnd, float XStep, float XOffset, char *XName,
		    float YStart, float YEnd, float YStep, float YOffset, char *YName,
		    int Type, char *Header1, char *Header2, char *Title = NULL,
		    int UpperLeftX = -1, int UpperLeftY = -1,
		    int WindowWidth = -1, int WindowHeight = -1, char *Font = NULL);

  void close_diagram (int GraphHandle);

  int init_curve (char *Name, COLORREF Color, int LineStyle, float Increment, int GraphHandle);

  int dia_set_point (int CurveHandle, float Y, float X = MINFLOAT);

  int dia_set_curve (int CurveHandle, int PointCount, float Y [], float X [],
                     float XStart = 0.0);

  int dia_set_xstart (int CurveHandle, float NewXStart);

  int dia_clear_curve (int CurveHandle);

  int dia_set_marker (int GraphHandle, float MarkerPosition, int MarkerOrientation,
		      int MarkerLineType, COLORREF MarkerColor, BOOL DisplayValue = TRUE);

  float calc_mean (float Values [], int ValueCount);

  float calc_sdev (float Values [], int ValueCount);

  float calc_mse (float Reference [], float Actual [], int ValueCount);

#endif