/*************************************************************************
*  Program:       NNWINGR.C                                              *
*  Description:   Graphic-functions for running VieNet2 on Windows 3.1   *
*  Version:       2.01                                                   *
*  Last modified: 04.01.94                                               *
*  Author:        LG                                                     *
*  Copyright (C) Austrian Research Institute for Artificial Intelligence *
**************************************************************************/

 /*
 =============================================================================
 Revision History:

  1.11.92 LG   1.) First release including Menus and Layer-drag

  10.11.92 LG 1.) Add Layer-Selection with Markers and Layer-Resizing
					    2.) Layer-Menu for right mousebutton

	19.11.92 LG 1.) msg_print() nur bei Init_OK ausfhren, ansonsten Message() verwenden

   8.3.93 LG  1.) msg_print() Ausgabe bei Init_OK verhindern

	26.4.93 LG  1.) EnableKBHandler() ermoeglicht funktionierende Button-Hotkeys
		und wechseln des Focus mit TAB
  18.11.93 LG 1) draw_scaletext() fuer Anzeige der Palettenendwerte bei Skalierung
  30.12.93 LG 1) implementation of DDE-Support (as server)
 	15.2.94  LG 1) Version 2.0 supporting full dynamic layers
 =============================================================================
 */

#define WIN31 // Def fr BC++ 3.1 noetig, laeuft nur ab Windows 3.1

#ifndef WINVER
#define WINVER 0x0310 // fuer BC++ 3.0 noetig, damit defs in windows.h stimmen
#define WS_EX_TOPMOST 0x00000008L // fuer Windows 3.0-Headerfiles noetig
#endif

#define MODULE
#define SOURCENAME "NNWINGR"

#include <alloc.h>
#include <owl.h>
#include <bchkbox.h>
#include <bgrpbox.h>
#include <button.h>
#include <bstatic.h>
#include <bdivider.h>
#include <bwcc.h>
#include <bwindow.h>
#include <dir.h>      // MAXPATH
#include <ddeml.h>    // DDE
#include <filedial.h> // FileDialog()
#include <scrollba.h>
#include <math.h>   // pow()
#include <module.h> // ValidWindow()
#include <string.h> // strcpy(), etc.
#include <stdio.h>  // sprintf() for using float outputs
#include <stdarg.h>
#include <stdlib.h> // atoi()

#include <commdlg.h>

#include "neurodef.h"
#include "neurons.h"
#include "nngraph.h"

#ifdef NNCONTROLS
#include "nncontrl.h" //defines additional controls
#else
/* dummydefs */
class TBar: public TBWindow
{public:
	void SetValue(float val, float lim1=0, float lim2=0,
				 float from=0, float to=0, char* Title="@") {};
};
#endif

#include "neurores.h" //Resource-IDs

#define ICON_MAIN "AI2"
#define ICON_MSG  "AI2"
#define ICON_INFO "AI2"

// Filetypes for Data-File
#define DAT_FILE 0
#define PAT_FILE 1

int BitMapWidth  = 800;
int BitMapHeight = 600;

#define ID_Knopf ID_FIRST
#define ID_NGroup ID_FIRST+400
#define ID_Text ID_FIRST+100
#define ID_Anzeige ID_FIRST+200
#define ID_Info ID_FIRST+300

#define max(x,y) ((x>y)?(x):(y))
#define min(x,y) ((x<y)?(x):(y))
#define NumColors 67
#define Black 66  /* Color# in colortable */
#define White 32
#define ARG_IS(i,x) (!strcmp(argv[i],x))

// Convert sliderposition to variablevalue and vice versa,
// internal resolution is 1/1000
#define SliderPosToVar(pos) (((float)(pos)*(buttons[i].val_max-buttons[i].val_min)/1000.0+(float)buttons[i].val_min))
#define SliderVarToPos(var) ((int)(((float)(var)-(float)buttons[i].val_min)*1000.0/((float)(buttons[i].val_max-buttons[i].val_min))))

#define SCALE_VAL(x) ((int)((1.0-(x-min)/scale)*height+R.top))

int black_white_display = False;
int use_colors = True;  // Farben oder nur Grauwerte
int use_gray   = True;  // positive Aktivierungen Grau, Negative Blau
												// (fr Bildverarbeitung besser geeignet)

#define BACK(col) ((col & 0x00f0) >> 4) /* Aufteilen des Farbwertes in */
#define FRONT(col) (col & 0x000f)       /* Hintergrund und Vordergrund */
					/* (col=front+16*back)     */

HWND hNWindow,hCWindow=NULL,hIWindow,hMWindow;
BOOL Show_MsgWindow, Show_InfoWindow, Show_CanvasWindow;

HDC      hCDC, hIDC;
HPALETTE hNPalette;
TMessage GMsg;
TEXTMETRIC tmMWindow, tmNWindow, tmCWindow;
BOOL EditLayout = FALSE;
BOOL BatchRun = FALSE;
BOOL Init_OK = FALSE; // Alle Windows initialisiert?
RECT Rect; //Drag-Rectangle
RECT RectExt; //Drag-Rect fuer extern-Layer
struct {RECT Rect;
				int Layer;
				BOOL Ext; 
				BOOL Select; 
				int Button;
				int DragStatus;
				int Resize;
				} far MouseMode;

// Mainwindow position and size, set in init_xgraph()
int NeuroWindow_X, NeuroWindow_Y, NeuroWindow_W, NeuroWindow_H;
HMENU hNMenu;

// User defined windows
#define MaxUserWindow CM_ShowLastUserWindow-CM_ShowFirstUserWindow+1
struct {
			 int ID;
			 PTWindowsObject ptr;
			 BOOL Show;
			 } UserWindow[MaxUserWindow];

BOOL bAbortRequested=FALSE;
BOOL in_break=FALSE;

#define MaxMsgLines 100
int MsgLines=0;
char* MsgText[MaxMsgLines];

struct TState {
	 HBITMAP   BitMap;
	 int       PenSize;
	 COLORREF  PenColor;
	 COLORREF  BrushColor;
};
typedef TState* PTState;

int X0=BUTTON_WIDTH+32;
int nr=0;
extern char dummy[];

/* DDE */
char DDE_DELIMITER='\t'; // default
static int act_conv=-1;

int SliderVarToPosNr(int i)
{ switch (buttons[i].val_typ) {
	case 'f':
		return(SliderVarToPos(*(float*)buttons[i].var));
		break;
	case 's':
		return(SliderVarToPos(*(short*)buttons[i].var));
		break;
	case 'i':
		return(SliderVarToPos(*(int*)buttons[i].var));
		break;
	case 'l':
		return(SliderVarToPos(*(long*)buttons[i].var));
		break;
	}
}

/* ========Globale Funktionen=========*/
/* Normale MessageBox */
void Message(char* fmt, ...)
{ char str[140];
	wvsprintf(str,fmt,(LPSTR)(&fmt+1));
	MessageBox(NULL, str, "Message", MB_OK);
}

/* ErrorMessage, float arguments are not supported! */
void ErrMsg(char* fmt, ...)
{ char str[128];
	wvsprintf(str,fmt,(LPSTR)(&fmt+1));
	MessageBox(NULL, str, "Error", MB_OK | MB_ICONEXCLAMATION | MB_APPLMODAL);
}

/* AlarmMessage, float arguments are not supported! */
void AlarmMsg(char* fmt, ...)
{ char str[128];
	wvsprintf(str,fmt,(LPSTR)(&fmt+1));
	MessageBox(NULL, str, "Attention", MB_OK | MB_ICONEXCLAMATION | MB_APPLMODAL);
}

void DrawLine(HDC MemDC, int x1, int y1,int x2,int y2)
{ RECT R;
	MoveTo(MemDC,x1,y1); LineTo(MemDC,x2,y2);
	R.left=x1;
	R.right=x2;
	R.top=y1;
	R.bottom=y2;
	InvalidateRect(hCWindow,&R,0);
}

/* ========Lokale Funktionen=========*/
// speziell fuer draw_unit()
void QuickFilledRect(HDC MemDC, int x1,int y1,int x2,int y2,COLORREF col,int update)
{ HBRUSH ABrush, OldBrush;
	RECT R;
	ABrush = CreateSolidBrush(col);
	OldBrush = (HBRUSH)SelectObject(MemDC, ABrush);
	HPEN hPen=CreatePen(PS_SOLID,1,0);
	HPEN OldPen = (HPEN)SelectObject(MemDC,hPen);
	Rectangle(MemDC, x1, y1, x2, y2);
	SelectObject(MemDC, OldBrush);
	DeleteObject(ABrush);
	SelectObject(MemDC, OldPen);
	DeleteObject(hPen);
 	if (update)
		{
		R.left=x1;
		R.right=x2;
		R.top=y1;
		R.bottom=y2;
		InvalidateRect(hCWindow,&R,0);
		}
}

// Draws a filled Rect with given Color and Borderstyle
void FilledRect(HDC MemDC, int x1,int y1,int x2,int y2,COLORREF col,COLORREF bordercol, int border)
{ HBRUSH ABrush, OldBrush;
	RECT R;
	ABrush = CreateSolidBrush(col);
	OldBrush = (HBRUSH)SelectObject(MemDC, ABrush);
	HPEN hPen=(border)?CreatePen(PS_SOLID,border,bordercol):CreatePen(PS_NULL,border,0);
	HPEN OldPen = (HPEN)SelectObject(MemDC,hPen);
	Rectangle(MemDC, x1, y1, x2, y2);
	SelectObject(MemDC, OldBrush);
	DeleteObject(ABrush);
	SelectObject(MemDC, OldPen);
	DeleteObject(hPen);
	R.left=x1;
	R.right=x2;
	R.top=y1;
	R.bottom=y2;
	InvalidateRect(hCWindow,&R,0);
}

// Same, but with Rect-Struct
#define FilledRect1(MemDC, R, col, bordercol, border) FilledRect(MemDC, R.left, R.top, R.right, R.bottom, col, bordercol, border)

// Sets Rect-Area for Layer l
void SetLayerRect(RECT& R, layer& L, int ext)
{ int x=L.x, y=L.y;
	if (ext)
    {
		x+=L.extsep_x;
		y+=L.extsep_y;
    }
	SetRect(&R, x, y,
			x+L.ux+(L.upr-1)*L.dx+5, y+L.uy+((L.size-1)/L.upr)*L.dy+5);
}

// Sets Rect und RectExt for Markersize including titel
void SetMarkerRect(pINT l)
{
	layer L = L_Tab[l];
	int tw=tmCWindow.tmMaxCharWidth;
  int th=tmCWindow.tmHeight;
	SetRect(&Rect, L.x-4, L.y-3-th-2,
//    max(L.x+strlen(L.titel)*tw+1, L.x+L.ux+L.dx*(L.upr-1)+6),
		  L.x+L.ux+L.dx*(L.upr-1)+6,
			L.y+L.uy+L.dy*((L.size-1)/L.upr)+6);
	if (L.ext)
		SetRect(&RectExt, Rect.left+L.extsep_x, Rect.top+L.extsep_y,
//     L.extsep_x+max(L.x+strlen(L.titel_ext)*tw+1,L.x+L.ux+L.dx*(L.upr-1)+6),
			 L.extsep_x+L.x+L.ux+L.dx*(L.upr-1)+6,
			 L.extsep_y+Rect.bottom+th+6);
	else
		SetRect(&RectExt,0,0,0,0);
}

// Layerposition am Screen updaten nach Aenderung von L_Tab[l]
void update_layer(pINT l)
{	layer *L=&L_Tab[l];
  reposition_layer(L);
  if (Init_OK)
		{draw_rahmen(l,0);
		draw_layer(l);
		draw_ext(l);}
}


void ClearLayerRect(pINT l)
{ RECT R;
  SetLayerRect(R,L_Tab[l],0);
  // clear including layertitle
  FilledRect(hCDC,R.left-1,R.top-5-tmCWindow.tmHeight,R.right+1,R.bottom+1,RGB(255,255,255),0,0);

//  FilledRect(hCDC,Rect.left-1,Rect.top-1,Rect.right+1,Rect.bottom+1,RGB(255,255,255),0,0);

	// Delete old normal layer
/*{
		SetRect(&R, MouseMode.Rect.left - L->extsep_x, MouseMode.Rect.top - L->extsep_y,
				 MouseMode.Rect.right - L->extsep_x, MouseMode.Rect.bottom - L->extsep_y +th+5);
		FilledRect(hCDC,R.left,R.top,R.right,R.bottom,RGB(255,255,255),0,0);
    }
*/
}

int LoadLayout(char* fname)
{ FILE *fp;
	int l=0,i,x,ver=1.0;
	pINT size;
	char str[200];
	char *p,*p1; 
	if ((fp = fopen(fname, "rt")) != NULL)
		{
		fgets(str,200,fp);
    if (!strncmp(str,"Version=V",9))
	    sscanf(str+9,"%u",&ver); // versionnumber
    else
			rewind(fp);

		while (!feof(fp) && l<L_anz)
			{layer *L=&L_Tab[l];
      strcpy(str,"");
			fgets(str,200,fp);
			p=strtok(str,",");
			if (!p) break;
			p=strchr(p,'"')+1;
      p1=strchr(p,'"');
			if (p1) *p1='\0';
			L->titel=(char*)realloc(L->titel,strlen(p)+1);
			if (L->titel) strcpy(L->titel,p); 
      else nn_alarm(SOURCENAME,"Zu wenig Speicher fuer Layertitel\n");

      if (ver>=2.0)
			  for (i=0;i<8;i++)
				  {
					p1=strtok(NULL,",");
		      p1=strchr(p1,'=');
		      if (p1)
		        {sscanf(p1+1," %u",&x);
					  switch (i) {
					 	  case 0: L->x=x; break;
					 	  case 1: L->y=x; break;
              case 2: size=x; break;
					 	  case 3: L->upr=x; break;
					 	  case 4: L->ux=x; break;
					 	  case 5: L->uy=x; break;
					 	  case 6: L->dx=x; break;
					 	  case 7: L->dy=x; break;
	   				  }
					  }
				  }
      else
        {
        size=L->size;
			  for (i=0;i<7;i++)
				  {
					p1=strtok(NULL,",");
		      p1=strchr(p1,'=');
		      if (p1)
		        {sscanf(p1+1," %u",&x);
					  switch (i) {
					 	  case 0: L->x=x; break;
					 	  case 1: L->y=x; break;
					 	  case 2: L->upr=x; break;
					 	  case 3: L->ux=x; break;
					 	  case 4: L->uy=x; break;
					 	  case 5: L->dx=x; break;
					 	  case 6: L->dy=x; break;
	   				  }
					  }
				  }
        }
			p=strtok(NULL,",");
			p=strchr(p,'"')+1;
      p1=strchr(p,'"');
      if (p1) *p1='\0';
			L->titel_ext=(char*)realloc(L->titel_ext,strlen(p)+1);
			if (L->titel_ext) strcpy(L->titel_ext,p);
			else nn_alarm(SOURCENAME,"Zu wenig Speicher fuer Extern-Layertitel\n");
			for (i=0;i<2;i++)
				{p1=strtok(NULL,",");
				p1=strchr(p1,'=');
		    if (p1)
		      {sscanf(p1+1," %u",&x);
					switch (i) {
					 case   0: L->extsep_x=x; break;
					 case   1: L->extsep_y=x; break;
					 }
					}
				}
      resize_layer(L,size,L->upr);
			l++;
			}
		fclose(fp);
    }
	else
		{
    wsprintf(dummy,"Error loading Layoutfile %s.",fname);
	  AlarmMsg(dummy);
    return IDCANCEL;
    }

  if (Init_OK)
		{RECT R;
	  GetClientRect(hCWindow,&R);
	  FilledRect(hCDC,0,50,R.right, R.bottom, RGB(255,255,255), 0, 0);
    }
	for (ALL_LAYERS(l))
		{//resize_layer(&L_Tab[l],L_Tab[l].size,L_Tab[l].upr);
		if (Init_OK)
			{
			draw_layer(l);
      draw_ext(l);
    	}
		}
	MouseMode.Select = FALSE;
	MouseMode.DragStatus = 0;
	MouseMode.Button = 0;
	MouseMode.Resize = 0;
	return IDOK;
}                   

int SaveLayout(char* fname)
{ FILE *fp;
	int i;
	if ((fp = fopen(fname, "wt")) != NULL)
		{
    fprintf(fp,"Version=%s\n",VERSION);
		for (ALL_LAYERS(i))
			{layer *L=&L_Tab[i];
			fprintf(fp,"\"%s\", x=%u, y=%u, size=%u, upr=%u, ux=%u, uy=%u, dx=%u, dy=%u, \"%s\", extx=%u, exty=%u\n",
				L->titel,L->x,L->y,L->size,L->upr,L->ux,L->uy,L->dx,L->dy,L->titel_ext,L->extsep_x,L->extsep_y);
      }
		fclose(fp);
    }
	else
		return ( MessageBox(hNWindow, fname, "Error saving layout:",MB_RETRYCANCEL|MB_ICONSTOP) );
	
	return IDOK;
}

// Markierungen im Editmodus anzeigen
void DrawMarkers(RECT& R)
{ HDC hdc = GetDC(hCWindow);
	HBRUSH hBrush = (HBRUSH)GetStockObject(BLACK_BRUSH);
	HBRUSH OldBrush = (HBRUSH)SelectObject(hdc, hBrush);
	HPEN hPen = (HPEN)GetStockObject(BLACK_PEN);
	//CreatePen(PS_DOT,1,0);
	HPEN OldPen = (HPEN)SelectObject(hdc, hPen);
	int OldROP = SetROP2(hdc, R2_NOT);
	RECT M1,M0={-4,-4,4,4};
  int x,y;
	for (x=0;x<3;x++)
		for(y=0;y<3;y++)
	if (! ((x==1) && (y==1)))
				{CopyRect(&M1,&M0);
				OffsetRect(&M1,R.left+x*(R.right-R.left)/2,R.top+y*(R.bottom-R.top)/2);
	Rectangle(hdc, M1.left, M1.top, M1.right, M1.bottom);
	}   
//      Rectangle(hdc, R.left, R.top, R.right, R.bottom);
	SetROP2(hdc,OldROP);
	SelectObject(hdc, OldBrush);
	SelectObject(hdc, OldPen);
  DeleteObject(hPen);
  DeleteObject(hBrush);
	ReleaseDC(hCWindow,hdc);
}

BOOL nn_break(void)
{// Let other processes run and check for user break
	MSG msg;
	bAbortRequested=FALSE;
  in_break=TRUE;
	if (PeekMessage((LPMSG)&msg,NULL,0,0,PM_REMOVE))
		{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
		}
	in_break=FALSE;
	if (bAbortRequested) {msg_print("\nUser break!\n");MessageBeep(0);}
	return bAbortRequested;
}

/* Register function as DDE executable */
void NN_DDE_RegisterFunction(void(*ptr)(),char*name)
{ static int i=5; // first 5 functions are reserved for default functions
  if (i<NN_DDE_MaxFcts)
		{
		NN_DDE_Fcts[i].ptr = ptr;
		NN_DDE_Fcts[i].name = name;
		i++;
    }
  else
  	AlarmMsg("To much DDE functions registered!");
}


/* Register variable as DDE accessable */
void NN_DDE_RegisterVariable(float*ptr,char*name)
{ static int i=0; // first 5 functions are reserved for default functions
  if (i<NN_DDE_MaxVars)
		{
		NN_DDE_Vars[i].ptr = ptr;
		NN_DDE_Vars[i].name = name;
		i++;
    }
  else
  	AlarmMsg("To much DDE variables registered!");
}


//------------------- Classes -------------------------------
_CLASSDEF(TCanvas)
_CLASSDEF(TInfoWindow)
_CLASSDEF(TNeuroWindow)
_CLASSDEF(TBar)


PTWindowsObject MsgWindow;
PTNeuroWindow NeuroWindow;
PTInfoWindow InfoWindow;
PTCanvas Canvas;

PTBar NNControl[MAXBUTTONS];

PTNeuroWindow pStaticThis = NULL;

HCURSOR hArrowCursor, hWaitCursor, hPosCursor, hSizeNSCursor, hSizeWECursor, hSizeNESWCursor, hSizeNWSECursor;

// Nr of DDE-topics
#define NR_TOPICS 4
// max. Nr of DDE conversations
#define MAX_CONV 10

class TNeuroWindow : public TBWindow
{
private:
	PTButton Knopf[MAXBUTTONS];
	PTScrollBar Slider[MAXBUTTONS];
	PTStatic Anzeige[MAXBUTTONS];
	PTBGroupBox NGroup;
	int button_row;

  /* DDE */
  DWORD idInst;
  HCONV hConv[MAX_CONV+1];
	HSZ ActualTopic[MAX_CONV], ActualItem[MAX_CONV];
  pINT ActualUnitStart,ActualUnitEnd;
	HSZ hszService, hszTopic[NR_TOPICS], hszItem0[MAXLAYER], hszItem1[1], hszItem2[1], hszItem3[NN_DDE_MaxVars]; // not optimal (dynamic!)
	FARPROC lpfnCallBack;
	unsigned int NR_ITEMS[NR_TOPICS];


public:
	BOOL tfIsHot[MAX_CONV], tfNewData;
  char LFileName[MAXPATH], WFileName[MAXPATH], AFileName[MAXPATH], DFileName[MAXPATH];

	TNeuroWindow(PTWindowsObject AParent, LPSTR ATitle);
	virtual ~TNeuroWindow();
	virtual LPSTR GetClassName() {return "TNeuroWindow";}
	virtual void GetWindowClass(WNDCLASS _FAR&);
	virtual void WMSize( RTMessage ) = [ WM_FIRST + WM_SIZE ];
	virtual void WMMove( RTMessage ) = [ WM_FIRST + WM_MOVE ];
	virtual void WMCreate( RTMessage )      = [ WM_FIRST + WM_CREATE ];
	virtual void WMCommand( RTMessage )  = [ WM_FIRST + WM_COMMAND ];
  virtual void WMChar( RTMessage )        = [ WM_FIRST + WM_CHAR ];
	// Scrollbars
	virtual void WMHScroll(RTMessage Msg)     = [ WM_FIRST + WM_HSCROLL ];
	virtual void WMVScroll(RTMessage Msg)     = [ WM_FIRST + WM_VSCROLL ];
  // MainMenue
	virtual void CMLoad( RTMessage Msg )    = [ CM_FIRST + CM_Load ];
	virtual void CMSave( RTMessage Msg )    = [ CM_FIRST + CM_Save ];
	virtual void CMSaveAs( RTMessage Msg )  = [ CM_FIRST + CM_SaveAs ];
	virtual void CMEdit( RTMessage Msg )    = [ CM_FIRST + CM_Edit ];
	virtual void CMDLoad( RTMessage Msg )   = [ CM_FIRST + CM_DLoad ];
	virtual void CMWNew( RTMessage Msg )    = [ CM_FIRST + CM_WNew ];
	virtual void CMWLoad( RTMessage Msg )   = [ CM_FIRST + CM_WLoad ];
	virtual void CMWSave( RTMessage Msg )   = [ CM_FIRST + CM_WSave ];
	virtual void CMWSaveAs( RTMessage Msg ) = [ CM_FIRST + CM_WSaveAs ];
	virtual void CMReset( RTMessage Msg )   = [ CM_FIRST + CM_Reset ];
	virtual void CMALoad( RTMessage Msg )   = [ CM_FIRST + CM_ALoad ];
	virtual void CMASave( RTMessage Msg )   = [ CM_FIRST + CM_ASave ];
	virtual void CMASaveAs( RTMessage Msg ) = [ CM_FIRST + CM_ASaveAs ];
	virtual void CMExit( RTMessage Msg )    = [ CM_FIRST + CM_EXIT ];
	virtual void CMOBatch( RTMessage Msg )    = [ CM_FIRST + CM_OBatch ];
	virtual void CMOScale( RTMessage Msg )    = [ CM_FIRST + CM_OScale ];
	virtual void CMShowMsgWindow( RTMessage Msg )    = [ CM_FIRST + CM_ShowMsgWindow ];
	virtual void CMShowInfoWindow( RTMessage Msg )    = [ CM_FIRST + CM_ShowInfoWindow ];
	virtual void CMShowCanvasWindow( RTMessage Msg )    = [ CM_FIRST + CM_ShowCanvasWindow ];
	virtual void CMUseHelp(RTMessage Msg)   = [ CM_FIRST + CM_UseHelp ];
	virtual void CMAbout( RTMessage Msg )   = [ CM_FIRST + CM_About ];

  // DDE-Support
  void DDESetup();
  void DDECleanUp();
  virtual BOOL MatchTopicAndService( HSZ, HSZ );
  virtual BOOL MatchTopicAndItem( HSZ, HSZ );
  virtual HDDEDATA WildConnect( HSZ, HSZ, WORD );
  virtual HDDEDATA DataRequested( WORD );
  virtual void UpdateData();
  static HDDEDATA FAR PASCAL _export CallBack( WORD, WORD, HCONV, HSZ, HSZ, HDDEDATA, DWORD, DWORD );
  int ActualConv(HCONV);

	// others
  char* MarkedText(int); 
	void CMShowUserWindow(int nr);

};

class TInfoWindow : public TBWindow
{
private:
	TEXTMETRIC tmIWindow;
	PTStatic Text[4];
	PTGroupBox Grp;

public:
	TInfoWindow(PTWindowsObject AParent, LPSTR ATitle);
	virtual ~TInfoWindow() {};
	virtual LPSTR GetClassName() {return "TInfoWindow";}
	virtual void GetWindowClass(WNDCLASS _FAR&);
	virtual void Paint(HDC, PAINTSTRUCT _FAR&);
	virtual void SetupWindow();
	virtual void WMCreate(RTMessage) = [ WM_FIRST + WM_CREATE ];
	virtual void WMSize( RTMessage ) = [ WM_FIRST + WM_SIZE ];
};

class TMsgWindow : public TWindow
{
private:
	char MFileName[MAXPATH];

public:
	TMsgWindow(PTWindowsObject AParent, LPSTR ATitle);
	virtual ~TMsgWindow();
	virtual LPSTR GetClassName() {return "TMsgWindow";}
	virtual void GetWindowClass(WNDCLASS _FAR&);
	virtual void Paint(HDC, PAINTSTRUCT _FAR&);
	virtual void SetupWindow();
	virtual void WMCreate(RTMessage) = [ WM_FIRST + WM_CREATE ];
	virtual void WMSize(RTMessage) = [ WM_FIRST + WM_SIZE ];
	// MsgMenue
	virtual void CMMsgClear( RTMessage Msg) = [ CM_FIRST + CM_MsgClear ];
	virtual void CMSave( RTMessage Msg )    = [ CM_FIRST + CM_Save ];
	virtual void CMSaveAs( RTMessage Msg )  = [ CM_FIRST + CM_SaveAs ];
	// others
	int SaveMsg( char* );
};

class TCanvas : public TWindow
{
private:
	PAINTSTRUCT ps;
	LPLOGPALETTE NNLogPalette;
	TState          State;
	POINT pt, ptOld, ptOrg;
	BOOL fCapture;
	BOOL fDrag, lDrag;
//      HCURSOR hArrowCursor, hWaitCursor, hPosCursor, hSizeNSCursor, hSizeWECursor, hSizeNESWCursor, hSizeNWSECursor;
	HMENU hCMenu, hLMenu, hBMenu;
	char AFileName[MAXPATH]; // lokale Layeractivations
						
public:
	TCanvas( PTWindowsObject AParent);
	virtual ~TCanvas();
	virtual void SetupWindow();
	virtual LPSTR GetClassName(void)  {return "TCanvas";};
	virtual void GetWindowClass(WNDCLASS _FAR&);
	virtual void Paint( HDC PaintDC, PAINTSTRUCT& PaintInfo );
	virtual void WMLButtonDown( RTMessage) = [ WM_FIRST + WM_LBUTTONDOWN ];
	virtual void WMLButtonUp( RTMessage )  = [ WM_FIRST + WM_LBUTTONUP ];
	virtual void WMRButtonDown( RTMessage) = [ WM_FIRST + WM_RBUTTONDOWN ];
	virtual void WMMouseMove( RTMessage ) = [ WM_FIRST + WM_MOUSEMOVE ];

  // LayerMenu
  // Activations
	virtual void CMLayerResetAct( RTMessage ) = [ CM_FIRST + CM_LayerResetAct ];
	virtual void CMLayerLoadAct( RTMessage ) = [ CM_FIRST + CM_LayerLoadAct ]; 
	virtual void CMLayerSaveAct( RTMessage ) = [ CM_FIRST + CM_LayerSaveAct ];
	virtual void CMLayerSaveAsAct( RTMessage ) = [ CM_FIRST + CM_LayerSaveAsAct ];
  // Data
	virtual void CMLayerDataClear( RTMessage ) = [ CM_FIRST + CM_LayerDataClear ];
	virtual void CMLayerDataLoad( RTMessage ) = [ CM_FIRST + CM_LayerDataLoad ]; 
  //Parameters
	virtual void CMLayerParams( RTMessage ) = [ CM_FIRST + CM_LayerParams ];

	// BackMenue
	virtual void CMLeftMouseFct1( RTMessage Msg )    = [ CM_FIRST + CM_LM_Weights ];
	virtual void CMLeftMouseFct2( RTMessage Msg )    = [ CM_FIRST + CM_LM_Toggle ];
	virtual void CMRightMouseFct1( RTMessage Msg )   = [ CM_FIRST + CM_RM_Weights ];
	virtual void CMRightMouseFct2( RTMessage Msg )   = [ CM_FIRST + CM_RM_Toggle ];

//  virtual void WMSetCursor( RTMessage) = [ WM_FIRST + WM_SETCURSOR ];
	virtual void WMCreate(RTMessage Msg) = [ WM_FIRST + WM_CREATE ];
// Sonstiges
//  void DrawLayer(HDC, pINT, int);
	int UnitClicked ( RTMessage Msg );
	int LayerClicked( RTMessage Msg );
	void DrawNotRect(RECT&);
};

class TNeuroApp: public TApplication
{
public:
	TNeuroApp( LPSTR name, HANDLE hInstance,
			HANDLE hPrevInstance, LPSTR lpCmd,
			int nCmdShow)
			: TApplication(name, (HINSTANCE)hInstance,
				 (HINSTANCE)hPrevInstance, lpCmd, nCmdShow) {};
	virtual void InitMainWindow( void );
	virtual void InitInstance( void );
	virtual void MessageLoop( void );
  virtual void IdleAction( void );
};

TNeuroApp* PNeuroApp;

class TLayerParamsDialog: public TDialog
{
public:
	TLayerParamsDialog::TLayerParamsDialog(PTWindowsObject hWin, LPSTR szTitel):
		TDialog(hWin,szTitel){};
	virtual void SetupWindow();
	virtual void Ok (RTMessage) = [ID_FIRST + IDOK];
};


class TAboutDialog: public TDialog
{
public:
	TAboutDialog::TAboutDialog(PTWindowsObject hWin, LPSTR szTitel):
		TDialog(hWin,szTitel){}
	virtual void WMLButtonDown( RTMessage) = [ WM_FIRST + WM_LBUTTONDOWN ];

};

int ScaleW=2;
int ScaleA=2;

class TScaleDialog: public TDialog
{
public:
	TScaleDialog::TScaleDialog(PTWindowsObject hWin, LPSTR szTitel):
		TDialog(hWin,szTitel){};
	virtual void SetupWindow();
	virtual void Ok (RTMessage) = [ID_FIRST + IDOK];
};

// Define our own button class to handle char-msg
class TNNButton: public TButton
{
public:
	TNNButton(PTWindowsObject,int,LPSTR,int,int,int,int,BOOL);
//      virtual void WMLButtonDown( RTMessage) = [ WM_FIRST + WM_LBUTTONDOWN ];
	virtual void WMChar( RTMessage Msg) = [ WM_FIRST + WM_CHAR ];
};

TNNButton::TNNButton(PTWindowsObject AParent,int Id,LPSTR Text,int x,int y,int w,int h,BOOL def):
	TButton(AParent,Id,Text,x,y,w,h,def)
{
//TButton::TButton(AParent,Id,Text,x,y,w,h,def);
}

void TNNButton::WMChar(RTMessage Msg)
{       if (Msg.WParam==VK_ESCAPE) bAbortRequested=TRUE; }


//*************************************************************************
//
// Memberfunctions of TNeuroWindow
//
//*************************************************************************
TNeuroWindow::TNeuroWindow(PTWindowsObject AParent, LPSTR ATitle) :
	TBWindow(AParent, ATitle)
{
	int i;
  char szAnzeige[30];
  DWORD TextExt;
	int TextWidth;
	if (NeuroWindow_X>-1) Attr.X=NeuroWindow_X;
	if (NeuroWindow_Y>-1) Attr.Y=NeuroWindow_Y;
	if (NeuroWindow_W>-1) Attr.W=NeuroWindow_W;
	if (NeuroWindow_H>-1) Attr.H=NeuroWindow_H;
	button_row=0;
	AssignMenu("MainMenu"); 
	strcpy(DFileName,DataFile);
	if (!strchr(DFileName,'.')) strcat(DFileName,".dat");
	strcpy(WFileName,WeightFile);
	if (!strchr(WFileName,'.')) strcat(WFileName,".nnw");
	strcpy(AFileName,ActFile);
	if (!strchr(AFileName,'.')) strcat(AFileName,".nna");
	strcpy(LFileName,LayoutFile);
	if (!strchr(LFileName,'.')) strcat(LFileName,".nnl");
	Show_MsgWindow = Show_InfoWindow = Show_CanvasWindow = TRUE;
	Canvas =  new TCanvas(this);
	HDC hDC=GetDC(HWindow);
	GetTextMetrics(hDC,&tmNWindow);
	int tw=tmNWindow.tmAveCharWidth;
  int th=tmNWindow.tmHeight;

	// Buttons erzeugen
	for (i=0;i<buttons_used;i++)
		{
		Slider[i]=NULL;
		switch (buttons[i].typ) {
		case BUTTON:
			Knopf[i]=new TNNButton(this, ID_Knopf+i,
				MarkedText(i),
				16, 16+button_row*(BUTTON_HEIGHT+BUTTON_DIST),
				BUTTON_WIDTH, BUTTON_HEIGHT, FALSE);
			button_row++;
			buttons[i].id=ID_Knopf+i;
			break;
		case SWITCH:
			new TBGroupBox(this,ID_NGroup,"",16,16+button_row*(BUTTON_HEIGHT+BUTTON_DIST),
				BUTTON_WIDTH,BUTTON_HEIGHT);
			Knopf[i]=new TBCheckBox(this, ID_Knopf+i,
				MarkedText(i),
				20, 18+button_row*(BUTTON_HEIGHT+BUTTON_DIST),
				BUTTON_WIDTH-8, BUTTON_HEIGHT-4, NULL);
			button_row++;
			buttons[i].id=ID_Knopf+i;
			break;
		case TWIN_1:
			Knopf[i]=new TNNButton(this, ID_Knopf+i, 
				MarkedText(i),
				16, 16+button_row*(BUTTON_HEIGHT+BUTTON_DIST), 
				BUTTON_WIDTH/2-2, BUTTON_HEIGHT, FALSE);
			buttons[i].id=ID_Knopf+i;
			break;
		case TWIN_2:
			Knopf[i]=new TNNButton(this, ID_Knopf+i, 
				MarkedText(i),
				16+BUTTON_WIDTH/2+2,
				16+button_row*(BUTTON_HEIGHT+BUTTON_DIST), 
				BUTTON_WIDTH/2-2, BUTTON_HEIGHT, FALSE);
			button_row++;
			buttons[i].id=ID_Knopf+i;
			break;
		case SLIDER:
			Knopf[i]=NULL;
			new TBGroupBox(this,ID_NGroup,"",16,16+button_row*(BUTTON_HEIGHT+BUTTON_DIST),
				BUTTON_WIDTH,2*BUTTON_HEIGHT+BUTTON_DIST);
			TextExt=GetTextExtent(hDC,buttons[i].text,strlen(buttons[i].text));
      TextWidth=LOWORD(TextExt);
			new TBStatic(this, ID_Text+i,
				MarkedText(i), 20,
				24+button_row*(BUTTON_HEIGHT+BUTTON_DIST),
				TextWidth,HIWORD(TextExt),strlen(buttons[i].text));
			switch (buttons[i].val_typ) {
				case 'f':
					sprintf(szAnzeige,buttons[i].val_fmt,*((float*)buttons[i].var));
					break;
				case 's':
					sprintf(szAnzeige,buttons[i].val_fmt,*((short*)buttons[i].var));
					break;
				case 'i':
					sprintf(szAnzeige,buttons[i].val_fmt,*((int*)buttons[i].var));
					break;
				case 'l':
					sprintf(szAnzeige,buttons[i].val_fmt,*((long*)buttons[i].var));
					break;
				}
			TextExt=GetTextExtent(hDC,szAnzeige,strlen(szAnzeige));
			Anzeige[i]=new TBStatic(this, ID_Anzeige+i,
				szAnzeige, 20+TextWidth+5,
				24+button_row*(BUTTON_HEIGHT+BUTTON_DIST),
				BUTTON_WIDTH-25-TextWidth,HIWORD(TextExt),strlen(szAnzeige));
			button_row++;
			Slider[i]=new TScrollBar(this, ID_Knopf+i, 
				20,     16+button_row*(BUTTON_HEIGHT+BUTTON_DIST), 
				BUTTON_WIDTH-8, BUTTON_HEIGHT-4, TRUE);
			button_row++;
			buttons[i].id=ID_Knopf+i;
			break;
		case METER3:
#ifdef NNCONTROLS
			NNControl[i]=new TBar(this, buttons[i].text,
				16, 16+button_row*(BUTTON_HEIGHT+BUTTON_DIST),
				BUTTON_WIDTH, 6*BUTTON_HEIGHT+5*BUTTON_DIST,
				buttons[i].limit1,buttons[i].limit2,buttons[i].value,buttons[i].val_min,buttons[i].val_max);
			button_row+=6;
			buttons[i].id=ID_Knopf+i;
#else
			nn_fatalerror(SOURCENAME,"You have to include 'nncontrl.cpp' into your\nproject and to define NNCONTROLS in 'neurodef.h'\nfor using METER3!");
#endif
			break;
			}
		}
	// Control-Group
	NGroup = new TBGroupBox(this,ID_NGroup,"",8,8,BUTTON_WIDTH+20,
		20-BUTTON_DIST+button_row*(BUTTON_HEIGHT+BUTTON_DIST)); 

	EnableKBHandler();
	ReleaseDC(HWindow,hDC);

  DDESetup();
}

TNeuroWindow::~TNeuroWindow()
{
	DDECleanUp();
}

void TNeuroWindow::GetWindowClass(WNDCLASS _FAR& winclass)
{ TBWindow::GetWindowClass(winclass);
	winclass.hIcon = LoadIcon(GetApplication()->hInstance, ICON_MAIN);
}

void TNeuroWindow::WMMove(RTMessage Msg)
{ static int neu=1;
	int x=Attr.X;
	int y=Attr.Y;
	TWindow::WMMove(Msg);
	if (!neu && !IsIconic(InfoWindow->HWindow))
		{
  	x=InfoWindow->Attr.X + Attr.X - x;
	  y=InfoWindow->Attr.Y + Attr.Y - y;
		SetWindowPos(InfoWindow->HWindow,0,x,y,0,0,SWP_NOZORDER|SWP_NOSIZE|SWP_NOACTIVATE);
		}
	neu=0;
}

void TNeuroWindow::WMSize(RTMessage Msg)
{ RECT R;
	int i;
  static int iconic=0;
  static int OldStat[MaxUserWindow+2];
	TBWindow::WMSize(Msg);
	if ( (Msg.WParam != SIZEICONIC) )
	{
		GetWindowRect(HWindow, &R);
		if (R.bottom-R.top<66+button_row*(BUTTON_HEIGHT+BUTTON_DIST))
			SetWindowPos(HWindow,0,0,0,R.right-R.left,
				66+button_row*(BUTTON_HEIGHT+BUTTON_DIST),SWP_NOMOVE);
		GetClientRect(HWindow, &R);
		MoveWindow(Canvas->HWindow, X0 , 8, R.right - X0 - 8, R.bottom - 16, TRUE);
		if (iconic)
			{
 			iconic=0;
			Show_InfoWindow=!OldStat[0];
			CMShowInfoWindow(Msg);
			Show_MsgWindow=!OldStat[1];
			CMShowMsgWindow(Msg);
	    for (i=0;i<MaxUserWindow;i++)
				if (UserWindow[i].ID>=0)
				  {
					UserWindow[i].Show=!OldStat[i+2];
  			  CMShowUserWindow(i);
				  }
			}
		}
  else
		{
 		iconic=1;
    OldStat[0]=Show_InfoWindow;
		Show_InfoWindow=1;
		CMShowInfoWindow(Msg);
    OldStat[1]=Show_MsgWindow;
		Show_MsgWindow=1;
		CMShowMsgWindow(Msg);
    for (i=0;i<MaxUserWindow;i++)
			if (UserWindow[i].ID>=0)
			  {
	      OldStat[i+2]=UserWindow[i].Show;
			  UserWindow[i].Show=1;
  		  CMShowUserWindow(i);
			  }
		}
}

void TNeuroWindow::WMCreate(RTMessage Msg)
{ TWindow::WMCreate(Msg);
	hNWindow=HWindow;
	hNMenu=GetMenu(HWindow);
  // Set Slider-Ranges
	for (int i=0; i<buttons_used; i++)
	  if (Slider[i])
			{Slider[i]->SetRange(0,1000);
			 Slider[i]->SetPosition(SliderVarToPosNr(i));
			}
}

char* TNeuroWindow::MarkedText(int b)
{ int   i=0,j=0;
	char* text = (char*)malloc(22);
	if (!text) return NULL;
	while (i<strlen(buttons[b].text) && i<20)
		{/* Hotkey im Buttontext unterstreichen, d.h. '&' einfuegen */
		if (buttons[b].text[i]==buttons[b].key)
			{text[j]='&'; j++;}
		text[j]=buttons[b].text[i];
    j++;
    i++;
		}
  text[j]='\0';
	return text;
}

void TNeuroWindow::WMCommand(RTMessage Msg)
{ int i=0;
	if (Msg.WParam>=CM_ShowFirstUserWindow && Msg.WParam<=CM_ShowLastUserWindow)
		{// Handle User Window ON/OFFs via Window-Menu
		for (i=0;i<MaxUserWindow;i++)
			if (UserWindow[i].ID==Msg.WParam)
				 CMShowUserWindow(i);
		}
	if (Msg.WParam<ID_FIRST) TWindowsObject::WMCommand(Msg);
  else
		if ( Msg.LP.Hi == BN_CLICKED && !in_break)
			{
			while( (i<buttons_used) && ((WORD)buttons[i].id != Msg.WParam) ) i++;
			if (i<buttons_used)
				{if (buttons[i].typ==SWITCH)
					buttons[i].status=!buttons[i].status;
				button_pressed = i;     // Buttonnr. merken
				SetCursor(hWaitCursor);
				if (buttons[i].call) (buttons[i].call)();
				SetCursor(hArrowCursor);
        }
			}
}

void TNeuroWindow::WMChar(RTMessage Msg) {}

void TNeuroWindow::WMHScroll(RTMessage Msg)
{ int i=0;
	char szAnzeige[30];
	float value=0.0;
	unsigned int id=GetWindowWord((HWND)Msg.LP.Hi, GWW_ID);
	while (i<buttons_used && id!=buttons[i].id) i++;
	if(i<buttons_used && Slider[i])
		{
		float res=buttons[i].val_res;
		switch (buttons[i].val_typ) {
				case 'f':
					value=*((float*)buttons[i].var);
					break;
				case 'c':
					value=(float)(*((short int*)buttons[i].var));
					break;
				case 'i':
					value=(float)(*((int*)buttons[i].var));
					break;
				case 'l':
					value=(float)(*((long*)buttons[i].var));
					break;
				}
    float oldvalue = value;
		switch (Msg.WParam)
			{
			case SB_PAGEUP:
				value-=10.0*res;
				break;
			case SB_LINEUP:
				value-=res;
				break;
			case SB_PAGEDOWN:
				value+=10.0*res;
				break;
			case SB_LINEDOWN:
				value+=res;
				break;
			case SB_TOP:
				value=(float)buttons[i].val_min;
				break;
			case SB_BOTTOM:
				value=(float)buttons[i].val_max;
				break;
			case SB_THUMBTRACK:
				break;
      case SB_THUMBPOSITION:
				value=SliderPosToVar(Msg.LP.Lo);
				break;
			case SB_ENDSCROLL:
				break;
      }
		value=max((float)buttons[i].val_min, min((float)buttons[i].val_max,value));
		Slider[i]->SetPosition(SliderVarToPosNr(i));
		if (value == oldvalue) return;
		switch (buttons[i].val_typ) {
				case 'f':
					sprintf(szAnzeige,buttons[i].val_fmt,value);
					*((float*)buttons[i].var)=value;
					break;
				case 'c':
					sprintf(szAnzeige,buttons[i].val_fmt,(short)value);
					*((short int*)buttons[i].var)=(short)value;
					break;
				case 'i':
					sprintf(szAnzeige,buttons[i].val_fmt,(int)value);
					*((int*)buttons[i].var)=(int)value;
					break;
				case 'l':
					sprintf(szAnzeige,buttons[i].val_fmt,(long)value);
					*((long*)buttons[i].var)=(long)value;
					break;
				}
		Anzeige[i]->SetText(szAnzeige);
		button_pressed = i;                     // Buttonnr. merken
		if (buttons[i].call) (buttons[i].call)();
		}
}

void TNeuroWindow::WMVScroll(RTMessage Msg)
{	pprintf("VScroll: %x, %x, %x\n",Msg.WParam,Msg.LP.Hi,Msg.LP.Lo);
}

//Menufunctions
void TNeuroWindow::CMLoad(RTMessage)
{ if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILEOPEN, strcpy(LFileName,"*.nnl"))) == IDOK )
	LoadLayout(LFileName);
}                                                    

void TNeuroWindow::CMSave(RTMessage Msg)
{if (strlen(LFileName)>5 || LFileName[0]!='*')
		{if (SaveLayout(LFileName)==IDRETRY)
			CMSaveAs(Msg);}
	else
		CMSaveAs(Msg);          
}

void TNeuroWindow::CMSaveAs(RTMessage Msg)
{ if(LFileName[0]=='*')
		{strcpy(LFileName,LayoutFile);
		if (!strchr(LFileName,'.')) strcat(LFileName,".nnl");
		}
	if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILESAVE, LFileName)) == IDOK )
		{if (SaveLayout(LFileName)==IDRETRY)
			CMSaveAs(Msg);}
}

void TNeuroWindow::CMEdit( RTMessage)
{ EditLayout = !EditLayout;       
	HMENU hMenu = GetMenu(HWindow);
	if (EditLayout)
		{
		CheckMenuItem(hMenu,CM_Edit,MF_BYCOMMAND|MF_CHECKED);
		EnableMenuItem(hMenu,CM_Load,MF_BYCOMMAND|MF_ENABLED);
		EnableMenuItem(hMenu,CM_Save,MF_BYCOMMAND|MF_ENABLED);
		EnableMenuItem(hMenu,CM_SaveAs,MF_BYCOMMAND|MF_ENABLED);
		MouseMode.Select = FALSE;
		MouseMode.DragStatus = 0;
		MouseMode.Button = 0;
		MouseMode.Resize = 0;
		}
	else
		{
		CheckMenuItem(hMenu,CM_Edit,MF_BYCOMMAND|MF_UNCHECKED);
		EnableMenuItem(hMenu,CM_Load,MF_BYCOMMAND|MF_GRAYED);
		EnableMenuItem(hMenu,CM_Save,MF_BYCOMMAND|MF_GRAYED);
		EnableMenuItem(hMenu,CM_SaveAs,MF_BYCOMMAND|MF_GRAYED);
		DrawMarkers(MouseMode.Rect); // Delete last markers
    MouseMode.Select = FALSE;
		MouseMode.DragStatus = 0;
		}
}


BOOL LoadFile(char* title,char* filter, char* FileName, int* Index)
{
 OPENFILENAME OpenStruct;

 char *DefaultExtension [] = {"#", "xy", "x", "y"};
 memset (&OpenStruct, 0, sizeof (OpenStruct));
 OpenStruct.lStructSize  = sizeof (OpenStruct);
 OpenStruct.hwndOwner    = hNWindow;
 OpenStruct.lpstrFilter  = filter;
 OpenStruct.nFilterIndex = *Index;
 OpenStruct.lpstrFile    = FileName;
 OpenStruct.nMaxFile     = MAXPATH;
 OpenStruct.lpstrDefExt  = DefaultExtension[*Index];
 OpenStruct.lpstrTitle   = title;
 OpenStruct.Flags        = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;

 if(GetOpenFileName(&OpenStruct))
 	{
  *Index=OpenStruct.nFilterIndex;
  return 1;
	}
 else return 0;
}

void TNeuroWindow::CMDLoad(RTMessage Msg)
{ pINT l=layer_clicked; //layer to load patterns into
  char drive[MAXDRIVE];
  char dir[MAXDIR];
  char file[MAXFILE];
  char ext[MAXEXT];
	int flags;
  int erg;
 char    OutputDirectory [MAXPATH];
 char    CurrentDirectory [MAXPATH];
 int     LoadType;
 BOOL    SelectedOnly = FALSE;

 strcpy (OutputDirectory, "C:\\");
 OutputDirectory [0] = getdisk () + 'A';
 getcurdir (0, CurrentDirectory);
 strcat (OutputDirectory, CurrentDirectory);

/*
 if (GetApplication ()->ExecDialog (new TDFileDialog (OutputDirectory, &LoadType, &SelectedOnly,
                                                  this)) == IDOK)
//	if( GetApplication()->ExecDialog(new TFileDialog(this, "DATA_OPEN", strcpy(DFileName,"*.pat"))) == IDOK )
*/
  char *title="Load Data";
	char *filter="DAT-Files\0*.dat\0PAT-Files\0*.pat\0All Files\0*.*\0";
  int Index=DAT_FILE;
  if (LoadFile(title,filter,DFileName,&Index))
    {
		free_patterns(l);
    switch (Index) {
		case DAT_FILE:  
		  erg=load_patterns(DFileName,l);
      break;
    case PAT_FILE:
			erg=load_data(DFileName,l);
      break;
    	}
    if (!erg)
			{if (MessageBox(HWindow, DFileName, "Error loading data:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMDLoad(Msg);
	  	}
		else
			{pprintf("Data loaded from file '%s'.\n",DFileName);
			draw_layers();
			}
		}
}

void TNeuroWindow::CMWNew(RTMessage)
{   connect_all_new(); }

void TNeuroWindow::CMWLoad(RTMessage Msg)
{ if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILEOPEN, strcpy(WFileName,"*.nnw"))) == IDOK )
		if (!load_weights(WFileName))
			{if (MessageBox(HWindow, WFileName, "Error loading weights:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMWLoad(Msg);
			}    
		else
			{pprintf("Weights loaded from file '%s'.\n",WFileName);
			draw_layers();
      }
}

void TNeuroWindow::CMWSave(RTMessage Msg)
{       if (strlen(WFileName)>5 || WFileName[0]!='*')
		if (!save_weights(WFileName))
			{if (MessageBox(HWindow, WFileName, "Error saving weights:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMWSaveAs(Msg);
      }
		else
	pprintf("Weights saved in file '%s'.\n",WFileName);
	else
		CMWSaveAs(Msg);         
}

void TNeuroWindow::CMWSaveAs(RTMessage Msg)
{ if(WFileName[0]=='*')
	 {strcpy(WFileName,WeightFile);
		if (!strchr(WFileName,'.')) strcat(WFileName,".nnw");
	 }
	if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILESAVE,WFileName)) == IDOK )
		if (!save_weights(WFileName))
			{if (MessageBox(HWindow, WFileName, "Error saving weights:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMWSaveAs(Msg);
      }
		else
	pprintf("Weights saved in file '%s'.\n",WFileName);
}

void TNeuroWindow::CMReset(RTMessage)
{ reset_layers();
  draw_layers();
}

void TNeuroWindow::CMALoad(RTMessage Msg)
{ if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILEOPEN, strcpy(AFileName,"*.nna"))) == IDOK )
		if (!load_activations(AFileName))
			{if (MessageBox(HWindow, AFileName, "Error loading activations:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMALoad(Msg);
      }
		else
			{pprintf("Activations loaded from file '%s'.\n",AFileName);
			draw_layers();
			}
}

void TNeuroWindow::CMASave(RTMessage Msg)
{       if (strlen(AFileName)>5 || AFileName[0]!='*')
		if (!save_activations(AFileName))
			{if (MessageBox(HWindow, AFileName, "Error saving Activations:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMASaveAs(Msg);
      }
		else
	pprintf("Activations saved in file '%s'.\n",AFileName);
	else
		CMASaveAs(Msg);         
}

void TNeuroWindow::CMASaveAs(RTMessage Msg)
{ if(AFileName[0]=='*')
	 {strcpy(AFileName,ActFile);
		if (!strchr(AFileName,'.')) strcat(AFileName,".nna");
	 }
	if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILESAVE,AFileName)) == IDOK )
		if (!save_activations(AFileName))
			{if (MessageBox(HWindow, AFileName, "Error saving Activations:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMASaveAs(Msg);
      }
		else
	pprintf("Activations saved in file '%s'.\n",AFileName);
}

void TNeuroWindow::CMExit(RTMessage)
{ shutdown();
}


void TNeuroWindow::CMOBatch(RTMessage)
{ BatchRun=!BatchRun;
	HMENU hMenu = GetMenu(HWindow);
	if (BatchRun)
		CheckMenuItem(hMenu,CM_OBatch,MF_BYCOMMAND|MF_CHECKED);
	else
		CheckMenuItem(hMenu,CM_OBatch,MF_BYCOMMAND|MF_UNCHECKED);
}

void TNeuroWindow::CMOScale(RTMessage)
{GetApplication()->ExecDialog(new TScaleDialog(this,"ScaleDlg"));
}

void TNeuroWindow::CMShowMsgWindow(RTMessage Msg)
{ Show_MsgWindow=!Show_MsgWindow;
	HMENU hMenu = GetMenu(HWindow);
	if (Show_MsgWindow)
		{
		CheckMenuItem(hMenu,CM_ShowMsgWindow,MF_BYCOMMAND|MF_CHECKED);
		MsgWindow->Show(SW_SHOWNORMAL);
		SetFocus(hMWindow);
		}
	else
		{
		CheckMenuItem(hMenu,CM_ShowMsgWindow,MF_BYCOMMAND|MF_UNCHECKED);
		MsgWindow->Show(SW_HIDE);
		SetFocus(HWindow);
		}
}

void TNeuroWindow::CMShowInfoWindow(RTMessage Msg)
{ Show_InfoWindow=!Show_InfoWindow;
	HMENU hMenu = GetMenu(HWindow);
	if (Show_InfoWindow)
		{
    InfoWindow->Attr.X=max(0,InfoWindow->Attr.X);
    InfoWindow->Attr.Y=max(0,InfoWindow->Attr.Y);
		SetWindowPos(InfoWindow->HWindow,0,InfoWindow->Attr.X,InfoWindow->Attr.Y,0,0,SWP_NOZORDER|SWP_NOSIZE);
		CheckMenuItem(hMenu,CM_ShowInfoWindow,MF_BYCOMMAND|MF_CHECKED);
		InfoWindow->Show(SW_SHOWNORMAL);
		}
	else
		{
		CheckMenuItem(hMenu,CM_ShowInfoWindow,MF_BYCOMMAND|MF_UNCHECKED);
		InfoWindow->Show(SW_HIDE);
		}
}

void TNeuroWindow::CMShowCanvasWindow(RTMessage Msg)
{ RECT R;
  static RECT R1;
	GetWindowRect(NGroup->HWindow,&R);
	Show_CanvasWindow=!Show_CanvasWindow;
	HMENU hMenu = GetMenu(HWindow);
	if (Show_CanvasWindow)
		{
		CheckMenuItem(hMenu,CM_ShowCanvasWindow,MF_BYCOMMAND|MF_CHECKED);
		SetWindowPos(hNWindow,0,0,0,R1.right-R1.left,R1.bottom-R1.top,SWP_NOMOVE);
		Canvas->Show(SW_SHOWNORMAL);
		}
	else
		{
		GetWindowRect(HWindow,&R1);
		CheckMenuItem(hMenu,CM_ShowCanvasWindow,MF_BYCOMMAND|MF_UNCHECKED);
		SetWindowPos(hNWindow,0,0,0,R.right-R.left+25,R.bottom-R.top+20,SWP_NOMOVE);
		Canvas->Show(SW_HIDE);
		}
}

void TNeuroWindow::CMShowUserWindow(int nr)
{ UserWindow[nr].Show=!UserWindow[nr].Show;
	HMENU hMenu = GetMenu(HWindow);
	if (UserWindow[nr].Show)
		{
		CheckMenuItem(hMenu,UserWindow[nr].ID,MF_BYCOMMAND|MF_CHECKED);
		UserWindow[nr].ptr->Show(SW_SHOWNORMAL);
		}
	else
		{
		CheckMenuItem(hMenu,UserWindow[nr].ID,MF_BYCOMMAND|MF_UNCHECKED);
		UserWindow[nr].ptr->Show(SW_HIDE);
		}
}

void TNeuroWindow::CMUseHelp(RTMessage)
{       WinHelp(HWindow,NULL,HELP_HELPONHELP,0L); }                                                    

void TNeuroWindow::CMAbout(RTMessage)
{       GetApplication()->ExecDialog(new TAboutDialog(this, "About")); }


//--------------- DDE Functions ----------------------
// (based on the OWL-Demo ddsvr.cpp)

// must be called before nn_init() !
void dde_init()
{  int i;
   for (i=0;i<NN_DDE_MaxFcts;i++)
    	{// Init up to 5 default functions
      switch (i) {
			  case 0:
				  NN_DDE_Fcts[i].name="SET_DELIMITER";
				  NN_DDE_Fcts[i].ptr=NULL;
					break;
        case 1:
				  NN_DDE_Fcts[i].name="DRAW_LAYERS";
				  NN_DDE_Fcts[i].ptr=draw_layers;
          break;
        case 2:
				  NN_DDE_Fcts[i].name="RESET_LAYERS";
				  NN_DDE_Fcts[i].ptr=reset_layers;
					break;

        default:
				  NN_DDE_Fcts[i].name="";
				  NN_DDE_Fcts[i].ptr=NULL;
      	  break;
        }
      }
   for (i=0;i<NN_DDE_MaxVars;i++)
    	{
		  NN_DDE_Vars[i].name="";
		  NN_DDE_Vars[i].ptr=NULL;
			}
}

/*
 * The code below sets up the DDE call back function that is used by the
 * DDE Management Library to carry out data transfers between applications.
*/
void TNeuroWindow::DDESetup()
{
	int i;
	char nrstr[12];
	pStaticThis = this;

	idInst = 0;
	for (i=0;i<MAX_CONV;i++)
			{
			hConv[i] = 0;
			ActualTopic[i]=ActualItem[i]=(HSZ)-1;
      ActualUnitStart=ActualUnitEnd=0;
			tfIsHot[i]=FALSE;
			}

	tfNewData = FALSE;

	hszService = hszTopic[0] = hszItem0[0] = 0;

	// Set Nr of DDE-Items for each topic
  NR_ITEMS[0]=L_anz;
	NR_ITEMS[1]=0; // not used
  NR_ITEMS[2]=1;
  NR_ITEMS[3]=NN_DDE_MaxVars;

  lpfnCallBack = MakeProcInstance( (FARPROC)TNeuroWindow::CallBack, GetApplication()->hInstance );
  if( lpfnCallBack != NULL )
  	{
    if( DdeInitialize( &idInst, (PFNCALLBACK)lpfnCallBack, 0, 0L ) == DMLERR_NO_ERROR )
    	{
/*
 The strings below are the 'Service', 'Topic' and 'Item' identifiers that
 this application makes available, through DDE, to other applications.
*/
        hszService = DdeCreateStringHandle( idInst, ProgName, CP_WINANSI );

        hszTopic[0] = DdeCreateStringHandle( idInst, "Layer", CP_WINANSI );
        hszTopic[1] = DdeCreateStringHandle( idInst, "Unit", CP_WINANSI );
        hszTopic[2] = DdeCreateStringHandle( idInst, "Network", CP_WINANSI );
        hszTopic[3] = DdeCreateStringHandle( idInst, "Variable", CP_WINANSI );

		    for (i=0;i<NR_ITEMS[0];i++)
	        hszItem0[i] = DdeCreateStringHandle( idInst, L_Tab[i].titel, CP_WINANSI );

		    for (i=0;i<NR_ITEMS[1];i++)
          {
          wsprintf(nrstr,"%d",i);
	        hszItem1[i] = DdeCreateStringHandle( idInst, nrstr , CP_WINANSI );
          }

        hszItem2[0] = DdeCreateStringHandle( idInst, "Function", CP_WINANSI );

		    for (i=0;i<NR_ITEMS[3];i++)
          { /* Variables */
	        hszItem3[i] = NULL;
          }

        if( (hszService != NULL) && (hszTopic[0] != NULL) && (hszTopic[1] != NULL) && (hszItem0[0] != NULL) )
        {
            if( DdeNameService( idInst, hszService, NULL, DNS_REGISTER ) != 0 )
            { // evtl. modify menu
            } else {
                MessageBox( HWindow, "DDE registration failed.", Title, MB_ICONSTOP );
//                    PostQuitMessage( 0 );
            }
        } else {
            MessageBox( HWindow, "DDE string creation failed.", Title, MB_ICONSTOP );
//                PostQuitMessage( 0 );
        }
	    }
		else
			{
      MessageBox( HWindow, "DDE initialization failed.", Title, MB_ICONSTOP );
//    PostQuitMessage( 0 );
    	}
    }
	else
		{
    MessageBox( HWindow, "DDE setup of callback failed.", Title, MB_ICONSTOP );
//  PostQuitMessage( 0 );
    }
}

/*
 * This clean up is required for those resources that were allocated during the
 * DDE conversation.
*/
void TNeuroWindow::DDECleanUp()
{   int i;
    for (i=0;i<MAX_CONV;i++)
	    if( hConv[i] != 0 )
	  	  {
        DdeDisconnect( hConv[i] );     // Let the other party know we are leaving
        hConv[i] = 0;
  		  }
    if( idInst != 0 )
    {
        DdeNameService( idInst, hszService, NULL, DNS_UNREGISTER );
        if( hszService != NULL )
        {
            DdeFreeStringHandle( idInst, hszService );
            hszService = NULL;
        }
        for (i=0;i<3;i++)
          if( hszTopic[i] != NULL )
          {
              DdeFreeStringHandle( idInst, hszTopic[i] );
              hszTopic[i] = NULL;
          }

        for (i=0;i<NR_ITEMS[0];i++)
          if( hszItem0[i] != NULL )
	          {
            DdeFreeStringHandle( idInst, hszItem0[i] );
            hszItem0[i] = NULL;
  	        }
        for (i=0;i<NR_ITEMS[1];i++)
          if( hszItem1[i] != NULL )
    		    {
            DdeFreeStringHandle( idInst, hszItem1[i] );
            hszItem1[i] = NULL;
        	  }

        for (i=0;i<NR_ITEMS[2];i++)
          if( hszItem2[i] != NULL )
            {
	          DdeFreeStringHandle( idInst, hszItem2[i] );
  	        hszItem2[i] = NULL;
            }

        DdeUninitialize(idInst);
        idInst = 0;
    }

    if( lpfnCallBack != NULL )
    {
        FreeProcInstance( lpfnCallBack );
        lpfnCallBack = NULL;
    }
}

/*
This function searches the conversation table for the actual conversation and
returns the associated nr
*/
int TNeuroWindow::ActualConv( HCONV hConv1)
{	int i;
	for (i=0;i<MAX_CONV;i++)
  	if (hConv1==hConv[i]) return i;
	return -1;
}
/*
This function is used to compare incoming Topic and Service requests.
*/
BOOL TNeuroWindow::MatchTopicAndService( HSZ hsz1, HSZ hsz2 )
{   int i;
		if( DdeCmpStringHandles( hszService, hsz2 ) == 0 )
			for (i=0;i<NR_TOPICS;i++)
				if( DdeCmpStringHandles( hszTopic[i], hsz1 ) == 0 )
					{
					ActualTopic[act_conv]=(HSZ)i;
					return TRUE;
					}

		return FALSE;
}

/*
This function is used to compare incoming Topic and Item pair requests.
*/
BOOL TNeuroWindow::MatchTopicAndItem( HSZ hsz1, HSZ hsz2 )
{   int i,j,s,len=0;
		char szItem[30], szItem1[30];
    for (i=0;i<NR_TOPICS;i++)
	    if( DdeCmpStringHandles( hszTopic[i], hsz1 ) == 0 )
  	  {
        switch (i) {
	        case 0: // Layername || Layername:from || Layername:from,to
  	      	for (j=0;j<NR_ITEMS[0];j++)
              {// Item=Layername
    	        if( DdeCmpStringHandles( hszItem0[j], hsz2 ) == 0 )
      	        {ActualTopic[act_conv]=(HSZ)i;
        	      ActualItem[act_conv]=(HSZ)j;
                ActualUnitStart=L_Tab[j].first;
                ActualUnitEnd=L_Tab[j].last;
	        	    return TRUE;
                }

              // Item=Layername:x[,y] ?
              DdeQueryString(idInst, hsz2, (LPSTR)&szItem, sizeof(szItem), CP_WINANSI);
              DdeQueryString(idInst, hszItem0[j], (LPSTR)&szItem1, sizeof(szItem1), CP_WINANSI);
              AnsiUpperBuff((LPSTR)&szItem,strlen(szItem));
              AnsiUpperBuff((LPSTR)&szItem1,strlen(szItem1));
    	        if(!strncmp(szItem,szItem1,len=strlen(szItem1))
								&& szItem[len]==':' )
      	        {
							  s=sscanf(szItem+len+1," %i , %i",&ActualUnitStart,&ActualUnitEnd);
                if (s==0)
                	{
									ActualUnitStart=ActualUnitEnd=0;
                  }
                else
                	{
									ActualTopic[act_conv]=(HSZ)i;
  	      	      ActualItem[act_conv]=(HSZ)j;
                  if (s<2)
                  	ActualUnitEnd=ActualUnitStart;

                  ActualUnitStart = max(ActualUnitStart,L_Tab[j].first);
									ActualUnitEnd = min(ActualUnitEnd,L_Tab[j].last);
		        	    return TRUE;
                  }
                }
              }
          	break;
          case 1:
          	break;
          case 2: // Topic=Network, Item=Function
  	      	for (j=0;j<NR_ITEMS[2];j++)
    	        if( DdeCmpStringHandles( hszItem2[j], hsz2 ) == 0 )
      	        {ActualTopic[act_conv]=(HSZ)i;
        	      ActualItem[act_conv]=(HSZ)j;
	        	    return TRUE;
                }
          	break;
          case 3: // Variables
          	break;
          }

    	}
    return FALSE;
}

/*
This function responds to 'system wide' polling of any available
Services with specific Topics, any Topics with specific Services or any
Services with any Topics.  It simply replies with a match, if there is
one, so that the polling application can open data discussions later, if
desired.
*/
HDDEDATA TNeuroWindow::WildConnect( HSZ hsz1, HSZ hsz2, WORD wFmt )
{
    HSZPAIR hszpTemp[NR_TOPICS+1]; 
    int i,j;
    
    if( (hsz1 == NULL) && ( (hsz2 == NULL) || (hsz2 == hszService ) ))
    	{
      j=0;
      for (i=0;i<NR_TOPICS;i++)
    	  if (!DdeCmpStringHandles( hsz1, hszTopic[i] ))
      	  {
          hszpTemp[j].hszSvc = hszService;
          hszpTemp[j].hszTopic = hszTopic[i];
          j++;
				  }
      hszpTemp[j].hszSvc = hszpTemp[j].hszTopic = NULL;
	  	return DdeCreateDataHandle( idInst, (LPBYTE)&hszpTemp[0], j*sizeof( HSZPAIR ), 0L, 0, wFmt, 0 );
      }
    else
      for (i=0;i<NR_TOPICS;i++)
			  if( (DdeCmpStringHandles( hsz1, hszTopic[i]) == 0) && (hsz2 == NULL) )
  		  	{
          hszpTemp[0].hszSvc = hszService;
          hszpTemp[0].hszTopic = hszTopic[i];
          hszpTemp[1].hszSvc = hszpTemp[1].hszTopic = NULL;
	  	    return DdeCreateDataHandle( idInst, (LPBYTE)&hszpTemp[0], sizeof( HSZPAIR ), 0L, 0, wFmt, 0 );
	  	  	}

		return NULL;
}

HDDEDATA TNeuroWindow::DataRequested( WORD wFmt )
{   int i;
    char str1[32];
    char *str=NULL;
    pINT l;
		HDDEDATA hData;
    if( wFmt == CF_TEXT )
      switch ((int)ActualTopic[act_conv]) {
				case 0:
          // Layer
			    l=(int)ActualItem[act_conv];
					str=(char *)malloc(L_Tab[l].size*13);
          str[0]='\0';
					for (i=ActualUnitStart;i<=ActualUnitEnd;i++) // ActualItem is layernr
						{
						if (i<ActualUnitEnd)
							sprintf(str1,"%f%c",X(l,i),DDE_DELIMITER);
						else
            	sprintf(str1,"%f",X(l,i));
            strcat(str,str1);
            }
                                                           
					hData = DdeCreateDataHandle( idInst, str, strlen(str)+1L, 0L, hszItem0[l], wFmt, 0 );
          free(str);
          return hData;
					break;
        case 1:
          /*
          // Layer:Unit
					sprintf(str1,"%f",X((pINT)ActualItem[act_conv],ActualUnit));
					return DdeCreateDataHandle( idInst, str1, strlen(str1)+1, 0L, hszItem1[(int)ActualItem[act_conv]], wFmt, 0 );
          */
					break;
        case 2:
					break;
      	case 3:
        	// Variable
			    l=(int)ActualItem[act_conv];
					str=(char*)malloc(NN_DDE_Vars[l].size*sizeof(float));
          str[0]='\0';
					for (i=0;i<NN_DDE_Vars[l].size;i++) // ActualItem is layernr
						{
						sprintf(str1,"%f%c",(*NN_DDE_Vars[l].ptr),DDE_DELIMITER);
            strcat(str,str1);
            }
					hData = DdeCreateDataHandle( idInst, str, strlen(str)+1L, 0L, hszItem0[l], wFmt, 0 );
          free(str);
          return hData;

        	break;
      }
    return NULL;
}

/*
This is triggered by the IdleAction() loop above whenever the user
enters an advise loop.
*/
void TNeuroWindow::UpdateData( void )
{/* int i;
  for (i=0;i<MAX_CONV;i++)
 		if (hConv[i]>0 && tfIsHot[i]) 
		  DdePostAdvise( idInst, ActualTopic[i], ActualItem[i] );*/
  DdePostAdvise( idInst, NULL, NULL);
}


/*
This call back function is the heart of interaction between this program
and DDE.  Because Windows doesn't pass C++ 'this' pointers to call
back functions, a static 'this' pointer was used.  If you wanted to
create a Server that would allow for more than one conversation, using a
List of conversations and their associated 'this' pointers would be one
possible method to try.  The XTYP_ constants are described in detail in
the online help.
*/

HDDEDATA FAR PASCAL TNeuroWindow::CallBack( WORD wType, WORD wFmt, HCONV hConv, HSZ hsz1, HSZ hsz2, HDDEDATA hData, DWORD, DWORD )
{
    char szTemp[128];
    int i;
    pINT l;
    char *p1;
    char str[128];

    if (hConv>0 && (wType!=XTYP_CONNECT_CONFIRM))
	    act_conv = pStaticThis->ActualConv(hConv);
    else
      { // new conversation, search for free conv-nr        	
			act_conv=-1;
      for (i=0;i<MAX_CONV;i++)
        if (pStaticThis->hConv[i]==0)
          {
					act_conv=i;
          i=MAX_CONV;
          }
      }
    if (act_conv>=0)
	    switch( wType )
  	  {
        case XTYP_ADVREQ :
        case XTYP_REQUEST :
		        // client needs a data item (for cold or warm links)
            if( pStaticThis->MatchTopicAndItem( hsz1, hsz2 ) == TRUE )
            	{
              return pStaticThis->DataRequested( wFmt );
            	}
            return NULL;
						break;

        case XTYP_ADVSTART :
            // start hot or warm link
            if( pStaticThis->MatchTopicAndItem( hsz1, hsz2 ) == TRUE) 
           	{
                pStaticThis->tfIsHot[act_conv] = TRUE;
                return (HDDEDATA)1;
            } else
						{
                return 0;
            }
						break;

        case XTYP_ADVSTOP :
            // ends hot or warm link
            if( pStaticThis->MatchTopicAndItem( hsz1, hsz2 ) == TRUE) 
	            pStaticThis->tfIsHot[act_conv] = FALSE;
            break;

        case XTYP_CONNECT :
            // a client wants to have a DDE connection 
            if( pStaticThis->MatchTopicAndService( hsz1, hsz2 ) == TRUE )
           		return (HDDEDATA)1;
            return 0;
						break;

        case XTYP_CONNECT_CONFIRM :
            pStaticThis->hConv[act_conv] = hConv;
            break;

        case XTYP_DISCONNECT :
            // connection has ended 
            for (i=0;i<MAX_CONV;i++)
	            if( hConv == pStaticThis->hConv[i] )
  	  	        {
                pStaticThis->hConv[i] = 0;
                pStaticThis->tfIsHot[i] = FALSE;
                pStaticThis->tfNewData = FALSE;
            		pStaticThis->ActualTopic[i]=pStaticThis->ActualItem[i]=(HSZ)-1;
    	  	      }
            break;

        case XTYP_ERROR :
        		// critical error (i.e. low memory or resources)
            MessageBox( pStaticThis->HWindow, "A critical DDE error has occured.", pStaticThis->Title, MB_ICONINFORMATION );
            break;

        case XTYP_EXECUTE :
        		// client gives an executable command
            char exestr[32];

            DdeGetData( hData, exestr, 32, 0);
            for (i=0;i<NN_DDE_MaxFcts;i++)
            	  if (strstr(exestr,NN_DDE_Fcts[i].name) && (NN_DDE_Fcts[i].name[0]!='\0'))
            	    {
                  if (strstr(exestr,"SET_DELIMITER"))
                  	{//special case for delimiter changing
				            p1=strtok(exestr," ");	//pointer to first field
						        if (p1)
											{
											p1=strtok(NULL," ");
	                    if (p1)
												{
												sscanf(p1,"%s",str);
                        if (str[0] == '\\')
                        	switch (str[1]) {
                           	case 't':
                            case 'T': str[0]='\t'; break;
                            case 'n':
                            case 'N': str[0]='\n'; break;
                            case 'r':
                            case 'R': str[0]='\r'; break;
                            }
                        DDE_DELIMITER=str[0];
                        }
                      }
										}
									else
                    // normal execution
                    if (NN_DDE_Fcts[i].ptr)
											{
											(*NN_DDE_Fcts[i].ptr)();
                      pStaticThis->tfNewData=TRUE;
                      }

									return (HDDEDATA)DDE_FACK;
                  }

						return DDE_FNOTPROCESSED;
						break;

        case XTYP_POKE :
        		// client sends data packages
            if( pStaticThis->MatchTopicAndItem( hsz1, hsz2 ) == TRUE )
              {
						  char* WHITESPACE="\t";
              long size=DdeGetData( hData, NULL, 0L, 0L );
              char far*str=(char far*)farmalloc(size+1);

              DdeGetData( hData, (LPBYTE)str, size, 0L );

			    	  if (!strchr(str,'\t')) /* override Whitespace */
                 if (strchr(str,';')) WHITESPACE=";";
                 else if (strchr(str,',')) WHITESPACE=",";
	                 else WHITESPACE=" ";

              p1=strtok(str,WHITESPACE);	//pointer to first field
              l=(int)pStaticThis->ActualItem[act_conv];
		          if (p1)
								for (i=pStaticThis->ActualUnitStart;i<=pStaticThis->ActualUnitEnd;i++) // ActualItem is layernr
									{
									sscanf(p1,"%f",&X(l,i));
								  p1=strtok(NULL,WHITESPACE);
				          if (!p1) break;
				          }
							return (HDDEDATA)DDE_FACK;
              }
            else
            	return DDE_FNOTPROCESSED;
						break;

        case XTYP_WILDCONNECT :
            return pStaticThis->WildConnect( hsz1, hsz2, wFmt );
						break;

        default :
            break;
    	}
    return NULL;
}

//--------------- TLayerParamsDialog Functions ----------------------
void TLayerParamsDialog::SetupWindow()
{ char str[6];
  SendDlgItemMsg(ID_LayerTitle,WM_SETTEXT,0,(DWORD)L_Tab[layer_clicked].titel);
	wsprintf(str,"%u",L_Tab[layer_clicked].size);
	SendDlgItemMsg(ID_LayerSize,WM_SETTEXT,0,(DWORD)str);
	wsprintf(str,"%u",L_Tab[layer_clicked].upr);
	SendDlgItemMsg(ID_LayerWidth,WM_SETTEXT,0,(DWORD)str);
}

void TLayerParamsDialog::Ok(RTMessage Msg)
{ int s,w;
  layer *L=&L_Tab[layer_clicked];
  char str[21],str2[6];
	TDialog::Ok(Msg);

	SendDlgItemMsg(ID_LayerTitle,WM_GETTEXT,20,(DWORD)str);
  if (!(L->titel = (char*)realloc(L->titel, strlen(str)+1)))
     nn_alarm(SOURCENAME,"Error allocating memory for titelstring",str);
  else
    strcpy(L->titel,str);

	SendDlgItemMsg(ID_LayerSize,WM_GETTEXT,5,(DWORD)str);
	SendDlgItemMsg(ID_LayerWidth,WM_GETTEXT,5,(DWORD)str2);
  s=atoi(str); w=atoi(str2);
  if (s!=L->size || w!=L->upr)
  	{
    if (L->ext==NULL)
    	{
      ClearLayerRect(layer_clicked);
      deconnect_layer(layer_clicked);
      resize_layer(L,atoi(str),atoi(str2));
      connect_all_new();
    	}
    else
    	AlarmMsg("You cannot resize a layer with external inputs!");
    }
  ClearLayerRect(layer_clicked);
  L->status=L_NEU; // zumindest fuer titel neu zeichnen
  draw_layer(layer_clicked);
}


//--------------- TAboutDialog Functions ----------------------
void TAboutDialog::WMLButtonDown(RTMessage Msg)
{  GetApplication()->ExecDialog(new TDialog(this, "Credits")); }

//--------------- TScaleDialog Functions ----------------------
void TScaleDialog::SetupWindow()
{ int i;
  // doesn't work in first group of dialog!
  // SendDlgItemMsg(ID_AScale+2,BM_SETCHECK,1,0);
  // not nice, but it works:
  for (i=0;i<5;i++)
    if (ScaleA-1 != i)
       SendDlgItemMsg(ID_AScale+1+i,BM_SETCHECK,0,0);

  SendDlgItemMsg(ID_WScale+ScaleW,BM_SETCHECK,1,0);
}

void TScaleDialog::Ok(RTMessage Msg)
{ int i;
	TDialog::Ok(Msg);
	for (i=1;i<6;i++)
		if (SendDlgItemMsg(ID_AScale+i,BM_GETCHECK,Msg.WParam,Msg.LParam))
			ScaleA=i;
  if (ScaleA<5)
		ScaleActivationDisplay = pow(10.0,ScaleA-2);
	else
		ScaleActivationDisplay = 0.0; /* autoscale */

	for (i=1;i<6;i++)
		if (SendDlgItemMsg(ID_WScale+i,BM_GETCHECK,Msg.WParam,Msg.LParam))
			ScaleW=i;
  if (ScaleW<5)
		ScaleWeightDisplay = pow(10.0,ScaleW-2);
	else
		ScaleWeightDisplay = 0.0; /* autoscale */

  draw_palette();
	draw_layers();
  show_overview();
}

//--------------- TInfoWindow Functions ----------------------
TInfoWindow::TInfoWindow(PTWindowsObject AParent, LPSTR ATitle) :
	TBWindow(AParent, ATitle)
{ Attr.Style = WS_DLGFRAME | WS_MINIMIZEBOX;
	Attr.ExStyle |= WS_EX_TOPMOST;
	Attr.X=200;
  Attr.Y=0;
	Attr.W=100;
	Attr.H=200;
}

void TInfoWindow::GetWindowClass(WNDCLASS _FAR& winclass)
{ TBWindow::GetWindowClass(winclass);
	winclass.hIcon = LoadIcon(GetApplication()->hInstance, ICON_INFO);
}

void TInfoWindow::SetupWindow()
{	TBWindow::SetupWindow(); }


void TInfoWindow::WMCreate(RTMessage Msg)
{ int i,dy=0, w=16;
	HDC DC=GetDC(HWindow);
	GetTextMetrics(DC,&tmIWindow);
	ReleaseDC(HWindow,DC);
	new TBGroupBox(this,0,"",5,5,10+tmIWindow.tmMaxCharWidth*w,10+6+tmIWindow.tmHeight*5);
	new TBDivider(this,0,"",5,10+tmIWindow.tmHeight,10+tmIWindow.tmMaxCharWidth*w,3,0,1);
	new TBDivider(this,0,"",7,10+3+tmIWindow.tmHeight*3,5+tmIWindow.tmMaxCharWidth*w,3,0,1);
  for (i=0;i<5;i++)
    {if (i==1) dy=3; if (i==3) dy=6;
		Text[i]=new TBStatic(this,0,"",10,10+dy+i*tmIWindow.tmHeight,tmIWindow.tmMaxCharWidth*w,tmIWindow.tmHeight,20);
		}
  TBWindow::WMCreate(Msg);
	MoveWindow(HWindow,300,0,22+tmIWindow.tmMaxCharWidth*w,40+6+tmIWindow.tmHeight*5,TRUE);
	hIWindow=HWindow;
}

void TInfoWindow::WMSize(RTMessage Msg)
{ TBWindow::WMSize(Msg);
}

void TInfoWindow::Paint(HDC hPaintDC, PAINTSTRUCT _FAR&)
{ int i;
	for (i=0;i<5;i++)
		Text[i]->SetText(info_msg[i]);
}

//--------------- TMsgWindow Functions ----------------------
TMsgWindow::TMsgWindow(PTWindowsObject AParent, LPSTR ATitle) :
	TWindow(AParent, ATitle)
{ Attr.Style |= WS_VSCROLL | WS_HSCROLL;
	Attr.X=0;
  Attr.Y=0;
	Attr.W=550;
	Attr.H=400;
	AssignMenu("MsgMenu");
	MsgText[0] = NULL;
	MsgLines=0;
}

TMsgWindow::~TMsgWindow()
{ int i=0;
	while(MsgText[i] && i<MaxMsgLines) farfree(MsgText[i++]);
}

void TMsgWindow::GetWindowClass(WNDCLASS _FAR& winclass)
{ TWindow::GetWindowClass(winclass);
	winclass.hIcon = LoadIcon(GetApplication()->hInstance, ICON_MSG);
}

void TMsgWindow::SetupWindow()
{ TWindow::SetupWindow();
	strcpy(MFileName,MsgFile);
	if (!strchr(MFileName,'.')) strcat(MFileName,".msg");
}

void TMsgWindow::WMCreate(RTMessage Msg)
{ TWindow::WMCreate(Msg);
	HDC DC=GetDC(HWindow);
	GetTextMetrics(DC,&tmMWindow);
	Scroller = new TScroller(this, tmMWindow.tmAveCharWidth,
		 tmMWindow.tmHeight+tmMWindow.tmExternalLeading, 10, MaxMsgLines);
	ReleaseDC(HWindow,DC);
	hMWindow=HWindow;
}

void TMsgWindow::WMSize(RTMessage Msg)
{ TWindow::WMSize(Msg); }

void TMsgWindow::CMMsgClear(RTMessage)
{ int i;
	for (i=0;i<MsgLines;i++)
		if (MsgText[i]) free(MsgText[i]);
	MsgText[0]=NULL;
  MsgLines=0;
	SetScrollPos(hMWindow,SB_VERT,0,TRUE);
	SetScrollPos(hMWindow,SB_HORZ,0,TRUE);
	InvalidateRect(hMWindow,NULL,TRUE);
}

void TMsgWindow::CMSave(RTMessage Msg)
{       if (strlen(MFileName)>5 || MFileName[0]!='*')
		switch (SaveMsg(MFileName)) {
		case IDOK:
			pprintf("Messages saved in file '%s'.\n",MFileName);
			break;
		case IDRETRY:
			CMSaveAs(Msg);
			break;
		case IDCANCEL:
		break;
		}
	else
		CMSaveAs(Msg); 
}

void TMsgWindow::CMSaveAs(RTMessage Msg)
{ FILE *fp;
	int i;
	if(MFileName[0]=='*')
		{strcpy(MFileName,MsgFile);
		if (!strchr(MFileName,'.')) strcat(MFileName,".msg");
		}
	if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILESAVE,MFileName)) == IDOK )
		{
		if ( (fp=fopen(MFileName,"r"))!=NULL )
	  {fclose(fp);
			wsprintf(dummy,"File %s already exists!"
			"OK to overwrite?",MFileName);
			if ( ( i = MessageBox(HWindow,dummy,"Save Messages to File",MB_ICONQUESTION|MB_YESNOCANCEL))==IDNO)
				{CMSaveAs(Msg);
	return;}
			else if (i==IDCANCEL)
	return;
			}
		switch (SaveMsg(MFileName)) {
		case IDOK:
			pprintf("Messages saved in file '%s'.\n",MFileName);
			break;
		case IDRETRY:
			CMSaveAs(Msg);
			break;
		case IDCANCEL:
	break;
			}
		} 
}

void TMsgWindow::Paint(HDC hPaintDC, PAINTSTRUCT _FAR&)
{ int i,y=0;
	for (i=0;i<MsgLines;i++)
		if (MsgText[i])
			{
			TextOut(hPaintDC,4,y,MsgText[i],strlen(MsgText[i]));
			y+=tmMWindow.tmHeight+tmMWindow.tmExternalLeading;
			}
}

int TMsgWindow::SaveMsg(char* fname)
{ FILE *fp;
	int i;
	if ((fp = fopen(fname, "wt")) != NULL)
		{
		for (i=0;i<MsgLines;i++)
			if (MsgText[i]) fprintf(fp,"%s\n",MsgText[i]);
		fclose(fp);
    }
	else
		return ( MessageBox(HWindow, MFileName, "Error saving messages:",MB_RETRYCANCEL|MB_ICONSTOP) );
	
	return IDOK;
}

//--------------- TCanvas Functions ----------------------
TCanvas::TCanvas(PTWindowsObject AParent) : TWindow(AParent,NULL)
{       Attr.Style = WS_BORDER | WS_CHILD | WS_VISIBLE;
	State.BitMap = 0;
	fDrag = FALSE;
	fCapture = FALSE;
	strcpy(AFileName,ActFile);
	if (!strchr(AFileName,'.')) strcat(AFileName,".nnb");
};

TCanvas::~TCanvas()
{ DeleteDC(hCDC);
	DeleteObject(State.BitMap);
	farfree(NNLogPalette);
	DeleteObject(hNPalette);
};

void TCanvas::SetupWindow()
{ TWindow::SetupWindow();
	HDC DC=GetDC(HWindow);
	GetTextMetrics(DC,&tmCWindow);
	ReleaseDC(HWindow,DC);
}

void TCanvas::WMCreate(RTMessage Msg)
{       TWindow::WMCreate(Msg);
	hCWindow=HWindow;
	HDC DC, MemDC;
	unsigned char RedVals[NumColors], BlueVals[NumColors], GreenVals[NumColors];
	int i,j;
	NNLogPalette = (LPLOGPALETTE)
		farmalloc(sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * NumColors);
	NNLogPalette->palVersion = 0x300;
	NNLogPalette->palNumEntries = NumColors;
	/* Setup Colortable to "visual linear" degrading */
	for (i=1; i<33; i++)
		{
		j = i<<3;
		/* Bluetable */
		RedVals[33-i]   = 256-j;
		BlueVals[33-i]  = (use_colors)? 255 : 256-j;
		GreenVals[33-i] = 256-j;
		/* Graytable */
		RedVals[33+i]   = (use_colors && !use_gray)? 255 : 256-j+4;
		BlueVals[33+i]  = 256-j+4;
		GreenVals[33+i] = 256-j+4;
		}
//      RedVals[0]  = GreenVals[0]  = BlueVals[0]  = 255;
	RedVals[33] = GreenVals[33] = BlueVals[33] = 255;
//      RedVals[66] = GreenVals[66] = BlueVals[66] = 0;
	for (i = 0; i <= NumColors; ++i)
		{
		NNLogPalette->palPalEntry[i].peRed = RedVals[i];
		NNLogPalette->palPalEntry[i].peGreen = GreenVals[i];
		NNLogPalette->palPalEntry[i].peBlue = BlueVals[i];
		NNLogPalette->palPalEntry[i].peFlags = NULL;
		}
	if (!(hNPalette = CreatePalette(NNLogPalette)))
			ErrMsg("Zu wenig Speicher fr Palette");
	DC = GetDC(0);
	if (!(State.BitMap = CreateCompatibleBitmap(DC, BitMapWidth, BitMapHeight)))
			ErrMsg("Zu wenig Speicher fr Bitmap");
	if (!(MemDC = CreateCompatibleDC(DC)))
			ErrMsg("Zu wenig Speicher fr DC");
	SelectObject(MemDC, State.BitMap);
  // Clear
	PatBlt(MemDC, 0, 0, BitMapWidth, BitMapHeight, WHITENESS);
//      SelectPalette(MemDC, hNPalette, FALSE);
//      RealizePalette(MemDC);
	hCDC = MemDC;

	draw_palette();
	draw_layers();

	ReleaseDC(0, DC);
}

void TCanvas::GetWindowClass(WNDCLASS _FAR& winclass)
{ TWindow::GetWindowClass(winclass);
	winclass.hCursor=0; // must be 0 to be changeable
	hCMenu = LoadMenu(GetApplication()->hInstance,"MainMenu"); 
  // Must be SubMenu for use in TrackPopUpMenu
	hLMenu = GetSubMenu(LoadMenu(GetApplication()->hInstance,"LayerMenu"),0);
	hBMenu = GetSubMenu(LoadMenu(GetApplication()->hInstance,"BackMenu"),0);

	hPosCursor=LoadCursor(NULL, IDC_SIZE);
	hArrowCursor=LoadCursor(NULL, IDC_ARROW);
	hWaitCursor=LoadCursor(NULL, IDC_WAIT);
	hSizeNSCursor=LoadCursor(NULL, IDC_SIZENS);
	hSizeWECursor=LoadCursor(NULL, IDC_SIZEWE);
	hSizeNESWCursor=LoadCursor(NULL, IDC_SIZENESW);
	hSizeNWSECursor=LoadCursor(NULL, IDC_SIZENWSE);
	switch (mouse_function[0]) {
		case M_WEIGHT:
			CheckMenuItem(GetSubMenu(hBMenu,2),CM_LM_Weights,MF_BYCOMMAND|MF_CHECKED);
			break;
		case M_TOGGLE:
			CheckMenuItem(GetSubMenu(hBMenu,2),CM_LM_Toggle,MF_BYCOMMAND|MF_CHECKED);
			break;
    }
	switch (mouse_function[1]) {
		case M_WEIGHT:
			CheckMenuItem(GetSubMenu(hBMenu,3),CM_RM_Weights,MF_BYCOMMAND|MF_CHECKED);
			break;
		case M_TOGGLE:
			CheckMenuItem(GetSubMenu(hBMenu,3),CM_RM_Toggle,MF_BYCOMMAND|MF_CHECKED);
			break;
		}
}

void TCanvas::Paint( HDC PaintDC, PAINTSTRUCT&)
{
	HPALETTE OldPalette;
	RECT R;
	OldPalette = SelectPalette(PaintDC, hNPalette, FALSE);
	RealizePalette(PaintDC);
	GetClientRect(HWindow, &R);
	BitBlt(PaintDC, 0, 0, R.right, R.bottom, hCDC, 0, 0, SRCCOPY);
	SelectPalette(PaintDC, OldPalette, FALSE);
	Init_OK=TRUE;
};

void TCanvas::CMLayerResetAct(RTMessage Msg)
{ reset(layer_clicked);
  draw_layer(layer_clicked);
}

void TCanvas::CMLayerLoadAct(RTMessage Msg)
{ if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILEOPEN, strcpy(AFileName,"*.nnb"))) == IDOK )
		if (!load_units(AFileName,layer_clicked))
			{if (MessageBox(HWindow, AFileName, "Error loading activations:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMLayerLoadAct(Msg);
      }
		else
			{pprintf("Activations loaded from file '%s'.\n",AFileName);
			draw_layer(layer_clicked);
			}
}

void TCanvas::CMLayerSaveAct(RTMessage Msg)
{  if (strlen(AFileName)>5 || AFileName[0]!='*')
		if (!save_units(AFileName,layer_clicked))
			{if (MessageBox(HWindow, AFileName, "Error saving Activations:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMLayerSaveAsAct(Msg);
      }
		else
	pprintf("Activations saved in file '%s'.\n",AFileName);
	else
		CMLayerSaveAsAct(Msg);          
}

void TCanvas::CMLayerSaveAsAct(RTMessage Msg)
{ if(AFileName[0]=='*')
		{strcpy(AFileName,ActFile);
		if (!strchr(AFileName,'.')) strcat(AFileName,".nnb");
		}
	if( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILESAVE, AFileName)) == IDOK )
		if (!save_units(AFileName,layer_clicked))
			{if (MessageBox(HWindow, AFileName, "Error saving Activations:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMLayerSaveAsAct(Msg);
      }
		else
	pprintf("Activations saved in file '%s'.\n",AFileName);
}

/* Free patterns of actual layer */
void TCanvas::CMLayerDataClear(RTMessage Msg)
{ pINT l=layer_clicked;
  free_patterns(l);
  pprintf("Patterns of layer %u (%s) freed.\n",l, L_Tab[l].titel);
}

/* Load new patterns and append to patternchain in actual layer */
void TCanvas::CMLayerDataLoad(RTMessage Msg)
{ pINT l=layer_clicked; //layer to load patterns into
  char drive[MAXDRIVE];
  char dir[MAXDIR];
  char file[MAXFILE];
  char ext[MAXEXT];
	int  flags;
  int  erg;
  char OutputDirectory [MAXPATH];
  char CurrentDirectory [MAXPATH];
  int  LoadType;
  BOOL SelectedOnly = FALSE;

  RECT R;
	POINT pt;
  GetCursorPos(&pt);

  strcpy (OutputDirectory, "C:\\");
  OutputDirectory [0] = getdisk () + 'A';
  getcurdir (0, CurrentDirectory);
  strcat (OutputDirectory, CurrentDirectory);

  char title[100];
  wsprintf(title,"Load Data into Layer %u (%s)",l,L_Tab[l].titel);
	char *filter="DAT-Files\0*.dat\0PAT-Files\0*.pat\0All Files\0*.*\0";
  int Index=DAT_FILE;
  if (LoadFile(title,filter,NeuroWindow->DFileName,&Index))
    {
//    pprintf("Index=%u, PAT_FILE=%i\n",Index,PAT_FILE);
    switch (Index) {
		  case DAT_FILE:
		    erg=load_patterns(NeuroWindow->DFileName,l);
        break;
      case PAT_FILE:
			  erg=load_data(NeuroWindow->DFileName,l);
        break;
      default: /* try to load float data */
			  erg=load_data(NeuroWindow->DFileName,l);
    }
    if (!erg)
			{if (MessageBox(HWindow, NeuroWindow->DFileName, "Error loading data:",MB_RETRYCANCEL|MB_ICONSTOP)==IDRETRY)
				CMLayerDataLoad(Msg);
	  	}
		else
			{pprintf("Data loaded from file '%s'.\n",NeuroWindow->DFileName);
			draw_layers();
			}
		}
}

void TCanvas::CMLayerParams(RTMessage Msg)
{
	if((GetApplication()->ExecDialog(new TLayerParamsDialog(this,"LayerParams"))) == IDOK )
    {/* Layer changed */ }
}

void TCanvas::DrawNotRect(RECT& R)
{
	HDC hdc = GetDC(HWindow);
	DrawFocusRect(hdc, &R);
	ReleaseDC(HWindow,hdc);
}

void TCanvas::WMLButtonDown( RTMessage Msg )
{	RECT R;
  pINT l,u;
	MouseMode.Button=1;
	if (EditLayout)
		{// Editmode
		pt = MAKEPOINT(Msg.LParam);
		fCapture=TRUE;
		SetCapture(HWindow);
		ptOld=pt;
		if (MouseMode.Resize)
			MouseMode.DragStatus=1;
		else
		 {
		 for (ALL_LAYERS(l))
			{
			GetWindowRect(HWindow,&R);
			layer L = L_Tab[l];
			SetMarkerRect(l);

      // Mouse in Layer?
			if (PtInRect(&Rect,pt))
				{
				MouseMode.DragStatus = 1;
				if (L.ext)
					R.top += pt.y-min(Rect.top,RectExt.top)+40; // don't destroy palettedisplay
				else
					R.top += pt.y-Rect.top+40;
				ClipCursor(&R); // restrict cursor to window

				if (MouseMode.Select)
					{
					if ( (l!=MouseMode.Layer) || ((l==MouseMode.Layer) && MouseMode.Ext) )
						{// Remove last selection
						DrawMarkers(MouseMode.Rect);
				    // Mark actual layer
						DrawMarkers(Rect);
						MouseMode.Rect = Rect;
						MouseMode.Ext = FALSE;
						MouseMode.Layer = l;
						MouseMode.Select = TRUE;
						ptOrg.x = pt.x - Rect.left; //Origin of Mousepoint in Rectangle
						ptOrg.y = pt.y - Rect.top;
						}
					}
				else
					{
					DrawMarkers(Rect);
					MouseMode.Rect = Rect;
					MouseMode.Ext = FALSE;
					MouseMode.Layer = l;
					MouseMode.Select = TRUE;
					ptOrg.x = pt.x - Rect.left; //Origin of Mousepoint in Rectangle
					ptOrg.y = pt.y - Rect.top;
					}
	}
			else
      // Mouse on Extern Layer?
			if (L.ext && PtInRect(&RectExt,pt))
				{
				MouseMode.DragStatus = 1;
				R.top += pt.y-RectExt.top+40; // don't destroy palettedisplay
				ClipCursor(&R); // restrict cursor to window

				if (MouseMode.Select)
					{
					if ( (l!=MouseMode.Layer) || ((l==MouseMode.Layer) && !MouseMode.Ext) )
						{// Remove last selection
						DrawMarkers(MouseMode.Rect);
						DrawMarkers(RectExt);
						MouseMode.Rect = RectExt;
						MouseMode.Ext = TRUE;
		    		MouseMode.Layer = l;
						MouseMode.Select = TRUE;
						ptOrg.x = pt.x - RectExt.left; //Origin of Mousepoint in Rectangle
						ptOrg.y = pt.y - RectExt.top;
						}
					}
				else
					{
					DrawMarkers(RectExt);
					MouseMode.Rect = RectExt;
					MouseMode.Ext = TRUE;
	    		MouseMode.Layer = l;
					MouseMode.Select = TRUE;
					ptOrg.x = pt.x - RectExt.left; //Origin of Mousepoint in Rectangle
					ptOrg.y = pt.y - RectExt.top;
					}
				}
			if (MouseMode.DragStatus>0) break; // keine weiteren Layer untersuchen
      } //end for all layers
		 } // end else
		} // end EditLayout
  else                   
  {// Normalmode
	mouse_button = 1;
	if(UnitClicked(Msg))
		(*action_when_unit_clicked)(layer_clicked,unit_clicked);
	else
		{/* Click on Background - redraw layers */
		draw_layers();
    show_overview();
		}
  }
};

void TCanvas::WMLButtonUp( RTMessage)
{ if (MouseMode.Button==1) MouseMode.Button=0;
	if (!EditLayout || !fCapture) return;
	int l,w;
	RECT R;
	int th=tmCWindow.tmHeight;
	int tw=tmCWindow.tmMaxCharWidth;
  ReleaseCapture();
  fCapture=FALSE;
	if (MouseMode.DragStatus>1)
		{DrawNotRect(Rect);
		layer *L = &L_Tab[MouseMode.Layer];
		// Delete layer on old position
    if (MouseMode.Ext && L->p && L->p->name)
			w=max(strlen(L->p->name)*tw,max(strlen(L->titel)*tw+1,MouseMode.Rect.right-MouseMode.Rect.left));
		else
			w=max(strlen(L->titel)*tw+1,MouseMode.Rect.right-MouseMode.Rect.left);
		FilledRect(hCDC,MouseMode.Rect.left,MouseMode.Rect.top,MouseMode.Rect.left+w,MouseMode.Rect.bottom,RGB(255,255,255),0,0);
		if (MouseMode.DragStatus==2)
			{// Normal Shift
			if (!MouseMode.Ext)
				{// Normallayer, Extern-Layer wurde mitverschoben
				L->x = Rect.left + 4;
				L->y = Rect.top  + th + 5;
				if (L->ext && L->p && L->p->name)
					{// Delete old extern layer
					w=max(strlen(L->p->name)*tw,max(strlen(L->titel_ext)*tw+1,MouseMode.Rect.right-MouseMode.Rect.left));
					SetRect(&R, MouseMode.Rect.left+L->extsep_x, MouseMode.Rect.top+L->extsep_y,
											MouseMode.Rect.left+L->extsep_x+w+1, MouseMode.Rect.bottom+L->extsep_y+1+20);
					FilledRect(hCDC,R.left,R.top,R.right,R.bottom,RGB(255,255,255),0,0);
			  	}
				}
		  else
				{
				// Nur Distanz zum Externlayer ndern, da Normallayer an alter Pos bleibt
				L->extsep_x = Rect.left + 4 - L->x;
				L->extsep_y = Rect.top + th + 5 - L->y ;
				}
      }
    else
      {// Resize
			int d_x = ((Rect.right - MouseMode.Rect.right) + (MouseMode.Rect.left - Rect.left))/(int)L->upr;
			int d_y = ((Rect.bottom - MouseMode.Rect.bottom) + (MouseMode.Rect.top - Rect.top))/(int)(L->size/L->upr);
			L->dx = max(1,(int)L->dx+d_x);
			L->ux = max(1,(int)L->ux+d_x);
			L->dy = max(1,(int)L->dy+d_y);
			L->uy = max(1,(int)L->uy+d_y);
			L->x = max(0,(int)L->x + Rect.left - MouseMode.Rect.left);
			L->y = max(0,(int)L->y + Rect.top  - MouseMode.Rect.top);
			SetMarkerRect(MouseMode.Layer); // Layergre neu berechnen wegen Rundungsfehler in dx,dy

			if (!MouseMode.Ext)
				{// Delete old extern layer
				if (L->ext)
					{SetRect(&R, MouseMode.Rect.left+L->extsep_x, MouseMode.Rect.top+L->extsep_y,
						 MouseMode.Rect.right+L->extsep_x, MouseMode.Rect.bottom+L->extsep_y+th+5);
					FilledRect(hCDC,R.left,R.top,R.right,R.bottom,RGB(255,255,255),0,0);
				  }
				}
			else
				{// Delete old normal layer
				SetRect(&R, MouseMode.Rect.left - L->extsep_x, MouseMode.Rect.top - L->extsep_y,
						 MouseMode.Rect.right - L->extsep_x, MouseMode.Rect.bottom - L->extsep_y +th+5);
				FilledRect(hCDC,R.left,R.top,R.right,R.bottom,RGB(255,255,255),0,0);
				CopyRect(&Rect,&RectExt); // ExternRect bernehmen fr Markers
				}
			}

		for (ALL_LAYERS(l))
			if ( l!=MouseMode.Layer )
			  { // Andere Layer sicherheitshalber neu zeichnen
				draw_rahmen(l,0);
				draw_layer(l);
				if (L_Tab[l].ext) draw_ext(l);
				}

		update_layer(MouseMode.Layer); // Verschobenen Layer anzeigen

		if (MouseMode.Ext)
			{// Zugehrigen Normallayer anzeigen
			draw_rahmen(MouseMode.Layer,0);
			draw_layer(MouseMode.Layer);
			}
		else
			{// zugehrigen Externlayer anzeigen
			if (L_Tab[l].ext) draw_ext(l);
			}

		InvalidateRect(HWindow,NULL,FALSE);
		UpdateWindow(HWindow);
		SetCursor(hArrowCursor);
		CopyRect(&MouseMode.Rect,&Rect);
		DrawMarkers(Rect); // Markers wieder anzeigen
		}
	else
		{
		}
	ClipCursor(NULL); // Cursor unrestricted
	MouseMode.DragStatus = 0;
	MouseMode.Resize = 0;
}

void TCanvas::WMMouseMove( RTMessage Msg )
{ RECT R;
  int dx,dy;
	pt = MAKEPOINT(Msg.LParam); // Aktuelle Mausposition
	GetClientRect(HWindow,&R);

	if (in_break)
		{//dont allow heavy actions while in nn_break()
		if (PtInRect(&R,pt)) SetCursor(hWaitCursor); //reset cursor to wait
		return;
    }

	if (MouseMode.Select && !MouseMode.Button)
		{//Check for Markers
		int Resize=0;
		int i=0;
		RECT M1,M0={-4,-4,4,4};
	int x,y;
		CopyRect(&R,&MouseMode.Rect);
		for (x=0;x<3;x++)
			for(y=0;y<3;y++)
		if (!((x==1) && (y==1)))
					{
					CopyRect(&M1,&M0);
					OffsetRect(&M1,R.left+x*(R.right-R.left)/2,R.top+y*(R.bottom-R.top)/2);
					i++;
					if (PtInRect(&M1,pt))
						{
						Resize = i;
						y=3;x=3;
						}
	  }
		if (Resize!=MouseMode.Resize)
			{
			MouseMode.Resize=Resize;
			switch (Resize) {
				case 0:
					SetCursor(hArrowCursor);
					break;
				case 1: // Upper Left Corner
				case 8: // Lower Right Corner
					SetCursor(hSizeNWSECursor);
					break;
				case 2: // Mid Left 
	    case 7: // Mid Right 
					SetCursor(hSizeWECursor);
					break;
				case 3: // Lower Left Corner
				case 6: // Upper Right Corner
					SetCursor(hSizeNESWCursor);
					break;
				case 4: // Upper Mid 
				case 5: // Lower Mid
					SetCursor(hSizeNSCursor);
					break;
		}
			}
		}

	if (MouseMode.Button==1)
    {// Button 1 pressed
		switch (MouseMode.DragStatus) {
		case 0:
	break;
		case 1:
			if (!MouseMode.Resize) SetCursor(hPosCursor);
			MouseMode.DragStatus=(MouseMode.Resize)?3:2;
			DrawMarkers(MouseMode.Rect); // Delete last markers
			CopyRect(&Rect,&MouseMode.Rect);
      DrawNotRect(Rect);
			if (MouseMode.DragStatus==3)
				{GetWindowRect(HWindow,&R);
	if (L_Tab[MouseMode.Layer].ext)
					R.top += pt.y-min(RectExt.top,Rect.top)+40;
				else
					R.top += pt.y-Rect.top+40;
				ClipCursor(&R); // restrict cursor to window for resize
				}
			break;
		case 2:
      // Move Layer
			DrawNotRect(Rect);
			dx = ptOld.x - pt.x;
			dy = ptOld.y - pt.y;
			Rect.left -= dx;
			Rect.top  -= dy;
			Rect.right -= dx;
			Rect.bottom -= dy;
	  if (L_Tab[MouseMode.Layer].ext)
				{
				RectExt.left -= dx;
				RectExt.top  -= dy;
				RectExt.right -= dx;
				RectExt.bottom -= dy;
				}
			ptOld = pt; 
			DrawNotRect(Rect);
	break;
		case 3:
	// Change Size

			DrawNotRect(Rect);
			dx = ptOld.x - pt.x;
			dy = ptOld.y - pt.y;
      CopyRect(&R,&Rect);
		  switch (MouseMode.Resize) {
				case 1:
					R.left -= dx;
					R.top  -= dy;
					break;
				case 2:
					R.left -= dx;
					break;
				case 3:
					R.left -= dx;
					R.bottom -= dy;
					break;
				case 4:
					R.top  -= dy;
					break;
				case 5:
					R.bottom -= dy;
					break;
				case 6:
					R.right -= dx;
					R.top  -= dy;
		break;
	case 7:
					R.right -= dx;
		break;
				case 8:
					R.right -= dx;
					R.bottom -= dy;
					break;
			}
			ptOld = pt;
      // Rect nicht zu klein werden lassen
			if ((R.right - R.left) > 20)
				{Rect.left = R.left;
				Rect.right = R.right;}
			if ((R.bottom - R.top) > 40)
				{Rect.bottom = R.bottom;
				Rect.top = R.top;}

			DrawNotRect(Rect);
      break;
		}
	}
}

void TCanvas::WMRButtonDown( RTMessage Msg )
{ pINT l,u;
	mouse_button=2;
	HMENU hmenu2;
	if(UnitClicked(Msg))
			(*action_when_unit_clicked)(layer_clicked,unit_clicked);
	else
	 {
	 POINT pt = MAKEPOINT(Msg.LParam);
	 ClientToScreen(HWindow,&pt);
	 if (LayerClicked(Msg))
			{
			// Layer-Menu
			// Reset, Load, Save, Statistics
			hmenu2 = hLMenu;
			}
	 else
		{/* Click on Background */
		//Show Pop-Up-Menu;
		hmenu2 = hBMenu;
		}

	 TrackPopupMenu(hmenu2,0,pt.x,pt.y,0,hCWindow,0L);
	 }
}

int TCanvas::UnitClicked( RTMessage event )
{ pINT l1,u1;
	mouse_x = event.LP.Lo;
	mouse_y = event.LP.Hi;
  for(ALL_LAYERS(l1))
		for(L_ALL(l1,u1))
			{
			layer* L=&L_Tab[l1];
			if ( ( mouse_x-PX(l1,u1) <= L->ux ) && ( mouse_y-PY(l1,u1) <= L->uy ) )
        {
        layer_clicked=l1; unit_clicked=u1;
			  return 1;
        }
		  }
	return 0;
}

int TCanvas::LayerClicked( RTMessage Msg )
{ pINT l1 = 0;
	pt = MAKEPOINT(Msg.LParam);
	while(l1<L_anz)
		{
    // use same Rect as in draw_rahmen
		SetMarkerRect(l1);
		if (PtInRect(&Rect,pt))
			{ layer_clicked=l1; return 1; }
		else l1++;
		}
	return 0;
}

void TCanvas::CMLeftMouseFct1(RTMessage)
{
	CheckMenuItem(GetSubMenu(hBMenu,2),CM_LM_Weights,MF_BYCOMMAND|MF_CHECKED);
	CheckMenuItem(GetSubMenu(hBMenu,2),CM_LM_Toggle,MF_BYCOMMAND|MF_UNCHECKED);
	mouse_function[0]=M_WEIGHT;
}

void TCanvas::CMLeftMouseFct2(RTMessage)
{
	CheckMenuItem(GetSubMenu(hBMenu,2),CM_LM_Weights,MF_BYCOMMAND|MF_UNCHECKED);
	CheckMenuItem(GetSubMenu(hBMenu,2),CM_LM_Toggle,MF_BYCOMMAND|MF_CHECKED);
	mouse_function[0]=M_TOGGLE;
}

void TCanvas::CMRightMouseFct1(RTMessage)
{
	CheckMenuItem(GetSubMenu(hBMenu,3),CM_RM_Weights,MF_BYCOMMAND|MF_CHECKED);
	CheckMenuItem(GetSubMenu(hBMenu,3),CM_RM_Toggle,MF_BYCOMMAND|MF_UNCHECKED);
	mouse_function[1]=M_WEIGHT;
}

void TCanvas::CMRightMouseFct2(RTMessage)
{
	CheckMenuItem(GetSubMenu(hBMenu,3),CM_RM_Weights,MF_BYCOMMAND|MF_UNCHECKED);
	CheckMenuItem(GetSubMenu(hBMenu,3),CM_RM_Toggle,MF_BYCOMMAND|MF_CHECKED);
	mouse_function[1]=M_TOGGLE;
}


//--------------- TNeuroApp Functions ----------------------
void TNeuroApp::InitMainWindow()
{ (void)BWCCGetVersion();  // Make sure BWCC gets initialized.
	sprintf(dummy,"VieNet2: %s",ProgName);
	MainWindow = NeuroWindow = new TNeuroWindow(NULL, dummy);
	hNWindow = NeuroWindow->HWindow;

	InfoWindow = new TInfoWindow(NULL,"VieNet2 - Infos");
	MakeWindow(InfoWindow);
	InfoWindow->Show(SW_SHOW);

	MsgWindow = new TMsgWindow(NULL, "VieNet2 - Messages");
	MakeWindow(MsgWindow);
	MsgWindow->Show(SW_SHOW);

	// Init UserWindow-Struct
	for (int i=0;i<MaxUserWindow;i++)
		{UserWindow[i].ID=-1;
		UserWindow[i].ptr=NULL;}
}

void TNeuroApp::InitInstance()
{ TApplication::InitInstance();}

void TNeuroApp::MessageLoop()
{ TApplication::MessageLoop();}

// -------------------- for DDE --------------------------
/*
There are 16 available timers in Windows.  This little trick uses the
OWL IdleAction() member function to save that scarce resource.
*/
void TNeuroApp::IdleAction()
{
/* update oncy a second
    static DWORD dwTime = GetTickCount();

    if( (NeuroWindow != NULL) && (NeuroWindow->tfLoop == TRUE) )
    {
        if( (GetTickCount() - dwTime) > 1000 )
        {
            dwTime = GetTickCount();
						NeuroWindow->UpdateData();
        }
    }
*/
	/* Update only once for each new data */
  if (NeuroWindow->tfNewData)
			{NeuroWindow->UpdateData();
      NeuroWindow->tfNewData=FALSE;
      }
}

//=============== public functions =========================
void g_acquire(void){}
void g_release(void){}
void bit_blt(pINT fromx,pINT fromy,pINT sizex,pINT sizey,
							pINT to_x,pINT to_y ){}
void bit_blt_inverted(pINT fromx,pINT fromy,pINT sizex,pINT sizey,
							pINT tox,pINT toy){}

// draws scaling text near palette
void draw_scaletext(REAL min,REAL max, char* title)
{ char str[10];
  int X1=5;
  RECT R;
  clear_box(0,0,0,200,16);
  clear_box(0,0,26,X1+65*6+100,26+16);
	TextOut(hCDC, X1+2, 0, title, strlen(title));
  sprintf(str,"%.2f",min);
	TextOut(hCDC, X1+2, 26,str,strlen(str));
	TextOut(hCDC, X1+32*6, 26, "0.00",4);
  sprintf(str,"%.2f",max);
	TextOut(hCDC, X1+65*6, 26, str,strlen(str));
	SetRect(&R,0,0,200,16);
	InvalidateRect(hCWindow,&R,0);
	SetRect(&R,0,26,X1+65*6+100,26+16);
	InvalidateRect(hCWindow,&R,0);
}

// draws palette into canvas window
void draw_palette()
{ int X1=5,i;
  draw_scaletext(-1.0/ScaleActivationDisplay,1.0/ScaleActivationDisplay,"Activation-Values:");
	MoveTo(hCDC,X1+200,15); LineTo(hCDC,X1+200,26);
	SelectPalette(hCDC, hNPalette, FALSE);
	RealizePalette(hCDC);
	for (i=1;i<NumColors-1;i++)
		{ FilledRect(hCDC,X1+i*6, 17, X1+i*6+6, 17+6, PALETTEINDEX(i), 0, 1); }
}

void show_overview()
{
	sprintf(info_msg[0],"Overview:");
	sprintf(info_msg[1],"%ld units in %d layers",units_used,L_anz);
	sprintf(info_msg[2],"%ld weights",weights_used);
	sprintf(info_msg[3],"display scaled by %.1f",ScaleActivationDisplay);
	sprintf(info_msg[4],"");
  draw_scaletext(-1.0/ScaleActivationDisplay,1.0/ScaleActivationDisplay,"Activation-Values:");
	InvalidateRect(hIWindow,NULL,0);
}

void rectangle(pINT x, pINT y, pINT x1, pINT y1)
{ if (BatchRun) return;
	HPEN Pen=CreatePen(PS_SOLID,1,RGB(0,0,0));
	HPEN OldPen=(HPEN)SelectObject(hCDC,Pen);
	Rectangle(hCDC,x,y,x1,y1);
	SelectObject(hCDC,OldPen);
	DeleteObject(Pen);
}

void rectangle_xor(pINT x, pINT y, pINT x1, pINT y1)
{ RECT R;
	HDC hdc = GetDC(hCWindow);
	SetRect(&R,x,y,x1,y1);
	DrawFocusRect(hdc, &R);
	ReleaseDC(hCWindow,hdc);
}

void box(pINT x, pINT y){}
void clear_box(pINT col, pINT x1, pINT y1, pINT x2, pINT y2)
{ PatBlt(hCDC, x1, y1, x2-x1, y2-y1, WHITENESS); }

void draw_unit(float val, pINT x, pINT y, pINT ux, pINT uy)
{ if (BatchRun) return;
	int negative=0, to_big=0;
  val*=ScaleActivationDisplay;
	if (val>1.0)  { val = 1.0; to_big=1; }  else
	if (val<-1.0) { val = -1.0; to_big=1; }
	if (val<0.0 && !use_colors) { val *= -1.0; negative = 1; }
	QuickFilledRect(hCDC, x, y, x+ux, y+uy, PALETTEINDEX(33+(int)(32.0*val)),1);
	if (to_big)
		{
		HPEN OldPen=(HPEN)SelectObject(hCDC,CreatePen(PS_SOLID,1,RGB(255,255,255)));
		DrawLine(hCDC,x+1,y+1,x+ux,y+uy-1);
		DrawLine(hCDC,x+ux-1,y+1,x,y+uy-1);
		DeleteObject(SelectObject(hCDC,OldPen));
		}
}

/* beschleunigte Ausgabe aller Units im Layer */
void x_draw_layer(pINT l,REAL maxabs)
{	pINT i,x1,y1,ux,uy;
  layer *L;
	int to_big;
	REAL val;
  RECT R;
	if (BatchRun) return;
	L = &L_Tab[l];
	for (L_ALL(l,i))
    {
    val=(X(l,i)/maxabs)*ScaleActivationDisplay;
    x1=PX(l,i); y1=PY(l,i); ux=L->ux; uy=L->uy;
		to_big=0.0;
	  if (val>1.0)  { val = 1.0; to_big=1; }  else
	  if (val<-1.0) { val = -1.0; to_big=1; }
	  if (val<0.0 && !use_colors) { val *= -1.0;}
	  QuickFilledRect(hCDC, x1, y1, x1+ux, y1+uy, PALETTEINDEX(33+(int)(32.0*val)),0);
	  if (to_big)
		  {
			HPEN OldPen=(HPEN)SelectObject(hCDC,CreatePen(PS_SOLID,1,RGB(255,255,255)));
		  DrawLine(hCDC,x1+1,y1+1,x1+ux,y1+uy-1);
		  DrawLine(hCDC,x1+ux-1,y1+1,x1,y1+uy-1);
		  DeleteObject(SelectObject(hCDC,OldPen));
		  }
    }
	SetRect(&R,PX(l,0),PY(l,0),PX(l,L->last)+L->dx,PY(l,L->last)+L->dy);
	InvalidateRect(hCWindow,&R,0);
}

void set_linestyle(pINT col, pINT style, pINT width)
{ HPEN hPen;
	hPen=(style)?CreatePen(PS_DOT,width,col):CreatePen(PS_SOLID,width,col);
	DeleteObject(SelectObject(hCDC,hPen));
}

void set_draw_op(pINT val){}

void draw_line(pINT x, pINT y, pINT x1, pINT y1)
{ DrawLine(hCDC,x,y,x1,y1);}

void draw_circle(pINT x, pINT y, pINT rd){}

void draw_bar(pINT col, pINT x, pINT y, pINT x1, pINT y1)
{ if (!BatchRun) FilledRect(hCDC,x,y,x1,y1,PALETTEINDEX(col),PALETTEINDEX(col),1);}

void draw_rahmen(pINT l1, pINT ext)
{ if (BatchRun) return;
	layer L=L_Tab[l1];
  RECT R;
	char titel[40];
	int x=L.x, y=L.y;
  int th=tmCWindow.tmHeight;
	COLORREF col=(ext)?RGB(255,0,255):RGB(0,255,255);
	if (ext)
			{ x+=L.extsep_x;
				y+=L.extsep_y;
				if (L.titel_ext) strcpy(titel,L.titel_ext);
								 else    sprintf(titel,"%s-X",L.titel);
			}
	else  strcpy(titel,L.titel);

	SetLayerRect(R,L_Tab[l1],ext);
 	FilledRect1(hCDC,R,col,0,2);

	TextOut(hCDC,x,y-th-2, titel, strlen(titel));
	InvalidateRect(hCWindow,NULL,0); // wichtig!
}

// Dummyfunktionen fr andere Grafik-Plattformen
char nnwait(void) {return 0;}
char x_wait_for_action(void) {return 0;}
void cursor_on(void){}
void cursor_off(void){}

void close_graph(void)
{	NeuroWindow->CloseWindow();
}

// Standard-Text-Ausgabe im Grafik-Window (Canvas)
int gprintf(pINT x, pINT y, int col, char *fmt, ...)
{ if (BatchRun) return 0; // Keine Grafikausgabe im Textmode
	RECT NRect;
	va_list args;
	int cnt;
	char str[140];
	int th=tmCWindow.tmHeight;
  int tw=tmCWindow.tmMaxCharWidth;
	va_start(args,fmt);
	cnt=vsprintf(str,fmt,(char*)args);
	TextOut(hCDC, x, y, str, strlen(str));
	NRect.left=x;
	NRect.top =y;
	NRect.right = x+strlen(str)*tw;
	NRect.bottom = y+th;
	InvalidateRect(hCWindow,&NRect,0); // wichtig!
	va_end(args);
	return(cnt);          /* Return the conversion count  */
}

/* Wird von draw_minmax_weights() aufgerufen */
void x_draw_minmax_weights()
{ // nur neu anzeigen
	InvalidateRect(hIWindow,NULL,0);
}

void show_button(pINT id, pINT stat)
{ if (stat==NEU || stat==AKTIV) stat=AUS;
	buttons[id].status=stat;
}

void show_shadow(pINT x, pINT y, pINT width, pINT height, pINT stat){}

void init_graph() {}

/* Textausgabe im Message-Window mit automatischem Scrolling,
	 wird von gof_printf() aufgerufen */
void msg_print(char *str)
{ if (BatchRun) return; // Keine Textausgabe im Batchmode
	char str1[200];
	RECT R;
	if (!str ) return;
	if (!strlen(str)) return;
	int MsgWidth=0;
	char* end=strchr(str,'\n');
	int len=(MsgText[MsgLines])?strlen(MsgText[MsgLines]):0;
	if (end)
		{strncpy(str1,str,end-str);
		str1[end-str]='\0';
		str=end+1;      // Reststring ohne '\n'
		}
	else
		{strcpy(str1,str);
		str1[strlen(str)]='\0';}

	if (!MsgText[MsgLines])
			{MsgText[MsgLines] = (char*)malloc(strlen(str1)+1);
			sprintf(MsgText[MsgLines],str1);}
	else
			{MsgText[MsgLines]=(char*)realloc(MsgText[MsgLines],\
																			len+strlen(str1)+1);
			strcat(MsgText[MsgLines],str1);
			}
	if (strlen(MsgText[MsgLines])>MsgWidth)
		MsgWidth=strlen(MsgText[MsgLines]);
	if (end)
		{MsgLines++;
		if (MsgLines>MaxMsgLines-1)
			{
			free(MsgText[0]);
			for (int i=1;i<MaxMsgLines;i++)
				MsgText[i-1]=MsgText[i];
			MsgLines=MaxMsgLines-1;
			}
		MsgText[MsgLines]=NULL;
		msg_print(str);

    if (Init_OK)
			{// Ausgabe erst wenn MsgWindow am Schirm
		GetClientRect(hMWindow,&R);
		if (MsgLines<MaxMsgLines-1 && MsgLines>R.bottom/(tmMWindow.tmHeight+tmMWindow.tmExternalLeading))
				SendMessage(hMWindow,WM_VSCROLL,SB_LINEDOWN,0);
      }
	else
		if (str1[0]=='*') Message(str1);

		}
	if (Init_OK) {InvalidateRect(hMWindow,NULL,1); /* UpdateWindow(hMWindow); */}
	else
		if (str1[0]=='*') Message(str1);
}

pINT load_layout(char* fname)
{ char fn[MAXPATH];
	strcpy(fn,fname);
	if (!strchr(fn,'.')) strcat(fn,".nnl");
	int i=LoadLayout(fn);
	return ((i==IDOK)?1:0);
}

// Meter-functions
void SetMeterValue(int Nr, float value)
{ buttons[Nr].value=value;
	NNControl[Nr]->SetValue(value); }

void SetMeterLimit(int Nr, float limit1, float limit2, float min, float max)
{ buttons[Nr].limit1=limit1;
	buttons[Nr].limit2=limit1;
	buttons[Nr].val_min=min;
	buttons[Nr].val_max=max;
	NNControl[Nr]->SetValue(buttons[Nr].value,limit1,limit2,min,max);
}

void SetMeterTitle(int Nr, char* Title)
{ buttons[Nr].text=Title;
	strcpy(NNControl[Nr]->Title,Title); }

// inserts Menu-Item at specified pos with menutext and callback-ID 
void insert_menu(int nMainPos,int nSubPos,char* text,int ID)
{       
	HMENU hMenu = GetSubMenu(GetMenu(hNWindow),nMainPos);
	InsertMenu(hMenu,nSubPos,MF_BYPOSITION,ID,text);
	DrawMenuBar(hNWindow);
}

// appends one entry in windows-menu for user-defined windows 
void AppendWindowMenu(void *WndPtr)
{ int i;
  PTWindowsObject Wnd=(PTWindowsObject)WndPtr;
	for (i=0;i<MaxUserWindow;i++)
		if (UserWindow[i].ptr==NULL) break;
	if (i==MaxUserWindow) return; //all user window entries used 
	insert_menu(5,-1, Wnd->Title, CM_ShowFirstUserWindow+i);
  // set data
	UserWindow[i].ptr  = Wnd;
	UserWindow[i].ID   = CM_ShowFirstUserWindow+i;
	UserWindow[i].Show = FALSE;
	// set checkmark in window-menu and toggle UserWindow[i].Show
	NeuroWindow->CMShowUserWindow(i);
}

void NNShowMenu(BOOL Show)
{ if (Show)
		SetMenu(hNWindow,hNMenu);
	else
		SetMenu(hNWindow,0);
}

void NNSetWindowPos(int WinNr,int ShowWindow, int X, int Y, int W, int H)
{ int Typ=(ShowWindow==NN_SHOWWINDOW)?SWP_SHOWWINDOW:SWP_HIDEWINDOW;
  TMessage Msg;
	if (ShowWindow==NN_SETWINDOWPOS)
		switch (WinNr) // SetWindowPos & Size
			{
			case MAIN_WINDOW:
				SetWindowPos(hNWindow,0,X,Y,W,H,SWP_NOZORDER);
				break;
			case INFO_WINDOW:
				SetWindowPos(hIWindow,0,X,Y,W,H,SWP_NOZORDER);
				break;
			case MSG_WINDOW:
				SetWindowPos(hMWindow,0,X,Y,W,H,SWP_NOZORDER);
				break;
			default:
				if (WinNr>=USER_WINDOW && WinNr<=USER_WINDOW+MaxUserWindow)
					if (UserWindow[WinNr-USER_WINDOW].ID>=0)
						SetWindowPos(UserWindow[WinNr-USER_WINDOW].ptr->HWindow,0,X,Y,W,H,SWP_NOZORDER);
				break;
			}
	else
		switch (WinNr) // Show/Hide window with changing menu selection
			{
			case MAIN_WINDOW:
				SetWindowPos(hNWindow,0,X,Y,W,H,Typ);
				break;
			case INFO_WINDOW:
				Show_InfoWindow=(ShowWindow==NN_HIDEWINDOW);
				NeuroWindow->CMShowInfoWindow(Msg);
				break;
			case MSG_WINDOW:
				Show_MsgWindow=(ShowWindow==NN_HIDEWINDOW);
				NeuroWindow->CMShowMsgWindow(Msg);
				break;
			case DRAW_WINDOW:
				Show_CanvasWindow=(ShowWindow==NN_HIDEWINDOW);
				NeuroWindow->CMShowCanvasWindow(Msg);
	break;
			default:
				if (WinNr>=USER_WINDOW && WinNr<=USER_WINDOW+MaxUserWindow
					&& UserWindow[WinNr-USER_WINDOW].ID>=0)
					{
					UserWindow[WinNr-USER_WINDOW].Show=(ShowWindow==NN_HIDEWINDOW);
					NeuroWindow->CMShowUserWindow(WinNr-USER_WINDOW);
	  }
				break;
			}
}

/* Check for special arguments */
void check_args(int argc,char** argv)
{ int i;
	if (argc<2) return;
	for (i=1; i<argc; i++)
		if (argv[i][0]=='-')
		{
		if (ARG_IS(i,"-h"))
			{
			printf("\n%s - A Neural Network Program",argv[0]);
			printf("\nusing the VieNet2-Toolkit %s.\n",VERSION);
			printf("\nFurther information will be available later...\n\n");
			exit(0);
			}
		if (ARG_IS(i,"-mono"))
			use_colors = False;
		}
}

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
			 LPSTR lpCmd, int nCmdShow)
{
	TNeuroApp NeuroApp( "VieNet2", hInstance, hPrevInstance,
		lpCmd, nCmdShow);
	PNeuroApp = &NeuroApp;
  dde_init();
  nn_setup(1,_argv);
	// Buttons und Layer definieren VOR NeuroWindow-Init
	//      main() ruft dann init_xgraph() auf
	nn_init(0,NULL); // not used in WIN
	NeuroApp.MessageLoop();
	return ( NeuroApp.Status );
}

void NNErrMsg(char* sn, char* str)
{	ErrMsg("In modul %s: %s",sn, str);}

void NNAlarmMsg(char* sn, char* str)
{	AlarmMsg("In modul %s: %s",sn, str);}

/* INIT Graphics, wird von main() aufgerufen */
void init_xgraph(int argc,char **argv,int UpperLeftX,int UpperLeftY,int WindowWidth,int WindowHeight)
{ /* Check and strip Special Arguments */
	NeuroWindow_X=UpperLeftX;
	NeuroWindow_Y=UpperLeftY;
	NeuroWindow_W=WindowWidth;
	NeuroWindow_H=WindowHeight;
	PNeuroApp->InitInstance(); // Init all windows
  NNGraphInit=TRUE;
}

