/*******************************************************************************
* Program:       NN_BP.C
* Project:	     VieNet2 V2.03
* Description:   General Backprop-Module for VieNet2
* Version:       2.03
* Last modified: 9.08.96
* Author:        LG
* Copyright (C) 1991-96 Austrian Research Institute for Artificial Intelligence
********************************************************************************/

/* This module implements functions for backprop networks.
   (with automatic bias-unit update, if defined & connected,
    because of use_weights=ALL in NNLearn(..) )


  I) 3-layer backprop networks

   Example:
     (1) Setup weights (layers must already be defined, wmin=-0.6, wmax=0.6):
           NN_BP_Init(IN,HIDDEN,OUT,BIAS,-0.6,0.6);

     (2) One update through network (without external layer)
           NN_BP_Update(IN,HIDDEN,OUT);

     (3) One Learnstep (rate=0.5)               
           NN_BP_Learn(IN,HIDDEN,OUT,0.5);

*/


#define MODULE

#include "neurodef.h"
#include "neurons.h"
#ifdef PC
#include <alloc.h>
#include "biblio\nn_bp.h"
#else
#include <malloc.h>
#include "biblio/nn_bp.h"
#endif

#define SOURCENAME "NN_BP"

/*====== functions for 3-layer network ======*/
/* Setup a 3-layer network with BIAS-Units */
void NN_BP_Init(pINT IN, pINT HIDDEN, pINT OUT, pINT BIAS, float wmin, float wmax)
{ nnconnect(     IN, HIDDEN, FULL,wmin,wmax,1.0);
  nnconnect(  BIAS, HIDDEN, FULL,wmin,wmax,1.0);
  nnconnect( HIDDEN, OUT,    FULL,wmin,wmax,1.0);
  nnconnect(  BIAS, OUT, FULL,wmin,wmax,1.0);
}

/* Update through 3-Layer network */
void NN_BP_Update(pINT IN, pINT HIDDEN, pINT OUT)
{ update(IN,HIDDEN,STANDARD,ALL);
  copy_layer(HIDDEN);
  update(HIDDEN,OUT,STANDARD,ALL);
  copy_layer(OUT);
}

/* simple learning for 3-Layer-Backprop */
void NN_BP_Learn(pINT IN, pINT HIDDEN, pINT OUT, float rate)
{
#ifndef ECANSE
  NNSetTarget(OUT,NULL);  /* calc target values, not for ECANSE! */
#endif
	/* OUT: delta[j] = f'(x[j]) * (target[j]-x[j]) */
  NNSetDelta(OUT,BACKPROP);

  /* HIDDEN: delta[i] = f'(x[i]) * SUMj (W(i,j)*delta[j]) */
  update(OUT,HIDDEN,BACKWARD,SINGLE);

	/* BIAS2,HIDDEN -> OUT: W(i,j) += x[i]*d[j]*rate */
  NNLearn(HIDDEN,OUT,BACKPROP_LEARN,rate,ALL);

	/* BIAS1,IN -> HIDDEN: W(i,j) += x[i]*delta[j]*rate */
  NNLearn(IN,HIDDEN,BACKPROP_LEARN,rate,ALL);
}

