/*************************************************************************
*  Program:       NEURONS.C                                              *
*  Description:   Core funtions for VieNet2                              *
*  Version:       2.03                                                   *
*  Last modified: 7.09.96                                                *
*  Author:        LG                                                     *
*  Copyright (C) Austrian Research Institute for Artificial Intelligence *
**************************************************************************
*  Tabsize 2 - Apollo: CV mit Domain-Command TS 1 3 -R umstellen         *
*              PC:     Turbo/Borland C - Optionen Tabsize 2 einstellen   *
**************************************************************************/

/* ===== Revision History =====
        2.12.92  LG  1) l_p() liest jetzt auch neg. Patternwerte im Integer-
                        bereich [-32768,+32767]
                     2) load_patterns(): Header kann auch Integer-Werte >255
                        enthalten, also Patterngroessen
                        bis (theoretisch) 65535x65535 zulaessig
                        patmax kann auch neg. sein im Bereich [-32768,+32767]

        28.2.93 LG  1)  pINT best[MAX], win[MAX];  wird nicht mehr unterstuetzt
                        REAL ss[MAX];

        8.3.93  LG  1)  Umstellung von x[], xnew[] und p[] auf dynamische Allokierung
                        -> max. 65535 Units in max. 65535 Layers moeglich (12 Bytes je Unit noetig),
                        da Unit-Indizes auf unsigned int (pINT) beschraenkt sind
                        max. 65535 Gewichte je nnconnect()-Aufruf, d.h. zw. je zwei Layern
        26.4.93 LG  1)  nn_setup(..) fuer Initialisierung der FileNamen
        20.9.93 LG  1)  optimiertes Patternmanagement durch Unterstuetzung von
                        Direktzugriff auf beliebige Patterns via Tabelle in
                        L_Tab[].p_tab und mitgefuehrter Patternanzahl
                        in L_Tab[].patanz.
  2.11.93 LG  1) NNLearn(), NNSetDelta(), NNSetTarget()
              2) load_data()
  11.11.93 LG 1) draw_minmax_weights() verbessert ->info_msg[]
  29.11.93 LG 1) L_ALL2() in update, damit keine negativen unitnr-differenzen auftreten
  10.02.94 LG 1) total, total_ext, change in struct layer nicht mehr unterstuetzt
              2) lokale Aktivierungsfunktionen je Layer (L_Tab[A].f bzw. L_Tab[A].df)

  15.02.94 LG 1) Version 2.0 mit:
                 a) dynamische, layerlokale Unitarrays statt x[],xnew[],p[]
                 b) Debugoption NNDEBUG fuer automatisches Rangechecking in allen Makros

   7.7.96 LG Update to V2.02 with:
             1) fixed free()-bug in free_patterns()
             2) limit f() to x>-227.0 to avoid overflows
             3) nn_break(long step) supports display of learning step and
                automatic weights saving after each step%AutoSaveWeights==0
             4) new makros min,max,sgn
             5) new Makro ALL_PATTERNS(l,i)
             6) new Makros X_ext(l,i), X_ext_next(l,i), X_ext_prev(l,i)
             7) new Makros X_target(l,i), X_delta(l,i)
             8) pattern parameters per pattern are calculated automatically while loading
                using calc_pattern_stat(pattern *p):
                  p->min, p->max, p->mean, p->var, p->length (=vectorlength)
             9) pattern parameters per layerunit (i.e. for each pattern column) are
                calculated automatically while loading using calc_pattern_stats(pINT l):
                  L->pmax[i], L->pmin[i], L->pmean[i], L->pvar[i]
             10) scale_patterns(pINT l) may be used to scale pattern-columns to [0,1]
                 after loading
             11) ignore comment-lines starting with ';' in load_data() and load_data2()
             12) nnmalloc(), nncalloc(), nnrealloc(), nnfree() for supporting
                 autom. logging of memory accesses to file "alloc.log"
                 if flag NNLOGALLOC is defined in neurodef.h
        ============================ */

#define MODULE
#define SOURCENAME "NEURONS"

/* Reihenfolge der Include-File muss eingehalten werden! */
#include "neurodef.h"
#include "neurons.h"
#include "nngraph.h"

#ifdef PC
#   include <dir.h>
#   include <direct.h>
#   include <alloc.h>
#   include <ctype.h>               /* toupper() */
#   include <stdlib.h>
#   include <stdarg.h>
#   include <string.h>
#   define getwd(x) getcwd(x,sizeof(x))
#   ifndef WIN
#     include <graphics.h>
#   endif
#endif

#include <stdio.h>
#include <math.h> /* fabs() */

/* Position for Extern-Layer Display-Position-Setting in init_layer() */
#define EXTSEP_X 0
#define EXTSEP_Y L.dy*units_ver+L.dy+20 /* Platz fuer Ueberschrift + Patternname */

FILE *pattern_file;
FILE *sched_file;
char dummy[100];

/* local prototypes */
void do_action(pINT);
pINT set_w_index(pINT);
int  push_scroll (char *);

/* Debugfunctions, should be inline */
void debug_layer(pINT l)
{(l>L_anz-1)?nn_error(__FILE__,"Line %u: Layernr (%u) >= L_anz (%u)",__LINE__,l,L_anz):0;}

void debug_unit(pINT l,pINT u)
{(u>L_Tab[l].last)?nn_error(__FILE__,"Line %u: Unitnr (%u) >= L.size (%u) in layer %u (%s)",__LINE__,u,L_Tab[l].size,l,L_Tab[l].titel):0;}

void debug_layer_L(layer L)
{(L.nr>L_anz-1)?nn_error(__FILE__,"Line %u: Layernr (%u) >= L_anz (%u)",__LINE__,L.nr,L_anz):0;}

void debug_unit_L(layer L,pINT u)
{(u>L.last)?nn_error(__FILE__,"Line %u: Unitnr (%u) >= L.size (%u) in layer %u (%s)",__LINE__,u,L.size,L.nr,L.titel):0;}

void debug_weight(pINT l1,pINT l2,pINT i,pINT j)
{debug_layer(l1);debug_layer(l2);debug_unit(l1,i);debug_unit(l2,j);
(w_ptr[l1][l2][W_LEVEL]->w && L_Tab[l2].size*i+j<w_ptr[l1][l2][W_LEVEL]->size)?0:nn_error(__FILE__,"Line %u: access to nonexisting W(%u,%u,%u,%u); (Layer %s->%s)",__LINE__,l1,l2,i,j,L_Tab[l1].titel,L_Tab[l2].titel);}

/*********************************************************************/
/*                                                                   */
/*       MISC. - MATHS, SYSTEM                                       */
/*                                                                   */
/*********************************************************************/

/* sets random starting point for random-number generator */
void init_randn(void)
{ long int t;
  srand((unsigned)time(&t));
}

REAL roundn(REAL x, pINT digits)
{
  double xx = (double)x; /* shiftable now */
  double shift = pow(10.0,(double)digits);
  return floor(xx*shift+0.5) / shift;
}

/* sichere Version von tolower()
  (Standardfunktion funktioniert nicht korrekt)
*/
char lower(char c)
{ return( (c>'Z' || c<'A')? c : tolower(c) ); }

/*********************************************************************/
/*                                                                   */
/*              ERROR - ALARM - INFORM                               */
/*                                                                   */
/*********************************************************************/

/* Errorausgabe: sn=SOURCENAME, fmt=Printformat */
int nn_error(char *sn, char *fmt, ...)
{
  va_list args;
  char str[GP_BUFFERSIZE];
  int cnt;

  if (!fmt || !sn)
    {/* fatal error in parameters */
    nn_fatalerror(SOURCENAME,"nn_error(%s, %s,...): wrong format",sn,fmt);
    close_graph();
    exit(1);
    }

  va_start(args,fmt);
  cnt = vsprintf(str,fmt,args);
  va_end(args);

  if (EG)
#ifndef WIN
    gp_printf("**ERROR** %s::%s\n",sn,str);
#else
    NNErrMsg(sn,str);
#endif
  if (EO) printf("**ERROR** %s::%s\n",sn,str);
  if (EF) pprintf_file(ErrorLogFile,"**ERROR** %s::%s\n",sn,str);

  if (cnt==EOF)
    nn_error(SOURCENAME,"nn_error(%s, %s,...): wrong format",sn,fmt);

  if (ERRORSTOP)
    { /* !f! close_files();*/
    close_graph();
    exit(1);
    }
  return cnt;
}

/* fatal error, supports message window */
void nn_fatalerror(char *sn, char *fmt, ...)
{
  va_list args;
  char str[GP_BUFFERSIZE];

  if (!fmt || !sn)
    {/* fatal error in parameters */
    nn_fatalerror(SOURCENAME,"nn_error(%s, %s,...): wrong format",sn,fmt);
    close_graph();
    exit(1);
    }

  va_start(args,fmt);
  vsprintf(str,fmt,args);
  va_end(args);

  if (EG)
#ifndef WIN
    gp_printf("**ERROR** %s::%s\n",sn,str);
#else
    NNErrMsg(sn,str);
#endif

  if (EO) printf("**ERROR** %s::%s\n",sn,str);
  if (EF) pprintf_file(ErrorLogFile,"**ERROR** %s::%s\n",sn,str);

  close_graph();
  exit(1);
}

int nn_alarm(char *sn, char *fmt, ...)
{
  va_list args;
  char str[GP_BUFFERSIZE];
  int cnt;

  va_start(args,fmt);
  cnt = vsprintf(str, fmt, args);
  va_end(args);

  if (AG)
#ifndef WIN
    gp_printf("**ALARM** %s::%s\n",sn,str);
#else
    NNAlarmMsg(sn,str);
#endif
  if (AO) printf("**ALARM** %s::%s\n",sn,str);
  if (AF) pprintf_file(ErrorLogFile,"**ALARM** %s::%s\n",sn,str);

  if (cnt==EOF)
     nn_error(SOURCENAME,"nn_alarm(%i,%i,%s,...): wrong format",sn,fmt);
  if (ALARMSTOP)
    {
    close_graph();
    exit(1);
    }
  return cnt;
}

int pprintf_e(pINT gof, char *fmt, ...)
{
  va_list args;
  char str[GP_BUFFERSIZE];
  int cnt;
  va_start(args, fmt);
  cnt = vsprintf(str,fmt,args);
  va_end(args);
  if (gof & PG) gp_printf(str);
  if (gof & PO) printf(str);
  if (gof & PF) pprintf_file(LogFile,str);

  if (cnt==EOF)
    nn_error(SOURCENAME,"pprintf_e(%i,%s,..): wrong format",gof,fmt);
  return cnt;
}

int pprintf(char *fmt, ...)
{
  va_list args;
  char str[GP_BUFFERSIZE];
  int cnt;
  va_start(args, fmt);
  cnt = vsprintf(str,fmt,args);
  va_end(args);

  pprintf_e(_GOF,str);

  if (cnt==EOF)
     nn_error(SOURCENAME,"pprintf(%s,..): wrong format",fmt);
  return cnt;
}


void inform(pINT about, REAL val)
{ /*!! needs a brushup */
  char text[20];
  switch (about) {
         case 1: sprintf(text,"er1%: %f ",val);
                 break;
         case 3: sprintf(text,"lrn: %i ",(pINT)val);
                 break;
         case 4: sprintf(text,"rep: %i ",(pINT)val);
                 break;
         case 5: sprintf(text,"LEV: %i ",(pINT)val);
                 break;
         case 6: sprintf(text,"ALARMW: %f ",val);
                 break;
         default: sprintf(text,"inform: '%i undefd'",about);
  }
  pprintf("\n%s\n",text);
}

/* prints to logfile, returns number of chars written
   or FALSE if error */
int pprintf_file(char *logfile, char *fmt, ...)
{
  va_list args;
  char str[GP_BUFFERSIZE];
  int cnt;
  FILE *fh;
  va_start(args, fmt);
  cnt = vsprintf(str,fmt,args);
  va_end(args);
  if (cnt==EOF)
     nn_error(SOURCENAME,"pprint_file(%s,..): wrong format",fmt);
  if (!(fh=fopen(logfile,"at")))
  return FALSE;
  fputs(str,fh);
  fclose(fh);
  return cnt;
}


/*********************************************************************/
/*                                                                   */
/*   DEFINING THE ARCHITECTURE                                       */
/*                                                                   */
/*********************************************************************/

/* defines unitpositions on screen */
void reposition_layer(layer *L)
{ int i, rx = L->x+2, ry = L->y+2;
  /* Init unit-arrays & calculate unitpositions */
  for (i=0;i<L->size;i++)
    {
    L->X[i]=L->XNEW[i]=(L->typ & BIAS_LAYER)?1.0:0.0; /* init BIAS to 1.0 */
    if ( (L->typ & HEX) && ((i%L->upr)==0) && ((i/L->upr)%2==0))
        rx += L->dx / 2;
    L->POS[i].x = rx;
    L->POS[i].y = ry;
    if ((i%L->upr) ==  L->upr - 1 )
      {
      ry = ry + L->dy;
      rx = L->x+2;
      }
    else rx += L->dx;
    }
  L->status = L_NEU; /* redraw necessary */
}

/* reallocates memory for unitarrays */
pINT resize_layer(layer *L, pINT newsize, pINT newwidth)
{
  if (newsize > 65535)
     nn_fatalerror(SOURCENAME,"resize_layer(): to many units defined (>65535) in layer %u!",L->nr);
  if (newwidth>newsize) newwidth=newsize;
  units_used -= L->size; units_used += newsize;
  L->size = newsize;
  L->last = newsize-1;
  L->upr  = newwidth;

  if (L->X==NULL)
     {/* unix requires allocation first */
     if (!(L->X=(REAL*)nnmalloc(L->size*sizeof(REAL))))
          nn_error(SOURCENAME,"init_layer(%s): Error allocating X[]",L->titel);
     if (!(L->XNEW=(REAL*)nnmalloc(L->size*sizeof(REAL))))
          nn_error(SOURCENAME,"init_layer(%s): Error allocating XNEW[]",L->titel);
     if (!(L->POS=(UPOS*)nnmalloc(L->size*sizeof(UPOS))))
          nn_error(SOURCENAME,"init_layer(%s): Error allocating POS[]",L->titel);
     }
  else
     {
     if (!(L->X=(REAL*)nnrealloc(L->X,L->size*sizeof(REAL))))
          nn_error(SOURCENAME,"init_layer(%s): Error reallocating X[]",L->titel);
     if (!(L->XNEW=(REAL*)nnrealloc(L->XNEW,L->size*sizeof(REAL))))
          nn_error(SOURCENAME,"init_layer(%s): Error reallocating XNEW[]",L->titel);
     if (!(L->POS=(UPOS*)nnrealloc(L->POS,L->size*sizeof(UPOS))))
          nn_error(SOURCENAME,"init_layer(%s): Error reallocating POS[]",L->titel);
     }
  reposition_layer(L);
}

/* allocate and initialize a layer */
pINT init_layer(pINT units_hor,
                pINT units_ver,
                pINT typ,   /* GRID or HEX, BIAS_LAYER, DELTA_LAYER */
                pINT start_x,
                pINT start_y,
                char *titel)
{
  layer L;
  pINT i;
  pINT size = units_hor * units_ver;
  if ((long)units_hor*(long)units_ver > 65535)
     nn_fatalerror(SOURCENAME,"initlayer(): to many units defined (>65535) in layer %u!",L_anz+1);
  L.first = 0;
  L.last = size - 1;
  L.size = size;
  if (L_anz==MAXLAYER)
     nn_fatalerror(SOURCENAME,"initlayer(): to many layers defined (>MAXLAYER (%u))!",MAXLAYER);
  L.nr = L_anz++;
  units_used += size;
  L.upr = units_hor;
  L.x = start_x;
  L.y = start_y;
  L.id = 0;
  L.ext = NULL;
  L.p = NULL;
  L.p_tab = NULL;
  L.patanz=0;
  L.status = L_NEU;
  L.typ = typ;
  L.ux = UWID;
  L.uy = UWID;
  L.dx = DX;
  L.dy = DX;
  L.extsep_x = EXTSEP_X;
  L.extsep_y = EXTSEP_Y;
  L.target = L.delta = NULL;
  L.X = L.XNEW = NULL;
  L.POS = NULL;
  L.f = f; /* default activation function */
  L.df = df;
  L.pmin=L.pmax=L.pmean=L.pvar=NULL;

  if (!(L.titel = (char*)nnmalloc(strlen(titel)+1)))
     nn_alarm(SOURCENAME,"init_layer(%s): Error allocating memory for titelstring",titel);
  else
     strcpy(L.titel,titel);
  if (!(L.titel_ext = (char*)nnmalloc(7+strlen(titel)+1)))
     nn_alarm(SOURCENAME,"init_layer(%s): Error allocating memory for default extern titelstring",titel);
  else
     {strcpy(L.titel_ext,"extern ");
     strcat(L.titel_ext,titel);}
  if (typ & DELTA_LAYER)
     {
     if (!(L.target = (REAL*)nnmalloc(size*sizeof(REAL))))
        nn_alarm(SOURCENAME,"init_layer(%s): Error allocating memory for target-array",titel);
     if (!(L.delta  = (REAL*)nnmalloc(size*sizeof(REAL))))
        nn_alarm(SOURCENAME,"init_layer(%s): Error allocating memory for delta-array",titel);
     }
  resize_layer(&L, size, units_hor);
  L_Tab[L.nr] = L;
  return L.nr;
}

void init_ext(pINT l1, int x, int y, char *titel_ext)
{
  L_Tab[l1].extsep_x = x;
  L_Tab[l1].extsep_y = y;
  if (titel_ext)
    {
    if (!L_Tab[l1].titel_ext)
       {
       if (!(L_Tab[l1].titel_ext = (char*)nnmalloc(strlen(titel_ext)+1)))
          nn_alarm(SOURCENAME,"init_ext(%s): Error allocating memory for extern titelstring",titel_ext);
       }
     else
       if (!(L_Tab[l1].titel_ext = (char*)nnrealloc(L_Tab[l1].titel_ext,strlen(titel_ext)+1)))
         nn_alarm(SOURCENAME,"init_ext(%s): Error reallocating memory for extern titelstring",titel_ext);

    strcpy(L_Tab[l1].titel_ext,titel_ext);
    }
}

/* connect_b() verbindet l2 ->l1  mit den Gewichten von l1->l2,
   dh.: W(l1, l2, i, j) = W(l2, l1, j, i) * Skalierung
*/
void connect_b(pINT l2, pINT l1, REAL scale)
{
  pINT i, j;
  connect_0(l2, l1);
  for(L_ALL(l1, i))
    for(L_ALL(l2, j))
      W(l2,l1,j,i) = W(l1,l2,i,j) * scale;
}

/* connect_n1() verbindet jede Unit aus l2 mit genau n (zufaellig
   gewaehlten) Units von l1, in Richtung l1->l2.
   Alle Gewichte werden mit 0 initialisiert.
*/
void connect_n1(pINT l1, pINT l2, pINT n, REAL val)
{
  pINT i,j,k,l1size = L_Tab[l1].size;
  connect_0(l1,l2);
  for(L_ALL(l2,j))
    for (k=0; k < n; k++)
      {
      do {
         i = (pINT) (random_nr() * (REAL)l1size);
         } while (W(l1,l2,i,j) != 0.0);
      W(l1,l2,i,j) = val;
      }
}

/* Verbindet jede Unit aus l1 mit genau n (zufaellig gewaehlten)
   Units von l2, in Richtung l1->l2.
   Alle Gewichte werden mit 0 initialisiert.
*/
void connect_1n(pINT l1, pINT l2, pINT n, REAL val)
{
  pINT i, j, k, l2size = L_Tab[l2].size;
  connect_0(l1, l2);
  for(L_ALL(l1, i))
    for(k=0; k < n; k++) {
      do {
         j = (pINT) (random_nr() * (REAL)l2size);
         } while (W(l1, l2, i, j) != 0.0);
      W(l1, l2, i, j) = val;
      }
}

/* connects all units in l1 with last unit in l2 */
void nnconnect_n1(pINT l1,
             pINT l2,
             pINT mode,
             REAL range_lo,
             REAL range_hi,
             REAL density)
{
  int i, j=L_Tab[l2].last;
  unsigned int size, size1=L_Tab[l1].size, size2=L_Tab[l2].size;
  weight_t* wptr = w_ptr[l1][l2][W_LEVEL];

  size2=1;
  switch (mode) {
    case ORDER2:
         size = (size1 * size1 / 2 - size1 / 2) * size2;
         break;
    case IDENT:
         size = size1;  /* Nur ein Gewichte je Unit in l1 */
         break;
    default:
         size = size1 * size2;
         break;
  }

  if (wptr == NULL) {
    /* erstmalige Initialisierung der Gewichtsmatrix */
    if ((wptr = w_ptr[l1][l2][W_LEVEL] =
          (weight_t *)nnmalloc(sizeof(weight_t))) == NULL)
        nn_error(SOURCENAME,"nnconnect_n1(%i,%i,%i,%g,%g,%g): Alloc-Fehler Weight-Structures",
                            l1,l2,mode,range_lo,range_hi,density);

    if ((wptr->w = (REAL *)nncalloc(size,sizeof(REAL))) == NULL)
        nn_error(SOURCENAME,"nnconnect_n1(%i,%i,%i,%g,%g,%g): Alloc-Fehler Weights",
                            l1,l2,mode,range_lo,range_hi,density);
    wptr->size = size;
    wptr->typ  = mode;
    weights_used+=size; /* nur beim ersten Mal! */
    }
  else
    {
    weights_used+=size;
    wptr->size += size;
    if ((wptr->w = (REAL *)nnrealloc(wptr->w,wptr->size*sizeof(REAL))) == NULL)
        nn_error(SOURCENAME,"nnconnect_n1(%i,%i,%i,%g,%g,%g): Alloc-Fehler Weights",
                            l1,l2,mode,range_lo,range_hi,density);
    }
  switch (mode) {
    case FULL: /* Direktes Setzen der linearisierten Matrix */
      for(L_ALL(l1,i))  /* Alle w ausser Verbindungen mit sich selbst */
        if (random_nr()<density)
           W(l1,l2,i,j) = range_lo+(range_hi-range_lo)*random_nr();
        else
           W(l1,l2,i,j) = 0.0;
        break;
   case COMP:
      for(L_ALL(l1,i))  /* Alle w ausser Verbindungen mit sich selbst */
        if (i!=j && random_nr()<density)
           W(l1,l2,i,j) = range_lo+(range_hi-range_lo)*random_nr();
        else
           W(l1,l2,i,j) = 0.0;
      break;
   case IDENT: /* Eins-zu-Eins-Verbindungen, als UPDATE geht nur W_COPY! */
        if (random_nr()<density)
           W_1(l1,l2,j) = range_lo+(range_hi-range_lo)*random_nr();
        else
           W_1(l1,l2,j) = 0.0;
      break;
   default:
        nn_alarm(SOURCENAME,"nnconnect_n1(%i,%i,%i,%g,%g,%g): unknown nnconnect-mode",
                            l1,l2,mode,range_lo,range_hi,density);
   }
}

/* connects last unit in l1 with all units in l2 */
void nnconnect_1n(pINT l1,
             pINT l2,
             pINT mode,
             REAL range_lo,
             REAL range_hi,
             REAL density)
{
  int i=L_Tab[l1].last,j;
  unsigned int size, size1=L_Tab[l1].size, size2=L_Tab[l2].size;
  weight_t* wptr = w_ptr[l1][l2][W_LEVEL];

  size1=1;
  switch (mode) {
    case ORDER2:
         size = (size1 * size1 / 2 - size1 / 2) * size2;
         break;
    case IDENT:
         size = size1;  /* Nur ein Gewichte je Unit in l1 */
         break;
    default:
         size = size1 * size2;
         break;
  }

  if (wptr == NULL) {
    /* erstmalige Initialisierung der Gewichtsmatrix */
    if ((wptr = w_ptr[l1][l2][W_LEVEL] =
           (weight_t *)nnmalloc(sizeof(weight_t))) == NULL)
       nn_error(SOURCENAME,"nnconnect_1n(%i,%i,%i,%g,%g,%g): Alloc-Fehler Weight-Structures",
                           l1,l2,mode,range_lo,range_hi,density);

    if ((wptr->w = (REAL *)nncalloc(size,sizeof(REAL))) == NULL)
       nn_error(SOURCENAME,"nnconnect_1n(%i,%i,%i,%g,%g,%g): Alloc-Fehler Weights",
                           l1,l2,mode,range_lo,range_hi,density);
    wptr->size = size;
    wptr->typ  = mode;
    weights_used+=size; /* nur beim ersten Mal! */
    }
  else
    {
    weights_used+=size;
    wptr->size += size;
    if ((wptr->w = (REAL *)nnrealloc(wptr->w,wptr->size*sizeof(REAL))) == NULL)
       nn_error(SOURCENAME,"nnconnect_1n(%i,%i,%i,%g,%g,%g): Alloc-Fehler Weights",
                           l1,l2,mode,range_lo,range_hi,density);
    }
  switch (mode) {
    case FULL: /* Direktes Setzen der linearisierten Matrix */
      for(L_ALL(l2,j))  /* Alle w ausser Verbindungen mit sich selbst */
        if (random_nr()<density)
          W(l1,l2,i,j) = range_lo+(range_hi-range_lo)*random_nr();
        else
          W(l1,l2,i,j) = 0.0;
        break;
   case COMP:
      for(L_ALL(l2,j))  /* Alle w ausser Verbindungen mit sich selbst */
        if (i!=j && random_nr()<density)
          W(l1,l2,i,j) = range_lo+(range_hi-range_lo)*random_nr();
        else
          W(l1,l2,i,j) = 0.0;
      break;
   case IDENT: /* Eins-zu-Eins-Verbindungen, als UPDATE geht nur W_COPY! */
      if (random_nr()<density)
         W_1(l1,l2,i) = range_lo+(range_hi-range_lo)*random_nr();
      else
         W_1(l1,l2,i) = 0.0;
      break;
   default:
      nn_alarm(SOURCENAME,"nnconnect_1n(%i,%i,%i,%g,%g,%g): unknown nnconnect-mode",
                          l1,l2,mode,range_lo,range_hi,density);
   }
}

/* Weight-Matrix zwischen Layer l1,l2 aufbauen und initialisieren
        init_rand() sollte einmalig vorher aufgerufen worden sein
*/
void nnconnect(pINT l1,
             pINT l2,
             pINT mode,
             REAL range_lo,
             REAL range_hi,
             REAL density)
{
  int i, j;
  unsigned int size, size1=L_Tab[l1].size, size2=L_Tab[l2].size;
  weight_t* wptr = w_ptr[l1][l2][W_LEVEL];

  if (wptr == NULL) {
    /* erstmalige Initialisierung der Gewichtsmatrix */
    switch (mode) {
      case ORDER2:
           size = (size1 * size1 / 2 - size1 / 2) * size2;
           break;
      case IDENT:
           size = size1;  /* Nur ein Gewichte je Unit in l1 */
           break;
      default:
           size = size1 * size2;
           break;
    }

    if ((wptr = w_ptr[l1][l2][W_LEVEL] =
               (weight_t *)nnmalloc(sizeof(weight_t))) == NULL)
       nn_error(SOURCENAME,"nnconnect(%i,%i,%i,%g,%g,%g): Alloc-Fehler Weight-Structures",
                           l1,l2,mode,range_lo,range_hi,density);

    if ((wptr->w = (REAL *)nncalloc(size,sizeof(REAL))) == NULL)
       nn_error(SOURCENAME,"nnconnect(%i,%i,%i,%g,%g,%g): Alloc-Fehler Weights",
                           l1,l2,mode,range_lo,range_hi,density);
    wptr->size = size;
    wptr->typ  = mode;
    weights_used+=size; /* nur beim ersten Mal! */
    }
  else
    size = wptr->size;

  switch (mode) {
    case FULL: /* Direktes Setzen der linearisierten Matrix */
      for (i = 0; i < size; i++)
        if (random_nr() < density)
          w_ptr[l1][l2][W_LEVEL]->w[i]=range_lo + (range_hi-range_lo)*random_nr();
        else
          w_ptr[l1][l2][W_LEVEL]->w[i]=0.0;
        break;
   case COMP:
      for(L_ALL(l1,i))  /* Alle w ausser Verbindungen mit sich selbst */
        for(L_ALL(l2,j))
          if (i!=j && random_nr()<density)
            W(l1,l2,i,j) = range_lo+(range_hi-range_lo)*random_nr();
          else
            W(l1,l2,i,j) = 0.0;
      break;
   case IDENT: /* Eins-zu-Eins-Verbindungen, als UPDATE geht nur W_COPY! */
      for(L_ALL(l1,i))
        if (random_nr()<density)
          W_1(l1,l2,i) = range_lo+(range_hi-range_lo)*random_nr();
        else
          W_1(l1,l2,i) = 0.0;
      break;
   default: nn_alarm(SOURCENAME,"nnconnect(%i,%i,%i,%g,%g,%g): unknown nnconnect-mode",
                                l1,l2,mode,range_lo,range_hi,density);
   }
}

/* free all weights for layer l */
void deconnect_layer(pINT l)
{ pINT lev, l1, l2;
  for (ALL_LEVELS(lev))
    {
    for (ALL_LAYERS(l1))
      {
      /* from l to others */
      if (w_ptr[l][l1][lev])
        {
        if (w_ptr[l][l1][lev]->w) nnfree(w_ptr[l][l1][lev]->w);
        nnfree(w_ptr[l][l1][lev]);
        w_ptr[l][l1][lev]=NULL;
        }
      /* from others to l */
      if (w_ptr[l1][l][lev])
        {
        if (w_ptr[l1][l][lev]->w) nnfree(w_ptr[l1][l][lev]->w);
        nnfree(w_ptr[l1][l][lev]);
        w_ptr[l1][l][lev]=NULL;
        }
      }
    /* intralayer l */
    if (w_ptr[l][l][lev])
      {
      if (w_ptr[l][l][lev]->w) nnfree(w_ptr[l][l][lev]->w);
      nnfree(w_ptr[l][l][lev]);
      w_ptr[l][l][lev]=NULL;
      }
    }
}

/* generates a new unit layer l */
pINT new_unit(pINT l)
{ layer *L=&L_Tab[l];
  if (L->size < 65535u)
    {
#ifdef WIN
    ClearLayerRect(l);
#endif
    L->size++;
    X(l,++L->last) = XNEW(l,L->last) = 0.0;
    resize_layer(L,L->size,L->upr);
    draw_layer(l);
    }
  return L->last;
}


/*********************************************************************/
/*                                                                   */
/*   CHANGE ACTIVATIONS, WEIGHTS, ...                                */
/*                                                                   */
/*********************************************************************/

void update(pINT from,
            pINT to,
            pINT mode,
            pINT use_weights)
{
  regpINT i,j,k,l, winner=0;
  REAL xj,net,min= MAX_REAL,max= -MAX_REAL;
  REAL* extinp;

  switch(mode) {
  case STANDARD: /* XNEW() = (1-RATE)*X() + RATE*(L.f(ALPHA * net + LAST * X())) */
      for (L_ALL(to,j))
        { net = 0.0;
        if (use_weights != ALL)
          { for(L_ALL(from,i))
              net += X(from,i) * W(from,to,i,j);
          }
        else
          { for(ALL_LAYERS(l))
              if (w_ptr[l][to][W_LEVEL] != NULL)
                {
                if (w_ptr[l][to][W_LEVEL]->typ!=IDENT)
                  for(L_ALL(l,i)) net+=X(l,i)*W(l,to,i,j);
                else
                  for(L_ALL2(l,to,i,j))
                    net+=X(l,i)*W_1(l,to,i);
                }
          }
        XNEW(to,j) = X(to,j) + RATE * (-X(to,j) + L_Tab[to].f(ALPHA * net + LAST * X(to,j)));
        }
      break;

  case IAC: /* Optimiertes IAC, XNEW()= xj + (XMAX-X())*net - DECAY*X(); */
      if (w_ptr[to][to][W_LEVEL] == NULL)
        nn_error(SOURCENAME,"update(%i,%i,IAC,%i): Units im Layer nicht verbunden",
                            from,to,use_weights);
      /* Externe Inputs fuer Layer to im voraus berechnen */
      if ((extinp=(REAL *)nncalloc(L_Tab[to].size,sizeof(REAL)))==NULL)
         nn_error(SOURCENAME,"update(%i,%i,IAC,%i): Alloc-Fehler",
                             from,to,use_weights);
      memset(extinp,0x000,sizeof(extinp)); /* Temporaerbereich initialisieren */
      if (use_weights != ALL)
        {
        if(from!=to)
           if(w_ptr[from][to][W_LEVEL]!=NULL)
             for (L_ALL(to,j)) for(L_ALL(from,i))
               extinp[j] += X(from,i)*W(from,to,i,j);
        }
      else
        for (ALL_LAYERS(l))
          if (w_ptr[l][to][W_LEVEL]!=NULL && l!=to)
            {
            if (w_ptr[l][to][W_LEVEL]->typ!=IDENT)
              for (L_ALL(to,j)) for(L_ALL(l,i))
                extinp[j] += X(l,i)*W(l,to,i,j);
            else
              for(L_ALL2(l,to,i,j))
                extinp[j]+=X(l,i)*W_1(l,to,i);
            }
      for (k=0;k<MAXIAC;k++)
        {/* IAC-Update, nurmehr Intralayer-Verbindungen neu berechnen */
        for(L_ALL(to,j))
          {
          net = 0.0;
          xj = X(to,j);
          for(L_ALL(to,i)) net += X(to,i)*W(to,to,i,j);
          net+=extinp[j];
          if (net>0.0) XNEW(to,j) = xj + (XMAX-xj)*net - DECAY*xj;
          else         XNEW(to,j) = xj + (xj-XMIN)*net - DECAY*xj;
          }
        copy_layer(to);
        }
      nnfree(extinp);
      break;

  case IAC2: /*  XNEW() = xj + (XMAX-X())*net - DECAY*X(); */
      for(L_ALL(to,j))
        {net = 0.0;
         xj = X(to,j);
         if (use_weights != ALL)
           for(L_ALL(from,i)) net += X(from,i)*W(from,to,i,j);
         else
           for (ALL_LAYERS(l))
             if (w_ptr[l][to][W_LEVEL]!=NULL)
               {
               if (w_ptr[l][to][W_LEVEL]->typ!=IDENT)
                  for(L_ALL(l,i)) net+=X(l,i)*W(l,to,i,j);
               else
                  for(L_ALL2(l,to,i,j))
                    net+=X(l,i)*W_1(l,to,i);
               }
             if (net>0.0)  XNEW(to,j) = xj + (XMAX-xj)*net - DECAY*xj;
             else          XNEW(to,j) = xj + (xj-XMIN)*net - DECAY*xj;
             }
      break;

  case KOHONEN: /* XNEW() = sqrt( SUMi( sqr( X()-W() ) ) ) */
      for(L_ALL(to,j))
        { net = 0.0;
        for(L_ALL(from,i))
          net += sqr(X(from,i) - W(from,to,i,j));
        XNEW(to,j) = sqrt(net);
        }
      break;

  case KOHONEN_WTA:
      for(L_ALL(to,j))
        { net = 0.0;
        for(L_ALL(from,i))
          net += sqr(X(from,i) - W(from,to,i,j));
        net = sqrt(net);
        if (net < min)
          { min = net;
          winner = j;
          }
        }
      for(L_ALL(to,j))
        XNEW(to,j) = (j==winner)? 1.0 : 0.0;
      break;

  case WTA: /* XNEW(to,j) = (j==winner)? 1.0 : 0.0; */
     for(L_ALL(to,j))
       { net = 0.0;
       if (use_weights != ALL)
         for(L_ALL(from,i))
           net += X(from,i)*W(from,to,i,j);
       else
         for (ALL_LAYERS(l))
           if (w_ptr[l][to][W_LEVEL]!=NULL)
             {
             if (w_ptr[l][to][W_LEVEL]->typ!=IDENT)
               for(L_ALL(l,i)) net+=X(l,i)*W(l,to,i,j);
             else
               for(L_ALL2(l,to,i,j))
                 net+=X(l,i)*W_1(l,to,i);
             }
       if (net>max) { max = net; winner = j; }
       }
     for(L_ALL(to,j))
       XNEW(to,j) = (j==winner)? 1.0: 0.0;
     break;

  case EXT: /* X() = XNEW() = actual_externpattern */
     /* copies the externals */
     if(L_Tab[to].p != NULL)
       for(L_ALL(to,i))
         X(to,i)=XNEW(to,i) = L_Tab[to].p->pat[i];
     break;

  case EXT_XNEW: /* XNEW() = actual_externpattern */
      /* copies the externals */
      if(L_Tab[to].p != NULL)
        for(L_ALL(to,i))
          XNEW(to,i) = L_Tab[to].p->pat[i];
      break;

  case ADDEXT: /* XNEW() = X() += actual_externpattern */
      /* Addiert Extern-Pattern */
      if(L_Tab[to].p != NULL)
        for(L_ALL(to,i))
          XNEW(to,i)=X(to,i)+=L_Tab[to].p->pat[i];
      break;

  case COPY_ADD_X: /* Assumes from.size==to.size !! */
      /* Addiert Aktivierungen aus Layer from zu Layer to */
      for(L_ALL2(from,to,i,j)) X(to,j)+=X(from,i);
      break;

  case COPY_ADD_XNEW: /* == L_COPY auf xnew Assumes from.size==to.size !! */
      /* Kopiert & addiert Aktivierungen aus Layer from auf Layer to (nur auf xnew!) */
      for(L_ALL2(from,to,i,j)) XNEW(to,j)+=X(from,i);
      break;

  case COPY_ADD_X_W: /* Weighted-Copy - wie COPY_ADD_X, aber mit Gewichtung (fuer IDENT-Verbindungen) */
      for(L_ALL2(from,to,i,j)) X(to,j)+=X(from,i)*W_1(from,to,i);
      break;

  case COPY_X_W: /* Weighted-Copy - wie COPY_X, aber mit Gewichtung (fuer IDENT-Verbindungen) */
      for(L_ALL2(from,to,i,j)) X(to,j)=X(from,i)*W_1(from,to,i);
      break;

  case COPY_X: /* Assumes from.size==to.size !! */
      /* Kopiert Aktivierungen aus Layer from auf Layer to (nur auf x!) */
      for(L_ALL2(from,to,i,j)) X(to,j)=X(from,i);
      break;

  case COPY_XNEW: /* Assumes from.size==to.size !! */
      /* Kopiert Aktivierungen aus Layer from auf Layer to (nur auf xnew!) */
      for(L_ALL2(from,to,i,j)) XNEW(to,j)=X(from,i);
      break;

  case SIMPLE: /* XNEW() = net */
      for(L_ALL(to,j))
        { net = 0.0;
        if (use_weights != ALL)
          for(L_ALL(from,i)) net+=X(from,i)*W(from,to,i,j);
        else
          for (ALL_LAYERS(l))
            if (w_ptr[l][to][W_LEVEL]!=NULL)
              {
              if (w_ptr[l][to][W_LEVEL]->typ!=IDENT)
                for(L_ALL(l,i)) net+=X(l,i)*W(l,to,i,j);
              else
                for(L_ALL2(l,to,i,j))
                  net+=X(l,i)*W_1(l,to,i);
              }
        XNEW(to,j) = net;
        }
     break;

  case SIMPLE2: /* XNEW() = X() + net */
      for(L_ALL(to,j))
        {net = 0.0;
        if (use_weights != ALL)
          for(L_ALL(from,i))  net+=X(from,i)*W(from,to,i,j);
        else
          for (ALL_LAYERS(l))
            if (w_ptr[l][to][W_LEVEL]!=NULL)
              {
              if (w_ptr[l][to][W_LEVEL]->typ!=IDENT)
                for(L_ALL(l,i)) net+=X(l,i)*W(l,to,i,j);
              else
                for(L_ALL2(l,to,i,j))
                  net+=X(l,i)*W_1(l,to,i);
              }
        XNEW(to,j) = X(to,j)+net;
        }
     break;

  case BACKWARD: /* Example: update(OUT,HIDDEN,BACKWARD,..) */
      for(L_ALL(to,j))
        {net=0.0;
        if (use_weights != ALL)
          for(L_ALL(from,i))
            net+=W(to,from,j,i)*L_Tab[from].delta[i];
        else
          for (ALL_LAYERS(l))
            if (w_ptr[to][l][W_LEVEL]!=NULL)
              {
              if (w_ptr[l][to][W_LEVEL]->typ!=IDENT)
                for(L_ALL(l,i))
                  net+=W(to,l,j,i)*L_Tab[l].delta[i];
              else
                for(L_ALL2(l,to,i,j))
                  net+=W_1(to,l,i)*L_Tab[l].delta[i];;
              }
        L_Tab[to].delta[j] = X(to,j)*(1.0-X(to,j))*net;
        }
    break;

  default: nn_alarm(SOURCENAME,"update(%i,%i,IAC,%i): upd-mode unbekannt",
                               from,to,use_weights);
  }
}

/*********************************************************************/
/*                                                                   */
/*   Standard-Lernfunktionen                                         */
/*                                                                   */
/*********************************************************************/
/* sets target values of backprop layer from extern pattern
  (*target=NULL) or directly */
void NNSetTarget(pINT l1, REAL target[])
{ pINT i;
  layer *L = &L_Tab[l1];
  if (!L->target)
     nn_fatalerror(SOURCENAME,"NNSetTarget(%d,..): Target-Array not defined, please use init_layer(..,DELTA_LAYER,..)!",l1);
  if (target)
    for (L_ALL(l1,i))
       L->target[i] = target[i];
  else
    if (L->p && L->p->pat)
       for (L_ALL(l1,i))
           L->target[i] = L->p->pat[i];
    else
       nn_fatalerror(SOURCENAME,"NNSetTarget(%d,NULL): No extern patterns available!",l1);
}

/* calculates deltavalues for backwardlayer */
void NNSetDelta(pINT l1, pINT mode)
{ pINT i;
  REAL *delta  = L_Tab[l1].delta;
  REAL *target = L_Tab[l1].target;
  if (!delta)
     nn_fatalerror(SOURCENAME,"NNSetDelta(%d,..): Delta-Array not defined, please use init_layer(..,DELTA_LAYER,..)!",l1);
  if (!target)
     nn_fatalerror(SOURCENAME,"NNSetDelta(%d,..): Target-Array not defined, please use init_layer(..,DELTA_LAYER,..)!",l1);
  switch (mode) {
    case DELTA:
         for(L_ALL(l1,i))
           delta[i] = target[i]-X(l1,i);
         break;
    case BACKPROP:
         for(L_ALL(l1,i))
           delta[i] = L_Tab[l1].df(X(l1,i))*(target[i]-X(l1,i));
         break;
    }
}
   
/* standard-learning functions */
void NNLearn(pINT l1, pINT l2, pINT typ, REAL rate, pINT use_weights)
{
  pINT i,j,l;
  REAL *delta = L_Tab[l2].delta;

  switch (typ) {

  case HEBB_LEARN: /* simple Hebb learning */
    if (use_weights != ALL)
      for (L_ALL(l1,i)) for (L_ALL(l2,j))
        W(l1,l2,i,j)+=rate*X(l1,i)*X(l2,j);
    else
      for (ALL_LAYERS(l))
        if (w_ptr[l][l2][W_LEVEL]!=NULL)
           for(L_ALL(l,i)) for (L_ALL(l2,j))
              W(l,l2,i,j)+=rate*X(l,i)*X(l2,j);
    break;

  case DELTA_LEARN: /* Delta+Backprop learning, assumes L.delta[] has already been set */
  case BACKPROP_LEARN:
    if (!delta)
       nn_fatalerror(SOURCENAME,"NNLearn(%d,%d,BACKPROP_LEARN,..): Delta-Array not defined, please use init_layer(..,DELTA_LAYER,..)!",l1,l2);
    /* multiply learnrate only once! */  
    for (L_ALL(l2,j)) 
        delta[j]*=rate;
    if (use_weights != ALL)
      for (L_ALL(l1,i)) for (L_ALL(l2,j))
        W(l1,l2,i,j)+=X(l1,i)*delta[j];
    else
      for (ALL_LAYERS(l))
        if (w_ptr[l][l2][W_LEVEL]!=NULL)
          for(L_ALL(l,i)) for (L_ALL(l2,j))
            W(l,l2,i,j)+=X(l1,i)*delta[j];
    break;
  }
}

/*********************************************************************/

void copy(void) 
{ regpINT i,l; for(ALL_LAYERS(l)) for(L_ALL(l,i)) X(l,i) = XNEW(l,i); }

void copy_layer(pINT l) 
{ regpINT i; for(L_ALL(l,i)) X(l,i) = XNEW(l,i); }

/* Aufsummieren von Aktivierungen (nach Updates auf verschiedenen Levels) */
void sum_layer(pINT l)
{ regpINT i; for(L_ALL(l,i)) X(l,i) += XNEW(l,i); }

void mult_layer(pINT l)
{ regpINT i; for(L_ALL(l,i)) X(l,i) *= XNEW(l,i); }



/*********************************************************************/
void reset_x(pINT l)
{ regpINT i; for(L_ALL(l,i)) X(l,i)=(L_Tab[l].typ & BIAS_LAYER)?1.0:0.0; }

void reset_xnew(pINT l) 
{ regpINT i; for(L_ALL(l,i)) XNEW(l,i)=(L_Tab[l].typ & BIAS_LAYER)?1.0:0.0; }

void reset(pINT l) 
{ regpINT i; for(L_ALL(l,i)) X(l,i)=XNEW(l,i)=(L_Tab[l].typ & BIAS_LAYER)?1.0:0.0; }

void reset_w(pINT l1, pINT l2)
{ regpINT i,j; for(L_ALL(l1,i)) for(L_ALL(l2,j)) W(l1,l2,i,j)=0.0; }
 
void reset_layers(void) 
{ regpINT l;
        for (l=0;l<L_anz;l++) reset(l);
}

/*********************************************************************/


#define border(val,max,tol)  ((max) + (tol) - (max)*(tol)/(val))

void border_xnew_e(pINT l, 
                   REAL xmin, 
                   REAL xmax,
                   REAL xmin_tol, 
                   REAL xmax_tol)
{
  regpINT i; 
  REAL val;
  for(L_ALL(l,i))
     { val = XNEW(l,i);
       if (val>xmax) XNEW(l,i) = border(val, xmax, xmax_tol); else
       if (val<xmin) XNEW(l,i) = border(val, xmin, xmin_tol);
     }
}

void border_x_e(pINT l, 
                REAL xmin, 
                REAL xmax,
                REAL xmin_tol, 
                REAL xmax_tol) 
{
  regpINT i; 
  REAL val;
  for(L_ALL(l,i))
     { val = X(l,i);
       if (val>xmax) X(l,i) = border(val, xmax, xmax_tol); else
       if (val<xmin) X(l,i) = border(val, xmin, xmin_tol);
     }
}

void border_w_e(pINT l1, 
                pINT l2, 
                REAL min, 
                REAL max,
                REAL min_tol, 
                REAL max_tol) 
{
  int i, j;
  REAL wght;
  for(L_ALL(l1,i)) for(L_ALL(l2,j))
     { wght = W(l1,l2,i,j);
       if (wght>max) W(l1,l2,i,j) = border(wght,max,max_tol); else 
       if (wght<min) W(l1,l2,i,j) = border(wght,min,min_tol);
     }
}

/*********************************************************************/


pINT best_n_e(pINT l, pINT n, REAL xmin, REAL xmax) 
{
  pINT i,k,best,count=0; 
  REAL max;
  for(k=0;k<n;k++)
     { max = -MAX_REAL;
       for(L_ALL(l,i)) if(X(l,i)>max) { max=X(l,i); best=i; }
       X(l,best) = -MAX_REAL;
     }
  for(L_ALL(l,i)) 
      if (X(l,i) <= -MAX_REAL) { X(l,i)=xmin; count++; } else X(l,i)=xmax;
  return count;
}


pINT thresh_e(pINT l, REAL min, REAL thresh, REAL max) 
{
  regpINT i,count=0;
  for(L_ALL(l,i)) if (X(l,i)>=thresh) { X(l,i)=max; count++; } else X(l,i)=min;
  return count;
}

pINT thresh_ask(pINT l, REAL val)
{
  regpINT i,count=0;
  for(L_ALL(l,i)) if(X(l,i)>=val) count++;
  return count;
}


void round_x(pINT l, pINT digits)
{ regpINT i; for(L_ALL(l,i)) X(l,i)=roundn(X(l,i),digits); }

void round_xnew(pINT l, pINT digits)
{ regpINT i; for(L_ALL(l,i)) XNEW(l,i)=roundn(XNEW(l,i),digits); }

void round_w(pINT l1, pINT l2, pINT digits)
{ regpINT i,j; 
  for(L_ALL(l1,i)) for(L_ALL(l2,j))
     W(l1,l2,i,j) = roundn(W(l1,l2,i,j),digits);
}



/*********************************************************************/
/*      Normalisierungen        */

/* the following functions are realized as macros in neurons.h:
void normalize_x(pINT l1);
void normalize_xnew(pINT l1);
void normalize_w(pINT l1, pINT l2);
*/

void normalize_x_e(pINT l, REAL minabs) /* minabs changes 'sensitivity' */
{
  regpINT i;
  REAL maxabs=0.0;
        if (minabs<0.0) nn_error(SOURCENAME,"normalize_x_e(): ARG2 mustbe >= 0.0");
  for(L_ALL(l,i)) if (fabs(X(l,i))>maxabs) maxabs=fabs(X(l,i));
  if (maxabs>minabs) for(L_ALL(l,i)) X(l,i)/=maxabs;
} 

void normalize_xnew_e(pINT l, REAL minabs)
{ 
  regpINT i;
  REAL maxabs=0.0;
        if (minabs<0.0) nn_error(SOURCENAME,"normalize_xnew_e(): ARG2 mustbe >= 0.0");
  for(L_ALL(l,i)) if (fabs(XNEW(l,i))>maxabs) maxabs=fabs(XNEW(l,i));
  if (maxabs>minabs) for(L_ALL(l,i)) XNEW(l,i)/=maxabs;
}

void normalize_w_e(pINT l1, pINT l2, REAL minabs)
{ 
  regpINT i,j;
  REAL maxabs=0.0;
        if (minabs<0.0) nn_error(SOURCENAME,"normalize_w_e(): ARG3 mustbe >= 0.0");
  for(L_ALL(l2,j)) for(L_ALL(l1,i)) 
     if (fabs(W(l1,l2,i,j))>maxabs) maxabs=fabs(W(l1,l2,i,j));
  if (maxabs>minabs) 
     for(L_ALL(l1,i)) for(L_ALL(l1,i)) W(l1,l2,i,j)/=maxabs;
}


/*********************************************************************/

/* the following functions are realized as Makros in neurons.h: 
REAL energy_x(pINT l1, REAL energy);
REAL energy_xnew(pINT l1, REAL energy);
REAL energy_w(pINT l1, pINT l2, REAL energy, pINT mode);
*/

REAL energy_x_e(pINT l, REAL en_thres, REAL energy)
{ 
  regpINT i;
  REAL en=0.0;
  for(L_ALL(l,i)) en+=fabs(X(l,i));
  if (en > en_thres)
     { en/=energy;
       for(L_ALL(l,i)) X(l,i)/=en;
     }
  return en; 
}   

REAL energy_xnew_e(pINT l, REAL en_thres, REAL energy)
{ 
  regpINT i;
  REAL en=0.0;
  for(L_ALL(l,i)) en+=fabs(X(l,i));
  if (en > en_thres)
     { en/=energy;
       for(L_ALL(l,i)) XNEW(l,i)/=en;
     }
  return en; 
}

REAL energy_w_e(pINT l1, pINT l2, REAL en_thres, REAL energy, pINT mode)
{
  regpINT i,j;
  REAL en;
  if (mode==ALL)
        {
    en=0.0;
    for(L_ALL(l1,i)) for(L_ALL(l2,j)) en+=fabs(W(l1,l2,i,j));
    if (en > en_thres)
                        { en/=energy;
                for(L_ALL(l1,i)) for(L_ALL(l2,j)) W(l1,l2,i,j)/=en;
                        } 
    }
  else
    {
    for(L_ALL(l2,j))
                        { en=0.0;
                for(L_ALL(l1,i)) en+=fabs(W(l1,l2,i,j));
                if (en > en_thres)
              { en/=energy;
                                for(L_ALL(l1,i)) W(l1,l2,i,j)/=en;
              }    
                        }
    } 
  return en;
}

/*********************************************************************/

void toggle_unit(pINT l, pINT unit) 
{
  layer *L = &L_Tab[l];
  X(l,unit) = (X(l,unit)<0.5)? 1.0 : 0.0;
  g_acquire();
  draw_unit(X(l,unit),PX(l,unit),PY(l,unit),L->ux,L->uy);
  g_release();
}

/*********************************************************************/
/*                                                                   */
/*   GET LAYER/UNIT-INFORMATION:                                     */
/*                                                                   */
/*********************************************************************/

void max_ux(pINT l, pINT *u, REAL *max) 
{
  regpINT i;
  *u= L_Tab[l].first;
  *max= -MAX_REAL;    
  for(L_ALL(l,i)) if (X(l,i)>*max) { *u=i; *max=X(l,i); }
}

REAL max_x(pINT l) 
{
  regpINT i;
  REAL max= -MAX_REAL;    
  for(L_ALL(l,i)) if (X(l,i)>max) max=X(l,i);
  return max;
}

pINT max_u(pINT l)
{
  pINT i,u=L_Tab[l].first;
  REAL max= -MAX_REAL;
  for(L_ALL(l,i))
    if (X(l,i)>max)
      {u=i;
      max=X(l,i); }
  return u;
}

REAL maxabs_x(pINT l) 
{
  regpINT i;
  REAL max=0.0;    
  for(L_ALL(l,i)) if (fabs(X(l,i))>max) max=fabs(X(l,i));
  return max;
}

pINT maxabs_u(pINT l) 
{
  regpINT i,u=L_Tab[l].first;
  REAL max=0.0;    
  for(L_ALL(l,i)) if (fabs(X(l,i))>max) { u=i; max=fabs(X(l,i)); }
  return u;
}

pINT wta(pINT l)
{ int i,j;
        i=max_u(l);
        for (L_ALL(l,j))
                X(l,j) = (i==j)?1.0:0.0;
        return i;
}


/*********************************************************************/
/*                                                                   */
/*   MISC. CONN:                                                     */
/*                                                                   */
/*********************************************************************/

/* globale Aktivierungsfunktion (default) */
/* standard logistic, returns 0.0 for x<-227.0 */
REAL f(REAL x)
{ return (REAL)((x+OFFSET< -227.0)? 0.0: 1.0 / (1.0 + exp((double)(-GAIN*(x + OFFSET))))); }

/* Ableitung der Aktivierungsfunktion */
REAL df(REAL x)
{  return (REAL)(GAIN*x*(1.0-x)); }

/* Setzen der Indextabelle fuer INVARI2-Netz, sodass alle Abstaende
 * deltax,deltay zwischen zwei Punkten im Vision-Layer, die den gleichen
 * Anstieg deltay/deltax ergeben, auf den gleichen Gewichtsindex gesetzt
 * abgebildet werden.
 */
pINT set_w_index(pINT l1) 
{
  int x,y,i,imax= -1;
  layer L;
  REAL *w,q;
  L=L_Tab[l1];
        if (w_ind!=NULL) nn_error(SOURCENAME,"set_w_index(%i): Nur ein INVARI2-Netz zulaessig",l1);
  if ((w_ind = (pINT *)nncalloc(L.size*2,sizeof(pINT))) == NULL)
                 nn_error(SOURCENAME,"set_w_index(%i): Alloc w_ind",l1);
  if ((w=(REAL *)nncalloc(L.size,sizeof(REAL))) == NULL)
                 nn_error(SOURCENAME,"set_w_index(%i): Alloc w[]",l1);
  for(x=1;x<L.upr;x++)                  /* Alle delta_x, delta_y mit neg.Steigung */
      for(y=0;y<L.size/L.upr;y++)  /* inkl. horizontal, excl. vertikal */
         {
         q=(REAL)y/x;
         i=0;
         while (i<=imax && w[i]!=q) i++; /*Suche nach Index mit gleicher Steigung*/
         if (i>imax) { imax=i; w[i]=q; }
         W_INDEX(l1,x,y,1)=i; /* fuer negative Steigungen */
         }
  for(x=0;x<L.upr;x++)
     for(y=1;y<L.size/L.upr;y++)     /* fuer positive Steigungen */
         W_INDEX(l1,x,y,0)=W_INDEX(l1,x,y,1)+imax+1; /* und vertikal */

/* 0..281 bei 16x16
  for (i=0;i<2;i++)
      {for (y=0;y<L.upr;y++)
         {
         for(x=0;x<L.size/L.upr;x++)
                        pprintf_e(2, "%3i ", W_INDEX(l1,x,y,i));
                 pprintf_e(2, "\n");
                 }
          pprintf_e(2, "\n");
          }
*/
  nnfree(w);
  return 2*imax+2;
}

/* Berechnung der Hiddenaktivierungen im Netzwerk 2.Ordnung Typ 1 */
void transform1(pINT l1, pINT l2)
{
  regpINT i,j;
  LOOP_2(l1,i,j)
    if (X(l1,i)!=0.0 && X(l1,j)!=0.0)
      X(l2,INDEX_I1(l1,i,j))+=X(l1,i)*X(l1,j);
}

/* Berechnung der Hiddenaktivierungen im Netzwerk 2.Ordnung Typ 2 */
void transform2(pINT l1, pINT l2)
{ regpINT i,j,ind;
  LOOP_2(l1,i,j)
    if (X(l1,i)!=0.0 && X(l1,j)!=0.0)
      {ind=W_INDEX(l1,DELTA_X(l1,i,j),DELTA_Y(l1,i,j),ANSTIEG(l1,i,j));
      X(l2,ind)+=X(l1,i)*X(l1,j);}
}




/*********************************************************************/
/*                                                                   */
/*   DISPLAY LAYER/UNIT-INFORMATION:                                 */
/*                                                                   */
/*********************************************************************/

/* Layer mit Titel anzeigen */
void draw_layer(pINT l) 
{ regpINT i;
  layer *L;
  L = &L_Tab[l];
  g_acquire();
  cursor_off();
  if (L->status==L_NEU) { draw_rahmen(l,0); L->status = L_AKTIV; } 
#ifdef WIN
   x_draw_layer(l,1.0); /* beschleunigte Ausgabe */
#else
  for(L_ALL(l,i))
     draw_unit(X(l,i),PX(l,i),PY(l,i),L->ux,L->uy);
#endif
  cursor_on();
  g_release();
} 


void draw_layer_n(pINT l) 
{
  regpINT i;  
  REAL maxabs;  
  layer *L;
  L = &L_Tab[l];
  g_acquire();
  cursor_off();
  if (L->status==L_NEU) { draw_rahmen(l,0); L->status = L_AKTIV; } 
  maxabs=maxabs_x(l);
  if (maxabs<1.0) maxabs=1.0;
#ifdef WIN
  x_draw_layer(l,maxabs); /* beschleunigte Ausgabe */
#else
  for(L_ALL(l,i))
     draw_unit(X(l,i)/maxabs,PX(l,i),PY(l,i),L->ux,L->uy);
#endif
  cursor_on();
  g_release();
}

/* Extern-Layer mit Mustername anzeigen */
void draw_ext(pINT l1)
{ layer *L;
  register int i;
  L = &L_Tab[l1];
  if(L->p != NULL)
    { g_acquire();
      cursor_off();
      draw_rahmen(l1,1);
      for(L_ALL(l1,i))
        draw_unit(L->p->pat[i],
                  PX(l1,i) + L->extsep_x,
                  PY(l1,i) + L->extsep_y,
                  L->ux,L->uy);
      show_pname(l1,L->p->name,EXT_PNAME);
      cursor_on();
      g_release();
    }
}

void draw_layers(void) 
{ regpINT i;
  for (i=0;i<L_anz;i++) { draw_layer(i); draw_ext(i); }
}

/* Ermittelt Muster mit kleinstem quadratischen Fehler zu gegebener
 * Unit-Aktivierung (= aehnlichstes Muster) und zeigt Musternamen an
 */
pINT interp_pat(pINT l1, REAL maxabw, pINT noname) 
{
  layer L;
  regpINT i,j,patn;
  pattern *pp;
  REAL sum,bestabw;
  char pname[PNAME_SIZE+1];

  bestabw=maxabw;
  L=L_Tab[l1];
  pp = L.ext;
  sprintf(pname,"%s",pp->name);
  for (patn=0; pp!=NULL; pp=pp->next, patn++)
      if (noname || strlen(pp->name))
         { for(sum=0.0,i=0,j=0; i<=L.last; i++,j++)
              sum += sqr(X(l1,i) - pp->pat[j]);
           if (sum<bestabw)
              { bestabw=sum; sprintf(pname,"%s",pp->name); }
         }
  if (bestabw<maxabw)
     { show_pname(l1,pname,INT_PNAME); return patn; }
  else
     return ((pINT)BAD_PATTERN);
}

/* Mustername anzeigen, typ=EXT_PNAME/INT_PNAME
 */
void show_pname(pINT l1, char *pname, pINT typ)
{
  layer *L=&L_Tab[l1];
  int pn_x,pn_y;

  if (typ==EXT_PNAME)
     { pn_x = PX(l1,0)+L->extsep_x;
       pn_y = PY(l1,L->last)+L->extsep_y+L->dy+2;
     }
  else
     { pn_x = PX(l1,0);
       pn_y = PY(l1,L->last)+L->dy+4;
     }
  gprintf(pn_x,pn_y,COL_TEXT,"%-*s",PNAME_SIZE,pname);
}


/*      Testet, ob ALLE Gewichte im Netzwerk im Bereich [min,max] liegen */

pINT alarm_all_w_e(REAL min, REAL max, pINT gof)
{
        pINT lev,l1,l2,alarm=0;
        regpINT i,j;
        REAL val;
        lev=W_LEVEL;
        pprintf("looking for weights outside [%g,..,%g] ...\n",min,max);
        for(ALL_LEVELS(W_LEVEL))
                for(ALL_LAYERS(l1))
                        for(ALL_LAYERS(l2))
                                if (w_ptr[l1][l2][W_LEVEL] != NULL)
                                        for(L_ALL(l1,i))
                                                for(L_ALL(l2,j))
                                                        {
                                                        val = W(l1,l2,i,j);
                                                        if (val>max)
                                                                 { pprintf("level %i: W(%i,%i,%i,%i) = %g\n",W_LEVEL,l1,l2,i,j,val);
                                                                 alarm++;
                                                                 }
                                                        else
                                                                if (val<min)
                                                                 { pprintf("level %i: W(%i,%i,%i,%i) = %g\n",W_LEVEL,l1,l2,i,j,val);
                                                                 alarm++;
                                                                 }
                                                        }
        pprintf("%i bad weight(s) found.\n",alarm);
        W_LEVEL=lev;
        return alarm;
}

pINT alarm_all_x_e(REAL min, REAL max, pINT gof)
{
  pINT l1,alarm=0;
  regpINT i; 
  REAL val;
  pprintf("looking for activations outside [%g,..,%g] ...\n",min,max);
  for(ALL_LAYERS(l1)) for(L_ALL(l1,i))
     { val=X(l1,i);
       if (val>max) { pprintf("layer %i: x[%i]=%g\n",l1,i,val);
                      alarm++; 
                    } else
       if (val<min) { pprintf("layer %i: x[%i]=%g\n",l1,i,val);
                      alarm++; 
                    }
     }
  pprintf("%i bad activation(s) found.\n",alarm);
  return alarm;
}



/* Anzeige aller Gewichte von allen Layern zu einer bestimmten Unit
   auf der aktuellen Weight-Ebene (W_LEVEL)
   -- und umgekehrt.
*/

int DRAWW_X = DRAWW_X_DEF;
int DRAWW_Y = DRAWW_Y_DEF;

void draw_minmax_weights(char*layerstr, char *direction, REAL min, REAL max, REAL scale_min, REAL scale_max)
{ int i;
  sprintf(info_msg[0],layerstr);
  sprintf(info_msg[1],direction);
  if(ScaleWeightDisplay==0.0)
    if (scale_min==scale_max)
       sprintf(info_msg[2],"display autoscaled by %.2f   ",scale_min);
    else
       sprintf(info_msg[2],"display autoscaled by %.2f - %.2f",scale_min,scale_max);
  else
    sprintf(info_msg[2],"display scaled by %.1f",scale_min);

  if (max>=min)
    {sprintf(info_msg[3],"wmin: %9.4f",min);
    sprintf(info_msg[4],"wmax: %9.4f",max);}
  else
    {sprintf(info_msg[3],"no connections");
    sprintf(info_msg[4],"--------------");
    sprintf(info_msg[2],"");}

#if !defined (sun4) && !defined(WIN)
  /* use messagewindow on DOS and Apollo (requires no additional screenspace) */
  pprintf("\n");
  for (i=0;i<5;i++)
    if (i!=2) pprintf("%s\n",info_msg[i]);
#else
  x_draw_minmax_weights();
  if (scale_max!=0.0)
    draw_scaletext(-1.0/scale_max,1.0/scale_max,"Weights:");
#endif
}

void draw_weights_to(pINT l2, pINT unit, REAL scale)
{ /* bei scale==0.0 erfolgt eine normalisierte Anzeige */
  regpINT j,others=0;
  pINT l1;
  layer L;
  REAL w,sc_min=MAX_REAL,sc_max=-MAX_REAL,sc=scale;
  REAL wmin=MAX_REAL, wmax=-MAX_REAL,min=MAX_REAL,max=-MAX_REAL;
  char str1[35],str2[35];
  g_acquire();
  cursor_off();
  for (ALL_LAYERS(l1))
    { L=L_Tab[l1];
    if (w_ptr[l1][l2][W_LEVEL]==NULL)
      /* Units ohne Verbindung loeschen */
      for (L_ALL(l1,j))
        draw_unit(0.0,PX(l1,j),PY(l1,j),L.ux,L.uy);
    else
      { /* Gewichte der Units mit Verbindung anzeigen */
      switch (w_ptr[l1][l2][W_LEVEL]->typ) {
        case IDENT:
             for(L_ALL(l1,j))
               if (j==unit)
                  {
                  w=W_1(l1,l2,unit);
                  others++;
                  if (w>max) max=w;
                  if (w<min) min=w;
                  if (scale!=0.0)
                    draw_unit(sc*w,PX(l1,j),PY(l1,j),L.ux,L.uy);
                  }
               else draw_unit(0.0,PX(l1,j),PY(l1,j),L.ux,L.uy);

               if (scale==0.0) /* bei autoscale erst jetzt anzeigen */
                 {
                 if (fabs(min)>fabs(max))
                   {if (min!=0.0) sc=1.0/fabs(min);}
                 else
                   {if (max!=0.0) sc=1.0/fabs(max); }

                 for(L_ALL(l1,j))
                   if (j==unit)
                     draw_unit(sc*W_1(l1,l2,j),PX(l1,j),PY(l1,j),L.ux,L.uy);
                 }
             break;
        default:
             for(L_ALL(l1,j))
               { w=W(l1,l2,j,unit);
               others++;
               if (w>max) max=w;
               if (w<min) min=w;
                 if (scale!=0.0)
                   draw_unit(sc*w,PX(l1,j),PY(l1,j),L.ux,L.uy);
               }
             if (scale==0.0)
               {
               if (fabs(min)>fabs(max))
                 {if (min!=0.0) sc=1.0/fabs(min);}
               else
                 {if (max!=0.0) sc=1.0/fabs(max); }
               for(L_ALL(l1,j))
                 draw_unit(sc*W(l1,l2,j,unit),PX(l1,j),PY(l1,j),L.ux,L.uy);
               }
             break;
        }
        if (sc<sc_min) sc_min=sc;
        if (sc>sc_max) sc_max=sc;
        if (min<wmin) wmin=min;
        if (max>wmax) wmax=max;
      }
    }
  if (others>0)
     {
     sprintf(str1,"Weights to layer %u:",l2);
     sprintf(str2,"%d others -> unit #%d",others,unit);
     }
  else
     {
     sprintf(str1,"Weights to layer %u:",l2);
     sprintf(str2,"no others -> unit #%d",unit);
     }
  draw_minmax_weights(str1,str2,wmin,wmax,sc_min,sc_max);
  cursor_on();
  g_release();
}

void draw_weights_from(pINT l1, pINT unit, REAL scale)
{
  /* bei scale==0.0 erfolgt eine normalisierte Anzeige */
  regpINT j,others=0;
  pINT l2;
  layer L;
  REAL w,sc_min=MAX_REAL,sc_max=-MAX_REAL,sc=scale;
  REAL wmin=MAX_REAL, wmax=-MAX_REAL,min=MAX_REAL,max=-MAX_REAL;
  char str1[35],str2[35];
  g_acquire();
  cursor_off();
  for (ALL_LAYERS(l2))
    { L=L_Tab[l2];
    if (w_ptr[l1][l2][W_LEVEL]==NULL)
      /* Units ohne Verbindung loeschen */
      for (L_ALL(l2,j)) draw_unit(0.0,PX(l2,j),PY(l2,j),L.ux,L.uy);
    else
      { /* Gewichte der Units mit Verbindung anzeigen */
      switch (w_ptr[l1][l2][W_LEVEL]->typ) {
        case ORDER2: break; /* not yet implemented */
        case INVARI2: break;
        case IDENT:
             for(L_ALL(l2,j))
               if (j==unit)
                 {w=W_1(l1,l2,unit);
                 others++;
                 if (w>max) max=w;
                 if (w<min) min=w;
                 if (scale!=0.0)
                    draw_unit(sc*w,PX(l2,j),PY(l2,j),L.ux,L.uy);
                 }
               else
                 draw_unit(0.0,PX(l2,j),PY(l2,j),L.ux,L.uy);
               if (scale==0.0) /* bei autoscale erst jetzt anzeigen */
                 {
                 if (fabs(min)>fabs(max))
                   {if (min!=0.0) sc=1.0/fabs(min);}
                 else
                   {if (max!=0.0) sc=1.0/fabs(max); }
                 for(L_ALL(l2,j))
                   if (j==unit)
                     draw_unit(sc*W_1(l1,l2,j),PX(l2,j),PY(l2,j),L.ux,L.uy);
                 }
             break;
        default:
             for(L_ALL(l2,j))
               { w=W(l1,l2,unit,j);
               others++;
               if (w>max) max=w;
               if (w<min) min=w;
               if (scale!=0.0)
                 draw_unit(sc*w,PX(l2,j),PY(l2,j),L.ux,L.uy);
               }
             if (scale==0.0)
               {
               if (fabs(min)>fabs(max))
                 {if (min!=0.0) sc=1.0/fabs(min);}
               else
                 {if (max!=0.0) sc=1.0/fabs(max); }
               for(L_ALL(l2,j))
                 draw_unit(sc*W(l1,l2,unit,j),PX(l2,j),PY(l2,j),L.ux,L.uy);
               }
             break;
        }
      if (sc<sc_min) sc_min=sc;
      if (sc>sc_max) sc_max=sc;
      if (min<wmin) wmin=min;
      if (max>wmax) wmax=max;
      }
    }

  if (others>0)
     {
     sprintf(str1,"Weights from layer %u:",l1);
     sprintf(str2,"unit #%d -> %d others",unit,others);
     }
  else
     {
     sprintf(str1,"Weights from layer %u:",l1);
     sprintf(str2,"unit #%d -> no others",unit);
     }

  draw_minmax_weights(str1,str2,wmin,wmax,sc_min,sc_max);
  cursor_on();
  g_release();
}


/* Mausfunktion bei Unit-Click aufrufen
 */
void unit_semantics(pINT l, pINT unit)
{
  cursor_off();
  switch (mouse_button) {
     case 1: /* left button */
       switch(mouse_function[0])
         {
         case M_TOGGLE: toggle_unit(l,unit); break;
         case M_WEIGHT: draw_weights_to(l,unit,ScaleWeightDisplay); break;
         }
       break;
     case 2: /* right button */
       switch(mouse_function[1])
         {
         case M_TOGGLE: toggle_unit(l,unit); break;
         case M_WEIGHT: draw_weights_from(l,unit,ScaleWeightDisplay); break;
         }
       break;
     case 3: /* middle button - of course this does not work in the PC version */
       switch(mouse_function[2])
         {
         case M_TOGGLE: toggle_unit(l,unit); break;
         case M_WEIGHT: draw_weights_from(l,unit,ScaleWeightDisplay); break;
         }
       break;
     }
  cursor_on();
}


/**********************************************************************/
/*                                                                    */
/*        PRINT INFORMATION:                                          */
/*                                                                    */
/**********************************************************************/

void print_act (pINT l)
{
  pINT i;
  pprintf("Activations of layer %u (%s):\n",l,L_Tab[l].titel);
  for ( L_ALL(l,i) )
      {
      pprintf("%6.2f ", X(l,i));
      if (i%L_Tab[l].upr==0) pprintf("\n");
      }
  pprintf("\n");
}

void print_weights(pINT l1,pINT l2)
{
  pINT i,j;
  pprintf("Weights from layer %u->%u (%s->%s):\n",l1,l2,L_Tab[l1].titel,L_Tab[l2].titel);
  if (w_ptr[l1][l2][W_LEVEL] && w_ptr[l1][l2][W_LEVEL]->w)
    for ( L_ALL(l1,i) )
      {
      pprintf("from Unit %u: ",i);
      if (w_ptr[l1][l2][W_LEVEL]->typ != IDENT)
            for (L_ALL(l2,j) )
              pprintf("%6.2f ", W(l1,l2,i,j));
      else
        pprintf("%6.2f ", W_1(l1,l2,i));
      pprintf("\n");
      }
  pprintf("\n");
}

/*********************************************************************/
/*                                                                   */
/*              GRAPHIC-PRINT: the scrolling area                    */
/*                                                                   */
/*********************************************************************/

/*
int GP_LINES = GP_LINES_DEF;
int GP_COLLS = GP_COLLS_DEF;
int GP_X = GP_X_DEF;
int GP_Y = GP_Y_DEF;
*/

char GP_buf[GP_LINES_DEF+1][GP_COLLS_DEF+1];
int GP_free=0;
int GP_bc=0;

void gp_border(void)
{
#if !defined(WIN) && !defined(sun4)
  /* Nur fuer DOS und Apollo */
  g_acquire();
  rectangle(GP_X,GP_Y,
  GP_X+GP_COLLS*GP_CHAR_SIZE+GP_BORD+2,
  GP_Y+GP_LINES*GP_LINE_SIZE+GP_BORD);
  g_release();
#endif
}

int displ (int iscroll)
{
  int i,lbc;

  for ( lbc = GP_bc; lbc < GP_COLLS; lbc++ )
    GP_buf[GP_free][lbc] = ' ';  /* buffer auffuellen */
  GP_buf[GP_free][GP_COLLS] = '\0';
  for (i=GP_LINES; i>0; i--)
    gprintf(GP_X + GP_BORD, GP_Y + GP_BORD + (i-1) * GP_LINE_SIZE,
                COL_TEXT, "%s", GP_buf[(GP_free+i) % GP_LINES]);
  if (iscroll)
    { GP_free++;
    GP_free %= GP_LINES;
    GP_bc = 0;
    }
  return iscroll;
}

int push_scroll (char *str)
{
  int lnl=0,es=0,of,nl,sc;
  for( sc=0; !es && sc < (GP_COLLS * GP_LINES * 5); )
    {
    nl = (str[sc] == '\n');
    of = (GP_bc >= GP_COLLS);
    es = (str[sc] == '\0');

    if ( nl || of )
      displ(1);
    if ( es && !lnl )
      displ(0);
    if ( !es )
      {
      GP_buf[GP_free][GP_bc++] = str[sc++];
      if ( nl ) GP_bc--;
      }
  lnl = nl || of;
  }
  return es? sc: -sc;
}

int gp_printf (char *fmt, ...)
{
   va_list args;
   char str[GP_BUFFERSIZE];
   int cnt;

   va_start(args, fmt);
   cnt = vsprintf(str,fmt,args);
   va_end(args);
#if !defined(sun4) && !defined(WIN)
   push_scroll(str);
#else
   msg_print(str);
#endif
   if (cnt==EOF) nn_error(SOURCENAME,"gp_printf(): wrong format");
   return cnt;
}


/*********************************************************************/
/*                                                                   */
/*         DEFINE BUTTONS:                                           */
/*                                                                   */
/*********************************************************************/
void make_control(float value, float limit1, float limit2)
{ buttons[buttons_used].value = value;
  buttons[buttons_used].limit1 = limit1;
  buttons[buttons_used].limit2 = limit2;
}

void make_key (char *text, void (*action)(void), char letter, char* fmt, int *var,
        REAL min, REAL max, REAL res, char val_typ, int typ)
{
  int x_offset = (typ==TWIN_2)? BUTTON_WIDTH/2+2: 0;
  buttons[buttons_used].x = button_x_pos+x_offset;
  buttons[buttons_used].y = button_y_pos;
  buttons[buttons_used].width =
         (typ==TWIN_1 || typ==TWIN_2) ? BUTTON_WIDTH/2-2: BUTTON_WIDTH;
  buttons[buttons_used].height = BUTTON_HEIGHT;
  buttons[buttons_used].call = action;
  buttons[buttons_used].key = letter;
  buttons[buttons_used].text = text;
  buttons[buttons_used].typ = typ;
  buttons[buttons_used].var = var;
  buttons[buttons_used].val_min = min;
  buttons[buttons_used].val_max = max;
  buttons[buttons_used].val_res = res;
  buttons[buttons_used].val_fmt = fmt;
  buttons[buttons_used].val_typ = val_typ;
  buttons[buttons_used].status = 0;
  show_button(buttons_used,NEU);
  if (typ!=TWIN_1) button_y_pos += buttons[buttons_used].height + BUTTON_DIST;
  buttons_used++;
}

void make_switch(char *text, void (*action)(void), char letter)
{ make_key(text,action,letter,NULL,NULL,0,0,0,0,SWITCH); }

void make_button(char *text, void (*action)(void), char letter)
{ make_key(text,action,letter,NULL,NULL,0,0,0,0,BUTTON); }

void make2button(char *text1, void(*action1)(void), char letter1,
                 char *text2, void(*action2)(void), char letter2)
{ make_key(text1,action1,letter1,NULL,NULL,0,0,0,0,TWIN_1);
  make_key(text2,action2,letter2,NULL,NULL,0,0,0,0,TWIN_2);
}

int make_meter3(char *text, float min, float max, float limit1, float limit2, float value)
{ make_control(value,limit1,limit2);
  make_key(text,NULL,' ',NULL,NULL,min,max,0,0,METER3);
  return(buttons_used-1);
}



/* Default Version */
void make_slider(char *text, void (*action)(void), char letter, long *var, long min, long max)
{  make_slider_e(text, action, letter, "%li",var,min,max,1L);}

/* Extended Version */
void make_slider_e(char *text, void (*action)(void), char letter, char* fmt, ...)
{
  va_list args;
  char typ;
  REAL min, max, res;
  int *var;
  va_start(args, fmt);
  if (strchr(fmt,'e') || strchr(fmt,'f') || strchr(fmt,'g'))
    { /* float args */
    var=(int*)va_arg(args,float*);
    min=va_arg(args,double);
    max=va_arg(args,double);
    res=va_arg(args,double);
    typ='f';
    }
  else
    if (strchr(fmt,'i') || strchr(fmt,'d') || strchr(fmt,'u') || strchr(fmt,'x'))
      if (strchr(fmt,'l'))
        {/* long args */
        var=(int*)va_arg(args,long*);
        min=(REAL)va_arg(args,long);
        max=(REAL)va_arg(args,long);
        res=(REAL)va_arg(args,long);
        typ='l';
        }
      else
        if (strchr(fmt,'h'))
          {/* short int args */
          var=(int*)va_arg(args,short int*);
          min=(REAL)va_arg(args,int);
          max=(REAL)va_arg(args,int);
          res=(REAL)va_arg(args,int);
          typ='s';
          }
        else
          {/* int args */
          var=(int*)va_arg(args,int*);
          min=(REAL)va_arg(args,int);
          max=(REAL)va_arg(args,int);
          res=(REAL)va_arg(args,int);
          typ='i';
          }
     else
       nn_error(SOURCENAME,"make_slider_e: wrong format");

    va_end(args);

#ifdef sun4
    make_key(text,action,letter,fmt,var,min,max,res,typ,SLIDER);
#endif
#ifdef WIN
    make_key(text,action,letter,fmt,var,min,max,res,typ,SLIDER);
#endif

}

/*********************************************************************/
/*                                                                   */
/*   PATTERN MANAGEMENT:                                             */
/*                                                                   */
/*********************************************************************/
pattern *make_single_pattern(char *s, char *pname)
{
  pattern *p;
  int i;
  p = (pattern *) nnmalloc(sizeof(pattern));
  if (p!=NULL)
     { p->pat = (REAL *) nncalloc(strlen(s),sizeof(REAL));
       if (p->pat != NULL)
             for(i=0;i<strlen(s);i++)
          switch(s[i]) {
          case '0': p->pat[i] = 0.0; break;
          case '1': p->pat[i] = 1.0; break;
          default: p->pat[i] = 0.0;
          }
       p->size = strlen(s);
       sprintf(p->name,"%s",pname);
       p->next = NULL;
       return (p->pat!=NULL)? p : NULL ;
     }
  else return NULL;
}


void make_pattern_e(pINT l, char *s, char *pname)
{
  pattern *p;
  layer *L = &L_Tab[l];
  if (L->ext == NULL)
    {
    L->ext = make_single_pattern(s, pname);
    if (L->ext == NULL)
      nn_alarm(SOURCENAME,"make_pattern_e(%i,%s,%s): first pattern failed",l,s,pname);
    else
      {
      if((L->p_tab = (pattern **)nnmalloc(512*sizeof(pattern*)))==NULL)
        nn_alarm(SOURCENAME,"make_pattern_e(%i,%s,%s): first pattern_tab failed",l,s,pname);
      L->p_tab[0] = L->ext;
      L->patanz=1;
      }
    }
  else
    {
    p = L->p_tab[L->patanz-1]; /* last pattern */
    p->next = make_single_pattern(s, pname);
    if (p->next == NULL)
      nn_alarm(SOURCENAME,"make_pattern_e(%i,%s,%s): nth pattern failed",l,s,pname);
    else
      {
      L->patanz++;
      if (!(L->patanz & 0x1ff)) /* extends p_tab[] by one block */
        if((L->p_tab = (pattern **)nnrealloc(L->p_tab,(L->patanz+512)*sizeof(pattern*)))==NULL)
          nn_alarm(SOURCENAME,"make_pattern_e(%i,%s,%s): nth pattern_tab failed",l,s,pname);
      L->p_tab[L->patanz-1]=p->next;
      }
    }
}

/* set pointer to first pattern and generate backward-chain */
void first_pattern(pINT l)
{
  layer *L = &L_Tab[l];
  pattern *p;
  pINT i=0;
  L->p = L->ext; /* set first */
  if ( L->ext != NULL )
    {
    L->ext->last = NULL;  /* wichtig !!!!! */
    for (p = L->ext; p && i < L->patanz; p = p->next)
      {
      if (p->next) p->next->last = p; /* set backward chain */
      }
    }
}

void prev_pattern(pINT l)
{
  layer *L= &L_Tab[l];
  if(L->p && L->p->last)  L->p = L->p->last;
  else last_pattern(l);
}

void next_pattern(pINT l)
{
  layer *L = &L_Tab[l];
  if(L->p && L->p->next) L->p = L->p->next;
  else L->p = L->ext;
}

BOOL next_pattern_exist(pINT l)
{ /* like next_pattern(), but returns 0 at end of pattern-list, 1 else */
  layer *L= &L_Tab[l];
  return (L->p && L->p->next)? (L->p = L->p->next,1):(L->p = L->ext,0);
}

BOOL pattern_ex(pINT l)
{  /* if end of pattern-list: 0 */
  return L_Tab[l].p->next != NULL;
}

void step_pattern(pINT l, int steps)
{
  pINT panz = patanz(l);
  layer *L= &L_Tab[l];
  long i; /* don't limit range! */

  if(panz==0)
    {nn_alarm(SOURCENAME,"step_pattern(%u,%i): patanz==0",l,steps);
    return;}

  while (steps>=panz) steps-=panz;

  i = L->patnr+steps;

  if (i>=panz) i-=panz;
  if (i<0)     i+=panz;

  L->patnr=(pINT)i;
  L->p = L->p_tab[(pINT)i];
}


pINT nth_pat(pINT l, pINT n)
{ /* n goes from 0 to patanz-1 */
  layer* L= &L_Tab[l];
  if (n >= L->patanz && L->patanz>0)
    {nn_alarm(SOURCENAME,"nth_pat(%u,%u): pattern number>=patanz",l,n);
     return BAD_PATTERN;}
  else
    {
    if (L->p_tab[n] == NULL)
      { nn_alarm(SOURCENAME,"nth_pat(%u,%u): pattern %i does not exist\n",l,n,n);
      return BAD_PATTERN;
      }
    L->p = L->p_tab[n];
    L->patnr=n;
    }
  return n;
}

void last_pattern(pINT l)
{ layer *L=&L_Tab[l];
  if (L->p && L->p_tab)
    L->p = L->p_tab[L->patanz-1];
}

pINT rand_pat(pINT l)
{ layer *L=&L_Tab[l];
  if (L->p && L->p_tab)
    return (nth_pat(l,(pINT)(random_nr()*(REAL)patanz(l))));
  return BAD_PATTERN; 
}

/* calculates pattern statistics for pattern p */
void calc_pattern_stat(pattern *p)
{ pINT i;
  p->min=MAX_REAL;
  p->max=-MAX_REAL;
  p->mean=0.0;
  p->var=0.0;
  p->length=0.0;
  for(i=0;i<p->size;i++)
	 {
	 p->min=min(p->min,p->pat[i]);
	 p->max=max(p->max,p->pat[i]);
	 p->mean+=p->pat[i];
	 p->length+=sqr(p->pat[i]);
	 }
  p->mean/=p->size;
  p->length=sqrt(p->length);
  for(i=0;i<p->size;i++)
	 p->var+=sqr(p->pat[i] - p->mean);
  p->var/=p->size;
}

/* calculates pattern statistics for all patterns in layer l */
void calc_pattern_stats(pINT l)
{ layer *L=&L_Tab[l];
  pINT i,j;
  pattern *p;
  for(L_ALL(l,i))
	{
	L->pmax[i]=-MAX_REAL;
	L->pmin[i]=MAX_REAL;
	L->pmean[i]=0.0;
	L->pvar[i]=0.0;
	}
  if (!L->p_tab) return;

  for (i=0;i<L->patanz;i++)
	 if ((p=L->p_tab[i]) && p->pat)
		 {
		 for(L_ALL(l,j))
			{
			L->pmax[j]=max(L->pmax[j],p->pat[j]);
			L->pmin[j]=min(L->pmin[j],p->pat[j]);
			L->pmean[j]+=p->pat[j];
			}
		 }
  for (i=0;i<L->patanz;i++)
	 if ((p=L->p_tab[i]) && p->pat)
		 for(L_ALL(l,j))
			L->pvar[j]+=sqr(p->pat[j] - L->pmean[j]);
  for(L_ALL(l,j))
	{
	L->pmean[j]/=L->patanz;
	L->pvar[j]/=L->patanz;
	}
}

void setup_pattern_stats(pINT l)
{ layer *L = &L_Tab[l];
  if (L->pmin==NULL)
	  {
	  if (!(L->pmin=(REAL*)nnmalloc(L->size*sizeof(REAL))))
			 nn_error(SOURCENAME,"setup_pattern_stats(%s): Error allocating pmin[]",L->titel);
	  if (!(L->pmax=(REAL*)nnmalloc(L->size*sizeof(REAL))))
			 nn_error(SOURCENAME,"setup_pattern_stats(%s): Error allocating pmax[]",L->titel);
	  if (!(L->pmean=(REAL*)nnmalloc(L->size*sizeof(REAL))))
			 nn_error(SOURCENAME,"setup_pattern_stats(%s): Error allocating pmean[]",L->titel);
	  if (!(L->pvar=(REAL*)nnmalloc(L->size*sizeof(REAL))))
			 nn_error(SOURCENAME,"setup_pattern_stats(%s): Error allocating pvar[]",L->titel);
	  }
  else
	  {
	  if (!(L->pmin=(REAL*)nnrealloc((char*)L->pmin,L->size*sizeof(REAL))))
			 nn_error(SOURCENAME,"setup_pattern_stats(%s): Error reallocating pmin[]",L->titel);
	  if (!(L->pmax=(REAL*)nnrealloc((char*)L->pmax,L->size*sizeof(REAL))))
			 nn_error(SOURCENAME,"setup_pattern_stats(%s): Error reallocating pmax[]",L->titel);
	  if (!(L->pmean=(REAL*)nnrealloc((char*)L->pmean,L->size*sizeof(REAL))))
			 nn_error(SOURCENAME,"setup_pattern_stats(%s): Error reallocating pmean[]",L->titel);
	  if (!(L->pvar=(REAL*)nnrealloc((char*)L->pvar,L->size*sizeof(REAL))))
			 nn_error(SOURCENAME,"setup_pattern_stats(%s): Error reallocating pvar[]",L->titel);
	  }
}


/* scales all patterns of layer l to [0,1] using min,max */
void scale_patterns(pINT l)
{ pINT i,j,nr=L_Tab[l].patnr;
  for (ALL_PATTERNS(l,j))
	 {
	 nth_pat(l,j);
	 for (L_ALL(l,i))
		X_ext(l,i)=(X_ext(l,i)-L_Tab[l].pmin[i])/(L_Tab[l].pmax[i]-L_Tab[l].pmin[i]);
	 }
  nth_pat(l,nr); /* reset to actual pattern */
}

/*********************************************************************/
/*                                                                   */
/*   LOAD PATTERNS:                                                  */
/*                                                                   */
/*********************************************************************/

void close_pf(pINT l) 
{ fclose(pattern_file); L_Tab[l].p=NULL; }

#define LINELENGTH 1023

pattern *l_p( pINT width, pINT patsteps, pattern *lastpat, pINT typ, pINT nr) 
{
  pattern *p;
  pINT i;
  long fp;
  pINT ipat;
  char pname[PNAME_SIZE+1], *p1, str[LINELENGTH+1];
  char* WHITESPACE="\t ";

  if ((p=(pattern *)nnmalloc(sizeof(pattern)))==NULL)
     nn_error(SOURCENAME,"l_p(%u,%i,.): Patternstruktur-Allocfehler",width,patsteps);
  if ((p->pat = (REAL *) nncalloc(width,sizeof(REAL)))==NULL)
     nn_error(SOURCENAME,"l_p(%u,%i,.): Patternwerte-Allocfehler",width,patsteps);
  fp=ftell(pattern_file);
  switch (typ) {
    case 0: /* Integer, .dat-Files */
       fscanf(pattern_file,"%s\n",pname);
       if (pname[0]>='0' && pname[0]<='9')
        { /*  Filepointer zurueckstellen, falls kein Name in Patternfile,
                                Name darf nicht mit Ziffer beginnen */
        sprintf(pname,"Pattern %i",nr); /* use default name */
        fseek(pattern_file,fp-1,0);
        }
       for(i=0;i<width;i++)
         if(!feof(pattern_file))
           {fscanf(pattern_file,"%i\n",&ipat);
           p->pat[i] = (REAL)(ipat) / (REAL)(patsteps-1);}
         else
           {nn_alarm(SOURCENAME,"*l_p(%u,%i,..): Length of %s is %u instead of %u\n",
                                 width,patsteps,pname,i,width);
           i = width;
           }
       break;

    case 1: /* Floats, .txt-Files, one pattern in one line, TAB-delimited */
      sprintf(pname,"Pattern %i",nr);
      i=0;
      do {
        fgets(str,LINELENGTH,pattern_file);

        if (!strchr(str,'\t')) /* override Whitespace */
          if (strchr(str,';')) WHITESPACE=";";
          else if (strchr(str,',')) WHITESPACE=";";
             else WHITESPACE=" ";

        p1=strtok(str,WHITESPACE);      /*pointer to first field*/

        if (p1)
          do {
           if (sscanf(p1,"%f",&(p->pat[i]))==1) i++;
           p1=strtok(NULL,WHITESPACE);
           if (!p1) break;
          } while (i<width);
        } while (i<width && !feof(pattern_file));
      break;
    }
  p->size = width;
  sprintf(p->name,pname);
  p->next = NULL;
  p->last = lastpat;
  return p;
}

void load_pattern(pINT patsteps, pINT l, pINT typ)
{
  layer *L= &L_Tab[l];
  pINT width = L->size;
  pattern *p;
  static pINT nr=0;

  if(L->ext == NULL)
    {
    nr=0;
    L->ext = l_p(width,patsteps,NULL,typ,0);
    /* alloc first block of p_tab[] (512*4 = 2kB) */
    if((L->p_tab = (pattern **)nnmalloc(512*sizeof(pattern*)))==NULL)
      nn_alarm(SOURCENAME,"load_pattern(%i,%i): first pattern_tab failed",patsteps,l);
    L->p_tab[0] = L->ext;
    L->patanz=1;
    }
  else
    {
    p = L->p_tab[L->patanz-1]; /* last pattern */
    p->next = l_p(width,patsteps,p,typ,++nr);
    L->patanz++;
    if (!(L->patanz & 0x1ff))
      if((L->p_tab = (pattern **)nnrealloc(L->p_tab,(L->patanz+512)*sizeof(pattern*)))==NULL)
        nn_alarm(SOURCENAME,"load_pattern(%i,%i): nth pattern_tab failed",patsteps,l);
    L->p_tab[L->patanz-1]=p->next;
    }
}


void load_pattern2(pINT patsteps, pINT l1, pINT l2, pINT typ)
{
  layer *L1= &L_Tab[l1], *L2= &L_Tab[l2];
  pINT width1 = L1->size, width2 = L2->size;
  pattern *p, *q;
  static pINT nr=0;

  if(L1->ext == NULL)
    {
    nr=0;
    L1->ext = l_p(width1,patsteps,NULL,typ,0);
    /* alloc first block of p_tab[] (512*4 = 2kB) */
    if((L1->p_tab = (pattern **)nnmalloc(512*sizeof(pattern*)))==NULL)
      nn_alarm(SOURCENAME,"load_pattern2(%i,%u,%u): first pattern_tab for L1 failed",patsteps,l1,l2);
    L1->p_tab[0] = L1->ext;
    L1->patanz=1;

    L2->ext = l_p(width2,patsteps,NULL,typ,0);
    if((L2->p_tab = (pattern **)nnmalloc(512*sizeof(pattern*)))==NULL)
      nn_alarm(SOURCENAME,"load_pattern2(%i,%u,%u): first pattern_tab for L2 failed",patsteps,l1,l2);
    L2->p_tab[0] = L2->ext;
    L2->patanz=1;
    }
  else
    {
    p = L1->p_tab[L1->patanz-1]; /* last pattern */
    p->next = l_p(width1,patsteps,p,typ,++nr);
    L1->patanz++;
    if (!(L1->patanz & 0x1ff))
      if((L1->p_tab = (pattern **)nnrealloc(L1->p_tab,(L1->patanz+512)*sizeof(pattern*)))==NULL)
        nn_alarm(SOURCENAME,"load_pattern2(%i,%u,%u): nth pattern_tab for L1 failed",patsteps,l1,l2);
    L1->p_tab[L1->patanz-1]=p->next;

    q = L2->p_tab[L2->patanz-1]; /* last pattern */
    q->next = l_p(width2,patsteps,p,typ,nr);
    L2->patanz++;
    if (!(L2->patanz & 0x1ff))
      if((L2->p_tab = (pattern **)nnrealloc(L2->p_tab,(L2->patanz+512)*sizeof(pattern*)))==NULL)
        nn_alarm(SOURCENAME,"load_pattern2(%i,%u,%u): nth pattern_tab for L2 failed",patsteps,l1,l2);
    L2->p_tab[L2->patanz-1]=q->next;
    }
}

BOOL load_patterns(char *fname, pINT l)
{
  layer *L = &L_Tab[l];
  pINT width = L->size,patx,paty,patsteps;
  if((pattern_file = fopen(fname,"rb")) == NULL)
    nn_error(SOURCENAME,"load_patterns(%s,%i): file open failure",fname,l);
  /* read header: */
  fscanf(pattern_file,"%u\n",&patx);
  fscanf(pattern_file,"%u\n",&paty);
  fscanf(pattern_file,"%i\n",&patsteps);
  if (width != patx*paty)
    nn_alarm(SOURCENAME,"load_patterns(%s,%u): wrong pattern size %u, shouldbe %u",
                            fname,l,patx*paty,width);
  for(;!feof(pattern_file);) load_pattern(patsteps,l,0);
  close_pf(l);
  first_pattern(l);
  return (pattern_file!=NULL);
}

/* new version V2.02, ignore comments, correctly detects feof */
BOOL load_data(char *fname, pINT l)
{ char linebuf[LINELENGTH+1];
  long fp;
  if((pattern_file = fopen(fname,"rb")) == NULL)
	  nn_error(SOURCENAME,"load_data(%s,%u): file open failure",fname,l);
  else
	 {
	 do { /* step over comment-lines */
		fp=ftell(pattern_file);
		fgets(linebuf,LINELENGTH,pattern_file);
	  } while (linebuf[0] == ';' || linebuf[0] == '\n');

	 if (fp>0L)
		fseek(pattern_file,fp-1,0);
	 else
		rewind(pattern_file);

	 for(;!feof(pattern_file);)
		/* get next char to ensure correct feof() recognition */
		{ load_pattern(2,l,1); ungetc(fgetc(pattern_file),pattern_file); }
	 close_pf(l);
	 first_pattern(l);
	 setup_pattern_stats(l);
	 calc_pattern_stats(l);
	 }
  return (pattern_file!=NULL);
}

/* general data loading for DAT and PAT without headers for 2 Layers */
/* old version
BOOL load_data2(char *fname, pINT l1, pINT l2)
{ if((pattern_file = fopen(fname,"rb")) == NULL)
    nn_error(SOURCENAME,"load_data2(%s,%u,%u): file open failure",fname,l1,l2);
  else
    {
    for(;!feof(pattern_file);) load_pattern2(2,l1,l2,1);
    fclose(pattern_file);
    L_Tab[l1].p=NULL;
    L_Tab[l2].p=NULL;
    first_pattern(l1);
    first_pattern(l2);
    }
  return (pattern_file!=NULL);
}
*/
/* general data loading for DAT and PAT without headers for 2 Layers */
BOOL load_data2(char *fname, pINT l1, pINT l2)
{ char linebuf[200];
  long fp;
  if((pattern_file = fopen(fname,"rb")) == NULL)
	 nn_error(SOURCENAME,"load_data2(%s,%u,%u): file open failure",fname,l1,l2);
  else
	 {
	 do { /* step over comment-lines */
		fp=ftell(pattern_file);
		fgets(linebuf,200,pattern_file);
	 } while (linebuf[0] == ';' || linebuf[0] == '\n');

	 if (fp>0L)
		fseek(pattern_file,fp-1,0);
	 else
		rewind(pattern_file);

	 for(;!feof(pattern_file);)
		/* get next char to ensure correct feof() recognition */
		{load_pattern2(2,l1,l2,1); ungetc(fgetc(pattern_file),pattern_file);}

	 fclose(pattern_file);
	 L_Tab[l1].p=NULL;
	 L_Tab[l2].p=NULL;
	 first_pattern(l1);
	 setup_pattern_stats(l1);
	 calc_pattern_stats(l1);
	 first_pattern(l2);
	 setup_pattern_stats(l2);
	 calc_pattern_stats(l2);
	 }
  return (pattern_file!=NULL);
}

/* free memory allocated for patterns of layer l */
void free_patterns(pINT l)
{ pINT i;
  layer *L = &L_Tab[l];
  if (L->p_tab)
	 {
	 for (i=0;i<L->patanz;i++)
		 if (L->p_tab[i])
			{
      if (L->p_tab[i]->pat) nnfree((char*)L->p_tab[i]->pat);
			nnfree((char*)L->p_tab[i]);
			}
   nnfree(L->p_tab);
   }
  L->ext = L->p = NULL;
  L->patanz=0;
  if (L->pmin)
  	{
    nnfree((char*)L->pmin);
	  nnfree((char*)L->pmax);
    nnfree((char*)L->pmean);
	  nnfree((char*)L->pvar);
    }

}

BOOL load_schedule(char *fname, pINT l)
{
  char pname[40];
  if((sched_file = fopen(fname,"r")) == NULL)
    nn_error(SOURCENAME,"load_schedule(%s,%i): file open failure",fname,l);
  for(;!feof(sched_file);)
    {fscanf(sched_file,"%s\n",pname);
    load_patterns(pname,l);
    }
  fclose(sched_file);
  return (sched_file!=NULL);
}

/* jump ahead "number" values in actual pattern file */
void advance(pINT number)
{
  int i;
  float dummy;
  for(i=0;i<number;i++)
    fscanf(pattern_file," %f",&dummy);
}

void rewind_pf(void)
{ rewind(pattern_file); }


/*********************************************************************/
/*                                                                   */
/*   SAVE + LOAD - WEIGHTS                                           */
/*                                                                   */
/*********************************************************************/

BOOL load_weights(char *fname)
{
  FILE *fp;
  pINT k,l,m,i;
  weight_t* wptr;

  if ((fp = fopen(fname,"r")) == NULL)
    nn_alarm(SOURCENAME,"load_weights(): could not open %s\nNO weights were loaded",fname);
  else
    { for(ALL_LAYERS(l))
        for(ALL_LAYERS(k))
          for(ALL_LEVELS(m))
            if ((wptr=w_ptr[k][l][m])!=NULL)
              for(i=0;i<wptr->size;i++)
                fscanf(fp,"%f,",wptr->w+i);
      fclose(fp);
    }
  return (fp!=NULL);
}

BOOL save_weights(char *fname)
{
  FILE *fp;
  pINT k,l,m,i;
  weight_t* wptr;
  if ((fp = fopen(fname, "w")) == NULL)
    nn_alarm(SOURCENAME,"save_weights(): could not open %s\nNO weights were saved",fname);
  else
    { for(ALL_LAYERS(l))
        for(ALL_LAYERS(k))
          for(ALL_LEVELS(m))
            if ((wptr=w_ptr[k][l][m])!=NULL)
              for(i=0;i<wptr->size;i++)
                fprintf(fp,"%f,",wptr->w[i]);
      fclose(fp);
    }
  return (fp!=NULL);
}


/* Laedt Unit-Aktivierungen in Layer l */
BOOL load_units(char *fname, pINT l)
{
  FILE *fp;
  pINT i;
  if ((fp = fopen(fname,"r")) == NULL)
    nn_alarm(SOURCENAME,"load_units(): could not open %s\nNO activations loaded",fname);
  else
    { for (L_ALL(l,i)) fscanf(fp,"%f,",&X(l,i));
    fclose(fp);
    }
  return (fp!=NULL);
}

/* Speichert Unit-Aktivierungen des Layers l */
BOOL save_units(char *fname, pINT l)
{
  FILE *fp;
  pINT i;

  if ((fp = fopen(fname,"w")) == NULL)
    nn_alarm(SOURCENAME,"save_units(): could not open %s\nNO activations saved",fname);
  else
    {
    for (L_ALL(l,i))
      { fprintf(fp,"%f,",X(l,i));
      if ((i-L_Tab[l].first+1)%L_Tab[l].upr==0) fprintf(fp,"\n");
      }
    fclose(fp);
    }
  return (fp!=NULL);
}

/* Laedt alle Aktivierungen im Netzwerk */
BOOL load_activations(char *fname)
{ pINT i,l;
  FILE *fp;
  if ((fp = fopen(fname,"r")) == NULL)
    nn_alarm(SOURCENAME,"load_activations(): could not open %s\nNO activations loaded",fname);
  else
    {
    for (ALL_LAYERS(l))
      for (L_ALL(l,i)) fscanf(fp,"%f,",&X(l,i));
    fclose(fp);
    }
  return (fp!=NULL);
}

/* Speichert alle Aktivierungen im Netzwerk */
BOOL save_activations(char *fname)
{ pINT i,l;
  FILE *fp;
  if ((fp = fopen(fname,"w")) == NULL)
    nn_alarm(SOURCENAME,"save_activations(): could not open %s\nNO activations saved",fname);
  else
     {
     for (ALL_LAYERS(l))
       {
       for (L_ALL(l,i))
         {
         fprintf(fp,"%f,",X(l,i));
         if ((i-L_Tab[l].first+1)%L_Tab[l].upr==0) fprintf(fp,"\n");
         }
       fprintf(fp,"\n");
       }
     fclose(fp);
     }
  return (fp!=NULL);
}

/*********************************************************************/
/*                                                                   */
/*   OTHER FILE STUFF:                                               */
/*                                                                   */
/*********************************************************************/

void change_dir(char *dname) 
{
#ifdef PC
  /* Laufwerkbez. gibts nur am PC */
  int lw=0;
  char dummy[1024]="ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  if (dname[1]==':')
     {/* Drive given in pathname */
     lw = (int)((char*)strchr(dummy,toupper(dname[0]))-(char*)dummy) + 1;
     if (lw<0 || lw>26 || (_chdrive(lw)))
         nn_error(SOURCENAME,"change_dir(): unable to change to drive %c.\n",dname[0]);
    }
#endif
  if(chdir(dname))
     nn_error(SOURCENAME,"change_dir(): %s not found - pwd was %s\n",dname,getwd(dummy));
}



/*********************************************************************/
/*                                                                   */
/*   MISC. SYSTEM:                                                   */
/*                                                                   */
/*********************************************************************/

/* Button-Function ausfuehren */
void do_action(pINT i) 
{
  if (buttons[i].typ==SWITCH)
    if (buttons[i].status==AUS) 
       show_button(i,EIN); 
    else                     
       show_button(i,AUS);
  else 
    show_button(i,EIN);
  gprintf(BFX,BFY,COL_TEXT,"BUSY");
  if (buttons[i].call) (*buttons[i].call)(); /* call only if functions exists */
  gprintf(BFX,BFY,COL_TEXT,"FREE");
  if(buttons[i].typ!=SWITCH) show_button(i,AUS);
}

#if !defined(sun4) && !defined(WIN)
BOOL UnitClicked()
{ pINT l,i;
  for (ALL_LAYERS(l))
    for (L_ALL(l,i))
      {
      if ( ( mouse_x-PX(l,i) <= L_Tab[l].ux )
         &&( mouse_y-PY(l,i) <= L_Tab[l].uy ) )
          {
          unit_clicked=i;layer_clicked=l;
          return 1;
          }
      }
  return 0;
}

char wait_for_action(void)
{
  char c;
  pINT i,found;
  draw_layers(); /* Layers erstmalig anzeigen */

  for (EVER) { /* bis Quit-Button angeklickt wird */
     c = nnwait();
     if (c=='@') /* a mouse click */
        { /* first check for buttons */
        found = 0; i=0;
        while(i<buttons_used && !found)
          if( (buttons[i].x<=mouse_x)
            &&(buttons[i].x+buttons[i].width  >= mouse_x)
            &&(buttons[i].y<=mouse_y)
            &&(buttons[i].y+buttons[i].height >= mouse_y) )
              found = 1;
          else i++;
        if (found)
           do_action(i);
        else /* no button found, check for unit-clicks */
             {
             if(UnitClicked())
                 (*action_when_unit_clicked)(layer_clicked,unit_clicked);
             else
                 {/* Click on Background - redraw layers */
                 draw_layers();
                 }
             }  /* end else */
        } /* end if mouse click section */
     else
         {/* Check for Hotkeys */
         found = 0; i = 0;
         while(i<buttons_used && !found)
           /*if(lower(buttons[i].key)==lower(c)) found = 1; else i++;*/
           if(buttons[i].key==c) found = 1; else i++;
         if (found) do_action(i);
         }
     }
  return c;
}
#else
#ifdef WIN
/* Windows - do nothing, eventloop is handled from WinMain() */
char wait_for_action(void){return 0;}
#else
/* SUN */
char wait_for_action(void)
{ x_wait_for_action();}
#endif
#endif

/* Init variables */
void nn_setup(int argc,char **argv)
{ char file[9];
 /* Init Filenames with programname */
#ifdef PC
  fnsplit(argv[0],NULL,NULL,file,NULL);
#else
/*	strncpy(file,argv[0],8);
  file[8]='\0'; */
  sprintf(file,"VieNet2"); /* sicher ist sicher */
#endif
  strcpy(ProgName,file);
  strcpy(MsgFile,ProgName);
  strcat(MsgFile,".msg");
  strcpy(ActFile,ProgName); /* no default extension for automatic Layer/Network-Extensions */
  strcpy(LayoutFile,ProgName);
  strcat(LayoutFile,".nnl");
  strcpy(WeightFile,ProgName);
  strcat(WeightFile,".nnw");
  strcpy(LogFile,ProgName);
  strcat(LogFile,".log");
  strcpy(ErrorLogFile,ProgName);
  strcat(ErrorLogFile,".err");
  strcpy(DataFile,ProgName);
  strcat(DataFile,".dat");
  check_args(argc,argv);
  init_randn(); /* init random generator */
  init_graph(); /* init graph for DOS and Apollo */
}

void shutdown(void)
{ layer* L;
  int l1,l2,lev;
  pattern *p1,*p2;

  /* Alle allokierten Gewichte freigeben */
  for (ALL_LEVELS(lev))
    for (ALL_LAYERS(l1))
      for (ALL_LAYERS(l2))
        if (w_ptr[l1][l2][lev])
          {
          if (w_ptr[l1][l2][lev]->w) nnfree(w_ptr[l1][l2][lev]->w);
          nnfree(w_ptr[l1][l2][lev]);
          }

        for (ALL_LAYERS(l1))
          {L=&L_Tab[l1];
          /* Alle externen Pattern freigeben */
          if (L->ext)
             {p1=L->ext;
             while (p1) {p2=p1->next; nnfree(p1); p1=p2;}
          }
          /* Unit-Arrays freigeben */
          if (L->X) nnfree(L->X);
          if (L->XNEW) nnfree(L->XNEW);
          if (L->POS) nnfree(L->POS);
        }
  switch((int)(4.0*random_nr()))
    {
    case 0: pprintf_e(4,"by by\n"); break;
    case 1: pprintf_e(4,"ciao\n"); break;
    case 2: pprintf_e(4,"adios\n"); break;
    case 3: pprintf_e(4,"Auf Wiedersehen!\n");break;
    }
  pprintf_e(2,"\n");
  close_graph();
  exit(0);
}

/*********************************************************************/
/*                                                                   */
/*                             THE END                               */
/*                                                                   */
/*********************************************************************/
#undef SOURCENAME
