/*************************************************************************
*  Program:       NEURONS.H                                              *
*  Description:   Variables and functiondefs for VieNet2 (neurons.c)     *
*  Version:       2.03                                                   *
*  Last modified: 4.9.96                                                 *
*  Author:        LG                                                     *
*  Copyright (C) Austrian Research Institute for Artificial Intelligence *
**************************************************************************/

#define SPRINTF_RETURNS_INT /* g++ <stdio> hat default *char, schalte auf int */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include <stdarg.h>
#include <values.h>
#include <stdlib.h>

#ifdef  __cplusplus
#ifdef sun4
#include <libc.h>
#include <osfcn.h>
#endif
#ifdef apollo
#define MAXFLOAT ((float)3.40282346638528860e+38)
#endif /*try to include values.h ! */
#endif

#ifdef PC
#define HUGE MAXINT
#endif

/*------------------------------------------------------------------------*/

#define TRUE  1
#define FAIL  0
#define FALSE 0


#define BAD_UNIT 65535    /* ungueltige Unitnummer fuer Returnvalues */
#define BAD_PATTERN 65535 /* ungueltige Patternnummer fuer Returnvalues */


/* Maustasten-Funktionen */
#define M_TOGGLE 0
#define M_WEIGHT 1

/* Types of Keys */
#define BUTTON 0
#define SWITCH 1
#define TWIN_1 2
#define TWIN_2 3
#define SLIDER 4
#define METER3 5

/* Types of Layer-Display-Position-Settings in init_layer(),
  DELTA_LAYER supports delta[] and target[] arrays */
#define GRID 0x01
#define HEX 0x02
#define STANDARD_LAYER 0x01
#define BIAS_LAYER 		0x08
#define DELTA_LAYER   0x10

/* Types of Connections in connect() */
#define FULL 0                /* All-to-All */
#define COMP 1                /* All-to-All-Others */
#define IDENT 2               /* One-to-One (Identity) */
#define ORDER2 3              /* All-to-All general 2.Order */

/* Types of use-weights in update(), energy...*/
#define SINGLE 0
#define ALL 1

/* Types of Update in update() */
#define STANDARD 0
#define IAC      1
#define IAC2     2
#define KOHONEN  3
#define KOHONEN_WTA  7
#define WTA      4
#define EXT      5
#define ADDEXT   6
#define COPY     8	/* old!! */
#define COPY_ADD_X 8 
#define L_ADD   9	/* old!! */
#define COPY_ADD_XNEW 9
#define W_COPY   10  /* old!! */
#define COPY_ADD_X_W  10
#define L_COPY  11	/* old!! */
#define COPY_XNEW 11
#define COPY_X  12
#define SIMPLE  13
#define SIMPLE2 14
#define SIMPLE3 15
#define INVARI2 16
#define EXT_XNEW 17
#define BACKWARD 18
#define COPY_X_W  20

/* types of learning for NNLearn() */
#define HEBB_LEARN 0
#define DELTA_LEARN 1
#define BACKPROP_LEARN 2
#define KOHONEN_LEARN 3

/* types for NNSetDelta() */
#define DELTA    1
#define BACKPROP 2

/* Types of show_pname */
#define EXT_PNAME 0
#define INT_PNAME 1

/* Types of Layer-Displaying */
#define L_NEU   0
#define L_AKTIV 1

/*****************************************************************************/
/*                                                                           */
/*                                      M A K R O S                          */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*  1.) Allgemeine Makros  */

#define sqr(x) ((x) * (x))
#define iabs(i) ( ((i)>0) ? (i) : -(i) )
#define round(x) (((x)>0.0)?((int)floor((x)+0.5)):((int)ceil((x)-0.5)))
#define max(x,y) ((x>y)?(x):(y))
#define min(x,y) ((x<y)?(x):(y))
#define sgn(x) (((REAL)x==0.0)?0:((REAL)x>0.0)?1:-1)

/* dynamic memory management with memory checking */
#ifdef NNLOGALLOC
#  define LogAlloc(str,nitems,size) pprintf_file(LogAllocName,"%s in %s, Line %i, %lu * %lu Bytes\n",str,__FILE__,__LINE__,long(nitems), long(size)),
#else
#  define LogAlloc(str,nitems,size)
#endif

#ifdef PC
/* farcalloc for long values */
#  define nncalloc(nitems,size) (LogAlloc("calloc ",nitems,size) farcalloc(nitems,size))
#  define nnmalloc(size) (LogAlloc("malloc ",1,size) farmalloc(size))
#  define nnrealloc(ptr,size) (LogAlloc("realloc",1,size-sizeof(*ptr)) farrealloc(ptr,size))
#else
/* farcalloc not available in UNIX */
#  define nncalloc(nitems,size) (LogAlloc("calloc ",nitems,size) calloc(nitems,size))
#  define nnmalloc(size) (LogAlloc("malloc ",nitems,size) malloc(size))
#  define nnrealloc(ptr,size) (LogAlloc("realloc ",nitems,size) realloc((char*)ptr,size))
#endif

#define nnfree(ptr) (LogAlloc("free   ",1,sizeof(*ptr)) free((char*)ptr),ptr=NULL)

/* Liefern Werte aus [0,1) */
#ifdef PC
#  define MOST_RANDOM 32768.0
#else
#  define MOST_RANDOM 2147483648.0
#endif 
#define random_nr()   ((REAL)rand()/MOST_RANDOM)
#define random_ev(x)  ( (random_nr() < (x)) ? 1 : 0 )


/*****************************************************************************/
/*  2.) Loop-Makros  */

#define EVER ;;

/* Iteration ueber alle Units im Layer (Struktur)
    var ist direkt Nummer der Unit, bei Gewichten wird automatisch ueber
    die Zugriffsmakros auf relative Nummern umgerechnet) */
#define LL_ALL(layer, var) var = layer.first; var <= layer.last; var++

/* wie oben, aber mit Layernummer */
#define L_ALL(l, var) DEBUG_LAYER(l) var = L_Tab[l].first; var <= L_Tab[l].last; var++

/* wie oben, aber mit Startwert */
#define L_ALL_ST(l, var, start) DEBUG_LAYER(l) var = start; var <= L_Tab[l].last; var++

/* Iteration ueber alle relativen Unitnummern von Layer l
   (z.B. fuer externe Patterns) */
#define L_ALL_R(l, var) DEBUG_LAYER(l) var = 0; var < L_Tab[l].size; var++

/* wie L_ALL, aber mit 2 Layers (fuer IDENT-Verbindungen) */
#define L_ALL2(l1, l2, var1, var2)\
     DEBUG_LAYER(l1) DEBUG_LAYER(l2)\
	   var1=0,var2=0;var1<=L_Tab[l1].last;var1++,var2++


/* Doppelsumme fuer 2.Order-Netze */
#define LOOP_2(l, var1, var2)\
   DEBUG_LAYER(l)0;\
   for (var1 = 0; var1 < L_Tab[l].size-1; var1++)\
		 for (var2 = var1 + 1; var2 < L_Tab[l].size; var2++)

/* Iteration ueber alle Gewichte zwischen Layer l1->l2 auf Ebene lev */
#define ALL_WEIGHTS(l1, l2, var) \
    (DEBUG_LAYER(l1) DEBUG_LAYER(l2) \
	  var = 0; var < w_ptr[l1][l2][W_LEVEL]->size; var++)

/* Iteration ueber alle Weight-Levels */
#define ALL_LEVELS(var) var = 0; var < MAX_W_LEVELS; var++

/* Iteration ueber alle definierten Layers */
#define ALL_LAYERS(var) var = 0; var < L_anz; var++

/* iterates var on all patterns of layer l */
#define ALL_PATTERNS(l,var) var = 0; var < L_Tab[l].patanz; var++

/*****************************************************************************/
/*  3.) Makros fuer Zugriff auf linearisierte Gewichtsmatrix                 */
/*      mit Pointer w_ptr[Layer1_Nr][Layer2_Nr][W_LEVEL]->w                  */


/******  allgemeine Hilfsmakros  ******/
#define AU(l, relpos) (relpos)      /* absolut-nummer */
#define RU(l, abspos) (abspos)      /* relative pos im Layer */
#define XPOS(l, i) (i % L_Tab[l].upr)  /* X-Pos im Layer */
#define YPOS(l, i) (i / L_Tab[l].upr)  /* Y-Pos im Layer */
#define ON(l,i) (X(l,i) > ONTHRES)  /* Praedikat ob Unit > ONTHRES  (->KONNSPRACH) */

/****** Normaler Zugriff auf Weights, i,j sind Unitnummern ******/
#define W(l1,l2,i,j) \
  (DEBUG_WEIGHT(l1,l2,i,j)\
	w_ptr[l1][l2][W_LEVEL]->w[L_Tab[l2].size * i + j])


/******  Zugriff auf Weights bei IDENT-Verbindung  ******/
/* i ist relative Unitnummer */
#define W_1(l1, l2, i) \
	(DEBUG_LAYER(l1) DEBUG_LAYER(l2)\
	w_ptr[l1][l2][W_LEVEL]->w[i])

/******  Netze 2.Ordnung  ******/
/* Hilfsmakro fuer Indexberechnung in Dreiecksmatrix fuer ORDER2-Netze */
#define INDEX_O2(l, i, j)\
	(DEBUG_LAYER(l)\
   ((2 * L_Tab[l].size - i + 1) * i / 2 + j))

/* Zugriff fuer allgemeine Netze 2.Ordnung (Dreiecksmatrix!)
    i,j sind 2 Units aus l1,  k ist Unit aus l2
*/
#define W_O2(l1, l2, i, j, k)\
   (DEBUG_LAYER(l1) DEBUG_LAYER(l2) \
   w_ptr[l1][l2][W_LEVEL]->w[INDEX_O2(l1, i, j) * L_Tab[l2].size + k])


/******  Invarianzen-Netz 2.Ordnung Typ 1 (PISONN)  ******/
/* Hilfsmakros fuer INVARI2-Netz */
#define DELTA_X(l, i, j) iabs(XPOS(l, i) - XPOS(l, j))
#define DELTA_Y(l, i, j) iabs(YPOS(l, i) - YPOS(l, j))

/* Indexnr. der Unit im Hiddenlayer fuer Produkt der Units i bzw. j in Layer l */  
#define INDEX_I1(l, i, j)\
   DELTA_Y(l, i, j) * ((L_Tab[l].upr - 1) * 2 + 1) + XPOS(l, j) \
   - XPOS(l, i) + L_Tab[l].upr - 1

/* wie INDEX_I1, nur in x-Richtung */
#define INDEX_I1_X(l, i, j)\
   DELTA_Y(l, i, j) * ((L_Tab[l].upr - 1) * 2 + 1) + XPOS(l, j) \
   - XPOS(l, i) + L_Tab[l].upr - 1

/* wie INDEX_I1_X, aber zyklisch */
#define INDEX_I1_X_Z(l, i, j)\
   DELTA_Y(l, i, j) * ((L_Tab[l].upr - 1) * 2 + 1) + XPOS(l, j) \
   - XPOS(l, i) + L_Tab[l].upr - 1


/****** Invarianzen-Netz 2.Ordnung Typ 2 (SISONN) ******/
/* Berechnung des Anstiegs zw. xi,xj;  Voraussetzung: i<j (!)
		Liefert 0 fuer pos. und vertikal, 1 fuer neg. und horizontal
*/
#define ANSTIEG(l, i, j) ( XPOS(l, i) < XPOS(l, j) )

/* Zugriff auf linearisierte Indextabelle */
#define W_INDEX(l, dx, dy, anst) w_ind[(dy * L_Tab[l].upr + dx) * 2 + anst]

/* Zugriff auf INVARI2-Netz (ueber Indextabelle - nur ein Netz moeglich!) */
#define W_I2(l1, l2, i, j, k)\
   (DEBUG_LAYER(l1) DEBUG_LAYER(l2) \
	 w_ptr[l1][l2][W_LEVEL]->w[w_indanz*k + \
	 W_INDEX(l1, DELTA_X(l1, i, j), DELTA_Y(l1, i, j), ANSTIEG(l1, i, j))])

/*****************************************************************************/
/*  4.) Makros fuer Zugriff auf linearisierte Matrix der Aktivierungswerte   */ 

/* Zugriff ueber Layer- und Unitnummer */
#define X(l,i) \
  (DEBUG_LAYER(l) DEBUG_UNIT(l,i)\
	L_Tab[l].X[i])
#define XNEW(l,i)\
	(DEBUG_LAYER(l) DEBUG_UNIT(l,i) \
	L_Tab[l].XNEW[i])

/* Zugriff ueber Layerstructure und Unitnummer */
#define X_L(L,i)\
 (DEBUG_LAYER_L(L) DEBUG_UNIT_L(L,i) \
 L.X[i])
#define XNEW_L(L,i)\
 (DEBUG_LAYER_L(L) DEBUG_UNIT_L(L,i) \
 L.XNEW[i])

/* Zugriff auf Bildschirm-Unitpositionen ueber Layer- und Unitnummer */
#define PX(l,i)\
	(DEBUG_LAYER(l) DEBUG_UNIT(l,i)\
	L_Tab[l].POS[i].x)
#define PY(l,i)\
	(DEBUG_LAYER(l) DEBUG_UNIT(l,i)\
	L_Tab[l].POS[i].y)


/* access to one unit of the actual external pattern of layer l */
#define X_ext(l,i) (L_Tab[l].p->pat[i])
/* access to one unit of the next external pattern of layer l */
#define X_ext_next(l,i) (L_Tab[l].p->next->pat[i])
/* access to one unit of the previous external pattern of layer l */
#define X_ext_prev(l,i) (L_Tab[l].p->last->pat[i])

/* access to one unit of the actual target pattern of layer l */
#define X_target(l,i) (L_Tab[l].target[i])

/* access to one unit of the actual delta pattern of layer l */
#define X_delta(l,i) (L_Tab[l].delta[i])

/*****************************************************************************/
/*  5.) Makros fuer Debugging  */

#ifdef NNDEBUG
#if defined(PC) && !defined(__cplusplus) && !defined(WIN)
#error Debug Option requires C++. Use C++ Option or #undef NNDEBUG in neurodef.h!
#endif

#define DEBUG_LAYER(l) debug_layer(l),
#define DEBUG_UNIT(l,u) debug_unit(l,u),
#define DEBUG_LAYER_L(L) debug_layer_L(L),
#define DEBUG_UNIT_L(L,u) debug_unit_L(L,u),
#define DEBUG_WEIGHT(l1,l2,i,j) debug_weight(l1,l2,i,j),
#else
#define DEBUG_LAYER(l)
#define DEBUG_UNIT(l,u)
#define DEBUG_LAYER_L(L)
#define DEBUG_UNIT_L(L,u)
#define DEBUG_WEIGHT(l1,l2,i,j)
#endif
/*****************************************************************************/
/*                                                                           */
/*    T Y P E N - D E F I N I T I O N E N                                    */
/*                                                                           */
/*****************************************************************************/
#ifndef NEURONS_H

typedef float real_t;                /* !!!  OLDIES */
typedef unsigned int int_t;
#define regint_t register int_t

typedef float  REAL;
#define MAX_REAL MAXFLOAT
#define MIN_REAL MINFLOAT

typedef unsigned int pINT; /* must not be short on SUNs!! */
typedef unsigned long plINT;
#define regpINT  register pINT
#define MAX_pINT ~((pINT)0)

#ifndef BOOL
typedef int BOOL;
#endif

typedef struct { pINT x, y; } UPOS;

typedef struct pattern
{
  REAL *pat;
  pINT size;
  char name[PNAME_SIZE+1];
  struct pattern *next, *last;
  REAL min,max,mean,var,length;
} pattern;

typedef struct 
{ pINT first,     /* Erste Unitnummer im Layer */
        last,     /* Letzte Unitnummer im Layer */
        size,     /* aktuelle Unitanzahl im Layer */
          nr,     /* Layernummer in Tabelle L_Tab[] */
          id,     /* ID-Status (siehe Lernregel) */
        x, y,     /* Layer-Position am Bildschirm (linke obere Ecke) */
         upr,     /* Anzahl Units je Reihe im Layer */
      ux, uy,     /* Unitbreite und -hoehe */
        dx, dy;   /* Unitabstand in X- und Y-Richtung */
 int  extsep_x,
      extsep_y;   /* relative Position des externen Layers */
   pINT status,   /* Layeranzeige: NEU/QUICK */
           typ;   /* Layertyp GRID/HEX; BACKPROP */
   char *titel,      /* Layername */
    *titel_ext;      /* Extern-Layer-name */
  pattern *ext,      /* Pointer auf Extern-Pattern-Kette */
            *p;      /* Pointer auf aktuelles Pattern */
  pINT  patanz;      /* aktuelle Patternanzahl */
  pINT  patnr;       /* aktuelle Patternnummer in p_tab[] */
  pattern *(*p_tab); /* Pointer auf Tabelle der Patternpointer fuer
                       schnellen Zugriff in nth_pat() */
  REAL *pmin, *pmax, *pmean, *pvar;

  REAL *target; /* target-array for NNLearn() */
  REAL *delta;  /* delta-array for NNLearn() */

  REAL (*f)(REAL);  /* local activation function */
  REAL (*df)(REAL); /* first derivation of activation function */

  REAL *X;   	    /* activation-values */
  REAL *XNEW;
  UPOS *POS;
} layer;

typedef struct
{
  int x,y;              /* Button-Position */
  int width,height;     /* Button-Groesse */
  void (*call)();       /* Button-Funktion */
  int param;            /* Zusatzparameter fuer call-function */
  char key;             /* Hotkey */
  char *text;           /* Button-Text */
  int typ;              /* Button-Typ (SWITCH/BUTTON/TWIN-1/TWIN-2/SLIDER) */
  int status;           /* Button-Status (EIN/AUS/PASSIV) */
  void *var;             /* Pointer auf Variable fuer Sliders */
  REAL val_min;         /* Minimalwert fuer var */
  REAL val_max;         /* Maximalwert fuer var */
  REAL val_res;         /* Slider-Auflsung von var */
  char val_typ;         /* Typ der Slider Variablen (f/s/i/l fFr float/short int/int/long) */
  char* val_fmt;        /* Druckformat fFr var im Sliderfeld */

  REAL value;           /* Werte fuer METER3 */
  REAL limit1;
  REAL limit2;
  long int id;          /* eindeutige ID # des Panelbuttons */
  void *ptr;
} button_t;

typedef struct 
{
	 REAL *w;          /* Pointer auf linearisiertes Weight-Array */
	 pINT  size;       /* Anzahl Weights */
	 pINT  typ;        /* Verbindungstyp (COMP,FULL,ORDER2,INVARI2) */
} weight_t;

#endif

/*****************************************************************************/
/*                                                                           */
/*    P R O T O T Y P E S                                                    */
/*                                                                           */
/*****************************************************************************/
#ifdef __cplusplus
extern "C" {
#endif
extern  void advance(pINT);
extern  pINT alarm_all_w_e(REAL,REAL,pINT);
extern  pINT alarm_all_x_e(REAL,REAL,pINT);
extern  pINT best_n_e(pINT,pINT,REAL,REAL);
extern  void border_w_e(pINT,pINT,REAL,REAL,REAL,REAL);
extern  void border_x_e(pINT,REAL,REAL,REAL,REAL);
extern  void border_xnew_e(pINT,REAL,REAL,REAL,REAL);
extern  void calc_pattern_stat(pattern *p);
extern  void calc_pattern_stats(pINT l);
extern  void change_dir(char*);
extern  void connect_all_new(void); /*Muss im Main-Programm definiert werden! */
extern  void close_pf(pINT);
extern  void connect_b(pINT,pINT,REAL);
extern  void connect_1n(pINT,pINT,pINT,REAL); 
extern  void connect_n1(pINT,pINT,pINT,REAL);
extern  void copy(void);
extern  void copy_layer(pINT);
extern  void debug_layer(pINT l);
extern  void debug_unit(pINT l,pINT u);
extern  void debug_layer_L(layer L);
extern  void debug_unit_L(layer L,pINT u);
extern  void debug_weight(pINT l1,pINT l2,pINT i,pINT j);
extern  void deconnect_layer(pINT);
extern  REAL df(REAL);
extern  void draw_ext(pINT);
extern  void draw_layer(pINT);
extern  void draw_layers(void);
extern  void draw_weights_from(pINT,pINT,REAL);
extern  void draw_weights_to(pINT,pINT,REAL);
extern  REAL energy_w_e(pINT,pINT,REAL,REAL,pINT);
extern  REAL energy_x_e(pINT,REAL,REAL);
extern  REAL energy_xnew_e(pINT,REAL,REAL);
extern  REAL f(REAL);
extern  void first_pattern(pINT);
extern  void free_patterns(pINT);
extern  void gp_border(void);
extern  int  gp_printf(char*,...);
extern  void init_ext(pINT,int,int,char*);
extern  pINT init_layer(pINT,pINT,pINT,pINT,pINT,char*);
extern  void init_randn(void);
extern  pINT interp_pat(pINT,REAL,pINT);
extern  void last_pattern(pINT);
extern  BOOL load_activations(char*);
extern  BOOL load_data(char*,pINT);
extern  BOOL load_data2(char*,pINT,pINT);
extern  pattern *l_p(pINT,pINT,pattern*, pINT, pINT);
extern  BOOL load_patterns(char*,pINT);
extern  BOOL load_schedule(char*,pINT);
extern  BOOL load_units(char*,pINT);
extern  BOOL load_weights(char*);
extern  void make_pattern_e(pINT,char*,char*);
extern  void make_button(char*,void(*)(),char);
extern  void make2button(char*,void(*)(),char,char*,void(*)(),char);
extern  int  make_meter3(char* text, float min, float max, float limit1, float limit2, float value);
extern  void make_switch(char*,void(*)(),char);
extern  void make_slider(char*,void(*)(),char,long*,long,long);
extern  void make_slider_e(char*,void(*)(),char,char*,...);
extern  pINT max_u(pINT);
extern  void max_ux(pINT,pINT*,REAL*);
extern  REAL max_x(pINT);
extern  pINT maxabs_u(pINT);
extern  REAL maxabs_x(pINT);
extern  void mult_layer(pINT);
extern  pINT new_unit(pINT);
extern  void next_pattern(pINT);
extern  BOOL next_pattern_exist(pINT);
extern  int  nn_alarm(char*,char*,...);
extern  int  nn_error(char*,char*,...);
extern  void nn_fatalerror(char*,char*,...);
extern  void nn_init(int,char**);
extern  void nn_setup(int,char**);
extern  void nnconnect(pINT,pINT,pINT,REAL,REAL,REAL);
extern  void nnconnect_1n(pINT,pINT,pINT,REAL,REAL,REAL);
extern  void nnconnect_n1(pINT,pINT,pINT,REAL,REAL,REAL);
extern  void NNSetDelta(pINT l,pINT mode);
extern  void NNSetTarget(pINT l, REAL* target);
extern  void NNLearn(pINT l1, pINT l2, pINT typ, REAL rate, pINT use_weights);
extern  void normalize_w_e(pINT,pINT,REAL);
extern  void normalize_x_e(pINT,REAL);
extern  void normalize_xnew_e(pINT,REAL);
extern  pINT nth_pat(pINT,pINT);
extern  void open_pf(pINT);
extern  BOOL pattern_ex(pINT);
extern  void print_act (pINT);
extern  void print_weights(pINT l1, pINT l2);
extern  int  pprintf(char*,...);
extern  int  pprintf_e(pINT,char*,...);
extern  int  pprintf_file(char*,char*,...);
extern  void prev_pattern(pINT);
extern  pINT rand_pat(pINT);
extern  void reposition_layer(layer *L);
extern  void reset(pINT);
extern  void reset_layers(void);
extern  void reset_x(pINT);
extern  void reset_xnew(pINT);
extern  pINT resize_layer(layer *L, pINT newsize, pINT newwidth);
extern  void rewind_pf(void);
extern  void round_w(pINT,pINT,pINT);
extern  void round_x(pINT,pINT);
extern  void round_xnew(pINT,pINT);
extern  BOOL save_units(char*,pINT);
extern  BOOL save_weights(char*);
extern  BOOL save_activations(char*);
extern  void scale_patterns(pINT l);
extern  void setup_pattern_stats(pINT l);
extern  void show_pname(pINT,char*,pINT);
extern  void shutdown(void);
extern  void step_pattern(pINT,int);
extern  pINT thresh_e(pINT,REAL,REAL,REAL);
extern  pINT thresh_ask(pINT,REAL);
extern  void toggle_unit(pINT,pINT);
extern  void unit_semantics(pINT,pINT);
extern  void update(pINT,pINT,pINT,pINT);
extern  char wait_for_action(void);
extern  pINT wta(pINT);

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
extern  void init_ext(pINT,int,int);
#endif

/*   abgekuerzte Funktionen */
#define alarm_all_w(min,max) alarm_all_w_e(min,max,6)
#define alarm_all_x(min,max) alarm_all_x_e(min,max,6)
#define best_n(l,n) best_n_e(l,n,XMIN,XMAX)
#define border_w(l1,l2) border_w_e(WMIN,WMAX,WMIN_TOL,WMAX_TOL)
#define border_x(l) border_x_e(l,XMIN,XMAX,XMIN_TOL,XMAX_TOL)
#define border_xnew(l) border_xnew_e(l,XMIN,XMAX,XMIN_TOL,XMAX_TOL)
#define connect_0(l1,l2) nnconnect(l1,l2,FULL,0.0,0.0,0.0)
#define connect_c(l1,l2,c) nnconnect(l1,l2,FULL,c,c,1.0)
#define energy_w(l1,l2,mode) energy_w_e(l1,l2,1.0,1.0,mode)
#define energy_x(l1) energy_x_e(l1,1.0,1.0)
#define energy_xnew(l1) energy_xnew_e(l1,1.0,1.0)
#define make_pattern(l,s) make_pattern_e(l,s,"")
#define normalize_w(l1,l2) normalize_w_e(l1,l2,0.0)
#define normalize_x(l1) normalize_x_e(l1,0.0)
#define normalize_xnew(l1) normalize_xnew_e(l1,0.0)
#define thresh(l,thresh) thresh_e(l,XMIN,thresh,XMAX)

/* optimierte Funktionen als Makros */
#define patanz(l) 			(L_Tab[l].patanz)

/* OLDIES: !!! */ 
#define WEIGHT(l1,l2,i,j) W(l1,l2,i,j)
#define WEIGHT_1(l1,l2,i) W_1(l1,l2,i)
#define WEIGHT_O2(l1,l2,i,j,k) W_O2(l1,l2,i,j,k)
#define WEIGHT_I2(l1,l2,i,j,k) W_I2(l1,l2,i,j,k)
#define make_pattern_i(n,a,h) make_pattern_e(n,a,h)
#define normalize(l1,l2) energy_w_e(l1,l2,0.0,WMAX,SINGLE);
#define w_level W_LEVEL
extern  void nth_pattern(pINT,pINT);
extern  void random_pattern(pINT,REAL);
extern  void make_pattern_i(pINT,char*,char*);
extern  void alarm_weights(REAL,REAL);
extern  short int detect_winner(pINT);
extern  void make_twin_buttons(char *,void(*)(),char,char *,void(*)(),char);
extern  void gerror(char *);
extern  void interp_pattern(pINT,pINT);




/*****************************************************************************/
/*                                                                           */
/*            V A R I A B L E N                                              */
/*                                                                           */
/*****************************************************************************/


/************** DEKLARATIONEN fuer MODULE **********************/
/* wird nur in neurons.c, nngraph.c, pcnngr, nnlayer definiert.
	 D.h. globale Variablen werden nur im Anwenderprogram gesetzt,
	 in den Modulen sind sie externe Variablen
*/

#ifndef NEURONS_H
#define NEURONS_H

#ifdef MODULE

  extern long units_used, weights_used;
  extern pINT unit_clicked, layer_clicked;
  extern pINT buttons_used, button_pressed;
  extern pINT button_x_pos, button_y_pos;
  extern button_t buttons[MAXBUTTONS];

  /* pointers for weight-arrays */
  extern weight_t *w_ptr[MAXLAYER][MAXLAYER][MAX_W_LEVELS];
  extern pINT W_LEVEL;
  extern layer L_Tab[MAXLAYER]; /* Layer-Structures */
  extern pINT L_anz;

  extern pINT *w_ind;   /* Indizis fuer w_ptr je delta-x, delta-y, Vorzeichen */
  extern pINT w_indanz; /*  Nur fuer INVARI2-Netz! */
  extern void (*action_when_unit_clicked)(pINT l, pINT u);
  extern int mouse_function[3];
  extern REAL ScaleWeightDisplay, ScaleActivationDisplay;

  extern REAL ONTHRES;
  extern int GP_LINES;
  extern int GP_COLLS;
  extern int GP_X;
  extern int GP_Y;

  /* a few standard parameters */
  extern REAL GAIN,
              OFFSET,
	      ALPHA,
	      LAST,
	      DECAY,
              EPSILON,
	      RATE;

  extern char ProgName[13];
  extern char LogAllocName[13];
  extern pINT MAXIAC;

  extern char LogFile[];
  extern char ErrorLogFile[];
  extern char WeightFile[];
  extern char ActFile[];
  extern char LayoutFile[];
  extern char MsgFile[];
  extern char DataFile[];

/****************  DEFINITIONEN fuer HAUPTPROGRAMM ***************************/

#else
  long units_used    = 0, weights_used   = 0;
  pINT layer_clicked = 0, unit_clicked   = 0;
  pINT buttons_used  = 0, button_pressed = 0;

  pINT button_x_pos,       /* Position des ersten Buttons, button_x_pos wird in */
       button_y_pos = 10;  /* init_graph() abhaengig von Schirmbreite gesetzt   */

  button_t buttons[MAXBUTTONS];

  /* Pointers for Weight-Structures */
  weight_t* w_ptr[MAXLAYER][MAXLAYER][MAX_W_LEVELS];
  pINT W_LEVEL = 0;            /* aktuelle Weight-Ebene fuer alle Zugriffe */
  layer L_Tab[MAXLAYER];       /* Layer-Structures */
  pINT L_anz = 0;              /* Anzahl der definierten Layer */

  pINT *w_ind;     /* Indizis fuer w_ptr je delta-x,delta-y,Vorzeichen */
  pINT w_indanz;   /* Nur fuer INVARI2-Netz! */

  void (*action_when_unit_clicked)(pINT l, pINT u) = unit_semantics;

  int mouse_function[3] = {M_WEIGHT, M_WEIGHT, M_TOGGLE}; /* Funktion der Maustasten (Left/Right/Middle) */
  REAL ScaleWeightDisplay = 1.0;    /* Skalierung der Gewichtsanzeige */
  REAL ScaleActivationDisplay = 1.0;/* Skalierung der Aktivierungsanzeige */

  REAL ONTHRES = ONTHRES_DEF;
  int GP_LINES = GP_LINES_DEF;
  int GP_COLLS = GP_COLLS_DEF;
  int GP_X = GP_X_DEF;
  int GP_Y = GP_Y_DEF;

  /* Parameter fuer f() */
  REAL GAIN    = GAIN_DEF,
       OFFSET  = OFFSET_DEF,
       ALPHA   = ALPHA_DEF,
       LAST    = LAST_DEF,
       DECAY   = DECAY_DEF,
       EPSILON = EPSILON_DEF,
       RATE    = RATE_DEF;

  /* IAC-Updates je Aufruf von update() */
  pINT MAXIAC  = MAXIAC_DEF;

  /* Default-Name */
  char ProgName[13] = "VieNet2";
  char LogAllocName[13] = "alloc.log";

  /* Filenames, Extensions werden automatisch generiert */
  char LogFile[32];
  char ErrorLogFile[32];
  char WeightFile[32];
  char ActFile[32];
  char LayoutFile[32];
  char MsgFile[32];
  char DataFile[32];

  long AutoSaveWeights=AutoSaveWeights_DEF;  /* number weights between autosaving (0=never) */

#endif

#endif



