/*************************************************************************
*  Program:       NNGRAPH.H                                              *
*  Description:   graphic- and systemfunctions for VieNet2               *
*  Version:       2.0                                                    *
*  Last modified: 10.02.94                                               *
*  Author:        LG                                                     *
*  Copyright (C) Austrian Research Institute for Artificial Intelligence *
**************************************************************************/
/* same for all Host-versions, */
/* but individal files implements the functions defined here */

#define SETSIZE (short) 256

/* Types of Button-Displaying */
#define AUS 0
#define EIN 1
#define NEU 2
#define PASSIV 3
#define AKTIV 4

/* Types for Linestyle */
#define SOLID 0
#define DOTTED 1

/* Window-Numbers */
#define MAIN_WINDOW 0
#define INFO_WINDOW 1
#define MSG_WINDOW 2
#define DRAW_WINDOW 3
#define USER_WINDOW 4

/* types for ShowWindow parameter in NNSetWindowPos() */
#define NN_HIDEWINDOW   0
#define NN_SETWINDOWPOS 1
#define NN_SHOWWINDOW   2

typedef struct {void(*ptr)(void); char* name;} NN_DDE_FctType;
typedef struct {float *ptr; char* name; unsigned int size;} NN_DDE_VarType;

#define NN_DDE_MaxFcts 25
#define NN_DDE_MaxVars 25

#ifdef MODULE
  extern int mouse_x, mouse_y, mouse_button;
  extern pINT DX;
  extern pINT UWID;
  extern pINT dither[4][4];
	extern char info_msg[5][40];
  extern NN_DDE_FctType NN_DDE_Fcts[];
  extern NN_DDE_VarType NN_DDE_Vars[];
  extern BOOL NNGraphInit;
#else
  int mouse_x,mouse_y,mouse_button;   /* these are defined in MAIN */
  pINT DX = DX_DEF;   /* Abstand zwischen Units in Pixel */
  pINT UWID = DX_DEF-2;  /* Unitgroessee in Pixel */
  /* Dither-Matrix (aus Zeitschrift c't 3/88 S.170)
     fuer Darstellung von 16 Grauwerten am S/W-Schirm */
  pINT dither[4][4]={{0, 8, 2, 10},
		     {12, 4, 14, 6},
		     {3, 11, 1, 9},
                     {15, 7, 13, 5}};
  /* fuer Infowindow */
  char info_msg[5][40]={"Usage:","Left-Click on unit to","show weights to it.","Right-Click a unit to","show weights from it."};
  NN_DDE_FctType NN_DDE_Fcts[NN_DDE_MaxFcts];
  NN_DDE_VarType NN_DDE_Vars[NN_DDE_MaxVars];
  BOOL NNGraphInit=FALSE; /* flag for graph init OK */
#endif

int gprintf(pINT,pINT,int,char*,...); /* wie printf(), aber im Grafikmodus */

BOOL nn_break(void);

/* functions only implemented for WIN */
void rectangle_xor(pINT,pINT,pINT,pINT);
extern BOOL bAbortRequested; /* Flag for nn_break() */
void draw_scaletext(REAL min,REAL max, char* title);
void x_draw_layer(pINT l,REAL maxabs);

#if defined(WIN) || !defined(PC)
  void rectangle(pINT,pINT,pINT,pINT);
#endif

/* functions only allowed for WIN and sun4 */
#ifdef __cplusplus
  void init_xgraph(int argc,char** argv,int UpperLeftX=-1,int UpperLeftY=-1,int WindowWidth=-1,int WindowHeight=-1);
#else
  void init_xgraph(int argc, char** argv);
#endif
  void x_draw_minmax_weights(void);
  void msg_print(char*);

/* ------------------------------------------*/
void check_args(int,char**);
void set_linestyle(pINT,pINT,pINT);
void draw_unit(REAL,pINT,pINT,pINT,pINT);
void draw_line(pINT,pINT,pINT,pINT);
void draw_circle(pINT,pINT,pINT);
void draw_bar(pINT,pINT,pINT,pINT,pINT);
void draw_rahmen(pINT,pINT);
void draw_palette(void);
void clear_box(pINT,pINT,pINT,pINT,pINT);
void set_draw_op(pINT);
char nnwait(void);
char x_wait_for_action(void);
void close_graph(void);
void init_graph(void);
void cursor_on(void);
void cursor_off(void);
void setMouseRange(int,int,int,int);
void setMousePosition(int,int);
void show_button(pINT,pINT);
void g_acquire(void);
void g_release(void);
void bit_blt(pINT,pINT,pINT,pINT,pINT,pINT);
pINT load_layout(char*);

#ifdef WIN
/* Vorlaeufig nur fuer Windows: */
void AppendWindowMenu(void*);
void insert_menu(int,int,char*,int);
void NNShowMenu(BOOL Show);
void NNSetWindowPos(int WinNr, int ShowWindow, int X=-1, int Y=-1, int W=-1, int H=-1);
void SetMeterValue(int Nr, float value);
void SetMeterTitle(int Nr, char* Title);
void SetMeterLimit(int Nr, float limit1, float limit2, float min, float max);
void NNErrMsg(char* sn, char* str);
void NNAlarmMsg(char* sn, char* str);
void NN_DDE_RegisterFunction(void(*ptr)(void),char *name);
void ClearLayerRect(pINT l);
void show_overview();
/* Soundfunctions */
void PlayWave(char*);
void StopWave();
#endif
